/*****************************************************************************
 *
 * This program is free software ; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 *
 * $Id$
 *
 ****************************************************************************/

#ifndef __AUDIO_H
#define __AUDIO_H

#define AUDIOFMT_PCM			1

#define SPEED_ONE			(1<<16)

// pcm flags
#define PCM_SWAPEDSTEREO	0x0001
#define PCM_PLANES			0x0002
#define PCM_UNSIGNED		0x0004
#define PCM_ONLY_LEFT		0x0008
#define PCM_ONLY_RIGHT		0x0010

typedef struct audio
{
	int Format;
	int Channels;	
	int SampleRate;
	int ByteRate;
	void* User;

	// pcm parameters
	int Flags;
	int Bits;		
	int FracBits;	

} audio;

typedef struct equalizer
{
	bool_t Enabled;
	int Amplify;	//-128..127
	int Eq[10];		//-128..127
} equalizer;

DLL void PCM_Init();
DLL void PCM_Done();

DLL void* PCMCreate(const audio* DstFormat, const audio* SrcFormat, bool_t Dither );
DLL void PCMReset(void* Handle);
DLL void PCMConvert(void* Handle, planes Dst, const planes Src, int* DstLength, int* SrcLength, int Speed );
DLL void PCMRelease(void* Handle);

//---------------------------------------------------------------
// audio output (abstract)

#define AOUT_CLASS			FOURCC('A','O','U','T')
// output volume (int 0..100)
#define AOUT_VOLUME			0x51
// volume mute (bool_t)
#define AOUT_MUTE			0x52
// quality (int 0..2)
#define AOUT_QUALITY		0x53
// swap stereo channels
#define AOUT_SWAP			0x55
// input pin (TYPE_AUDIO)
#define AOUT_INPUT			0x56
// output format (TYPE_AUDIO readonly)
#define AOUT_OUTPUT			0x57
// background (release resources)
#define AOUT_BACKGROUND		0x58
// stereo mode (int)
#define AOUT_STEREO			0x59
// buffer length (tick_t)
#define AOUT_BUFFER			0x5A

extern DLL const datadef AOutputParams[];

//---------------------------------------------------------------
// null audio

#define NULLAUDIO_ID		FOURCC('N','U','L','A')

extern DLL void NullAudio_Init();
extern DLL void NullAudio_Done();

#endif
