/*****************************************************************************
 *
 * This program is free software ; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 *
 * $Id$
 *
 ****************************************************************************/

#include "stdafx.h"
#include "../dyncode/dyncode.h"
#include "blit_soft.h"

#if defined( ARM ) 

typedef struct stack
{
	int EndOfLine;
	int EndOfRect;
	int DstPitch; 
	int DstNext;
	int SrcNext;
	int UVNext;

	int StackFrame[STACKFRAME];

	//void* this   R0
	//char* Dst    R1
	//char* Src    R2
	//int DstPitch R3 can be signed
	int SrcPitch;
	int Width; 
	int Height;
	int Src2SrcLast;
} stack;

void Fix_RGB_UV_LoadUV(blit_soft* p)
{
	//set R4 =           RVMul*v + RAdd
	//set R5 = GUMul*u + GVMul*v + GAdd
	//set R6 = BUMul*u +           BAdd

	if (p->FX.Flags & BLITFX_DITHER)
	{
		//R0,R4,R8 for temporary
		//R5(DiffMask),R6(Src2SrcLast),R7(EndOfLine)

		if (p->OnlyDiff)
		{
			Half(); I3(LDR,R0,R12,R6);
			Half(); I2C(LDR_POST,R8,R12,2);
			Half(); I3(LDR,R4,R14,R6);
			//xscale stall
			I3(EOR,R0,R0,R8);
			Half(); I2C(LDR_POST,R8,R14,2);
			S(); I3(TST,NONE,R0,R5);
			Byte(); I3(LDR,R0,R10,R6);
			I3(EOR,R4,R4,R8);
			Byte(); I2C(LDR_POST,R8,R10,1); //u
			C(EQ); S(); I3(TST,NONE,R4,R5);
			Byte(); I3(LDR,R4,R11,R6);
			I3(EOR,R0,R0,R8);
			Byte(); I2C(LDR_POST,R8,R11,1); //v
			C(EQ); S(); I3(TST,NONE,R0,R5);
			//xscale stall
			I3(EOR,R4,R4,R8);
			C(EQ); S(); I3(TST,NONE,R4,R5);

			if (p->SwapXY)
			{
				MB(); I2C(LDR,R0,SP,OFS(stack,DstPitch));
				C(EQ); I3S(ADD,R9,R9,R0,LSL,1+p->DstDoubleY);
			}
			else
			{
				C(EQ); I2C(ADD,R9,R9,p->DstStepX);
			}
			I0P(B,EQ,p->Skip);
		}
		else
		{
			Byte(); I2C(LDR_POST,R8,R11,1); //v
		}

		I1P(LDR,R5,p->GAdd,0);
		I1P(LDR,R0,p->GVMul,0);
		I1P(LDR,R4,p->RAdd,0);
		I1P(LDR,R6,p->RVMul,0);
		I4(MLA,R5,R0,R8,R5);
		
		Byte(); 
		if (p->OnlyDiff) //already incremented
			I2C(LDR,R0,R10,-1);	  //u
		else
			I2C(LDR_POST,R0,R10,1); //u

		I4(MLA,R4,R6,R8,R4);
		I1P(LDR,R7,p->GUMul,0);
		I1P(LDR,R6,p->BAdd,0);
		I1P(LDR,R8,p->BUMul,0);
		I4(MLA,R5,R7,R0,R5);

		//R7 Y read will moveback here (this will prevent stall by R8)

		I4(MLA,R6,R8,R0,R6);

		I1P(LDR,R8,p->YMul,0); //restore R8
	}
	else
	{
		//R0,R1,R2,R3,R4 for temporary
		//R5(DiffMask),R6(Src2SrcLast),R7(EndOfLine)

		if (p->OnlyDiff)
		{
			Half(); I3(LDR,R0,R12,R6);
			Half(); I2C(LDR_POST,R1,R12,2);
			Half(); I3(LDR,R4,R14,R6);	
			Half(); I2C(LDR_POST,R2,R14,2);
			I3(EOR,R0,R0,R1);
			S(); I3(TST,NONE,R0,R5);
			I3(EOR,R4,R4,R2);
			C(EQ); S(); I3(TST,NONE,R4,R5);

			Byte(); I3(LDR,R0,R10,R6);
			Byte(); I2C(LDR_POST,R1,R10,1); //u
			Byte(); I3(LDR,R4,R11,R6);
			Byte(); I2C(LDR_POST,R2,R11,1); //v
			I3(EOR,R0,R0,R1);
			C(EQ); S(); I3(TST,NONE,R0,R5);
			I3(EOR,R4,R4,R2);
			C(EQ); S(); I3(TST,NONE,R4,R5);

			if (p->SwapXY)
			{
				MB(); I2C(LDR,R3,SP,OFS(stack,DstPitch));
				C(EQ); I3S(ADD,R9,R9,R3,LSL,1+p->DstDoubleY);
			}
			else
			{
				C(EQ); I2C(ADD,R9,R9,p->DstStepX);
			}
			I0P(B,EQ,p->Skip);
		}
		else
		{
			Byte(); I2C(LDR_POST,R1,R10,1); //u
			Byte(); I2C(LDR_POST,R2,R11,1); //v
		}

		I1P(LDR,R5,p->GAdd,0);
		I1P(LDR,R0,p->GVMul,0);
		I1P(LDR,R4,p->RAdd,0);
		I1P(LDR,R7,p->RVMul,0);
		I1P(LDR,R6,p->BAdd,0);
		I4(MLA,R5,R0,R2,R5); 
		I1P(LDR,R0,p->BUMul,0);
		I1P(LDR,R3,p->GUMul,0);
		I4(MLA,R4,R7,R2,R4); 
		I4(MLA,R6,R0,R1,R6); 
		I4(MLA,R5,R3,R1,R5); 
	}
}

void Fix_RGB_UV_Pixel(blit_soft* p, int Col, int Row )
{
	int SatBit = p->QAdd ? 32 : 24;
	int RPos = p->RPos;
	int GPos = p->GPos;
	int BPos = p->BPos;

	p->Upper = (p->DirX<0) ^ (Col>0);

	if ((p->Dst.Flags & PF_PALETTE) || (p->Upper && p->DstBPP==8 && p->DstDoubleX))
	{
		RPos += 16;
		GPos += 16;
		BPos += 16;
	}
	else
	if (p->Upper && p->DstBPP<=16)
	{
		RPos += p->DstBPP;
		GPos += p->DstBPP;
		BPos += p->DstBPP;
	}

	//load Y
	MB();
	Byte(); 

	if (p->OnlyDiff) //is R12,R14 already incremented?
	{
		if (p->SwapXY)
			I2C(LDR,R7,(reg)(Col==0?R12:R14),-2+Row);
		else
			I2C(LDR,R7,(reg)(Row==0?R14:R12),-2+Col);
	}
	else
		if (p->SwapXY)
			I2C(LDR_POST,R7,(reg)(Col==0?R12:R14),1);
		else
			I2C(LDR_POST,R7,(reg)(Row==0?R14:R12),1);

	if (p->FX.Flags & BLITFX_DITHER)
	{
		I3S(ADD,R1,R4,R1,LSR,32-SatBit+p->RSize);
		I3S(ADD,R2,R5,R2,LSR,32-SatBit+p->GSize);
		I3S(ADD,R3,R6,R3,LSR,32-SatBit+p->BSize);
		I4(MLA,R1,R8,R7,R1);
		I4(MLA,R2,R8,R7,R2);
		I4(MLA,R3,R8,R7,R3);
	}
	else
	{
		I4(MLA,R1,R8,R7,R4);
		I4(MLA,R2,R8,R7,R5);
		I4(MLA,R3,R8,R7,R6);
	}

	if (p->QAdd)
	{
		I3(QDADD,R1,R1,R1);
		I3(QDADD,R2,R2,R2);
		I3(QDADD,R3,R3,R3);
	}
	else
	{
		I2C(TST,NONE,R1,0xFF000000);
		C(NE);I2C(MVN,R1,NONE,0xFF000000);
		C(MI);I2C(MOV,R1,NONE,0x00000000);
		I2C(TST,NONE,R2,0xFF000000);
		C(NE);I2C(MVN,R2,NONE,0xFF000000);
		C(MI);I2C(MOV,R2,NONE,0x00000000);
		I2C(TST,NONE,R3,0xFF000000);
		C(NE);I2C(MVN,R3,NONE,0xFF000000);
		C(MI);I2C(MOV,R3,NONE,0x00000000);
	}

	if (p->InvertMask && p->Pos<0)
	{
		p->Pos = RPos;
		MB(); I1P(LDR,R0,p->InvertMask,0);
	}
	if (p->Pos!=RPos && p->Pos>=0) I3S(MOV,R0,NONE,R0,ROR,RPos-p->Pos);
	I3S(p->Pos<0?MOV:EOR,R0,(reg)(p->Pos<0?NONE:R0),R1,LSR,SatBit-p->RSize);
	I3S(MOV,R0,NONE,R0,ROR,GPos-RPos);
	I3S(EOR,R0,R0,R2,LSR,SatBit-p->GSize);
	I3S(MOV,R0,NONE,R0,ROR,BPos-GPos);
	I3S(EOR,R0,R0,R3,LSR,SatBit-p->BSize);
	p->Pos = BPos;

	if (p->FX.Flags & BLITFX_DITHER)
	{
		MB(); I3S(MOV,R1,NONE,R1,LSL,32-SatBit+p->RSize);
		MB(); I3S(MOV,R2,NONE,R2,LSL,32-SatBit+p->GSize);
		MB(); I3S(MOV,R3,NONE,R3,LSL,32-SatBit+p->BSize);
	}
}

void Fix_RGB_UV(blit_soft* p)
{
	dyninst* LoopY;
	dyninst* LoopX;
	int Invert = 0;
	int Mask = 0;

	p->DstStepX = p->DirX * ((p->DstBPP*2) >> 3) << p->DstDoubleX;
	p->QAdd = DSP && !(p->FX.Flags & BLITFX_DITHER);

	p->PalPtr = NULL;
	p->DiffMask = NULL;
	p->InvertMask = NULL;

	if (p->Dst.Flags & PF_PALETTE)
		p->PalPtr = InstCreate32((int)&p->PalLookUp,NONE,NONE,NONE,0,0);
	else
	if (p->Dst.Flags & PF_INVERTED)
		Invert = -1;

	CalcColor(p);

	if (p->QAdd)
	{
		int Mask2;
		int Shift;

		Mask |= 1 << (BitMaskPos(p->Dst.RBitMask)+BitMaskSize(p->Dst.RBitMask)-1);
		Mask |= 1 << (BitMaskPos(p->Dst.GBitMask)+BitMaskSize(p->Dst.GBitMask)-1);
		Mask |= 1 << (BitMaskPos(p->Dst.BBitMask)+BitMaskSize(p->Dst.BBitMask)-1);

		Mask2 = Mask;
		Shift = 0;
		if ((p->Dst.Flags & PF_PALETTE) || (p->DstBPP==8 && p->DstDoubleX))
			Mask2 |= Mask << 16;
		else
		if (p->DstBPP <= 16)
		{
			if (p->DirX<0) Shift = p->DstBPP;
			Mask2 |= Mask << p->DstBPP;
		}
		
		Invert ^= RotateRight(Mask2,Shift+BitMaskPos(p->Dst.RBitMask));
	}

	p->YMul = InstCreate32(p->_YMul,NONE,NONE,NONE,0,0);
	p->RVMul = InstCreate32(p->_RVMul,NONE,NONE,NONE,0,0);
	p->RAdd = InstCreate32(p->_RAdd,NONE,NONE,NONE,0,0);
	p->GUMul = InstCreate32(p->_GUMul,NONE,NONE,NONE,0,0);
	p->GVMul = InstCreate32(p->_GVMul,NONE,NONE,NONE,0,0);
	p->GAdd = InstCreate32(p->_GAdd,NONE,NONE,NONE,0,0);
	p->BUMul = InstCreate32(p->_BUMul,NONE,NONE,NONE,0,0);
	p->BAdd = InstCreate32(p->_BAdd,NONE,NONE,NONE,0,0);

	if (Invert)
		p->InvertMask = InstCreate32(Invert,NONE,NONE,NONE,0,0);

	if (p->OnlyDiff)
		p->DiffMask = InstCreate32(0xFCFCFCFC,NONE,NONE,NONE,0,0);

	CodeBegin();
	I2C(SUB,SP,SP,OFS(stack,StackFrame));

	I2C(LDR,R9,R1,0); //Dst[0] RGB
	I2C(LDR,R10,R2,4); //Src[1] U
	I2C(LDR,R11,R2,8); //Src[2] V
	I2C(LDR,R12,R2,0); //Src[0] Y
	I2C(STR,R3,SP,OFS(stack,DstPitch));

	I3(MOV,R6,NONE,R3); //DstPitch
	I2C(LDR,R7,SP,OFS(stack,SrcPitch));
	I2C(LDR,R0,SP,OFS(stack,Height));
	I2C(LDR,R4,SP,OFS(stack,Width));
	I1P(LDR,R8,p->YMul,0);

	//SrcNext = 2*Src->Pitch - (Width >> SrcDoubleX)
	I3S(MOV,R1,NONE,R7,LSL,1);
	I3S(SUB,R1,R1,R4,LSR,p->SrcDoubleX); 
	I2C(STR,R1,SP,OFS(stack,SrcNext));

	//UVNext = (Src->Pitch >> 1) - (Width >> SrcDoubleX >> 1);
	I3S(MOV,R2,NONE,R7,LSR,1);
	I3S(SUB,R2,R2,R4,LSR,p->SrcDoubleX+1); 
	I2C(STR,R2,SP,OFS(stack,UVNext));

	if (p->DirX<0) //adjust reversed destination for block size
		I2C(SUB,R9,R9,-p->DstStepX-(p->DstBPP >> 3));

	if (p->SwapXY)
	{
		I3S(SUB,R9,R9,R6,LSL,p->DstDoubleY);
		I2C(MOV,R1,NONE,p->DstBPP * p->DirX);
		I3(MUL,R0,R1,R0);
		I3S(ADD,R0,R9,R0,ASR,3);
		I2C(STR,R0,SP,OFS(stack,EndOfRect));

		//DstNext = DstStepX - Width*DstPitch;
		MB(); I3(MUL,R2,R6,R4);
		I2C(MOV,R0,NONE,p->DstStepX);
		I3(SUB,R0,R0,R2); 
		I2C(STR,R0,SP,OFS(stack,DstNext));
	}
	else
	{
		I3(MUL,R0,R6,R0);
		I3(ADD,R0,R9,R0);
		I2C(STR,R0,SP,OFS(stack,EndOfRect));

		//DstNext = ((DstPitch*2 << DstDoubleY) - DirX * Width << DstBPP2;
		I3S(MOV,R2,NONE,R6,LSL,p->DstDoubleY+1);
		I3S(p->DirX>0?SUB:ADD,R2,R2,R4,LSL,p->DstBPP2); 
		I2C(STR,R2,SP,OFS(stack,DstNext));
	}
	I3(ADD,R14,R12,R7);

	if (p->FX.Flags & BLITFX_DITHER)
	{
		I2C(MVN,R1,NONE,0x80000000);
		I2C(MVN,R2,NONE,0x80000000);
		I2C(MVN,R3,NONE,0x80000000);
	}

	LoopY = Label(1);

	if (p->SwapXY)
	{
		I3(MUL,R4,R6,R4); //R6=dstpitch
		I3(ADD,R7,R9,R4);
	}
	else
	{
		if (p->DirX > 0)
			I3S(ADD,R7,R9,R4,LSL,p->DstBPP2);
		else
			I3S(SUB,R7,R9,R4,LSL,p->DstBPP2);
	}
	I2C(STR,R7,SP,OFS(stack,EndOfLine));
    if (p->OnlyDiff)
	{
		MB(); I1P(LDR,R5,p->DiffMask,0);
		MB(); I2C(LDR,R6,SP,OFS(stack,Src2SrcLast));
		p->Skip = Label(0);
	}

	LoopX = Label(1);
	{
		int PitchDouble;
		reg Pitch;

		Fix_RGB_UV_LoadUV(p);

		p->Pos = -1;

		Fix_RGB_UV_Pixel(p,0,0);
	
		if (p->Dst.Flags & PF_PALETTE)
		{
			if (p->Pos) 
				I3S(MOV,R0,NONE,R0,ROR,-p->Pos);
			p->Pos = 0;
			MB(); I1P(LDR,R7,p->PalPtr,0);
			Byte(); I3S(LDR,R0,R7,R0,LSR,16);
		}

		Fix_RGB_UV_Pixel(p,1,0);
		if (p->Pos) 
			I3S(MOV,R0,NONE,R0,ROR,-p->Pos);

		if (p->Dst.Flags & PF_PALETTE)
		{
			reg Tmp = (reg)((p->FX.Flags & BLITFX_DITHER) ? R7:R2); 
			MB(); I1P(LDR,Tmp,p->PalPtr,0);
			if (Mask) I2C(EOR,R0,R0,Mask << 16);
			Byte(); I3S(LDR,Tmp,Tmp,R0,LSR,16);
			if (p->DstDoubleX) I2C(AND,R0,R0,255);
			if (p->Upper)
				I3S(ORR,R0,R0,Tmp,LSL,8 << p->DstDoubleX);
			else
				I3S(ORR,R0,Tmp,R0,LSL,8 << p->DstDoubleX);
		}

		Pitch = (reg)((p->FX.Flags & BLITFX_DITHER) ? R7:R1);
		MB(); I2C(LDR,Pitch,SP,OFS(stack,DstPitch));

		if (p->DstBPP==8 && p->DstDoubleX) 
			I3S(ORR,R0,R0,R0,LSL,8);

		if (p->DstBPP==16 && p->DstDoubleX) 
		{
			I2C(ADD,R9,R9,4);
			I3S(MOV,R3,NONE,R0,LSR,16);
			I3S(MOV,R0,NONE,R0,LSL,16);
			I3S(ORR,R3,R3,R3,LSL,16);
			I3S(ORR,R0,R0,R0,LSR,16);

			if (p->DstDoubleY)
			{
				I3S(ADD,R2,Pitch,Pitch,LSL,1); //R2=3*DstPitch
				I3(STR,R3,R9,R2);
			}
			I3S(STR,R3,R9,Pitch,LSL,p->DstDoubleY);
			I2C(SUB,R9,R9,4);
		}

		if (p->DstDoubleY)
		{
			I3S(ADD,R2,Pitch,Pitch,LSL,1); //R2=3*DstPitch
			if (p->DstBPP==8 && !p->DstDoubleX) Half();
			I3(STR,R0,R9,R2);
		}
		
		PitchDouble = p->DstDoubleY;
		if (p->DstBPP==8 && !p->DstDoubleX) 
		{
			if (PitchDouble) // can't use STR with Half() and LSL,#1 at the same time
			{
				PitchDouble = 0;
				I3(ADD,Pitch,Pitch,Pitch);
			}
			Half();
		}
		I3S(STR,R0,R9,Pitch,LSL,PitchDouble);
		if (p->SwapXY)
			I3S(ADD,R9,R9,Pitch,LSL,1+PitchDouble);

		p->Pos = -1;
		Fix_RGB_UV_Pixel(p,0,1);

		if (p->Dst.Flags & PF_PALETTE)
		{
			if (p->Pos) 
				I3S(MOV,R0,NONE,R0,ROR,-p->Pos);
			p->Pos = 0;
			MB(); I1P(LDR,R7,p->PalPtr,0);
			Byte(); I3S(LDR,R0,R7,R0,LSR,16);
		}

		Fix_RGB_UV_Pixel(p,1,1);
		if (p->Pos) 
			I3S(MOV,R0,NONE,R0,ROR,-p->Pos);

		if (p->Dst.Flags & PF_PALETTE)
		{
			reg Tmp = (reg)((p->FX.Flags & BLITFX_DITHER) ? R7:R2);
			MB(); I1P(LDR,Tmp,p->PalPtr,0);
			//last pal lookup cleared the invert bits -> reinvert
			if (Mask) I2C(EOR,R0,R0,Mask << 16); 
			Byte(); I3S(LDR,Tmp,Tmp,R0,LSR,16);
			if (p->DstDoubleX) I2C(AND,R0,R0,255);
			if (p->Upper)
				I3S(ORR,R0,R0,Tmp,LSL,8 << p->DstDoubleX);
			else
				I3S(ORR,R0,Tmp,R0,LSL,8 << p->DstDoubleX);
		}

		if (p->DstBPP==8 && p->DstDoubleX) 
			I3S(ORR,R0,R0,R0,LSL,8);

		if (p->DstDoubleY)
		{
			MB(); I2C(LDR,R1,SP,OFS(stack,DstPitch));
		}

		if (p->DstBPP==16 && p->DstDoubleX) 
		{
			I3S(MOV,R3,NONE,R0,LSR,16);
			I3S(MOV,R0,NONE,R0,LSL,16);
			I3S(ORR,R3,R3,R3,LSL,16);
			I3S(ORR,R0,R0,R0,LSR,16);

			if (p->DstDoubleY)
			{
				I2C(ADD,R2,R1,4); //DstPitch+4
				I3(STR,R3,R9,R2);
			}
			I2C(STR,R3,R9,4);
		}

		if (p->DstDoubleY)
		{
			if (p->DstBPP==8 && !p->DstDoubleX) Half();
			I3(STR,R0,R9,R1);
		}

		if (p->DstBPP==8 && !p->DstDoubleX) Half();
		if (p->SwapXY)
			I2(STR,R0,R9);
		else
			I2C(STR_POST,R0,R9,p->DstStepX);

		MB(); I2C(LDR,R7,SP,OFS(stack,EndOfLine));
		if (p->OnlyDiff)
		{
			MB(); I1P(LDR,R5,p->DiffMask,0);
			MB(); I2C(LDR,R6,SP,OFS(stack,Src2SrcLast));
			InstPost(p->Skip);
		}

		I3(CMP,NONE,R9,R7);
		I0P(B,NE,LoopX);
	}

	I2C(LDR,R0,SP,OFS(stack,SrcNext));
	I2C(LDR,R4,SP,OFS(stack,DstNext));
	I2C(LDR,R6,SP,OFS(stack,UVNext));
	I2C(LDR,R5,SP,OFS(stack,EndOfRect));
	
	//increment pointers
	I3(ADD,R12,R12,R0);
	I3(ADD,R14,R14,R0);
	I3(ADD,R9,R9,R4);
	I3(ADD,R10,R10,R6);
	I3(ADD,R11,R11,R6);

	//prepare registers for next row
	if (p->SwapXY) I2C(LDR,R6,SP,OFS(stack,DstPitch));
	I2C(LDR,R4,SP,OFS(stack,Width));

	I3(CMP,NONE,R9,R5);
	I0P(B,NE,LoopY);

	I2C(ADD,SP,SP,OFS(stack,StackFrame));
	CodeEnd();

	InstPost(p->YMul);
	InstPost(p->RVMul);
	InstPost(p->RAdd);
	InstPost(p->GUMul);
	InstPost(p->GVMul);
	InstPost(p->GAdd);
	InstPost(p->BUMul);
	InstPost(p->BAdd);
	if (p->InvertMask) InstPost(p->InvertMask);
	if (p->DiffMask) InstPost(p->DiffMask);
	if (p->PalPtr) InstPost(p->PalPtr);
}

#endif