/*****************************************************************************
 *
 * This program is free software ; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 *
 * $Id$
 *
 ****************************************************************************/

#include "stdafx.h"
#include "../dyncode/dyncode.h"
#include "blit_soft.h"

#if defined( MIPS )

typedef struct stack
{
	int64_t DiffMask;
	int64_t SatNegMask; // 0x7FFF 
	int64_t SatPosMask; // 0x0400 (0x100 << 3)
	int64_t AddR,AddG,AddB;		//AddY,Const128 when low quality
	int64_t MaskR,MaskAccR;
	int64_t MaskG,MaskAccG;
	int64_t MaskB,MaskAccB;
} stack;

typedef struct param
{
	void* This;   //R4
	char* Dst;    //R5
	char* Src;    //R6
	int DstPitch; //R7
	int SrcPitch;
	int Width; 
	int Height;
	int Src2SrcLast;
} param;

// mutual exclusive flags: scale vs double vs diff
// other options: dither, swapxy

// R1,R2,R3		row0 acc (if dither)
// R1,R2,R3		tmp (if !dither)
// R9,R10,R11	row1 acc (if !swapxy && dither)
// R12,R13,R14	row2 acc (if !swapxy && dither)
// R15,R24,R25	row3 acc (if !swapxy && dither)
// R15,R24,R25	maskr,maskg,maskb (if !dither && (!swaxp || !scale))
// R14,R15,R24,R25 u1,v1,u3,v3 (if swapxy && scale)
// R9,R10,R11   y2,y3,y4 (if swapxy)
// R12,R13		u2,v2 (if swapxy)
// R4 y
// R5 u
// R6 v
// R7 dst
// R8 dst2 (if doubley)
// R8 src2srclast (if diff)
// R16 tmp
// R17 tmp
// R18 tmp
// R19 tmp
// R20,R21,R22,R23 tmp for diff compare (if diff)
// R20 y inc resum	(if scale)
// R21 uv inc resum (if scale)
// R22 y add fix	(if scale)
// R23 uv add fix	(if scale)
// R20 lastrgb0 (if double)
// R21 lastrgb1 (if double & !swapxy)
// R22 lastrgb2 (if double & !swapxy)
// R23 lastrgb3 (if double & !swapxy)
// R28 endofline
// R30 srcpitch 
// R31 dstpitch * double


//input:
// src offset
// acc array
// dst
// dst2 (if double)
// y,u,v
// y2,y3,y4,u2,u3,u4,v2,v3,v4 (if swapxy)
//
//dst write:
// 2x1 pixel (if 32bit)
// 4x2 pixel (if 32bit & double)
// 4x1 pixel (if 64bit)
// 8x2 pixel (if 64bit & double)

void Universal_RGB_UV_Pixel( blit_soft* p, int No )
{
	dyninst* Skip = NULL;
	int ShiftR,ShiftG,ShiftB;
	int Ofs[4];
	reg AccR,AccG,AccB,Last;
	reg Y[4];
	reg U[4];
	reg V[4];

	int Width = 8;
	int _LD = LD;
	int _SD = SD;
	int _LW = LW;
	int _LHU = LHU;
	int _SRL = DSRL;
	int _SLL = DSLL;
	int _ADDU = DADDU;
	int _SUBU = DSUBU;
	
	if (!M64)
	{
		Width = 4;
		_LD = LW;
		_SD = SW;
		_LW = LHU;
		_LHU = LBU;
		_SRL = SRL;
		_ADDU = ADDU;
		_SLL = SLL;
		_SUBU = SUBU;
	}

	if (p->OnlyDiff2)
		Skip = Label(0);

	ShiftR = p->RPos + p->RSize - 8;
	ShiftG = p->GPos + p->GSize - 8;
	ShiftB = p->BPos + p->BSize - 8;

	if (p->FX.Flags & BLITFX_QUALITYCOLOR)
	{
		ShiftR -= 3;
		ShiftG -= 3;
		ShiftB -= 3;
	}

	AccR = R1;
	AccG = R2;
	AccB = R3;
	Last = R20;

	if (p->SwapXY)
	{
		Ofs[0] = No;
		Ofs[1] = No;
		Ofs[2] = No;
		Ofs[3] = No;

		Y[0] = R4; Y[1] = R9;  Y[2] = R10; Y[3] = R11;
		U[0] = R5; U[1] = R14; U[2] = R12; U[3] = R24;
		V[0] = R6; V[1] = R15; V[2] = R13; V[3] = R25;

		if (p->OnlyDiff2 && No==0) // only available when RScaleX2 == 16 && RScaleY2 == 16
		{
			// 2x2 comapre (32bit)
			// 4x4 compare (64bit)

			I3(ADDU,R21,Y[0],R8);
			I2C(_LW,R16,Y[0],0);
			I2C(_LW,R20,R21,0);
			I3(XOR,R16,R16,R20);

			MB(); I3(ADDU,R21,Y[1],R8);
			I2C(_LW,R17,Y[1],0);
			I2C(_LW,R20,R21,0);
			I3(XOR,R17,R17,R20);
			I3(OR,R16,R16,R17);

			MB(); I3(ADDU,R21,Y[2],R8);
			I2C(_LW,R17,Y[2],0);
			I2C(_LW,R20,R21,0);
			I3(XOR,R17,R17,R20);
			I3(OR,R16,R16,R17);

			MB(); I3(ADDU,R21,Y[3],R8);
			I2C(_LW,R17,Y[2],0);
			I2C(_LW,R20,R21,0);
			I3(XOR,R17,R17,R20);
			I3(OR,R16,R16,R17);

			MB(); I3(ADDU,R21,U[0],R8);
			I2C(_LHU,R17,U[0],0);
			I2C(_LHU,R20,R21,0);
			I3(XOR,R17,R17,R20);
			I3(OR,R16,R16,R17);

			MB(); I3(ADDU,R21,U[2],R8);
			I2C(_LHU,R17,U[2],0);
			I2C(_LHU,R20,R21,0);
			I3(XOR,R17,R17,R20);
			I3(OR,R16,R16,R17);

			MB(); I3(ADDU,R21,V[0],R8);
			I2C(_LHU,R17,V[0],0);
			I2C(_LHU,R20,R21,0);
			I3(XOR,R17,R17,R20);
			I3(OR,R16,R16,R17);

			MB(); I3(ADDU,R21,V[2],R8);
			I2C(_LHU,R17,V[2],0);
			I2C(_LHU,R20,R21,0);
			I3(XOR,R17,R17,R20);
			I3(OR,R16,R16,R17);

			MB(); I2C(LD,R21,SP,OFS(stack,DiffMask));
			I3(AND,R16,R16,R12);
			I2P(BEQ,R16,ZERO,Skip);
			// delay slot doesn't matter
		}
	}
	else
	{
		switch (No)
		{
		case 1:
			if (p->FX.Flags & BLITFX_DITHER)
			{
				AccR = R9;
				AccG = R10;
				AccB = R11;
			}
			Last = R21;
			break;
		case 2:
			if (p->FX.Flags & BLITFX_DITHER)
			{
				AccR = R12;
				AccG = R13;
				AccB = R14;
			}
			Last = R22;
			break;
		case 3:
			if (p->FX.Flags & BLITFX_DITHER)
			{
				AccR = R15;
				AccG = R24;
				AccB = R25;
			}
			Last = R23;
			break;
		}

		Ofs[0] = ((No*4+0)*p->RScaleX2) >> 4;
		Ofs[1] = ((No*4+1)*p->RScaleX2) >> 4;
		Ofs[2] = ((No*4+2)*p->RScaleX2) >> 4;
		Ofs[3] = ((No*4+3)*p->RScaleX2) >> 4;

		Y[0] = Y[1] = Y[2] = Y[3] = R4;
		U[0] = U[1] = U[2] = U[3] = R5;
		V[0] = V[1] = V[2] = V[3] = R6;

		if (p->OnlyDiff2) // only available when RScaleX2 == 16 && RScaleY2 == 16
		{
			// 2x1 compare (32bit)
			// 4x1 compare (64bit)

			I3(ADDU,R21,Y[0],R8);
			I2C(_LW,R16,Y[0],Ofs[0]);
			I2C(_LW,R20,R21,Ofs[0]);
			I3(XOR,R16,R16,R20);

			MB(); I3(ADDU,R21,U[0],R8);
			I2C(_LHU,R17,U[0],Ofs[0] >> 1);
			I2C(_LHU,R20,R21,Ofs[0] >> 1);
			I3(XOR,R17,R17,R21);
			I3(OR,R16,R16,R17);

			MB(); I3(ADDU,R21,V[0],R8);
			I2C(_LHU,R17,V[0],Ofs[0] >> 1);
			I2C(_LHU,R20,R21,Ofs[0] >> 1);
			I3(XOR,R17,R17,R20);
			I3(OR,R16,R16,R17);

			MB(); I2C(LD,R21,SP,OFS(stack,DiffMask));
			I3(AND,R16,R16,R21);
			I2P(BEQ,R16,ZERO,Skip);
			// delay slot doesn't matter
		}
	}

	if (M64)
	{
		// load 4 pixel 
		I2C(LBU,R16,Y[0],Ofs[0]);
		I2C(LBU,R17,Y[1],Ofs[1]);
		I2C(LBU,R18,Y[2],Ofs[2]);
		I3(DSLL,R17,R17,16);
		I3(OR,R16,R16,R17);
		I2C(LBU,R17,Y[3],Ofs[3]);
		I2C(DSLL,R18,R18,32);
		I3(OR,R16,R16,R18);
		I2C(DSLL,R17,R17,48);
		I3(OR,R16,R16,R17);

		I2C(LBU,R19,V[2],Ofs[2] >> 1);
		I2C(LBU,R18,V[0],Ofs[0] >> 1);
		I2C(DSLL,R19,R19,32);
		I3(OR,R18,R18,R19);

		if (p->RScaleX2 != 16) // V not same for Y01 and Y23
		{
			I2C(LBU,R19,V[1],Ofs[1] >> 1);
			I2C(LBU,R17,V[3],Ofs[3] >> 1);
			I2C(DSLL,R19,R19,16);
			I2C(DSLL,R17,R17,48);
			I3(OR,R18,R18,R19);
			I3(OR,R18,R18,R17);
		}
		else
		{
			I2C(DSLL,R19,R18,32);
			I3(OR,R18,R18,R19);
		}

		I2C(LBU,R19,U[2],Ofs[2] >> 1);
		I2C(LBU,R17,U[0],Ofs[0] >> 1);
		I2C(DSLL,R19,R19,32);
		I3(OR,R17,R17,R19);

		if (p->RScaleX2 != 16) // U not same for Y01 and Y23
		{
			I2C(LBU,R19,U[1],Ofs[1] >> 1);
			I2C(DSLL,R19,R19,16);
			I3(OR,R17,R17,R19);
			I2C(LBU,R19,U[3],Ofs[3] >> 1);
			I2C(DSLL,R19,R19,48);
			I3(OR,R17,R17,R19);
		}
		else
		{
			I2C(DSLL,R19,R17,32);
			I3(OR,R17,R17,R19);
		}
	}
	else
	{
		// load 2 pixel 
		I2C(LBU,R17,Y[1],Ofs[1]);
		I2C(LBU,R16,Y[0],Ofs[0]);
		I2C(DSLL,R17,R17,16);
		I3(OR,R16,R16,R17);

		MB(); I2C(LBU,R18,V[0],Ofs[0] >> 1);

		if (p->RScaleX2 != 16) // V not same for Y01
		{
			MB(); I2C(LBU,R19,V[1],Ofs[1] >> 1);
			I2C(DSLL,R19,R19,16);
			I3(OR,R18,R18,R19);
		}
		else
		{
			I2C(DSLL,R19,R18,16);
			I3(OR,R18,R18,R19);
		}

		MB(); I2C(LBU,R17,U[0],Ofs[0] >> 1);

		if (p->RScaleX2 != 16) // U not same for Y01 
		{
			MB(); I2C(LBU,R19,U[1],Ofs[1] >> 1);
			I2C(DSLL,R19,R19,16);
			I3(OR,R17,R17,R19);
		}
		else
		{
			I2C(DSLL,R19,R17,16);
			I3(OR,R17,R17,R19);
		}
	}

	// R16,R17,R18 (Y,U,V)

	if (!(p->FX.Flags & BLITFX_QUALITYCOLOR))
	{
		// we doesn't care about carry to neighbour pixel

		// Tmp = Y + AddR (-144+Brightness)
		// R = Tmp + V 
		// G = Y + AddG (-16+Brightness)
		// B = Tmp + U

		if (p->FX.Flags & BLITFX_DITHER)
		{
			I2C(_LD,R19,SP,OFS(stack,AddR));
			I3(_ADDU,R18,R18,AccR);
			I3(_ADDU,R17,R17,AccB);
			I3(_ADDU,R19,R19,R16);
			I3(_ADDU,R17,R17,R19);
			I3(_ADDU,R18,R18,R19);
			I2C(_LD,R19,SP,OFS(stack,AddG));
			I3(_ADDU,R16,R16,AccG);
			I3(_ADDU,R16,R16,R19);
		}
		else
		{
			// R1,R2,R3 is free
			I2C(_LD,R19,SP,OFS(stack,AddG));
			I2C(_LD,R1,SP,OFS(stack,AddR));
			I3(_ADDU,R19,R19,R16);
			I3(_ADDU,R17,R17,R19);
			I3(_ADDU,R18,R18,R19);
			I3(_ADDU,R16,R16,R1);
		}

		// negative saturate
		I3(NOR,R19,R16,R16);
		I2C(_SRL,R19,R19,8);
		I3(AND,R16,R16,R19);

		I3(NOR,R19,R17,R17);
		I2C(_SRL,R19,R19,8);
		I3(AND,R17,R17,R19);

		I3(NOR,R19,R18,R18);
		I2C(_SRL,R19,R19,8);
		I3(AND,R18,R18,R19);
	}
	else
	{
		// R = 9*Y        + 13*V + AddR 
		// G = 9*Y -  3*U -  7*V + AddG 
		// B = 9*Y + 16*U        + AddB 

		if (!(p->FX.Flags & BLITFX_DITHER))
		{
			I2C(_LD,AccR,SP,OFS(stack,AddR));
			I2C(_LD,AccG,SP,OFS(stack,AddG));
			I2C(_LD,AccB,SP,OFS(stack,AddB));
		}

		I2C(_SLL,R19,R16,3);
		I3(_ADDU,R19,R19,R16);	// R19 = 9*Y

		I3(_ADDU,R16,R19,AccG);
		I2C(_SLL,AccG,R17,1);
		I3(_SUBU,R16,R16,AccG);		
		I3(_SUBU,R16,R16,R17);	// 9*Y - 3*U + AddG

		I2C(_SLL,R17,R17,4);		
		I3(_ADDU,R17,R17,R19);	
		I3(_ADDU,R17,R17,AccB);	// 9*Y + 16*U + AddB

		I2C(_SLL,AccG,R18,3);
		I3(_SUBU,AccG,AccG,R18);	// 7*V
		I3(_SUBU,R16,R16,AccG);		// 9*Y - 3*U - 7*U + AddG
		I3(_ADDU,AccG,AccG,AccG);		
		I3(_SUBU,AccG,AccG,R18);	// 13*V
		I3(_ADDU,R18,AccG,R19);	
		I3(_ADDU,R18,R18,AccR);		// 9*Y + 13*V + AddR

		if (p->FX.Flags & BLITFX_DITHER)
		{
			I2C(_LD,AccR,SP,OFS(stack,AddR));
			I2C(_LD,AccG,SP,OFS(stack,AddG));
			I2C(_LD,AccB,SP,OFS(stack,AddB));
			I3(_ADDU,R18,R18,AccR);
			I3(_ADDU,R16,R16,AccG);
			I3(_ADDU,R17,R17,AccB);
		}

		// negative sat
		MB(); I2C(_LD,AccB,SP,OFS(stack,SatNegMask));

		I3(NOR,AccR,R16,AccB);
		I2C(_SRL,AccG,AccR,15-3);
		I3(_SUBU,AccR,AccR,AccG);
		I3(AND,R16,R16,AccR);

		I3(NOR,AccR,R17,AccB);
		I2C(_SRL,AccG,AccR,15-3);
		I3(_SUBU,AccR,AccR,AccG);
		I3(AND,R17,R17,AccR);

		I3(NOR,AccR,R18,AccB);
		I2C(_SRL,AccG,AccR,15-3);
		I3(_SUBU,AccR,AccR,AccG);
		I3(AND,R18,R18,AccR);

		// positive sat
		MB(); I2C(_LD,AccB,SP,OFS(stack,SatPosMask));

		I3(AND,AccR,R16,AccB);
		I2C(_SRL,AccG,AccR,8);
		I3(_SUBU,AccR,AccR,AccG);
		I3(OR,R16,R16,AccR);

		I3(AND,AccR,R17,AccB);
		I2C(_SRL,AccG,AccR,8);
		I3(_SUBU,AccR,AccR,AccG);
		I3(OR,R17,R17,AccR);

		I3(AND,AccR,R18,AccB);
		I2C(_SRL,AccG,AccR,8);
		I3(_SUBU,AccR,AccR,AccG);
		I3(OR,R18,R18,AccR);
	}

	// R16,R17,R18 (G,B,R)

	if (p->FX.Flags & BLITFX_DITHER)
	{
		I2C(_LD,R19,SP,OFS(stack,MaskR));
		I2C(_LD,AccR,SP,OFS(stack,MaskAccR));
		I3(AND,R19,R18,R19);
		I3(AND,AccR,R18,AccR);

		I2C(_LD,R18,SP,OFS(stack,MaskG));
		I2C(_LD,AccG,SP,OFS(stack,MaskAccG));
		I3(AND,R18,R16,R18);
		I3(AND,AccG,R16,AccG);

		I2C(_LD,R16,SP,OFS(stack,MaskB));
		I2C(_LD,AccB,SP,OFS(stack,MaskAccB));
		I3(AND,R16,R17,R16);
		I3(AND,AccB,R17,AccB);

		// R18,R16,R19 (G,B,R)

		if (ShiftR) I2C(_SLL,R19,R19,ShiftR);
		if (ShiftG) I2C(_SLL,R18,R18,ShiftG);
		if (ShiftB) I2C(_SLL,R16,R16,ShiftB);

		I3(OR,R16,R16,R19);
		I3(OR,R16,R16,R18);
	}
	else
	{
		if (!(p->SwapXY) || !Scale)
		{
			//R15,R24,R25 MaskR,MaskG,MaskB
			I3(AND,R16,R16,R15);
			I3(AND,R17,R17,R24);
			I3(AND,R18,R18,R25);
		}
		else
		{
			I2C(_LD,AccR,SP,OFS(stack,MaskR));
			I2C(_LD,AccG,SP,OFS(stack,MaskG));
			I2C(_LD,AccB,SP,OFS(stack,MaskB));
			I3(AND,R18,R18,AccR);
			I3(AND,R16,R16,AccG);
			I3(AND,R17,R17,AccB);
		}

		// R16,R17,R18 (G,B,R)

		if (ShiftR) I2C(_SLL,R18,R18,ShiftR);
		if (ShiftG) I2C(_SLL,R16,R16,ShiftG);
		if (ShiftB) I2C(_SLL,R17,R17,ShiftB);

		I3(OR,R16,R16,R18);
		I3(OR,R16,R16,R17);
	}

	//R16 rgb 16bit

	//!!! paletta lookup and 8bit rgb support

	if (p->Double)
	{
		//!! average with last


		I2C(_SD,R16,R7,No*Width*2);
		I2C(_SD,R16,R7,No*Width*2+Width);
		I2C(_SD,R16,R8,No*Width*2);
		I2C(_SD,R16,R8,No*Width*2+Width);
	}
	else
		I2C(_SD,R16,R7,No*Width);

	if (p->OnlyDiff2 && (!(p->SwapXY) || No==3))
		InstPost(Skip);
}

void Universal_RGB_UV( blit_soft* p )
{
	return;

	p->Double = p->SrcDoubleX && p->SrcDoubleY;
	p->Scale = !p->Double && (p->RScaleX!=16 || p->RScaleY!=16);
	p->OnlyDiff2 = p->OnlyDiff && !p->Scale && !p->Double;
	p->RScaleX2 = p->RScaleX << p->Double;
	p->RScaleY2 = p->RScaleY << p->Double;

	CodeBegin(10,sizeof(stack),M64);




	CodeEnd(10,sizeof(stack),M64);
}

#endif