/*****************************************************************************
 *
 * This program is free software ; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 *
 * $Id$
 *
 ****************************************************************************/

#include "stdafx.h"
#include "../dyncode/dyncode.h"
#include "blit_soft.h"

#if defined( SH3 )

typedef struct stack
{
	int DiffMask;
	int SatNegMask; // 0x7FFF 
	int SatPosMask; // 0x0400 (0x100 << 3)
	int AddR,AddG,AddB;		//AddY,Const128 when low quality
	int MaskR,MaskAccR;
	int MaskG,MaskAccG;
	int MaskB,MaskAccB;
} stack;

typedef struct param
{
	void* This;   //R4
	char* Dst;    //R5
	char* Src;    //R6
	int DstPitch; //R7
	int SrcPitch;
	int Width; 
	int Height;
	int Src2SrcLast;
} param;

// NO SCALE
// mutual exclusive flags: double vs diff
// other options: dither, swapxy

// R0 tmp
// R1,R2,R3		acc (if dither)
// R1,R2,R3		tmp (if !dither)
// R4 y
// R5 u
// R6 v
// R7 dst
// R8 y2 (if swapxy)
// R9 dstpitch
// R10 endofline
// R11 tmp
// R12 tmp
// R13 tmp
// R14

//input:
// src offset
// acc array
// dst
// dst2 (if double)
// y,u,v
// y2,y3,y4,u2,u3,u4,v2,v3,v4 (if swapxy)
//
//dst write:
// 2x1 pixel (if 32bit)
// 4x2 pixel (if 32bit & double)

void Universal_RGB_UV_Pixel( int No )
{
/*
	dyninst* Skip = NULL;

	if (OnlyDiff2)
		Skip = Label(false);

	int ShiftR,ShiftG,ShiftB;
	int Ofs[4];
	reg AccR,AccG,AccB,Last;
	reg Y[4];
	reg U[4];
	reg V[4];

	ShiftR = RPos + RSize - 8;
	ShiftG = GPos + GSize - 8;
	ShiftB = BPos + BSize - 8;

	if (p->FX.Flags & BLITFX_QUALITYCOLOR)
	{
		ShiftR -= 3;
		ShiftG -= 3;
		ShiftB -= 3;
	}

	AccR = R1;
	AccG = R2;
	AccB = R3;
	Last = R20;

	if (SwapXY)
	{
		Ofs[0] = No;
		Ofs[1] = No;
		Ofs[2] = No;
		Ofs[3] = No;

		Y[0] = R4; Y[1] = R9;  Y[2] = R10; Y[3] = R11;
		U[0] = R5; U[1] = R14; U[2] = R12; U[3] = R24;
		V[0] = R6; V[1] = R15; V[2] = R13; V[3] = R25;

		if (OnlyDiff2 && No==0) // only available when RScaleX2 == 16 && RScaleY2 == 16
		{
			// 2x2 comapre (32bit)
			// 4x4 compare (64bit)

			I(ADDU,R21,Y[0],R8);
			I(_LW,R16,Y[0],0);
			I(_LW,R20,R21,0);
			I(XOR,R16,R16,R20);

			MB(); I(ADDU,R21,Y[1],R8);
			I(_LW,R17,Y[1],0);
			I(_LW,R20,R21,0);
			I(XOR,R17,R17,R20);
			I(OR,R16,R16,R17);

			MB(); I(ADDU,R21,Y[2],R8);
			I(_LW,R17,Y[2],0);
			I(_LW,R20,R21,0);
			I(XOR,R17,R17,R20);
			I(OR,R16,R16,R17);

			MB(); I(ADDU,R21,Y[3],R8);
			I(_LW,R17,Y[2],0);
			I(_LW,R20,R21,0);
			I(XOR,R17,R17,R20);
			I(OR,R16,R16,R17);

			MB(); I(ADDU,R21,U[0],R8);
			I(_LHU,R17,U[0],0);
			I(_LHU,R20,R21,0);
			I(XOR,R17,R17,R20);
			I(OR,R16,R16,R17);

			MB(); I(ADDU,R21,U[2],R8);
			I(_LHU,R17,U[2],0);
			I(_LHU,R20,R21,0);
			I(XOR,R17,R17,R20);
			I(OR,R16,R16,R17);

			MB(); I(ADDU,R21,V[0],R8);
			I(_LHU,R17,V[0],0);
			I(_LHU,R20,R21,0);
			I(XOR,R17,R17,R20);
			I(OR,R16,R16,R17);

			MB(); I(ADDU,R21,V[2],R8);
			I(_LHU,R17,V[2],0);
			I(_LHU,R20,R21,0);
			I(XOR,R17,R17,R20);
			I(OR,R16,R16,R17);

			MB(); I(LD,R21,SP,OFS(stack,DiffMask));
			I(AND,R16,R16,R12);
			I(BEQ,R16,ZERO,Skip);
			// delay slot doesn't matter
		}
	}
	else
	{
		switch (No)
		{
		case 1:
			if (FX.Flags & BLITFX_DITHER)
			{
				AccR = R9;
				AccG = R10;
				AccB = R11;
			}
			Last = R21;
			break;
		case 2:
			if (FX.Flags & BLITFX_DITHER)
			{
				AccR = R12;
				AccG = R13;
				AccB = R14;
			}
			Last = R22;
			break;
		case 3:
			if (FX.Flags & BLITFX_DITHER)
			{
				AccR = R15;
				AccG = R24;
				AccB = R25;
			}
			Last = R23;
			break;
		}

		Ofs[0] = ((No*4+0)*RScaleX2) >> 4;
		Ofs[1] = ((No*4+1)*RScaleX2) >> 4;
		Ofs[2] = ((No*4+2)*RScaleX2) >> 4;
		Ofs[3] = ((No*4+3)*RScaleX2) >> 4;

		Y[0] = Y[1] = Y[2] = Y[3] = R4;
		U[0] = U[1] = U[2] = U[3] = R5;
		V[0] = V[1] = V[2] = V[3] = R6;

		if (OnlyDiff2) // only available when RScaleX2 == 16 && RScaleY2 == 16
		{
			// 2x1 compare (32bit)
			// 4x1 compare (64bit)

			I(ADDU,R21,Y[0],R8);
			I(_LW,R16,Y[0],Ofs[0]);
			I(_LW,R20,R21,Ofs[0]);
			I(XOR,R16,R16,R20);

			MB(); I(ADDU,R21,U[0],R8);
			I(_LHU,R17,U[0],Ofs[0] >> 1);
			I(_LHU,R20,R21,Ofs[0] >> 1);
			I(XOR,R17,R17,R21);
			I(OR,R16,R16,R17);

			MB(); I(ADDU,R21,V[0],R8);
			I(_LHU,R17,V[0],Ofs[0] >> 1);
			I(_LHU,R20,R21,Ofs[0] >> 1);
			I(XOR,R17,R17,R20);
			I(OR,R16,R16,R17);

			MB(); I(LD,R21,SP,OFS(stack,DiffMask));
			I(AND,R16,R16,R21);
			I(BEQ,R16,ZERO,Skip);
			// delay slot doesn't matter
		}
	}

	if (M64)
	{
		// load 4 pixel 
		I(LBU,R16,Y[0],Ofs[0]);
		I(LBU,R17,Y[1],Ofs[1]);
		I(LBU,R18,Y[2],Ofs[2]);
		I(DSLL,R17,R17,16);
		I(OR,R16,R16,R17);
		I(LBU,R17,Y[3],Ofs[3]);
		I(DSLL,R18,R18,32);
		I(OR,R16,R16,R18);
		I(DSLL,R17,R17,48);
		I(OR,R16,R16,R17);

		I(LBU,R19,V[2],Ofs[2] >> 1);
		I(LBU,R18,V[0],Ofs[0] >> 1);
		I(DSLL,R19,R19,32);
		I(OR,R18,R18,R19);

		if (RScaleX2 != 16) // V not same for Y01 and Y23
		{
			I(LBU,R19,V[1],Ofs[1] >> 1);
			I(LBU,R17,V[3],Ofs[3] >> 1);
			I(DSLL,R19,R19,16);
			I(DSLL,R17,R17,48);
			I(OR,R18,R18,R19);
			I(OR,R18,R18,R17);
		}
		else
		{
			I(DSLL,R19,R18,32);
			I(OR,R18,R18,R19);
		}

		I(LBU,R19,U[2],Ofs[2] >> 1);
		I(LBU,R17,U[0],Ofs[0] >> 1);
		I(DSLL,R19,R19,32);
		I(OR,R17,R17,R19);

		if (RScaleX2 != 16) // U not same for Y01 and Y23
		{
			I(LBU,R19,U[1],Ofs[1] >> 1);
			I(DSLL,R19,R19,16);
			I(OR,R17,R17,R19);
			I(LBU,R19,U[3],Ofs[3] >> 1);
			I(DSLL,R19,R19,48);
			I(OR,R17,R17,R19);
		}
		else
		{
			I(DSLL,R19,R17,32);
			I(OR,R17,R17,R19);
		}
	}
	else
	{
		// load 2 pixel 
		I(LBU,R17,Y[1],Ofs[1]);
		I(LBU,R16,Y[0],Ofs[0]);
		I(DSLL,R17,R17,16);
		I(OR,R16,R16,R17);

		MB(); I(LBU,R18,V[0],Ofs[0] >> 1);

		if (RScaleX2 != 16) // V not same for Y01
		{
			MB(); I(LBU,R19,V[1],Ofs[1] >> 1);
			I(DSLL,R19,R19,16);
			I(OR,R18,R18,R19);
		}
		else
		{
			I(DSLL,R19,R18,16);
			I(OR,R18,R18,R19);
		}

		MB(); I(LBU,R17,U[0],Ofs[0] >> 1);

		if (RScaleX2 != 16) // U not same for Y01 
		{
			MB(); I(LBU,R19,U[1],Ofs[1] >> 1);
			I(DSLL,R19,R19,16);
			I(OR,R17,R17,R19);
		}
		else
		{
			I(DSLL,R19,R17,16);
			I(OR,R17,R17,R19);
		}
	}

	// R16,R17,R18 (Y,U,V)

	if (!(p->FX.Flags & BLITFX_QUALITYCOLOR))
	{
		// we doesn't care about carry to neighbour pixel

		// Tmp = Y + AddR (-144+Brightness)
		// R = Tmp + V 
		// G = Y + AddG (-16+Brightness)
		// B = Tmp + U

		if (FX.Flags & BLITFX_DITHER)
		{
			I(_LD,R19,SP,OFS(stack,AddR));
			I(_ADDU,R18,R18,AccR);
			I(_ADDU,R17,R17,AccB);
			I(_ADDU,R19,R19,R16);
			I(_ADDU,R17,R17,R19);
			I(_ADDU,R18,R18,R19);
			I(_LD,R19,SP,OFS(stack,AddG));
			I(_ADDU,R16,R16,AccG);
			I(_ADDU,R16,R16,R19);
		}
		else
		{
			// R1,R2,R3 is free
			I(_LD,R19,SP,OFS(stack,AddG));
			I(_LD,R1,SP,OFS(stack,AddR));
			I(_ADDU,R19,R19,R16);
			I(_ADDU,R17,R17,R19);
			I(_ADDU,R18,R18,R19);
			I(_ADDU,R16,R16,R1);
		}

		// negative saturate
		I(NOR,R19,R16,R16);
		I(_SRL,R19,R19,8);
		I(AND,R16,R16,R19);

		I(NOR,R19,R17,R17);
		I(_SRL,R19,R19,8);
		I(AND,R17,R17,R19);

		I(NOR,R19,R18,R18);
		I(_SRL,R19,R19,8);
		I(AND,R18,R18,R19);
	}
	else
	{
		// R = 9*Y        + 13*V + AddR 
		// G = 9*Y -  3*U -  7*V + AddG 
		// B = 9*Y + 16*U        + AddB 

		if (!(FX.Flags & BLITFX_DITHER))
		{
			I(_LD,AccR,SP,OFS(stack,AddR));
			I(_LD,AccG,SP,OFS(stack,AddG));
			I(_LD,AccB,SP,OFS(stack,AddB));
		}

		I(_SLL,R19,R16,3);
		I(_ADDU,R19,R19,R16);	// R19 = 9*Y

		I(_ADDU,R16,R19,AccG);
		I(_SLL,AccG,R17,1);
		I(_SUBU,R16,R16,AccG);		
		I(_SUBU,R16,R16,R17);	// 9*Y - 3*U + AddG

		I(_SLL,R17,R17,4);		
		I(_ADDU,R17,R17,R19);	
		I(_ADDU,R17,R17,AccB);	// 9*Y + 16*U + AddB

		I(_SLL,AccG,R18,3);
		I(_SUBU,AccG,AccG,R18);	// 7*V
		I(_SUBU,R16,R16,AccG);	// 9*Y - 3*U - 7*U + AddG
		I(_ADDU,AccG,AccG,AccG);		
		I(_SUBU,AccG,AccG,R18);	// 13*V
		I(_ADDU,R18,AccG,R19);	
		I(_ADDU,R18,R18,AccR);	// 9*Y + 13*V + AddR

		if (FX.Flags & BLITFX_DITHER)
		{
			I(_LD,AccR,SP,OFS(stack,AddR));
			I(_LD,AccG,SP,OFS(stack,AddG));
			I(_LD,AccB,SP,OFS(stack,AddB));
			I(_ADDU,R18,R18,AccR);
			I(_ADDU,R16,R16,AccG);
			I(_ADDU,R17,R17,AccB);
		}

		// negative sat
		MB(); I(_LD,AccB,SP,OFS(stack,SatNegMask));

		I(NOR,AccR,R16,AccB);
		I(_SRL,AccG,AccR,15-3);
		I(_SUBU,AccR,AccR,AccG);
		I(AND,R16,R16,AccR);

		I(NOR,AccR,R17,AccB);
		I(_SRL,AccG,AccR,15-3);
		I(_SUBU,AccR,AccR,AccG);
		I(AND,R17,R17,AccR);

		I(NOR,AccR,R18,AccB);
		I(_SRL,AccG,AccR,15-3);
		I(_SUBU,AccR,AccR,AccG);
		I(AND,R18,R18,AccR);

		// positive sat
		MB(); I(_LD,AccB,SP,OFS(stack,SatPosMask));

		I(AND,AccR,R16,AccB);
		I(_SRL,AccG,AccR,8);
		I(_SUBU,AccR,AccR,AccG);
		I(OR,R16,R16,AccR);

		I(AND,AccR,R17,AccB);
		I(_SRL,AccG,AccR,8);
		I(_SUBU,AccR,AccR,AccG);
		I(OR,R17,R17,AccR);

		I(AND,AccR,R18,AccB);
		I(_SRL,AccG,AccR,8);
		I(_SUBU,AccR,AccR,AccG);
		I(OR,R18,R18,AccR);
	}

	// R16,R17,R18 (G,B,R)

	if (FX.Flags & BLITFX_DITHER)
	{
		I(_LD,R19,SP,OFS(stack,MaskR));
		I(_LD,AccR,SP,OFS(stack,MaskAccR));
		I(AND,R19,R18,R19);
		I(AND,AccR,R18,AccR);

		I(_LD,R18,SP,OFS(stack,MaskG));
		I(_LD,AccG,SP,OFS(stack,MaskAccG));
		I(AND,R18,R16,R18);
		I(AND,AccG,R16,AccG);

		I(_LD,R16,SP,OFS(stack,MaskB));
		I(_LD,AccB,SP,OFS(stack,MaskAccB));
		I(AND,R16,R17,R16);
		I(AND,AccB,R17,AccB);

		// R18,R16,R19 (G,B,R)

		if (ShiftR) I(_SLL,R19,R19,ShiftR);
		if (ShiftG) I(_SLL,R18,R18,ShiftG);
		if (ShiftB) I(_SLL,R16,R16,ShiftB);

		I(OR,R16,R16,R19);
		I(OR,R16,R16,R18);
	}
	else
	{
		if (!(SwapXY) || !Scale)
		{
			//R15,R24,R25 MaskR,MaskG,MaskB
			I(AND,R16,R16,R15);
			I(AND,R17,R17,R24);
			I(AND,R18,R18,R25);
		}
		else
		{
			I(_LD,AccR,SP,OFS(stack,MaskR));
			I(_LD,AccG,SP,OFS(stack,MaskG));
			I(_LD,AccB,SP,OFS(stack,MaskB));
			I(AND,R18,R18,AccR);
			I(AND,R16,R16,AccG);
			I(AND,R17,R17,AccB);
		}

		// R16,R17,R18 (G,B,R)

		if (ShiftR) I(_SLL,R18,R18,ShiftR);
		if (ShiftG) I(_SLL,R16,R16,ShiftG);
		if (ShiftB) I(_SLL,R17,R17,ShiftB);

		I(OR,R16,R16,R18);
		I(OR,R16,R16,R17);
	}

	//R16 rgb 16bit

	//!!! paletta lookup and 8bit rgb support

	if (Double)
	{
		//!! average with last


		I(_SD,R16,R7,No*Width*2);
		I(_SD,R16,R7,No*Width*2+Width);
		I(_SD,R16,R8,No*Width*2);
		I(_SD,R16,R8,No*Width*2+Width);
	}
	else
		I(_SD,R16,R7,No*Width);

	if (OnlyDiff2 && (!(SwapXY) || No==3))
		Post(Skip);
*/
}

void Universal_RGB_UV( blit_soft* p )
{
	return;

	p->Double = p->SrcDoubleX && p->SrcDoubleY;
	p->Scale = !p->Double && (p->RScaleX!=16 || p->RScaleY!=16);
	p->OnlyDiff2 = p->OnlyDiff && !p->Scale && !p->Double;
	p->RScaleX2 = p->RScaleX << p->Double;
	p->RScaleY2 = p->RScaleY << p->Double;

	CodeBegin(6,sizeof(stack));




	CodeEnd(6,sizeof(stack));
}

#endif