/*****************************************************************************
 *
 * This program is free software ; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 *
 * $Id$
 *
 ****************************************************************************/

#ifndef __BLIT_SOFT_H
#define __BLIT_SOFT_H

struct blit_soft;
typedef void (*blitsoftentry)(struct blit_soft* This, 
							  uint8_t** DstPtr,uint8_t** SrcPtr,
							  int DstPitch,int SrcPitch,
							  int Width,int Height,int Src2SrcLast);

typedef struct blit_soft
{
	blitsoftentry Entry;

	dyncode Code;
	pixel Dst;
	pixel Src;
	blitfx FX;

	int SrcBPP;
	int DstBPP;
	int DstBPP2;
	uint8_t PalLookUp[512];
	bool_t OnlyDiff;
	int SrcType;
	int DstType;
	bool_t SwapXY;
	bool_t Slices;
	int DirX; // destination space
	int RSize;
	int GSize;
	int BSize;
	int RPos;
	int GPos;
	int BPos;
	int RScaleX; // reverse scale, unit=16
	int RScaleY; // reverse scale, unit=16
	const rgb* SrcPalette;
	const rgb* DstPalette;

	int SrcDoubleX;
	int SrcDoubleY;
	int DstDoubleX;
	int DstDoubleY;

	bool_t QAdd;
	int DstStepX;

	bool_t Scale;
	int Double;
	int RScaleX2;  // adjusted by double
	int RScaleY2;  // adjusted by double
	bool_t OnlyDiff2;// adjusted OnlyDiff (diff mode not always available)

	dyninst* PalPtr;
	dyninst* DiffMask;
	dyninst* InvertMask;
	dyninst* Skip;
	dyninst* YMul;
	dyninst* GAdd;
	dyninst* RAdd;
	dyninst* BAdd;
	dyninst* GVMul;
	dyninst* RVMul;
	dyninst* GUMul;
	dyninst* BUMul;

	int _YMul;
	int _GAdd;
	int _RAdd;
	int _BAdd;
	int _GVMul;
	int _RVMul;
	int _GUMul;
	int _BUMul;
	bool_t IsYNextRow;
	bool_t IsUVNextRow;

	bool_t Upper;
	int Pos;
} blit_soft;

bool_t BlitCompile(blit_soft* This,
			  	   const video* Dst,const video* Src,
				   const blitfx* FX,bool_t OnlyDiff);

void BlitUniversal(blit_soft*,uint8_t** Dst,uint8_t** Src,int DstPitch,int SrcPitch,
                   int Width,int Height,int Src2SrcLast);

void BuildPalLookUp(blit_soft* This);
void Stretch_RGB_UV(blit_soft* This);
void Double_RGB_UV(blit_soft* This);
void Fix_RGB_UV(blit_soft* This);
void Fix_Gray_UV(blit_soft* This);
void Half_RGB_UV(blit_soft* This);
void Universal_RGB_UV(blit_soft* This);
void Fix_YUV_YUV(blit_soft* This);

void CalcColor( blit_soft* This );

#endif
