/*****************************************************************************
 *
 * This program is free software ; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 *
 * $Id$
 *
 ****************************************************************************/

#include "stdafx.h"

#define SYNCCHUNKS		8

int Format_Enum(format_base* p,int No,datadef* Param)
{
	if (NodeEnumType(&No,Param,NodeParams,FormatParams,FormatMediaParams,NULL) == ERR_NONE)
		return ERR_NONE;

	if (No < p->StreamCount)
	{
		memset(Param,0,sizeof(datadef));
		Param->Class = FORMAT_CLASS;
		Param->Size = sizeof(pin);
		Param->No = FORMAT_STREAM + No;
		Param->Type = TYPE_PIN;
		Param->Format1 = p->Streams[No]->FormatType;
		Param->Flags = DF_OUTPUT;
		return ERR_NONE;
	}
	No -= p->StreamCount;

	if (No < p->StreamCount)
	{
		memset(Param,0,sizeof(datadef));
		Param->Class = FORMAT_CLASS;
		Param->Size = sizeof(pin);
		Param->No = FORMAT_COMMENT + No;
		Param->Type = TYPE_PIN;
		Param->Format1 = PACKET_COMMENT;
		Param->Flags = DF_OUTPUT;
		return ERR_NONE;
	}

	return ERR_INVALID_PARAM;
}

INLINE void Format_BufferInsert(format_base* p,format_buffer* Buffer)
{
	Buffer->Next = NULL;
	if (p->BufferLast)
		p->BufferLast->Next = Buffer;
	else
		p->BufferFirst = Buffer;
	p->BufferLast = Buffer;
}

format_buffer* Format_BufferAlloc(format_base*);

int Format_InputSeek(format_base* p,int FilePos)
{
	if (p->Input->Seek(p->Input,FilePos,SEEK_SET) < 0)
		return ERR_NOT_SUPPORTED;

	p->BufferFilePos = FilePos;
	Format_ReleaseBuffers(p);
	return ERR_NONE;
}

int Format_SeekByPacketTime(format_base* p,tick_t Time,int FilePos)
{
	format_buffer* Buffer[SYNCCHUNKS];
	int No;

	if (FilePos < 0)
	{
		Time -= SHOWAHEAD; // we should go a little earlier
		if (Time > 0)
		{
			if (p->FileSize<0 || p->Duration<0)
				return ERR_NOT_SUPPORTED;

			FilePos = Scale(p->FileSize,Time,p->Duration);
		}
		else
			FilePos = 0;
	}

	memset(Buffer,0,sizeof(Buffer));

	for (No=0;No<3;++No) // maximum iterations
	{
		if (Format_InputSeek(p,FilePos) != ERR_NONE)
			return ERR_NOT_SUPPORTED;

		if (Time>0 && p->ReadPacketTime)
		{
			// check if really seeked to the right time (and adjust if needed)

			void* State = NULL;
			int No;

			for (No=0;No<SYNCCHUNKS;++No)
			{
				tick_t Min = -1;

				if (!Buffer[No])
					Buffer[No] = Format_BufferAlloc(p);
				
				if (!Buffer[No])
					break;

				Buffer[No]->Length = p->Input->Read(p->Input,Buffer[No]->Buffer,CHUNKSIZE);
				if (Buffer[No]->Length<=0)
					break;

				State = p->ReadPacketTime(p,Buffer[No],&Min,NULL,State);

				if (Min >= 0)
				{
					// found a packet with reftime

					if (Time>=Min && (Time-Min)*100 <= p->Duration)
						FilePos = -1; // accept 1% earlier 
					else
						FilePos += Scale(p->FileSize,Time-Min,p->Duration)*15/16;

					if (FilePos < 0 || FilePos >= p->FileSize)
					{
						// insert buffers to chain and stop

						int i;
						for (i=0;i<=No;++i)
						{
							Format_BufferInsert(p,Buffer[i]);
							Buffer[i] = NULL;
						}
						No = SYNCCHUNKS;
					}
					break;
				}
			}

			p->ReadPacketTime(p,NULL,NULL,NULL,State); // free state

			if (No<SYNCCHUNKS)
				continue; // try again
		}
		break;
	}

	for (No=0;No<SYNCCHUNKS;++No)
		if (Buffer[No])
			Format_BufferRelease(p,Buffer[No]);

	return ERR_NONE;
}

format_packet* Format_PacketAlloc( format_base* );
void Format_PacketRelease( format_base*, format_packet*);

void* Format_ReadPacketTime(format_base* p,format_buffer* Buffer,tick_t* Min,tick_t* Max,void* s)
{
	// emulate ReadPacket enviroment (FilePos will be not valid)

	typedef struct bufferstate
	{
		bool_t NeedMoreInput;
		int BufferPos;
		format_buffer* Buffer;

	} bufferstate;

	int Result;
	format_buffer** Ptr;
	bufferstate* State = (bufferstate*)s;

	if (!Buffer)
	{
		if (State)
		{
			p->NeedMoreInput = State->NeedMoreInput;
			Free(State);
		}
		return NULL;
	}

	if (!State)
	{
		State = (bufferstate*)Alloc(sizeof(bufferstate));
		if (!State)
			return NULL;
		
		State->NeedMoreInput = p->NeedMoreInput;
		State->BufferPos = 0;
		State->Buffer = NULL;
	}

	// add new buffer to State->Buffer chain

	p->NeedMoreInput = 0;
	Ptr = &State->Buffer;
	while (*Ptr) Ptr = &(*Ptr)->Next;
	*Ptr = Buffer;

	while (!p->NeedMoreInput)
	{
		tick_t RefTime;
		format_packet* Packet = Format_PacketAlloc(p);
		if (!Packet)
			break;

		// reading is final only after success
		p->ReadPos = State->BufferPos;
		p->ReadBuffer = State->Buffer;

		Result = p->ReadPacket(p,Packet);

		if (p->NeedMoreInput)
			Result = ERR_NEED_MORE_DATA;

		if (Result != ERR_NONE && Result != ERR_DATA_NOT_FOUND)
		{
			Format_PacketRelease(p,Packet);
			break;
		}

		while (State->Buffer != p->ReadBuffer)
			State->Buffer = State->Buffer->Next;
		State->BufferPos = p->ReadPos;

		RefTime = Packet->RefTime;
		Format_PacketRelease(p,Packet);

		if (RefTime >= 0)
		{
			if (Min && (*Min<0 || RefTime < *Min))
				*Min = RefTime;

			if (!Max) // we can stop if maximum is not need

				break;

			if (*Max<0 || RefTime > *Max)
				*Max = RefTime;
		}
	}
	return State;
}

format_buffer* Format_BufferAlloc(format_base* p)
{
	format_buffer* Buffer = p->FreeBuffers;
	if (!Buffer)
	{
		Buffer = (format_buffer*) Alloc(sizeof(format_buffer));
		if (!Buffer)
			return NULL;
	}
	else
	{
		assert(Buffer->RefCount==0);
		p->FreeBuffers = Buffer->Next;
	}

	Buffer->Id = p->BufferId++;
	Buffer->RefCount = 1;
	Buffer->Next = NULL;
	++p->BufferUsed;

	return Buffer;
}

void CalcDuration(format_base* p)
{
	if (p->Duration<0 && p->FileSize>=0)
	{
		if (p->ReadPacketTime)
		{
			// try to read the last packets and figure out Duration
			int SavePos;

			LockEnter(p->InputLock);
			SavePos = p->Input->Seek(p->Input,0,SEEK_CUR);
			if (SavePos >= 0)
			{
				format_buffer* Buffer[SYNCCHUNKS];
				int No,i;
				tick_t Max=-1;
				memset(Buffer,0,sizeof(Buffer));

				for (No=0;No<SYNCCHUNKS;++No)
				{
					void* State = NULL;
					int Pos = p->FileSize - (No+1)*CHUNKSIZE;
					if (Pos<0)
						break;

					Buffer[No] = Format_BufferAlloc(p);
					if (!Buffer[No])
						break;

					if (p->Input->Seek(p->Input,Pos,SEEK_SET)<0)
						break;

					Buffer[No]->Length = p->Input->Read(p->Input,Buffer[No]->Buffer,CHUNKSIZE);
					if (Buffer[No]->Length < 0)
						break;

					for (i=No;i>=0;--i)
						State = p->ReadPacketTime(p,Buffer[i],NULL,&Max,State);

					p->ReadPacketTime(p,NULL,NULL,NULL,State); // free state

					if (Max >= 0)
					{
						p->Duration = Max;
						break;
					}
				}

				for (No=0;No<SYNCCHUNKS;++No)
					if (Buffer[No])
						Format_BufferRelease(p,Buffer[No]);

				p->Input->Seek(p->Input,SavePos,SEEK_SET);
			}
			LockLeave(p->InputLock);
		}

		if (p->Duration<0 && p->StreamCount)
		{
			// try to sum stream byterates

			int No;
			int SumByteRate = 0;
			format_stream* Stream;

			for (No=0;No<p->StreamCount;++No)
			{
				int ByteRate = 0;
				Stream = p->Streams[No];

				switch (Stream->FormatType)
				{
				case PACKET_VIDEO:
					ByteRate = ((video*)Stream->Format)->ByteRate;
					if (!ByteRate && Stream->Pin.Node)
					{
						// maybe codec has some information about byterate
						Stream->Pin.Node->Get(Stream->Pin.Node,Stream->Pin.No|PACKET_FORMAT,Stream->Format,Stream->FormatLength);
						ByteRate = ((video*)Stream->Format)->ByteRate;
					}
					break;

				case PACKET_AUDIO:
					ByteRate = ((audio*)Stream->Format)->ByteRate;
					if (!ByteRate && Stream->Pin.Node)
					{
						// maybe codec has some information about byterate
						Stream->Pin.Node->Get(Stream->Pin.Node,Stream->Pin.No|PACKET_FORMAT,Stream->Format,Stream->FormatLength);
						ByteRate = ((audio*)Stream->Format)->ByteRate;
					}
					break;
				}

				if (ByteRate>0)
					SumByteRate += ByteRate;
				else
					break;
			}

			if (No == p->StreamCount)
				p->Duration = Scale(p->FileSize,TICKSPERSEC,SumByteRate);
		}
	}
}

void Format_BufferRelease( format_base* p, format_buffer* Buffer )
{
	LockEnter(p->BufferLock);

	//DEBUG_MSG4("Buffer:%d release %d->%d (%d)",Buffer->Id,Buffer->RefCount,Buffer->RefCount-1,p->BufferUsed);

	assert(Buffer->RefCount>0 && Buffer->RefCount<100);

	if (--Buffer->RefCount <= 0)
	{
		--p->BufferUsed;
		Buffer->Next = p->FreeBuffers;
		p->FreeBuffers = Buffer;
	}

	LockLeave(p->BufferLock);
}

void Format_RefRelease( format_base* p, format_ref* Ref )
{
	if (Ref->Buffer)
		Format_BufferRelease(p,Ref->Buffer);
	Ref->Next = p->FreeRefs;
	p->FreeRefs = Ref;
}

void Format_PacketRelease( format_base* p, format_packet* Packet )
{
	format_ref* Ref;
	while (Packet->Data)
	{
		Ref = Packet->Data;
		Packet->Data = Ref->Next;
		Format_RefRelease(p,Ref);
	}

	Packet->Next = p->FreePackets;
	p->FreePackets = Packet;
}

void Format_ReleaseStream(format_base* p,int No)
{
	format_stream* Stream = p->Streams[No];
	format_packet* Packet;

	if (p->DropStream == Stream)
		p->DropStream = NULL;

	Stream->Packet.DropLevel = 0;
	Stream->Pending = 0;
	Stream->LastTime = -1;

	while (Stream->PacketFirst)
	{
		Packet = Stream->PacketFirst;
		Stream->PacketFirst = Packet->Next;
		if (Stream->PacketLast == Packet)
			Stream->PacketLast = NULL;

		Format_PacketRelease(p,Packet);
	}

	if (p->ReleaseStream)
		p->ReleaseStream(p,Stream);
}

void Format_ReleaseBuffers(format_base* p)
{
	format_buffer* Buffer;
	int No;

	// release stream packets

	for (No=0;No<p->StreamCount;++No)
		Format_ReleaseStream(p,No);

	// release read buffers

	p->NeedMoreInput = 0;
	p->WaitForInput = 0;
	p->NoMoreInput = 0;
	p->DropStream = NULL;
	p->BufferPos = 0;
	p->ReadBuffer = NULL;

	while (p->BufferFirst)
	{
		Buffer = p->BufferFirst;
		p->BufferFirst = Buffer->Next;
		if (p->BufferLast == Buffer)
			p->BufferLast = NULL;

		Format_BufferRelease(p,Buffer);
	}
}

int Format_Reset(format_base* p)
{
	int Result = ERR_NONE;
	format_stream* Stream;
	format_packet* Packet;
	format_buffer* Buffer;
	format_ref*	Ref;
	int No;

	if (p->Done)
		p->Done(p);

	Format_ReleaseBuffers(p);

	while (p->FreePackets)
	{
		Packet = p->FreePackets;
		p->FreePackets = Packet->Next;
		Free(Packet);
	}

	while (p->FreeBuffers)
	{
		Buffer = p->FreeBuffers;
		p->FreeBuffers = Buffer->Next;
		Free(Buffer);
	}

	while (p->FreeRefs)
	{
		Ref = p->FreeRefs;
		p->FreeRefs = Ref->Next;
		Free(Ref);
	}

	for (No=0;No<p->StreamCount;++No)
	{
		if (p->FreeStream)
			p->FreeStream(p,p->Streams[No]);
		Stream = p->Streams[No];
		Free(Stream->Format);
		MergeDone(&Stream->Merge);
		Free(Stream);
	}

	p->StreamCount = 0;

	LockDelete(p->BufferLock);
	LockDelete(p->InputLock);
	p->BufferLock = NULL;
	p->InputLock = NULL;

	if (p->Input)
	{
		if (!NodeIsClass((node*)p->Input,STREAM_CLASS))
			return ERR_INVALID_PARAM;

		if (p->Input->Get(p->Input,STREAM_LENGTH,&p->FileSize,sizeof(p->FileSize)) != ERR_NONE)
			p->FileSize = -1;

		p->InputLock = LockCreate();
		p->BufferLock = LockCreate();
		p->Duration = -1;
		p->NeedDuration = 1;
		p->NeedMoreInput = 0;
		p->WaitForInput = 0;
		p->NoMoreInput = 0;
		p->DropStream = NULL;
		p->SyncMode = 0;

		p->BufferId = 0;
		p->BufferFilePos = 0;
		p->BufferUsed = 0;
		p->BufferPos = 0;
		p->BufferFirst = NULL;
		p->BufferLast = NULL;
		p->ReadBuffer = NULL;
		p->PreLoad = 0;

		if (p->Init)
		{
			Result = p->Init(p);
			if (p->StreamCount)
				p->PreLoad = PRELOAD; // we don't need preloading
		}
	}
	return Result;
}

int Format_Get(format_base* p,int No,void* Data,int Size)
{
	int Result = ERR_INVALID_PARAM;
	if (No >= FORMAT_COMMENT && No < FORMAT_COMMENT+p->StreamCount)  
	{
		GETVALUE(p->Streams[No-FORMAT_COMMENT]->Comment,pin)
	}
	else
	if (No >= FORMAT_STREAM && No < FORMAT_STREAM+p->StreamCount)
	{
		GETVALUE(p->Streams[No-FORMAT_STREAM]->Pin,pin)
	}
	else
	if (No >= (FORMAT_STREAM|PACKET_FORMAT) && No < (FORMAT_STREAM|PACKET_FORMAT)+p->StreamCount)
	{
		No -= FORMAT_STREAM|PACKET_FORMAT;
		if (Size == p->Streams[No]->FormatLength)
		{
			format_stream* s = p->Streams[No];
			memcpy(Data,s->Format,s->FormatLength);

			if (s->Pin.Node)
			{
				video Video;
				audio Audio;

				//try to fill unkown fields from codec
				switch (p->Streams[No]->FormatType)
				{
				case PACKET_VIDEO:
					if (s->Pin.Node->Get(s->Pin.Node,s->Pin.No|PACKET_FORMAT,&Video,sizeof(Video))==ERR_NONE)
						CombineVideo((video*)Data,&Video);
					break;
				case PACKET_AUDIO:
					if (s->Pin.Node->Get(s->Pin.Node,s->Pin.No|PACKET_FORMAT,&Audio,sizeof(Audio))==ERR_NONE)
						CombineAudio((audio*)Data,&Audio);
					break;
				}
			}
			Result = ERR_NONE;
		}
	}
	else
	switch (No)
	{
	case NODE_ID: GETVALUE(p->Id,int); break;
	case FORMAT_BUFFERSIZE: GETVALUE(p->BufferSize*CHUNKSIZE,int); break;
	case FORMAT_INPUT: GETVALUE(p->Input,stream*); break;
	case FORMAT_EXTS: GETSTRING(p->Exts); break;
	case FORMAT_DURATION: GETVALUECOND(p->Duration,tick_t,p->Duration>=0); break;
	case FORMAT_FILEPOS: GETVALUECOND(p->BufferFilePos,int,p->BufferFilePos>=0); break;
	case FORMAT_DROPTOL: GETVALUE(p->DropTolerance,tick_t); break;
	case FORMAT_SKIPTOL: GETVALUE(p->SkipTolerance,tick_t); break;
	case FORMAT_AVOFFSET: GETVALUE(p->AVOffset,tick_t); break;
	}
	return Result;
}

int Format_Set(format_base* p,int No,const void* Data,int Size)
{
	int Result = ERR_INVALID_PARAM;

	if (No >= FORMAT_COMMENT && No < FORMAT_COMMENT+p->StreamCount)
	{
		No -= FORMAT_COMMENT;
		SETVALUE(p->Streams[No]->Comment,pin,ERR_NONE);
	}
	else
	if (No >= FORMAT_STREAM && No < FORMAT_STREAM+p->StreamCount)
	{
		No -= FORMAT_STREAM;
		if (Size == sizeof(pin))
		{
			Result = NodeSetPin(&p->Streams[No]->Pin,(const pin*)Data,p->Streams[No]->Format,p->Streams[No]->FormatLength);
			if (!p->Streams[No]->Pin.Node)
				Format_ReleaseStream(p,No);
		}
	}
	else
	switch (No)
	{
	case FORMAT_INPUT: SETVALUE(p->Input,stream*,Format_Reset(p)); break;
	case FORMAT_DROPTOL: SETVALUE(p->DropTolerance,tick_t,ERR_NONE); break;
	case FORMAT_SKIPTOL: SETVALUE(p->SkipTolerance,tick_t,ERR_NONE); break;
	case FORMAT_AVOFFSET: SETVALUE(p->AVOffset,tick_t,ERR_NONE); break;
	case FORMAT_BUFFERSIZE: 
		if (Size == sizeof(int))
		{
 			p->BufferSize = *(int*)Data / CHUNKSIZE;
			if (p->BufferSize < PRELOAD/CHUNKSIZE)
				p->BufferSize = PRELOAD/CHUNKSIZE;
			Result = ERR_NONE;
			break;
		}
	}
	return Result;
}

format_packet* Format_PacketAlloc( format_base* p)
{
	format_packet* Packet = p->FreePackets;

	if (!Packet)
	{
		Packet = (format_packet*) Alloc(sizeof(format_packet));
		if (!Packet)
			return NULL;
	}
	else
		p->FreePackets = Packet->Next;

	Packet->Data = NULL;
	Packet->RefTime = -1;
	Packet->Stream = NULL;
	Packet->Next = NULL;
	return Packet;
}

int Format_Sync(format_base* p,tick_t Time,int FilePos)
{
	if (!p->Seek || !p->Input)
		return ERR_NOT_SUPPORTED;

	if (Time>=0 || FilePos>=0)
	{
		int Result = p->Seek(p,Time,FilePos);
		if (Result != ERR_NONE)
			return Result;
	}

	p->SyncMode = 1;
	p->SyncStream = NULL;
	return ERR_NONE;
}

int Format_Send(format_base* p,format_stream* Stream)
{
	int Result;
	tick_t CurrTime = p->ProcessTime;

	Stream->Packet.CurrTime = CurrTime;

	if (Stream->Packet.RefTime >= 0 && !Stream->DisableDrop)
	{
		if (p->DropStream && CurrTime >= 0 && Stream->Packet.RefTime > CurrTime + TICKSPERSEC)
		{
			if (p->DropStream != Stream)
				return ERR_BUFFER_FULL; // dropmode: postpone future packets (example buffered audio stream)
			else
				p->DropStream = NULL;
		}

		if (CurrTime >= 0 && Stream->Packet.RefTime < CurrTime - p->DropTolerance)
		{
			if (Stream->Packet.RefTime < CurrTime - p->SkipTolerance - p->DropTolerance)
				Stream->Packet.DropLevel = 2;
			else
			if (!Stream->Packet.DropLevel) // keep DropLevel=2 until fully recovered
				Stream->Packet.DropLevel = 1;

			p->DropStream = Stream; // try to save time in other tasks
		}
		else
			Stream->Packet.DropLevel = 0;
	}

	DEBUG_MSG4(T("Send stream:%d size:%d time:%d drop:%d"),Stream->No,Stream->Packet.Length,Stream->Packet.RefTime,Stream->Packet.DropLevel);

	// sending packet
	Result = Stream->Pin.Node->Set(Stream->Pin.Node,Stream->Pin.No,&Stream->Packet,sizeof(packet));

	// packet not processed?
	if (Result == ERR_BUFFER_FULL)
	{
		if (p->DropStream == Stream)
			p->DropStream = NULL;

		DEBUG_MSG1(T("Send stream:%d BufferFull"),Stream->No);
		return Result;
	}

	if (p->Sended)
		p->Sended(p,Stream);

	Stream->Pending = 0;

	// sync: found a valid frame?
	if (p->SyncMode && Result == ERR_NONE)
	{
		p->SyncMode = 0;
		p->SyncTime = Stream->Packet.RefTime;
		if (p->SyncTime < 0)
			p->SyncTime = Stream->LastTime;
		Result = ERR_SYNCED;
	}

	return Result;
}

void Format_Sended(format_base* p, format_stream* Stream)
{
	format_packet* Packet = Stream->PacketFirst;

	if (Stream->Pending == 2 && Packet->Data)
	{
		// only first ref is released

		format_ref* Ref = Packet->Data;
		Packet->Data = Ref->Next;

		Format_RefRelease(p,Ref);

		// packet still not finished?
		if (Packet->Data)
		{
			Packet->RefTime = -1;
			return;
		}
	}

	// full packet is released

	Stream->PacketFirst = Packet->Next;
	if (Stream->PacketLast == Packet)
		Stream->PacketLast = NULL;

	Format_PacketRelease(p,Packet);
}

int Format_ProcessStream(format_base* p,format_stream* Stream)
{
	format_packet* Packet;
	format_buffer* Buffer;
	format_ref* Ref;
	int Result;
	int No;
	int Burst = 1;

	if (Stream)
	{
		Burst = Stream->PacketBurst;

		if (Stream->Pending)
		{
			Result = Format_Send(p,Stream);

			if (Result == ERR_BUFFER_FULL || Result == ERR_SYNCED)
				return Result;
		}
	}

	// process a limited number of packets at once (other streams need cpu time too)

	for (No=0;No<Burst;++No)
	{
		while ((Stream && !Stream->PacketFirst) || p->PreLoad < PRELOAD)
		{
			// read packets until we find one for this stream

			if (p->WaitForInput)
				return ERR_NEED_MORE_DATA; // don't even try to read packet

			Packet = Format_PacketAlloc(p);
			if (!Packet)
				return ERR_OUT_OF_MEMORY;

			// reading is final only after success
			LockEnter(p->BufferLock);
			p->NeedMoreInput = 0;
			p->FilePos = p->BufferFilePos;
			p->ReadPos = p->BufferPos;
			p->ReadBuffer = p->BufferFirst;
			LockLeave(p->BufferLock);

			Result = p->ReadPacket(p,Packet);

			if (p->NeedMoreInput)
			{
				p->WaitForInput = 1;
				Result = ERR_NEED_MORE_DATA;
			}

			if (Result != ERR_NONE && Result != ERR_DATA_NOT_FOUND)
			{
				Format_PacketRelease(p,Packet);
				return Result; // something wrong
			}

			// finalize reading and release finished buffers
			LockEnter(p->BufferLock);
			while (p->BufferFirst != p->ReadBuffer)
			{
				Buffer = p->BufferFirst;
				p->PreLoad += Buffer->Length;
				p->BufferFirst = Buffer->Next;
				if (p->BufferLast == Buffer)
					p->BufferLast = NULL;
				Format_BufferRelease(p,Buffer);
			}
			//Packet->FilePos = p->BufferFilePos;
			p->BufferPos = p->ReadPos;
			p->BufferFilePos = p->FilePos;
			LockLeave(p->BufferLock);

			if (Packet->Stream && Packet->Stream->Pin.Node)
			{
				format_stream* s = Packet->Stream;

				if (Packet->RefTime>=0)
				{
					if (s->FormatType == PACKET_AUDIO)
						Packet->RefTime += p->AVOffset;

					if (s->LastTime<0)
						s->LastTime = Packet->RefTime; // found reftime
				}

				if (s->LastTime >= 0)
				{
					// add packet to stream
					if (!s->PacketFirst)
						s->PacketFirst = Packet;
					else
						s->PacketLast->Next = Packet;
					s->PacketLast = Packet;

					if (p->Bench && Stream && (No || s != Stream))
						return ERR_NONE;
				}
				else
					Format_PacketRelease(p,Packet);
			}
			else
			{
				Format_PacketRelease(p,Packet);

				if (Result == ERR_DATA_NOT_FOUND)
					return ERR_NONE;
			}
		}

		if (!Stream)
			return ERR_NONE;

		Packet = Stream->PacketFirst;

		// sync: all non SyncStream packets are stored (or dropped)
		if (p->SyncMode && Stream != p->SyncStream)
		{
			// catch up to syncstream (drop packets)
			while (Packet && Packet->RefTime < p->SyncStream->LastTime)
			{
				Stream->PacketFirst = Packet->Next;
				if (Stream->PacketLast == Packet)
					Stream->PacketLast = NULL;

				Format_PacketRelease(p,Packet);

				Packet = Stream->PacketFirst;
			}
			return ERR_NEED_MORE_DATA; // allow SyncStream to need more data
		}

		//DEBUG_MSG3(T("Packet stream:%d time:%d pos:%d"),Stream->No,Packet->RefTime,Packet->FilePos);

		// build output packet

		if (Packet->RefTime >= 0)
			Stream->LastTime = Packet->RefTime;

		Stream->Packet.RefTime = Packet->RefTime;
		Ref = Packet->Data;

		if (!Stream->Fragmented && Ref && Ref->Next)
		{
			// merge fragments

			MergeClear(&Stream->Merge);

			for (;Ref;Ref=Ref->Next)
				MergeAdd(&Stream->Merge,Ref->Buffer->Buffer + Ref->Begin, Ref->Length);

			Stream->Packet.Data[0] = Stream->Merge.Data;
			Stream->Packet.Length = Stream->Merge.Pos;
			Stream->Pending = 1;
		}
		else
		{
			// use only first reference (if there is one)

			if (Ref)
			{
				Stream->Packet.Data[0] = Ref->Buffer->Buffer + Ref->Begin;
				Stream->Packet.Length = Ref->Length;
			}
			else
			{
				Stream->Packet.Data[0] = NULL;
				Stream->Packet.Length = 0;
			}

			Stream->Pending = 2;
		}

		Result = Format_Send(p,Stream);

		if (Result == ERR_BUFFER_FULL || Result == ERR_SYNCED)
			break;

		if (Stream->Packet.DropLevel==2)
			Burst = Stream->PacketBurst*2; // try to catch up
	}

	// buffer full: there was possibly some processing so don't allow sleeping
	// need data: burst count exceeded, but it doesn't mean there is no data left in buffers
	if (Result == ERR_BUFFER_FULL || Result == ERR_NEED_MORE_DATA)
		Result = ERR_NONE;

	return Result;
}

format_stream* Format_DefSyncStream( format_base* p )
{
	format_stream* Found = NULL;
	format_stream* Stream;
	int No;

	// prefer video streams

	for (No=0;No<p->StreamCount;++No)
	{
		Stream = p->Streams[No];
		if (Stream->Pin.Node)
		{
			if (Stream->FormatType == PACKET_VIDEO)
				return Stream;

			if (!Found)
				Found = Stream;
		}
	}

	return Found;
}

int Format_Process( format_base* p,tick_t* Time)
{
	int No;
	int Result;
	bool_t NoStreams = 1;
	bool_t AllNeedData = 1;
	bool_t AllEndOfFile = 1;

	p->ProcessTime = *Time;
	p->Bench = p->ProcessTime < 0;

	if (p->PreLoad < PRELOAD)
	{
		// load format begining to find streams

		Result = p->Process(p,NULL);

		if (Result != ERR_NONE &&
			Result != ERR_END_OF_FILE &&
			(Result != ERR_NEED_MORE_DATA || !p->NoMoreInput))
			return Result;

		p->PreLoad = PRELOAD;	
	}

	if (p->SyncMode)
	{
		p->ProcessTime = -1;
		// we need to choose one stream (to find a non dropped frame)
		if (!p->SyncStream)
		{
			p->SyncStream = Format_DefSyncStream(p);
			if (!p->SyncStream)
			{
				// no sync needed
				p->SyncMode = 0;
				// leave time as it is
				return ERR_SYNCED;
			}
		}
	}

	// if all streams are full: caller can go to sleep
	Result = ERR_BUFFER_FULL;

	for (No=0;No<p->StreamCount;++No)
		if (p->Streams[No]->Pin.Node)
		{
			int i = p->Process(p,p->Streams[No]);

			if (!p->SyncMode || p->SyncStream == p->Streams[No])
			{
				if (i != ERR_NEED_MORE_DATA)
					AllNeedData = 0;

				if (i != ERR_END_OF_FILE)
					AllEndOfFile = 0;
			}

			// immediately stop so the player can add codec
			if (i == ERR_STREAM_ADDED)
				return i;

			if (i == ERR_SYNCED)
			{
				// try to calc duration once (after first sync)
				if (p->NeedDuration)
				{
					p->NeedDuration = 0;
					CalcDuration(p);
				}

				*Time = p->SyncTime;
				return i;
			}

			if (i != ERR_BUFFER_FULL && 
				i != ERR_END_OF_FILE && 
				i != ERR_NEED_MORE_DATA)
				Result = ERR_NONE; // no sleep

			NoStreams = 0;
		}

	if (NoStreams)
	{
		// release buffers
		LockEnter(p->BufferLock);

		while (p->BufferFirst)
		{
			format_buffer* Buffer = p->BufferFirst;
			p->BufferFilePos += Buffer->Length;
			p->BufferFirst = Buffer->Next;
			if (p->BufferLast == Buffer)
				p->BufferLast = NULL;
			Format_BufferRelease(p,Buffer);
		}

		LockLeave(p->BufferLock);

		AllEndOfFile = 0;
		AllNeedData = 0;

		*Time = -1;

		// return buffer full so fill mode doesn't go though file
		Result = p->NoMoreInput ? ERR_END_OF_FILE:ERR_BUFFER_FULL;
	}

	if (AllEndOfFile)
		Result = ERR_END_OF_FILE;
	else
	if (AllNeedData)
		Result = p->NoMoreInput ? ERR_END_OF_FILE:ERR_NEED_MORE_DATA;

	if (p->SyncMode && Result == ERR_END_OF_FILE)
	{
		p->SyncMode = 0;
		if (p->Duration < 0)
			*Time = p->SyncStream->LastTime;
		else
			*Time = p->Duration;
		Result = ERR_SYNCED;
	}

	return Result;
}

format_ref* Format_RefAlloc( format_base* p, format_buffer* To, int Begin, int Length )
{
	format_ref* Ref = p->FreeRefs;

	if (!Ref)
	{
		Ref = (format_ref*) Alloc(sizeof(format_ref));
		if (!Ref)
			return NULL;
	}
	else
		p->FreeRefs = Ref->Next;

	LockEnter(p->BufferLock);

	//DEBUG_MSG3("Buffer:%d addref %d->%d",To->Id,To->RefCount,To->RefCount+1);
	++To->RefCount;

	LockLeave(p->BufferLock);

	Ref->Length = Length;
	Ref->Begin = Begin;
	Ref->Buffer = To;
	Ref->Next = NULL;
	return Ref;
}

format_ref* Format_ReadAsRef( format_base* p, int Length )
{
	int n;
	format_ref* Ref;
	format_ref** RefPtr;

	Ref = NULL;
	RefPtr = &Ref;

	if (Length < 0) // read as much as possible from one buffer
	{
		while (p->ReadBuffer)
		{
			n = p->ReadBuffer->Length - p->ReadPos;
			if (n>0)
				break;

			p->ReadPos -= p->ReadBuffer->Length;
			p->ReadBuffer = p->ReadBuffer->Next;
		}

		if (!p->ReadBuffer)
		{
			p->NeedMoreInput = 1;
			return NULL;
		}

		if (n > -Length)
			n = -Length;
		Length = n;
	}

	while (Length > 0)
	{
		if (!p->ReadBuffer)
		{
			p->NeedMoreInput = 1;
			break;
		}

		n = p->ReadBuffer->Length - p->ReadPos;
		if (n > Length)
			n = Length;

		if (n>0)
		{
			*RefPtr = Format_RefAlloc(p,p->ReadBuffer,p->ReadPos,n);
			if (!*RefPtr)
				break;

			RefPtr = &(*RefPtr)->Next;

			Length -= n;
			p->FilePos += n;
			p->ReadPos += n;
		}

		if (p->ReadBuffer->Length <= p->ReadPos)
		{
			p->ReadPos -= p->ReadBuffer->Length;
			p->ReadBuffer = p->ReadBuffer->Next;
		}
	}

	if (Length)
	{
		// something went wrong, release references
		while (Ref)
		{
			format_ref* i = Ref->Next;
			Format_RefRelease(p,Ref);
			Ref = i;
		}
	}

	return Ref;
}

void Format_Read( format_base* p, void* Data, int Length )
{
	uint8_t* Ptr = (uint8_t*)Data;
	int n;
	while (Length > 0)
	{
		if (!p->ReadBuffer)
		{
			p->NeedMoreInput = 1;
			break;
		}

		n = p->ReadBuffer->Length - p->ReadPos;
		if (n > Length)
			n = Length;

		if (n>0)
		{
			memcpy(Ptr,p->ReadBuffer->Buffer+p->ReadPos,n);
			Ptr += n;
			Length -= n;
			p->FilePos += n;
			p->ReadPos += n;
		}

		if (p->ReadBuffer->Length <= p->ReadPos)
		{
			p->ReadPos -= p->ReadBuffer->Length;
			p->ReadBuffer = p->ReadBuffer->Next;
		}
	}
}

void Format_Skip(format_base* p,int n)
{
	if (n > 0)
	{
		p->FilePos += n;
		p->ReadPos += n;
	}
}

int Format_Read8(format_base* p)
{
	while (p->ReadBuffer && p->ReadPos >= p->ReadBuffer->Length)
	{
		p->ReadPos -= p->ReadBuffer->Length;
		p->ReadBuffer = p->ReadBuffer->Next;
	}
	if (!p->ReadBuffer)
	{
		p->NeedMoreInput = 1;
		return -1;
	}
	p->FilePos++;
	return p->ReadBuffer->Buffer[p->ReadPos++];
}
int Format_ReadBE16(format_base* p)
{
	int v = Format_Read8(p) << 8;
	v |= Format_Read8(p);
	return v;
}

int Format_ReadLE16(format_base* p)
{
	int v = Format_Read8(p);
	v |= Format_Read8(p) << 8;
	return v;
}

int Format_ReadBE32(format_base* p)
{
	int v = Format_Read8(p) << 24;
	v |= Format_Read8(p) << 16;
	v |= Format_Read8(p) << 8;
	v |= Format_Read8(p);
	return v;
}

int Format_ReadLE32(format_base* p)
{
	int v = Format_Read8(p);
	v |= Format_Read8(p) << 8;
	v |= Format_Read8(p) << 16;
	v |= Format_Read8(p) << 24;
	return v;
}

int Format_ReadInput(format_base* p,int BufferMaxSize,int BufferPercent,int* BufferUsed)
{
	format_buffer* Buffer;

	LockEnter(p->BufferLock);

//!!! is it really a good idea???
//	if (p->DropStream && BufferPercent>PERCENT_ONE/2) 
//		BufferPercent = PERCENT_ONE/2; // dropmode: don't load buffer (only when it's less then half)

	BufferMaxSize *= BufferPercent;

	if (BufferMaxSize < (2*PRELOAD/1024*PERCENT_ONE))
		BufferMaxSize = (2*PRELOAD/1024*PERCENT_ONE);

	if ((p->BufferUsed+1)*(CHUNKSIZE/1024*PERCENT_ONE) > BufferMaxSize)
	{
		if (BufferUsed)
			*BufferUsed = p->BufferUsed*(CHUNKSIZE/1024);
		LockLeave(p->BufferLock);
		return ERR_BUFFER_FULL;
	}

	// allocate buffer
	Buffer = Format_BufferAlloc(p);

	if (BufferUsed)
		*BufferUsed = p->BufferUsed*(CHUNKSIZE/1024);

	LockLeave(p->BufferLock);

	if (!Buffer)
		return ERR_BUFFER_FULL;

	// read buffer
	LockEnter(p->InputLock);
	Buffer->Length = p->Input->Read(p->Input,Buffer->Buffer,CHUNKSIZE);
	LockLeave(p->InputLock);

	if (Buffer->Length < 0)
	{
		Format_BufferRelease(p,Buffer);
		return ERR_DEVICE_ERROR;
	}

	p->NoMoreInput = Buffer->Length == 0;

	if (p->NoMoreInput)
	{
		Format_BufferRelease(p,Buffer);
		return ERR_END_OF_FILE;
	}

	// insert buffer to buffer chain

	//DEBUG_MSG2("Buffer:%d input %d",Buffer->Id,Buffer->RefCount);

	LockEnter(p->BufferLock);
	Format_BufferInsert(p,Buffer);
	LockLeave(p->BufferLock);

	p->WaitForInput = 0;
	return ERR_NONE;
}

format_stream* Format_AddStream(format_base* p, int Length, int FormatLength )
{
	format_stream* s;

	if (p->StreamCount == MAXSTREAM)
		return NULL;

	s = (format_stream*) CAlloc(Length,1);
	if (!s)
		return NULL;

	MergeInit(&s->Merge);
	s->No = p->StreamCount;
	s->LastTime = -1;
	s->FormatLength = FormatLength;
	s->Format = CAlloc(FormatLength,1);
	s->PacketBurst = DEFAULT_BURST;

	p->Streams[p->StreamCount++] = s;
	return s;
}

format_base FormatBase =
{
	Format_Enum,
	Format_Get,
	Format_Set,

	NULL, // Probe
	Format_Sync,
	Format_ReadInput,
	Format_Process,
	NULL, //ReadList

	NULL, //Init
	NULL, //Done
	NULL, //Seek
	NULL, //ReadPacket
	NULL, //FreeStream
	NULL, //ReleaseStream
	NULL, //ReadPacketTime

	Format_ProcessStream,
	Format_Sended,
};
