/*****************************************************************************
 *
 * This program is free software ; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 *
 * $Id$
 *
 ****************************************************************************/

#ifndef __FORMAT_BASE_H
#define __FORMAT_BASE_H

// base format implementation

#define MAXSTREAM	256
#define CHUNKSIZE	16384
#define PRELOAD		256*1024
#define SHOWAHEAD	(TICKSPERSEC/25)

#define VIDEO_BURST		2
#define DEFAULT_BURST	6

typedef struct format_buffer
{
	int Id;
	int Length;
	int RefCount;
	struct format_buffer* Next;
	uint8_t Buffer[CHUNKSIZE];

} format_buffer;

typedef struct format_ref
{
	format_buffer* Buffer;
	int Begin;
	int Length;
	struct format_ref* Next;

} format_ref;

struct format_stream;
typedef struct format_packet
{
	format_ref* Data;
	tick_t RefTime;
	struct format_stream* Stream;
	struct format_packet* Next;
	//int FilePos;
} format_packet;

typedef struct format_stream
{
	int No;
	int Id;
	pin Pin;
	pin Comment;
	int FormatType;
	void* Format;
	int FormatLength;
	bool_t Fragmented; //don't bother to merge buffers
	bool_t DisableDrop; // no dropping allowed
	int PacketBurst;

	// buffer for merging fragmented packets 
	// to a single continous memory
	merge Merge;

	bool_t Pending; // 1:full packet 2:only first reference
	tick_t LastTime; // last known refrence time (-1 out of sync)
	packet Packet;  // pending packet

	// buffered packets waiting for processing
	format_packet* PacketFirst;
	format_packet* PacketLast;

} format_stream;

typedef struct format_base
{
	format VMT;

	int (*Init)(void*);
	void (*Done)(void*);
	int (*Seek)(void*, tick_t, int );
	int (*ReadPacket)(void*, format_packet* Packet);
	void (*FreeStream)(void*, void* Stream);
	void (*ReleaseStream)(void*, void* Stream);
	void* (*ReadPacketTime)(void*,format_buffer* Buffer,tick_t* Min,tick_t* Max,void* State);
	int (*Process)(void* p, void* Stream);
	void (*Sended)(void* p, void* Stream);

	int Id;
	const tchar_t* Exts;
	bool_t NeedDuration;
	tick_t Duration;
	int FileSize;
	stream* Input;
	tick_t DropTolerance;
	tick_t SkipTolerance;
	tick_t AVOffset;
	
	int StreamCount;
	format_stream* Streams[MAXSTREAM];

	bool_t SyncMode;
	format_stream* SyncStream;
	tick_t SyncTime;

	int BufferId;
	format_packet* FreePackets;
	format_buffer* FreeBuffers; 
	format_ref*	   FreeRefs;

	bool_t NoMoreInput;
	bool_t NeedMoreInput;
	bool_t WaitForInput;
	int FilePos;
	int ReadPos;
	format_buffer* ReadBuffer;
	int PreLoad; // for headerless formats when to stop reading after open
	format_stream* DropStream; // one stream is dropping: try saving time on input and other streams

	int BufferFilePos;
	void* BufferLock; // buffer handling must be multithread safe
	int BufferUsed; // number of buffers used
	int BufferSize;
	int BufferPos;
	format_buffer* BufferFirst; 
	format_buffer* BufferLast;  

	tick_t ProcessTime;
	bool_t Bench;

	void* InputLock;

} format_base;

DLL format_base FormatBase;

DLL void Format_BufferRelease( format_base* p, format_buffer* Buffer );
DLL int Format_Send(format_base* p,format_stream* Stream);
DLL format_stream* Format_AddStream(format_base*, int Length, int FormatLength ); 
DLL format_stream* Format_DefSyncStream(format_base*);
DLL int Format_InputSeek(format_base*,int FilePos);
DLL	void* Format_ReadPacketTime(format_base*,format_buffer* Buffer,tick_t* Min, tick_t* Max,void* State);
DLL int Format_SeekByPacketTime(format_base*,tick_t Time,int FilePos);
DLL void Format_ReleaseBuffers(format_base* p);

DLL void Format_Skip(format_base*,int);
DLL void Format_Read(format_base*,void*,int);
DLL int Format_Read8(format_base*);
DLL int Format_ReadBE16(format_base*);
DLL int Format_ReadLE16(format_base*);
DLL int Format_ReadBE32(format_base*);
DLL int Format_ReadLE32(format_base*);
DLL format_ref* Format_ReadAsRef(format_base*, int Length);

#endif
