/*****************************************************************************
 *
 * This program is free software ; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 *
 * $Id$
 *
 ****************************************************************************/

#include "stdafx.h"

const datadef NodeParams[] = 
{
	{ NODE_CLASS,	TYPE_CLASS },
	{ NODE_ID,		TYPE_INT, DF_HIDDEN },

	DATADEF_END
};

pin NodeErrorPin = { NULL, 0 };
static int NodeCount = 0;
static node* Node[MAXNODE];
static int NodePri[MAXNODE];
static int NodeId[MAXNODE];

static int IsClassCounter = 0;
static int EnumClass = -1;
static int EnumCount = 0;
static node* EnumCache[MAXNODE];

int NodeRegister( node* p, int Priority )
{
	int No;
	int Id;

	if (!p || NodeCount >= MAXNODE || 
		p->Get(p,NODE_ID,&Id,sizeof(Id))!=ERR_NONE)
		return 0;

	DEBUG_MSG2(T("NodeRegister %08x (%08x)"),Id,p);

	for (No=0;No<NodeCount;++No)
	{
		assert(NodeId[No] != Id);
		if (NodeId[No] == Id)
			return 0;
	}

	assert(p->Enum && p->Get && p->Set);

	for (No=0;No<NodeCount;++No)
		if (NodePri[No] < Priority)
			break;

	memmove(Node+No+1,Node+No,(NodeCount-No)*sizeof(node*));
	memmove(NodePri+No+1,NodePri+No,(NodeCount-No)*sizeof(int));
	NodeCount++;
	EnumClass = -1;

	Node[No] = p;
	NodePri[No] = Priority;
	NodeId[No] = Id;

	NodeLoadSetup(p);
	return 1;
}

int NodeUnRegister( node* p )
{
	int i;
	for (i=0;i<NodeCount;++i)
		if (Node[i]==p)
		{
			NodeCount--;
			EnumClass = -1;
		
			memmove(Node+i,Node+i+1,sizeof(node*)*(NodeCount-i));
			return 1;
		}
	return 0;
}

bool_t NodeDataDef( node* p, int No, datadef* Out )
{
	int i;
	for (i=0;p->Enum(p,i,Out)==ERR_NONE;++i)
		if (Out->No == No)
			return 1;
	return 0;
}

node* NodeEnum( int Class, int No )
{
	int i;

	if (Class != EnumClass)
	{
		EnumClass = Class;
		EnumCount = 0;

		for (i=0;i<NodeCount;++i)
			if (NodeIsClass(Node[i],Class))
				EnumCache[EnumCount++] = Node[i];
	}

	if (No<0 || No >= EnumCount)
		return NULL;
	return EnumCache[No];
}

bool_t NodeInsideIsClass()
{
	return IsClassCounter!=0;
}

bool_t NodeIsClass( node* p, int Class )
{
	datadef DataDef;
	int No;
	int Result = 0;

	if (!p)
		return 1;
	
	IsClassCounter++;

	for (No=0;p->Enum(p,No,&DataDef)==ERR_NONE;++No)
		if (DataDef.Type == TYPE_CLASS && DataDef.No == Class)
		{
			Result = 1;
			break;
		}

	IsClassCounter--;

	return Result;
}

node* NodeFind( int FindId )
{
	int Id;
	int No;

	for (No=0;No<NodeCount;++No)
		if (Node[No]->Get(Node[No],NODE_ID,&Id,sizeof(Id))==ERR_NONE && Id==FindId)
			return Node[No];

	return NULL;
}

void NodeError( node* Node, int Class, int No,... )
{
	tchar_t s[1024];
	int Id;
	if (Node)
	{
		Node->Get(Node,NODE_ID,&Id,sizeof(Id));
		stprintf( s, T("%s: "),String(Id,0));
	}
	else
		s[0] = 0;

	vstprintf( s+tcslen(s), String(Class,No), (va_list)(&No+1) );

	if (NodeErrorPin.Node)
		NodeErrorPin.Node->Set(NodeErrorPin.Node,NodeErrorPin.No,s,sizeof(s));
	else
		ShowMessage(String(PLATFORM_ID,PLATFORM_ERROR),s);
}

static void DumpPtr( FILE* File, tchar_t* s, node* Ptr, int No )
{
	if (Ptr)
	{
		int Id;
		tchar_t v[32];

		if (No)
			stprintf(v,T(":%d"),No);
		else
			v[0] = 0;

		Id = 0;
		Ptr->Get(Ptr,NODE_ID,&Id,sizeof(Id));
		ftprintf(File,T("%s%s%s (0x%08x)\n"),s,StringDef(Id,0),v,Ptr);
	}
	else
		ftprintf(File,T("%sNULL\n"),s);
}

static void DumpFormat( FILE* File, tchar_t* s, node* Node, datadef* DataDef )
{
	audio Audio;
	video Video;
	tchar_t v[8];

	switch (DataDef->Format1)
	{
	case PACKET_VIDEO:
		ftprintf(File,T("%svideo"),s);
		if (Node->Get(Node,DataDef->No|PACKET_FORMAT,&Video,sizeof(Video))==ERR_NONE)
		{
			if (Video.Pixel.Flags & PF_FOURCC) 
			{
				FourCCToString(v,Video.Pixel.FourCC,sizeof(v));
				ftprintf(File,T(" fourcc=%s"),v);
			}
			if (Video.Width) ftprintf(File,T(" width=%d"),Video.Width);
			if (Video.Height) ftprintf(File,T(" height=%d"),Video.Height);
			if (Video.Direction) ftprintf(File,T(" dir=%d"),Video.Direction);
		}
		ftprintf(File,T("\n"),s);
		break;
	case PACKET_AUDIO:
		ftprintf(File,T("%saudio"),s);
		if (Node->Get(Node,DataDef->No|PACKET_FORMAT,&Audio,sizeof(Audio))==ERR_NONE)
		{
			if (Audio.Format) ftprintf(File,T(" fmt=%d"),Audio.Format);
			if (Audio.SampleRate) ftprintf(File,T(" rate=%d"),Audio.SampleRate);
			if (Audio.Channels) ftprintf(File,T(" ch=%d"),Audio.Channels);
			if (Audio.Bits) ftprintf(File,T(" bits=%d"),Audio.Bits);
			if (Audio.ByteRate) ftprintf(File,T(" bitrate=%d"),Audio.ByteRate*8);
		}
		ftprintf(File,T("\n"),s);
		break;
	case PACKET_COMMENT:
		ftprintf(File,T("%scomment\n"),s);
		break;
	default:
		ftprintf(File,T("%sunknown packet type (%d)\n"),s,DataDef->Format1);
	}
}

void NodeDump( FILE* File )
{
	int No;
	for (No=0;No<NodeCount;++No)
	{
		TRY_BEGIN 
		{
			tchar_t s[256];
			tchar_t Buffer[512];
			const tchar_t* Name;
			rect Rect;
			pin Pin;
			point Point;
			bool_t Bool;
			tick_t Tick;
			fraction Fraction;
			node* Ptr;
			int Int;

			int i;
			int Id = 0;
			node* p = Node[No];
			datadef DataDef;
	
			p->Get(p,NODE_ID,&Id,sizeof(Id));
			ftprintf(File,T("%s (0x%08x)\n"),StringDef(Id,0),p);

			for (i=0;p->Enum(p,i,&DataDef)==ERR_NONE;++i)
			{
				TRY_BEGIN 
				{
					Name = StringDef(DataDef.Class,DataDef.No);
					if (!Name[0]) Name = DataDef.Name;
					stprintf(s,T("  %s(%d)="),Name,DataDef.No);

					if (DataDef.Class!=NODE_CLASS||DataDef.No!=NODE_ID)
					switch (DataDef.Type)
					{
					case TYPE_BOOL:
						if (p->Get(p,DataDef.No,&Bool,sizeof(bool_t))==ERR_NONE)
						{
							ftprintf(File,T("%s%s\n"),s,StringDef(PLATFORM_ID,Bool ? PLATFORM_YES:PLATFORM_NO));
						}
						break;
					case TYPE_TICK:
						if (p->Get(p,DataDef.No,&Tick,sizeof(tick_t))==ERR_NONE)
						{
							TickToString(Buffer,Tick,sizeof(Buffer),Tick<TICKSPERSEC,1);
							ftprintf(File,T("%s%s\n"),s,Buffer);
						}
						break;
					case TYPE_INT:
						if (p->Get(p,DataDef.No,&Int,sizeof(int))==ERR_NONE)
						{
							IntToString(Buffer,Int,sizeof(Buffer),(DataDef.Flags & DF_HEX)!=0);
							ftprintf(File,T("%s%s\n"),s,Buffer);
						}
						break;
					case TYPE_HOTKEY:
						if (p->Get(p,DataDef.No,&Int,sizeof(int))==ERR_NONE)
						{
							HotKeyToString(Buffer,Int,sizeof(Buffer));
							ftprintf(File,T("%s%s\n"),s,Buffer);
						}
						break;
					case TYPE_FRACTION:
						if (p->Get(p,DataDef.No,&Fraction,sizeof(fraction))==ERR_NONE)
						{
							FractionToString(Buffer,&Fraction,sizeof(Buffer),1,2);
							ftprintf(File,T("%s%s\n"),s,Buffer,2);
						}
						break;
					case TYPE_STRING:
						if (p->Get(p,DataDef.No,Buffer,sizeof(Buffer))==ERR_NONE)
							ftprintf(File,T("%s%s\n"),s,Buffer);
						break;
					case TYPE_RECT:
						if (p->Get(p,DataDef.No,&Rect,sizeof(rect))==ERR_NONE)
							ftprintf(File,T("%s%d:%d:%d:%d\n"),s,Rect.x,Rect.y,Rect.Width,Rect.Height);
						break;
					case TYPE_POINT:
						if (p->Get(p,DataDef.No,&Point,sizeof(point))==ERR_NONE)
							ftprintf(File,T("%s%d:%d\n"),s,Point.x,Point.y);
						break;
					case TYPE_RGB:
						if (p->Get(p,DataDef.No,&Int,sizeof(int))==ERR_NONE)
							ftprintf(File,T("%s0x%08X\n"),s,Int);
						break;
					case TYPE_PIN:
						if (p->Get(p,DataDef.No,&Pin,sizeof(Pin))==ERR_NONE)
						{
							DumpPtr(File,s,Pin.Node,Pin.No);
							DumpFormat(File,s,p,&DataDef);
						}
						break;
					case TYPE_PACKET:
						DumpFormat(File,s,p,&DataDef);
						break;
					case TYPE_NODE:
						if (p->Get(p,DataDef.No,&Ptr,sizeof(Ptr)) == ERR_NONE)
							DumpPtr(File,s,Ptr,0);
						break;
					}
				}
				TRY_END
			}
		}
		TRY_END
	}
}