/*****************************************************************************
 *
 * This program is free software ; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 *
 * $Id$
 *
 ****************************************************************************/

#ifndef __PLATFORM_H
#define __PLATFORM_H

#define PLATFORM_ID			FOURCC('P','L','A','T')

#define PLATFORM_LANG		0x10
#define PLATFORM_OLDSHELL	0x1F
#define PLATFORM_NOBACKLIGHT 0x20
#define PLATFORM_PROCESSOR	0x11
#define PLATFORM_MODEL		0x12
#define PLATFORM_CAPS		0x13
#define PLATFORM_ICACHE		0x14
#define PLATFORM_DCACHE		0x15
#define PLATFORM_TYPE		0x17
#define PLATFORM_VER		0x19
#define PLATFORM_TYPENO		0x1A
#define PLATFORM_OEMINFO	0x18
#define PLATFORM_VERSION	0x1B

#define PLATFORM_BATTERYTIMEOUT	0x2000
#define PLATFORM_ACTIMEOUT		0x2001
#define PLATFORM_SCREENSAVER	0x2002
#define PLATFORM_DISPPOWER		0x2003

// models
#define	MODEL_DEFAULT			0
#define	MODEL_IPAQ_H3600		1
#define	MODEL_IPAQ_H3700		2
#define	MODEL_IPAQ_H3800		3
#define	MODEL_IPAQ_H3900		4
#define	MODEL_O2_XDA			5
#define	MODEL_TOSHIBA_E740		6
#define	MODEL_TOSHIBA_E750		7
#define	MODEL_CASIO_E105		8
#define	MODEL_CASIO_E115		9
#define	MODEL_CASIO_E125		10
#define	MODEL_CASIO_EM500		11
#define	MODEL_CASIO_E200		12
#define MODEL_AMIGO_600C		13
#define MODEL_JVC_C33			14
#define MODEL_JORNADA_680		15
#define MODEL_JORNADA_690		16
#define MODEL_JORNADA_710		17
#define MODEL_JORNADA_720		18
#define MODEL_SIGMARION			19
#define MODEL_SIGMARION2		20
#define MODEL_SIGMARION3		21
#define MODEL_MOBILEGEAR2_450	22
#define MODEL_MOBILEGEAR2_430	23
#define MODEL_MOBILEGEAR2_520	24
#define MODEL_MOBILEGEAR2_530	25
#define MODEL_MOBILEGEAR2_550	26
#define MODEL_MOBILEGEAR2_700	27
#define MODEL_MOBILEGEAR2_730	28
#define MODEL_MOBILEPRO_770		29
#define MODEL_MOBILEPRO_780		30
#define MODEL_INTERMEC_6651		31
#define MODEL_JORNADA_540		32
#define	MODEL_TOSHIBA_E800		33

// cpus
#define CPU_UNKNOWN			0
#define CPU_X86				1			
#define CPU_ARM				2
#define CPU_MIPS			3
#define CPU_SH3				4

// special cpu and framebuffer features
#define CAPS_ARM_5E				0x0001
#define CAPS_ARM_WMMX			0x0002
#define CAPS_MIPS_VR41XX		0x0004
#define CAPS_X86_MMX			0x0008
#define CAPS_X86_SSE			0x0010
#define CAPS_ONLY12BITRGB		0x0020
#define CAPS_LANDSCAPE			0x0040
#define CAPS_PORTRAIT			0x0080

// platform types
#define TYPE_UNKNOWN			0
#define TYPE_POCKETPC			1
#define TYPE_SMARTPHONE			2

DLL void Platform_Init();
DLL void Platform_Done();

DLL void Dump();

DLL int QueryPlatform( int Param );
DLL void QueryDesktop( video* );
DLL int QueryOrientation();
DLL bool_t IsOrientationChanged();
DLL void GlobalInvalidate( const rect* );

DLL void* LockFrameBuffer( video* );
DLL void UnlockFrameBuffer();

DLL const tchar_t* GetDocumentPath();
DLL const tchar_t* GetSystemPath();

DLL void SoftReset();
DLL bool_t KernelMode(bool_t);
DLL bool_t GetDisplayPower();
DLL void SetDisplayPower(bool_t State);
DLL void BacklightTimeout(bool_t Disable);
DLL void IdleTimerReset();

DLL void SetFileAssociation(const tchar_t* Ext,const tchar_t* Type,bool_t State);
DLL bool_t GetFileAssociation(const tchar_t* Ext,const tchar_t* Type);
DLL void SetFileType(const tchar_t* Type, const tchar_t* Open, const tchar_t* Icon );

DLL void* CodeAlloc(int Size);
DLL void CodeFree(void*,int Size);
DLL void CodeLock(void*,int Size);
DLL void CodeUnlock(void*,int Size);
DLL void CodeFindPages(void* Ptr,unsigned char** Min,unsigned char** Max,int* PageSize);

DLL	void* LockCreate();
DLL void LockDelete(void*);
DLL	void LockEnter(void*);
DLL void LockLeave(void*);

DLL void WaitBegin();
DLL void WaitEnd();

DLL void* ThreadCreate( int(*Start)(void*) ,void* Parameter,int Quantum);

DLL void* Alloc(int);
DLL void* CAlloc(int,int);
DLL void* ReAlloc(void*,int);
DLL void Free(void*);
DLL void CheckHeap();
DLL bool_t ShowMemoryLow;

DLL int GetTime(); // in millisec

DLL void ShowMessage(const tchar_t* Title,const tchar_t* Msg,...);
DLL void DebugMessage(const tchar_t*,...);

#ifndef NDEBUG
  #define DEBUG_MSG(x) DebugMessage(x)
  #define DEBUG_MSG1(x,a) DebugMessage(x,a)
  #define DEBUG_MSG2(x,a,b) DebugMessage(x,a,b)
  #define DEBUG_MSG3(x,a,b,c) DebugMessage(x,a,b,c)
  #define DEBUG_MSG4(x,a,b,c,d) DebugMessage(x,a,b,c,d)
  #define DEBUG_MSG5(x,a,b,c,d,e) DebugMessage(x,a,b,c,d,e)
  #define DEBUG_MSG6(x,a,b,c,d,e,f) DebugMessage(x,a,b,c,d,e,f)
#else
  #define DEBUG_MSG(x)
  #define DEBUG_MSG1(x,a)
  #define DEBUG_MSG2(x,a,b)
  #define DEBUG_MSG3(x,a,b,c)
  #define DEBUG_MSG4(x,a,b,c,d)
  #define DEBUG_MSG5(x,a,b,c,d,e)
  #define DEBUG_MSG6(x,a,b,c,d,e,f)
#endif

#if defined(_WIN32)

DLL int SafeException(void*);

#define SAFE_BEGIN __try {
#define SAFE_END ;} __except (SafeException(GetExceptionInformation())) {}
#define TRY_BEGIN __try {
#define TRY_END ;} __except (1) {}

#else

DLL int SafeException(int Code);

#define SAFE_BEGIN try {
#define SAFE_END ;} catch (int Code) { SafeException(Code); }
#define TRY_BEGIN try {
#define TRY_END ;} catch (int) {}
#endif

#define PLATFORM_CRASH_MESSAGE		100
#define PLATFORM_CRASH_TITLE		101
#define PLATFORM_YES				102
#define PLATFORM_NO					103
#define PLATFORM_DUMP_MESSAGE		104
#define PLATFORM_DUMP_TITLE			105
#define PLATFORM_OK					106
#define PLATFORM_CANCEL				107
#define PLATFORM_ERROR				108
#define PLATFORM_MEMORY_LOW			109
#define PLATFORM_KEY_ENTER			110
#define PLATFORM_KEY_UP				111
#define PLATFORM_KEY_DOWN			112
#define PLATFORM_KEY_LEFT			113
#define PLATFORM_KEY_RIGHT			114
#define PLATFORM_KEY_SPACE			115
#define PLATFORM_KEY_APP			116
#define PLATFORM_KEY_CTRL			117
#define PLATFORM_KEY_SHIFT			118
#define PLATFORM_KEY_ALT			119
#define PLATFORM_KEY_WIN			120
#define PLATFORM_KEY_ACTION			123
#define PLATFORM_ASSIGN				121
#define PLATFORM_CLEAR				122
#define PLATFORM_RESET				130

#define HOTKEY_MASK					0x0FFFF
#define HOTKEY_SHIFT				0x10000
#define HOTKEY_CTRL					0x20000
#define	HOTKEY_ALT					0x40000
#define	HOTKEY_WIN					0x80000

#endif
