/*****************************************************************************
 *
 * This program is free software ; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 *
 * $Id$
 *
 ****************************************************************************/

#ifndef __PORTAB_H
#define __PORTAB_H

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#if _MSC_VER > 1000
#pragma once
#pragma warning( push, 4 )
#pragma warning( disable : 4100 4710 4514 4201 4714 4115 4206 4055 4214 4998 )
#endif

#undef INLINE
#undef LITTLE_ENDIAN
#undef BIG_ENDIAN

#if defined(ARM) || defined(MIPS) || defined(SH3) || defined(_M_IX86)
#define LITTLE_ENDIAN
#else
#define BIG_ENDIAN
#endif

#define TICKSPERSEC			16384

#define MAX_INT	0x7FFFFFFF

typedef int int32_t;
typedef unsigned int uint32_t;  
typedef short int16_t; 
typedef unsigned short uint16_t; 
typedef char int8_t;
typedef unsigned char uint8_t;

#ifdef _MSC_VER
typedef __int64 int64_t;
typedef unsigned __int64 uint64_t;
#define INLINE static __forceinline
#define STDCALL __stdcall
#else
typedef long long int64_t;
typedef unsigned long long uint64_t;
#define INLINE static inline
#define STDCALL 
#endif

typedef int bool_t;
typedef int tick_t;

typedef struct fraction
{
	int Num;
	int Den;
} fraction;

#ifdef UNICODE
typedef unsigned short tchar_t;
#define tcsstr wcsstr
#define tcslen wcslen
#define tcscmp wcscmp
#define tcscpy wcscpy
#define tcscat wcscat
#define tcsncpy wcsncpy
#define tcschr wcschr
#define tcsrchr wcsrchr
#define tcsicmp wcsicmp
#define tcstostr wcstombs
#define tcsupr wcsupr
#define strtotcs mbstowcs
#define wcstotcs wcsncpy
#define stscanf swscanf
#define stprintf swprintf
#define vstprintf vswprintf
#define vftprintf vfwprintf
#define ftprintf fwprintf
#define tfopen	_wfopen
#define T(a) L ## a
#else
typedef char tchar_t;
#define tcsstr strstr
#define tcslen strlen
#define tcscpy strcpy
#define tcscmp strcmp
#define tcscat strcat
#define tcsncpy strncpy
#define tcschr strchr
#define tcsrchr strrchr
#define tcsicmp stricmp
#define tcstostr strncpy
#define tcsupr strupr
#define strtotcs strncpy
#define wcstotcs wcstombs
#define stscanf sscanf
#define stprintf sprintf
#define vstprintf vsprintf
#define vftprintf vfprintf
#define ftprintf fprintf
#define tfopen	fopen
#define T(a) a
#endif

#define DLLEXPORT __declspec(dllexport)
#define DLLIMPORT __declspec(dllimport)

#if !defined(SH3) && !defined(MIPS)
#define _INLINE INLINE
#define _CONST const
#else
#define _INLINE
#define _CONST
#endif

#define MAXPLANES 4
typedef void* planes[MAXPLANES];

#define MAXPATH 256

#define FOURCCBE(a,b,c,d) \
	(((unsigned char)a << 24) | ((unsigned char)b << 16) | \
	((unsigned char)c << 8) | ((unsigned char)d << 0))

#define FOURCCLE(a,b,c,d) \
	(((unsigned char)a << 0) | ((unsigned char)b << 8) | \
	((unsigned char)c << 16) | ((unsigned char)d << 24))

#ifdef BIG_ENDIAN
#define FOURCC(a,b,c,d) FOURCCBE(a,b,c,d)
#else
#define FOURCC(a,b,c,d) FOURCCLE(a,b,c,d)
#endif

#endif
