/*****************************************************************************
 *
 * This program is free software ; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 *
 * $Id$
 *
 ****************************************************************************/
 
#include "stdafx.h"
#include "softidct.h"

//#define STAT

#ifdef STAT

int Len[16] = {0};
int Count = 0;
int Type[2][3] = {0};
int Spec[2][5] = {0};
int Skip = 0;
int Max = 0;

void Statistics( idct_block_t *Block, int Length, int ScanType, int Add )
{
	int i;
	for (i=0;i<Length;++i)
	{
		int v = Block[i];
		if (v<0) v=-v;
		if (v>Max) Max=v;
	}

	++Len[Length >> 2];
	++Count;
	++Type[Add][ScanType];
	if (Length==1)
		Spec[Add][0]++;
	else
	if (ScanType==0 && Length==2)
		Spec[Add][1]++;
	else
	if (!Add && ScanType!=1 && Length < 15)
	{
		Spec[Add][2]++;
	}
	else
	if (Add && ScanType==0 && (Length < 15 || (Length<26 && ((uint32_t*)Block)[2]==0 && ((uint32_t*)Block)[6]==0)))
	{
		Spec[Add][2]++;
	}
}
#else
#define Statistics(a,b,c,d)
#endif

INLINE void SetPtr(softidct* p,int x,int y)
{
	x <<= 4;
	y <<= 4;
	x += EDGE;
	y += EDGE;
	y *= p->BufWidth;
	p->CurrPitch = p->BufWidth;
	p->DstPtr = p->Dst+x+y;
	p->BackPtr = p->Back+x+y;
	p->FwdPtr = p->Fwd+x+y;
	p->Ptr = p->Tab;
	p->Tab[3] = 1 | 2*(-y-x-8*p->BufWidth-8 + 
		p->YToU + (x>>1) + (y>>2));	//Y[1;1] -> U
}

INLINE void IncPtr(softidct* p)
{
	int v = *(p->Ptr++);
	p->CurrPitch >>= v & 1;
	v >>= 1;
	p->DstPtr += v;
	p->BackPtr += v;
	p->FwdPtr += v;
}

INLINE void IncPtrLum(softidct* p)
{
	int v = (p->Tab[0]+p->Tab[1]+p->Tab[2]+p->Tab[3]) >> 1;
	p->CurrPitch >>= 1;
	p->DstPtr += v;
	p->BackPtr += v;
	p->FwdPtr += v;
	p->Ptr += 4;
}

void Copy(softidct* p,int x,int y,int Forward)
{
#ifdef MIPS64_WIN32
	DisableInterrupts();
#endif

	SetPtr(p,x,y);
	Copy16x16(Forward ? p->FwdPtr:p->BackPtr,p->DstPtr,p->CurrPitch,p->CurrPitch);
	IncPtrLum(p);
	CopyBlock(Forward ? p->FwdPtr:p->BackPtr,p->DstPtr,p->CurrPitch,p->CurrPitch);
	IncPtr(p);
	CopyBlock(Forward ? p->FwdPtr:p->BackPtr,p->DstPtr,p->CurrPitch,p->CurrPitch);

#ifdef MIPS64_WIN32
	EnableInterrupts();
#endif
	EMMS();
}

void Data(softidct* p,idct_block_t *Block,int Length,int ScanType)
{
	if (Length)
	{
		Statistics(Block,Length,ScanType,0);

#ifdef IDCT8X4
		if (ScanType!=IDCTSCAN_ALT_HORI && Length < 11)
			IDCT_Block8x4(Block,p->DstPtr,p->CurrPitch,NULL,0);
#else
		if (ScanType!=IDCTSCAN_ALT_HORI && Length < 15)
			IDCT_Block4x8(Block,p->DstPtr,p->CurrPitch,NULL,0);
#endif
		else
			IDCT_Block8x8(Block,p->DstPtr,p->CurrPitch,NULL,0);

		EMMS();
	}
	
	IncPtr(p);
}

#ifdef MIPS64

void AddData(softidct* p,int MVBack,int MVFwd,idct_block_t *Block,int Length,int ScanType)
{
	// mcomp part is done by Process()
	if (Length)
	{
		if (Length == 1)
			IDCT_Const8x8(p->DstPtr,p->DstPtr,p->CurrPitch,p->CurrPitch,(Block[0]+4) >> 3);
#ifdef IDCT8X4
		else if (ScanType==IDCTSCAN_ZIGZAG && (Length < 11 || (Length<20 && Block[32]==0)))
			IDCT_Block8x4(Block,p->DstPtr,p->CurrPitch,p->DstPtr,p->CurrPitch);
#else
		else if (ScanType==IDCTSCAN_ZIGZAG && (Length < 15 || (Length<26 && ((uint32_t*)Block)[2]==0 && ((uint32_t*)Block)[6]==0)))
			IDCT_Block4x8(Block,p->DstPtr,p->CurrPitch,p->DstPtr,p->CurrPitch);
#endif
		else
			IDCT_Block8x8(Block,p->DstPtr,p->CurrPitch,p->DstPtr,p->CurrPitch);
	}
	
	IncPtr(p);
}

void MCompBackFwd(softidct* p,int MVBack,int MVFwd,idct_block_t *Block,int Length, int ScanType)
{
	uint8_t* Ptr;

#ifdef MIPS64_WIN32
	DisableInterrupts();
#endif

	if (MVBack != NOMV)
	{
		Ptr = p->BackPtr + ((MVBack<<16)>>17) + p->CurrPitch * (MVBack>>17);
		if (Ptr >= p->BackMin && Ptr < p->BackMax)
			p->CopyBlock[(MVBack&1)+((MVBack>>15)&2)](Ptr,p->DstPtr,p->CurrPitch,p->CurrPitch);
	}

	if (MVFwd != NOMV)
	{
		Ptr = p->FwdPtr + ((MVFwd<<16)>>17) + p->CurrPitch * (MVFwd>>17);
		if (Ptr >= p->FwdMin && Ptr < p->FwdMax)
			if (MVBack != NOMV)
			{
				p->CopyBlock[(MVFwd&1)+((MVFwd>>15)&2)](Ptr,p->Tmp,p->CurrPitch,8);
				Add8x8(p->Tmp,p->DstPtr,8,p->CurrPitch);
			}
			else
				p->CopyBlock[(MVFwd&1)+((MVFwd>>15)&2)](Ptr,p->DstPtr,p->CurrPitch,p->CurrPitch);
	}

#ifdef MIPS64_WIN32
	EnableInterrupts();
#endif

	if (Length)
		IDCT_Block8x8(Block,p->DstPtr,p->CurrPitch,p->DstPtr,p->CurrPitch);

	IncPtr(p);
}

void Process(softidct* p,int x,int y,const int *MVBackHint,const int *MVFwdHint )
{
	SetPtr(p,x,y);

	// using optional hint and doing all the mcomp here
	if (MVBackHint && MVFwdHint)
	{
		uint8_t* Ptr;
		int MVBack = MVBackHint[0];
		int MVFwd = MVFwdHint[0];
		int i=0;

#ifdef MIPS64_WIN32
		DisableInterrupts();
#endif
		// same mcomp for all lum parts? use 16x16 routines?
		if (((MVBack ^ MVBackHint[1])|(MVBack ^ MVBackHint[2])|
			 (MVBack ^ MVBackHint[3])|(MVFwd  ^ MVFwdHint[1]) | 
			 (MVFwd  ^ MVFwdHint[2]) |(MVFwd  ^ MVFwdHint[3]))==0)
		{
			if (MVBack != NOMV)
			{
				Ptr = p->BackPtr + ((MVBack<<16)>>17) + p->CurrPitch * (MVBack>>17);
				if (Ptr >= p->BackMin && Ptr < p->BackMax)
					p->CopyMBlock[(MVBack&1)+((MVBack>>15)&2)](Ptr,p->DstPtr,p->CurrPitch,p->CurrPitch);
			}

			if (MVFwd != NOMV)
			{
				Ptr = p->FwdPtr + ((MVFwd<<16)>>17) + p->CurrPitch * (MVFwd>>17);
				if (Ptr >= p->FwdMin && Ptr < p->FwdMax)
					if (MVBack != NOMV)
					{
						p->CopyMBlock[(MVFwd&1)+((MVFwd>>15)&2)](Ptr,p->Tmp,p->CurrPitch,16);
						Add16x16(p->Tmp,p->DstPtr,16,p->CurrPitch);
					}
					else
						p->CopyMBlock[(MVFwd&1)+((MVFwd>>15)&2)](Ptr,p->DstPtr,p->CurrPitch,p->CurrPitch);
			}

			// skipping lum part in 8x8 cycle
			IncPtrLum(p);
			i += 4;
			MVBackHint += 4;
			MVFwdHint += 4;
		}

		// mcomp for 8x8 blocks
		for (;i<6;++i,++MVBackHint,++MVFwdHint)
		{
			if (*MVBackHint != NOMV)
			{
				Ptr = p->BackPtr + ((*MVBackHint<<16)>>17) + p->CurrPitch * (*MVBackHint>>17);
				if (Ptr >= p->BackMin && Ptr < p->BackMax)
					p->CopyBlock[(*MVBackHint&1)+((*MVBackHint>>15)&2)](Ptr,p->DstPtr,p->CurrPitch,p->CurrPitch);
			}

			if (*MVFwdHint != NOMV)
			{
				Ptr = p->FwdPtr + ((*MVFwdHint<<16)>>17) + p->CurrPitch * (*MVFwdHint>>17);
				if (Ptr >= p->FwdMin && Ptr < p->FwdMax)
					if (*MVBackHint != NOMV)
					{
						p->CopyBlock[(*MVFwdHint&1)+((*MVFwdHint>>15)&2)](Ptr,p->Tmp,p->CurrPitch,8);
						Add8x8(p->Tmp,p->DstPtr,8,p->CurrPitch);
					}
					else
						p->CopyBlock[(*MVFwdHint&1)+((*MVFwdHint>>15)&2)](Ptr,p->DstPtr,p->CurrPitch,p->CurrPitch);
			}

			IncPtr(p);
		}

#ifdef MIPS64_WIN32
		EnableInterrupts();
#endif
		SetPtr(p,x,y);

		p->VMT.MCompData = AddData;
	}
	else
		p->VMT.MCompData = MCompBackFwd; //no hint: compability mode, but this is slower
}

#else

void Process(softidct* p,int x,int y,const int *MVBackHint,const int *MVFwdHint )
{
	SetPtr(p,x,y);
}

void MCompBackFwd(softidct* p,int MVBack,int MVFwd,idct_block_t *Block,int Length, int ScanType)
{
	uint8_t* Ptr;

	if (Length)
	{
		// mcomp and idct (using tmp buffer)

		if (MVBack != NOMV)
		{
			Ptr = p->BackPtr + ((MVBack<<16)>>17) + p->CurrPitch * (MVBack>>17);
			if (Ptr >= p->BackMin && Ptr < p->BackMax)
				p->CopyBlock[(MVBack&1)+((MVBack>>15)&2)](Ptr,p->Tmp,p->CurrPitch,8);
		}
		if (MVFwd != NOMV)
		{
			Ptr = p->FwdPtr + ((MVFwd<<16)>>17) + p->CurrPitch * (MVFwd>>17);
			if (Ptr >= p->FwdMin && Ptr < p->FwdMax)
				if (MVBack != NOMV)
				{
#if defined(MIPS32)
					p->CopyBlock[(MVFwd&1)+((MVFwd>>15)&2)](Ptr,p->Tmp+64,p->CurrPitch,8);
					Add8x8(p->Tmp+64,p->Tmp,8,8);
#else
					p->AddBlock[(MVFwd&1)+((MVFwd>>15)&2)](Ptr,p->Tmp,p->CurrPitch);
#endif
				}
				else
					p->CopyBlock[(MVFwd&1)+((MVFwd>>15)&2)](Ptr,p->Tmp,p->CurrPitch,8);
		}

		if (Length == 1)
			IDCT_Const8x8(p->Tmp,p->DstPtr,8,p->CurrPitch,(Block[0]+4) >> 3);
#ifdef IDCT8X4
		else if (ScanType!=IDCTSCAN_ALT_HORI && Length < 11)
			IDCT_Block8x4(Block,p->DstPtr,p->CurrPitch,p->Tmp,8);
#else
		else if (ScanType!=IDCTSCAN_ALT_HORI && Length < 15)
			IDCT_Block4x8(Block,p->DstPtr,p->CurrPitch,p->Tmp,8);
#endif
		else
			IDCT_Block8x8(Block,p->DstPtr,p->CurrPitch,p->Tmp,8);
	}
	else 
	if (MVBack != NOMV && MVFwd != NOMV)
	{
		// interpolate back and foward (using tmp buffer)

		Ptr = p->BackPtr + ((MVBack<<16)>>17) + p->CurrPitch * (MVBack>>17);
		if (Ptr >= p->BackMin && Ptr < p->BackMax)
			p->CopyBlock[(MVBack&1)+((MVBack>>15)&2)](Ptr,p->Tmp,p->CurrPitch,8);

		Ptr = p->FwdPtr + ((MVFwd<<16)>>17) + p->CurrPitch * (MVFwd>>17);
		if (Ptr >= p->FwdMin && Ptr < p->FwdMax)
		{
#if defined(MIPS32)
			p->CopyBlock[(MVFwd&1)+((MVFwd>>15)&2)](Ptr,p->DstPtr,p->CurrPitch,p->CurrPitch);
			Add8x8(p->Tmp,p->DstPtr,8,p->CurrPitch);
#else
			p->AddBlock[(MVFwd&1)+((MVFwd>>15)&2)](Ptr,p->Tmp,p->CurrPitch);
			// copy Tmp to Dst
			CopyBlock(p->Tmp,p->DstPtr,8,p->CurrPitch);
#endif
		}
	}
	else 
	if (MVBack != NOMV)
	{
		// only back mcomp
		Ptr = p->BackPtr + ((MVBack<<16)>>17) + p->CurrPitch * (MVBack>>17);
		if (Ptr >= p->BackMin && Ptr < p->BackMax)
			p->CopyBlock[(MVBack&1)+((MVBack>>15)&2)](Ptr,p->DstPtr,p->CurrPitch,p->CurrPitch);
	}
	else 
	if (MVFwd != NOMV)
	{
		// only forward mcomp
		Ptr = p->FwdPtr + ((MVFwd<<16)>>17) + p->CurrPitch * (MVFwd>>17);
		if (Ptr >= p->FwdMin && Ptr < p->FwdMax)
			p->CopyBlock[(MVFwd&1)+((MVFwd>>15)&2)](Ptr,p->DstPtr,p->CurrPitch,p->CurrPitch);
	}

	EMMS();
	IncPtr(p);
}

void MCompBack(softidct* p,int MVBack,int MVFwd,idct_block_t *Block,int Length, int ScanType )
{
	uint8_t* Ptr;

	if (Length)
	{
		// mcomp and idct (using tmp buffer)

		Ptr = p->BackPtr + ((MVBack<<16)>>17) + p->CurrPitch * (MVBack>>17);
		if (Ptr >= p->BackMin && Ptr < p->BackMax)
			p->CopyBlock[(MVBack&1)+((MVBack>>15)&2)](Ptr,p->Tmp,p->CurrPitch,8);
		Statistics(Block,Length,0,1);

		if (Length == 1)
			IDCT_Const8x8(p->Tmp,p->DstPtr,8,p->CurrPitch,(Block[0]+4) >> 3);
#ifdef IDCT8X4
		else if (ScanType==IDCTSCAN_ZIGZAG && (Length < 11 || (Length<20 && Block[32]==0)))
			IDCT_Block8x4(Block,p->DstPtr,p->CurrPitch,p->Tmp,8);
#else
		else if (ScanType==IDCTSCAN_ZIGZAG && (Length < 15 || (Length<26 && ((uint32_t*)Block)[2]==0 && ((uint32_t*)Block)[6]==0)))
			IDCT_Block4x8(Block,p->DstPtr,p->CurrPitch,p->Tmp,8);
#endif
		else
			IDCT_Block8x8(Block,p->DstPtr,p->CurrPitch,p->Tmp,8);
	}
	else 
	{
		// only back mcomp
		Ptr = p->BackPtr + ((MVBack<<16)>>17) + p->CurrPitch * (MVBack>>17);
		if (Ptr >= p->BackMin && Ptr < p->BackMax)
			p->CopyBlock[(MVBack&1)+((MVBack>>15)&2)](Ptr,p->DstPtr,p->CurrPitch,p->CurrPitch);
	}

	EMMS();
	IncPtr(p);
}

#endif

