/*****************************************************************************
 *
 * This program is free software ; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 *
 * $Id$
 *
 ****************************************************************************/
 
#include "stdafx.h"
#include "softidct.h"

#ifdef MIPS64

static void edge_copy(uint8_t *p_border_top,uint8_t *p_border_top_ref,int width){
	__asm (	"srl	$8,$6,4;"
			"bucle: ldr	$10,0($5);"
			"ldr	$11,8($5);"
#ifdef MIPSVR41XX
			".set noreorder;"
			"cache	13,0($4);"
			".set reorder;"
#endif
			"addi	$5,$5,16;"
			"sdr	$10,0($4);"
			"sdr	$11,8($4);"
			"addi	$8,$8,-1;"
			"addi	$4,$4,16;"
			"bgtz	$8,bucle;"
			"andi	$6,$6,0x08;"
			"beq	$6,$0,final;"
			"ldr	$10,0($5);"
			"sdr	$10,0($4);"
			"final:"
			);
}
static void edge_set(uint8_t *p_border_top,int p_border_top_ref,int width) {
	__asm (	"srl	$8,$6,4;"
			"dsll	$10,$5,56;"
			"dsrl	$11,$10,8;"
			"or		$10,$10,$11;"
			"dsrl	$11,$10,16;"
			"or		$10,$10,$11;"
			"dsrl	$11,$10,32;"
			"or		$10,$10,$11;"
			"bucle2:"
#ifdef MIPSVR41XX
			".set noreorder;"
			"cache	13,0($4);"
			".set reorder;"
#endif
			"sdr	$10,0($4);"
			"sdr	$10,8($4);"
			"addi	$8,$8,-1;"
			"addi	$4,$4,16;"
			"bgtz	$8,bucle2;"
			);
}
#else
#define edge_copy memcpy
#define edge_set memset
#endif

static void FillEdge(uint8_t *Ptr, int Width, int Height, int Edge)
{
	int n;
	uint8_t *p;
	int InnerWidth = Width - Edge*2;
	int InnerHeight = Height - Edge*2;
		
	// left and right
	p = Ptr + Edge + Edge * Width;
	for (n=0;n<InnerHeight;++n)
	{
		edge_set(p-Edge, p[0], Edge);
		edge_set(p+InnerWidth,p[InnerWidth-1], Edge);
		p += Width;
	}

	// top
	p = Ptr;
	for (n=0;n<Edge;++n)
	{
		edge_copy(p, Ptr+Edge*Width, Width);
		p += Width;
	}

	// bottom
	p = Ptr + (Height-Edge)*Width;
	for (n=0;n<Edge;++n)
	{
		edge_copy(p, Ptr+(Height-Edge-1)*Width, Width);
		p += Width;
	}
}

static void FillEdgeYUV(softidct* p,uint8_t *Ptr)
{
	FillEdge(Ptr, p->BufWidth, p->BufHeight, EDGE );
	Ptr += p->YToU;
	FillEdge(Ptr, p->BufWidth>>1, p->BufHeight>>1, EDGE>>1);
	Ptr += p->YToU >> 2;
	FillEdge(Ptr, p->BufWidth>>1, p->BufHeight>>1, EDGE>>1);
}

#if defined( MIPS64_WIN32 ) 
static int TouchPages(softidct* p)
{
	uint8_t* i;
	int Sum = 0;
	for (i=p->CodeBegin;i<p->CodeEnd;i+=p->CodePage)
		Sum += *i;
	return Sum;
}
#endif

static int Lock(softidct* p,int No,planes Planes,int* Brightness,int* Pitch)
{
	if (Brightness)
		*Brightness = 0;
	if (Pitch)
		*Pitch = p->OutputFormat.Pitch;

	if (No<0 || No>=p->BufCount)
	{
		Planes[0] = NULL;
		Planes[1] = NULL;
		Planes[2] = NULL;
		return ERR_INVALID_PARAM;
	}

	Planes[0] = p->Buffer[No] + EDGE + EDGE*p->BufWidth;
	Planes[1] = p->Buffer[No] + p->YToU + (EDGE >> 1) + ((EDGE*p->BufWidth) >> 2);
	Planes[2] = (uint8_t*)Planes[1] + (p->YToU >> 2);
	return ERR_NONE;
}

static void Unlock(softidct* p,int No)
{
}

static int SetRounding(softidct* p,int Rounding)
{
	p->CopyBlock = p->AllCopyBlock[0][Rounding];
	p->CopyMBlock = p->AllCopyBlock[1][Rounding];
	p->Rounding = Rounding;
	return ERR_NONE;
}

static int SetBufferCount(softidct* p, int n, int Temp)
{
	int i;
	n += Temp;
	if (n>p->MaxCount) return ERR_NOT_SUPPORTED;

	for (i=n;i<p->BufCount;++i)
	{
		Free(p->_Buffer[i]);
		p->_Buffer[i] = NULL;
		p->Buffer[i] = NULL;
	}
	if (p->BufCount > n)
		p->BufCount = n;

	for (i=p->BufCount;i<n;++i)
	{
#ifndef MIPS
		int Align = 32;
		int Offset = 0;
#else
		int Align = 8192;
		int Offset = ((i>>1)+(i&1)*2)*2048;
#endif
		p->_Buffer[i] = (uint8_t*) Alloc(((p->BufWidth*p->BufHeight*3) >> 1)+Align+Offset);
		if (!p->_Buffer[i]) 
		{
			p->MaxCount = p->BufCount;
			break;
		}
		p->Buffer[i] = (uint8_t*)(((int)p->_Buffer[i] + Align-1) & ~(Align-1));
		p->Buffer[i] += Offset;
		p->BufBorder[i] = 0;
		p->BufFrameNo[i] = -1;
		p->BufCount = i+1;
	}

	if (p->ShowNext >= p->BufCount)
		p->ShowNext = -1;
	if (p->ShowCurr >= p->BufCount)
		p->ShowCurr = -1;
	p->LastTemp = Temp && p->BufCount == n;

	if (p->BufCount != n)
		return ERR_OUT_OF_MEMORY;

	return ERR_NONE;
}

static int SetFormat(softidct* p, const video* Format )
{
	SetBufferCount(p,0,0);
	memset(&p->OutputFormat,0,sizeof(p->OutputFormat));

	if (Format)
	{
		if (Format->Pixel.Flags != PF_YUV420)
			return ERR_NOT_SUPPORTED;

		p->OutputFormat = *Format;
		if (p->OutputFormat.Direction & DIR_SWAPXY)
			SwapInt(&p->OutputFormat.Width,&p->OutputFormat.Height);

		p->MaxCount = MAXBUF;
		p->BufWidth = ((p->OutputFormat.Width+15)&~15) + 2*EDGE;
		p->BufHeight = ((p->OutputFormat.Height+15)&~15) + 2*EDGE;
		p->YToU = p->BufHeight * p->BufWidth;

		p->OutputFormat.Direction = 0;
		p->OutputFormat.Pitch = p->BufWidth;

		p->Tab[0] = 2*(8);					//Y[0;0] -> Y[0;1]
		p->Tab[1] = 2*(8*p->BufWidth-8);	//Y[0;1] -> Y[1;0]
		p->Tab[2] = 2*(8);					//Y[1;0] -> Y[1;1]
		p->Tab[4] = 2*(p->YToU >> 2);		//U->V
		p->Tab[5] = 2*(0);	

		if (p->Output.Node)
		{
			int Result = p->Output.Node->Set(p->Output.Node,p->Output.No | PACKET_FORMAT,&p->OutputFormat,sizeof(video));
			if (Result != ERR_NONE)
				return Result;
		}
	}
	else
	if (p->Output.Node)
		p->Output.Node->Set(p->Output.Node,p->Output.No | PACKET_FORMAT,NULL,0);

	return ERR_NONE;
}

static int Set(softidct* p, int No, const void* Data, int Size )
{
	int Result = ERR_INVALID_PARAM;
	switch (No)
	{
	case IDCT_SHOW: SETVALUE(p->ShowNext,int,ERR_NONE); break;

	case FLOW_EMPTY:
		p->ShowCurr = -1;
		p->ShowNext = -1;
		return ERR_NONE;

	case IDCT_FORMAT:
		//only block format is supported by softidct
		if (Size == sizeof(int))
			Result = (*(const int*)Data == IDCTFORMAT_BLOCK) ? ERR_NONE:ERR_NOT_SUPPORTED;
		break;

	case IDCT_OUTPUT|PACKET_FORMAT:
		if (!Size || Size == sizeof(video))
		{
			if (!Size) Data = NULL;
			Result = SetFormat(p,(const video*)Data);
		}
		break;

	case IDCT_OUTPUT:
		if (Size == sizeof(pin))
			Result = NodeSetPin(&p->Output,(pin*)Data,&p->OutputFormat,sizeof(video));
		break;

	case IDCT_ROUNDING:
		if (Size == sizeof(bool_t))
			Result = SetRounding(p,*(const bool_t*)Data);
		break;

	case IDCT_BUFFERCOUNT:
		if (Size == sizeof(int))
			Result = SetBufferCount(p,*(const int*)Data,0);
		break;
	}
	return Result;
}

static int Get(softidct* p, int No, void* Data, int Size )
{
	int Result = ERR_INVALID_PARAM;
	switch (No)
	{
	case NODE_ID: GETVALUE(p->Id,int); break;
	case IDCT_NATIVEFORMAT: GETVALUE(IDCTFORMAT_BLOCK,int); break;
	case IDCT_FORMAT: GETVALUE(IDCTFORMAT_BLOCK,int); break;
	case IDCT_OUTPUT|PACKET_FORMAT: GETVALUE(p->OutputFormat,video); break;
	case IDCT_OUTPUT: GETVALUE(p->Output,pin); break;
	case IDCT_ROUNDING: GETVALUE(p->Rounding,bool_t); break;
	case IDCT_BUFFERCOUNT: GETVALUE(p->BufCount - p->LastTemp,int); break;
	case IDCT_SHOW: GETVALUE(p->ShowCurr,int); break;
	}
	return Result;
}

static void Drop(softidct* p)
{
	int i;
	for (i=0;i<p->BufCount;++i)
		p->BufFrameNo[i] = -1;
}

static int Send(softidct* p,packet* Packet,int* Dropped)
{
	int Result;

	if (Packet->DropLevel)
	{
		if (Dropped)
			++(*Dropped);
		return ERR_NONE;
	}

	if (p->ShowNext < 0)
		return ERR_NEED_MORE_DATA;

	if (p->ShowCurr == p->ShowNext)
		p->ShowCurr = -1;

	//DEBUG_MSG2("IDCT Show Next:%d Curr:%d",p->ShowNext,p->ShowCurr);

	Lock(p,p->ShowNext,Packet->Data,NULL,NULL);
	Lock(p,p->ShowCurr,Packet->LastData,NULL,NULL);

	Result = p->Output.Node->Set(p->Output.Node,p->Output.No,Packet,sizeof(packet));

	Unlock(p,p->ShowNext);
	Unlock(p,p->ShowCurr);

	if (Result != ERR_BUFFER_FULL)
		p->ShowCurr = p->ShowNext;

	return Result;
}

static int FrameStart(softidct* p,int FrameNo,int DstNo,int BackNo,int FwdNo,int ShowNo,bool_t Drop)
{
	int OldFrameNo;
#if defined( MIPS64_WIN32 ) 
	// win32 doesn't support 64bit natively. thread and process
	// changes won't save upper part of 64bit registers.
	// we have to disable interrupt and make sure all 1KB
	// code pages are in memory so no page loading is called
	// during code execution 
	TouchPages(p);	
	DisableInterrupts();
#endif

	// ShowCurr will be the next "LastData" in output packet
	// we don't want to lose it's content
	if (p->ShowCurr == DstNo)
	{
		int SwapNo;
		for (SwapNo=0;SwapNo<p->BufCount;++SwapNo)
			if (SwapNo != DstNo && SwapNo != BackNo && SwapNo != FwdNo && SwapNo != ShowNo)
				break;

		if (SwapNo < p->BufCount || SetBufferCount(p,p->BufCount,1)==ERR_NONE)
		{
			//DEBUG_MSG2("IDCT Swap %d,%d",SwapNo,DstNo);

			SwapPByte(&p->Buffer[SwapNo],&p->Buffer[DstNo]);
			SwapPByte(&p->_Buffer[SwapNo],&p->_Buffer[DstNo]);
			SwapInt(&p->BufFrameNo[SwapNo],&p->BufFrameNo[DstNo]);
			SwapBool(&p->BufBorder[SwapNo],&p->BufBorder[DstNo]);

			p->ShowCurr = SwapNo;
		}
	}

	p->ShowNext = ShowNo;

	OldFrameNo = p->BufFrameNo[DstNo];
	p->BufFrameNo[DstNo] = FrameNo;

	p->Dst = p->Buffer[DstNo];
	p->Fwd = NULL;
	p->Back = NULL;

	if (BackNo>=0)
	{
		p->Back = p->Buffer[BackNo];
		if (!p->BufBorder[BackNo])
		{
			p->BufBorder[BackNo] = 1;
			FillEdgeYUV(p,p->Back);
		}
		p->BackMin = p->Back;
		p->BackMax = p->Back + ((p->BufWidth*p->BufHeight*3) >> 1) -8-8*(p->BufWidth>>1);
	}

	if (FwdNo>=0)
	{
		p->Fwd = p->Buffer[FwdNo];
		if (!p->BufBorder[FwdNo])
		{
			p->BufBorder[FwdNo] = 1;
			FillEdgeYUV(p,p->Fwd);
		}
		p->FwdMin = p->Fwd;
		p->FwdMax = p->Fwd + ((p->BufWidth*p->BufHeight*3) >> 1) -8-8*(p->BufWidth>>1);
	}

#ifdef MIPS64_WIN32
	EnableInterrupts();
#endif

	p->BufBorder[DstNo] = 0; //invalidate border for dst

	//mips64 version will set MCompData in Process()
#ifndef MIPS64
	p->VMT.MCompData = p->Fwd ? MCompBackFwd : MCompBack;
#endif

	return OldFrameNo;
} 

static void FrameEnd(softidct* p)
{
}

static int Enum( softidct* p, int No, datadef* Param )
{
	return NodeEnumType(&No,Param,NodeParams,FlowParams,IDCTParams,NULL);
}

static softidct SoftIDCT = 
{
	// VMT
	{
		Enum,
		Get,
		Set,

		Send,
		Drop,
		Lock,
		Unlock,
		FrameStart,
		FrameEnd,
		Copy,
		Process,
		Data,
		NULL,
	},

	//AllCopyBlock[2][2][4]
	{
	   {{ CopyBlock, CopyBlockHor, CopyBlockVer, CopyBlockHorVer },
		{ CopyBlock, CopyBlockHorRound, CopyBlockVerRound, CopyBlockHorVerRound }},
#ifdef MIPS64
	   {{ CopyMBlock, CopyMBlockHor, CopyMBlockVer, CopyMBlockHorVer },
		{ CopyMBlock, CopyMBlockHorRound, CopyMBlockVerRound, CopyMBlockHorVerRound }},
#endif
	},

#if !defined(MIPS64) && !defined(MIPS32)
	//AddBlock[4]
	{AddBlock, AddBlockHor, AddBlockVer, AddBlockHorVer }
#endif
};

void SoftIDCT_Done()
{
	NodeUnRegister((node*)&SoftIDCT);
}

idct* SoftIDCT_Init(int Id)
{
	softidct* p = &SoftIDCT;

	// configure IDCT for current hardware configuration

#if defined(MMX)
	if (!(QueryPlatform(PLATFORM_CAPS) & CAPS_X86_MMX))
		return NULL;

#elif defined(ARM)

	if (QueryPlatform(PLATFORM_ICACHE) >= 32768)
	{
	   p->AllCopyBlock[0][0][1] = FastCopyBlockHor;
	   p->AllCopyBlock[0][0][2] = FastCopyBlockVer;
	   p->AllCopyBlock[0][0][3] = FastCopyBlockHorVer;
	   p->AllCopyBlock[0][1][1] = FastCopyBlockHorRound;
	   p->AllCopyBlock[0][1][2] = FastCopyBlockVerRound;
	   p->AllCopyBlock[0][1][3] = FastCopyBlockHorVerRound;
	   p->AddBlock[0] = FastAddBlock;
	   p->AddBlock[1] = FastAddBlockHor;
	   p->AddBlock[2] = FastAddBlockVer;
	   p->AddBlock[3] = FastAddBlockHorVer;
	}

#elif defined(MIPSVR41XX)
	if (!(QueryPlatform(PLATFORM_CAPS) & CAPS_MIPS_VR41XX))
		return NULL;
#endif

#if defined( MIPS64_WIN32 ) 
	CodeFindPages(*(void**)&p->VMT.Enum,&p->CodeBegin,&p->CodeEnd,&p->CodePage);
#endif

	p->Id = Id;
	p->Tmp = (uint8_t*)(((int)p->_Tmp + 15) & ~15);
	p->MaxCount = MAXBUF;
	p->BufCount = 0;
	p->LastTemp = 0;
	p->ShowCurr = -1;
	p->ShowNext = -1;

	SetRounding(p,0);

	NodeRegister((node*)p,PRI_MAXIMUM);

	return &p->VMT;
}

