/*****************************************************************************
 *
 * This program is free software ; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 *
 * $Id$
 *
 ****************************************************************************/

#include "stdafx.h"

#define MAXSTRING	1024
#define MAXBUF		64

#define BUFSIZE		4096

typedef struct stringtable
{
	const tchar_t* Empty;

	int BufferLeft;
	tchar_t** BufferEnd;
	tchar_t* Buffer[MAXBUF];

	stringdef* TableEnd;
	stringdef Table[MAXSTRING];

	stringdef* TableDefEnd;
	stringdef TableDef[MAXSTRING];

} stringtable;

int StringLang = LANG_DEFAULT; // current language is filled by platform
static stringtable Table;

void String_Init()
{
	stringtable* p = &Table;

	p->Empty = T("");
	p->BufferEnd = p->Buffer;
	p->BufferLeft = 0;
	p->TableEnd = p->Table;
	p->TableDefEnd = p->TableDef;
}

void String_Done()
{
	stringtable* p = &Table;
	tchar_t** i;

	for (i=p->Buffer;i!=p->BufferEnd;++i)
		Free(*i);
}

bool_t StringAdd( int Lang, int Class, int Id, const tchar_t* s )
{
	stringtable* p = &Table;
	stringdef** q;
	tchar_t* d;
	int Size;

	if (Lang == LANG_DEFAULT)
	{
		q = &p->TableDefEnd;
		if (*q == &p->TableDef[MAXSTRING])
			return 0;
	}
	else
	if (Lang == StringLang)
	{
		q = &p->TableEnd;
		if (*q == &p->Table[MAXSTRING])
			return 0;
	}
	else
		return 0;

	Size = tcslen(s)+1;
	if (Size > BUFSIZE)
		Size = BUFSIZE;

	if (Size > p->BufferLeft && p->BufferEnd != &p->Buffer[MAXBUF])
	{
		*p->BufferEnd = (tchar_t*) Alloc(BUFSIZE*sizeof(tchar_t));
		if (*p->BufferEnd)
		{
			p->BufferLeft = BUFSIZE;
			++p->BufferEnd;
		}
	}

	if (Size > p->BufferLeft)
		return 0;

	d = p->BufferEnd[-1]+(BUFSIZE-p->BufferLeft);
	tcsncpy(d,s,Size);
	d[Size-1] = 0;
	p->BufferLeft -= Size;

	(*q)->Class = Class;
	(*q)->Id = Id;
	(*q)->String = d;
	(*q)++;

	return 1;
}

bool_t StringAddMore( int Lang, stringdef* Def )
{
	if (Lang != StringLang && Lang != LANG_DEFAULT)
		return 0;

	for (;Def->Class >= 0;++Def)
		StringAdd(Lang,Def->Class,Def->Id,Def->String);

	return 1;
}

const stringdef* StringEnum( int Class, int No )
{
	stringtable* p = &Table;
	stringdef* i;
	int LastId = -1;
	for (i=p->TableDef;i!=p->TableDefEnd;++i)
	{
		if (i->Class==Class && i->Id!=LastId && No--==0)
			return i;
		LastId = i->Id;
	}
	return NULL;
}

const tchar_t* String( int Class, int Id )
{
	stringtable* p = &Table;
	stringdef* i;
	for (i=p->Table;i!=p->TableEnd;++i)
		if (i->Class==Class && i->Id==Id)
			return i->String;
	for (i=p->TableDef;i!=p->TableDefEnd;++i)
		if (i->Class==Class && i->Id==Id)
			return i->String;
	return p->Empty;
}

const tchar_t* StringDef( int Class, int Id )
{
	stringtable* p = &Table;
	stringdef* i;
	for (i=p->TableDef;i!=p->TableDefEnd;++i)
		if (i->Class==Class && i->Id==Id)
			return i->String;
	for (i=p->Table;i!=p->TableEnd;++i)
		if (i->Class==Class && i->Id==Id)
			return i->String;
	return p->Empty;
}

