/*****************************************************************************
 *
 * This program is free software ; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 *
 * $Id$
 *
 ****************************************************************************/

#include "stdafx.h"

typedef struct systimer
{
	node VMT;

	void* Section;

	bool_t Play;
	fraction Speed;
	fraction SpeedTime;
	tick_t TickStart;
	int TimeRef;

} systimer;

static int Enum( systimer* p, int No, datadef* Param )
{
	return NodeEnumType(&No,Param,NodeParams,TimerParams,NULL);
}

static int Get( systimer* p, int No, void* Data, int Size )
{
	int Result = ERR_INVALID_PARAM;
	switch (No)
	{
	case NODE_ID: GETVALUE(SYSTIMER_ID,int); break;
	case TIMER_PLAY: GETVALUE(p->Play,bool_t); break;
	case TIMER_SPEED: GETVALUE(p->Speed,fraction); break;
	case TIMER_TIME:											
		if (Size == sizeof(tick_t))
		{
			LockEnter(p->Section);
			if (p->Speed.Num==0) // benchmark mode
				*(tick_t*)Data = -1;
			else
			if (p->Play)
				*(tick_t*)Data = p->TickStart + Scale(GetTime()-p->TimeRef,p->SpeedTime.Num,p->SpeedTime.Den);
			else
				*(tick_t*)Data = p->TickStart;
			LockLeave(p->Section);
			Result = ERR_NONE;
		}
		break;
	}
	return Result;
}

static int Set(systimer* p, int No, const void* Data, int Size )
{
	int Result = ERR_INVALID_PARAM;
	switch (No)
	{
	case TIMER_TIME:
		if (Size == sizeof(tick_t))
		{
			LockEnter(p->Section);
			p->TickStart = *(tick_t*)Data;
			p->TimeRef = GetTime();
			LockLeave(p->Section);
			Result = ERR_NONE;
		}
		break;

	case TIMER_SPEED:
		if (Size == sizeof(fraction))
		{
			Result = ERR_NONE;
			if (!EqFrac( &p->Speed, (const fraction*)Data ))
			{
				LockEnter(p->Section);
				if (p->Play)
				{
					int t = GetTime();
					p->TickStart += Scale(t-p->TimeRef,p->SpeedTime.Num,p->SpeedTime.Den);
					p->TimeRef = t;
				}
				p->Speed = *(fraction*)Data;
				p->SpeedTime = p->Speed;
				p->SpeedTime.Num *= TICKSPERSEC;
				p->SpeedTime.Den *= 1000;
				LockLeave(p->Section);
			}
		}
		break;
			
	case TIMER_PLAY:
		if (Size == sizeof(bool_t))
		{
			if (p->Play != *(bool_t*)Data)
			{
				int t;

				LockEnter(p->Section);
				t = GetTime();
				p->Play = *(bool_t*)Data;
				if (!p->Play) // save time
					p->TickStart += Scale(t-p->TimeRef,p->SpeedTime.Num,p->SpeedTime.Den);
				p->TimeRef = t;
				LockLeave(p->Section);
			}
			Result = ERR_NONE;
		}
	}
	return Result;
}

static systimer SysTimer = 
{
	Enum,
	Get,
	Set,
};

void SysTimer_Init()
{
	systimer* p = &SysTimer;
	p->Play = 0;
	p->TickStart = 0;
	p->Speed.Den = p->Speed.Num = 1;
	p->SpeedTime.Num = TICKSPERSEC;
	p->SpeedTime.Den = 1000;
	p->Section = LockCreate();
	NodeRegister(&p->VMT,PRI_MAXIMUM);
}

void SysTimer_Done()
{
	systimer* p = &SysTimer;
	LockDelete(p->Section);
	NodeUnRegister(&p->VMT);
}

