/*****************************************************************************
 *
 * This program is free software ; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 *
 * $Id$
 *
 ****************************************************************************/

#ifndef __TOOLS_H
#define __TOOLS_H

//some helper functions

INLINE bool_t EqInt( const int* a, const int* b ) { return *a == *b; }
INLINE bool_t EqBool( const bool_t* a, const bool_t* b ) { return *a == *b; }
INLINE bool_t EqPtr( const void** a, const void** b ) { return *a == *b; }

DLL bool_t EqPoint( const point* a, const point* b );
DLL bool_t EqRect( const rect* a, const rect* b );
DLL bool_t EqPixel(const pixel* a, const pixel* b );
DLL bool_t EqVideo(const video* a, const video* b );
DLL bool_t EqBlitFX( const blitfx* a, const blitfx* b );
DLL bool_t EqFrac(const fraction* a, const fraction* b );
DLL bool_t EqAudio(const audio* a, const audio* b );

DLL int AnyAlign(rect* DstRect, rect* SrcRect, const blitfx* FX, 
				 int DstAlignSize, int DstAlignPos, 
				 int MinScale, int MaxScale );

DLL void CombineVideo(video* Acc, video* New );
DLL void CombineAudio(audio* Acc, audio* New );
DLL int CombineDir(int SrcDir,int FXDir,int DstDir);
DLL void FillColor(unsigned char* Dst,int DstPitch,int Width,int Height,int BPP,int Value);

DLL void FourCCToString( tchar_t* Out, int FourCC, int OutSize );
DLL int StringToFourCC( const tchar_t* In, bool_t Upper );
DLL void FractionToString( tchar_t* Out, const fraction*, int OutSize, bool_t Percent, int Decimal );
DLL int StringToInt( const tchar_t* In, bool_t Hex );
DLL void IntToString( tchar_t* Out, int, int OutSize, bool_t Hex );
DLL void TickToString( tchar_t* Out, tick_t, int OutSize, bool_t MS, bool_t Extended );
DLL void HotKeyToString( tchar_t* Out, int HotKey, int OutSize );
DLL int SurfaceAlloc(planes, const video* );
DLL void SurfaceFree(planes );
DLL int SurfaceCopy(const video* SrcFormat, const video* DstFormat, 
					const planes Src, planes Dst, blitfx* FX );

DLL void Simplify( fraction*, int MaxNum, int MaxDen );

DLL void SwapPChar( tchar_t**, tchar_t** );
DLL void SwapPByte( uint8_t**, uint8_t** );
DLL void SwapInt( int*, int* );
DLL void SwapLong( long*, long* );
DLL void SwapBool( bool_t*, bool_t* );
DLL void SwapPoint( point* );
DLL void SwapRect( rect* );

DLL void UpperUntil( tchar_t* s, tchar_t Delimiter );

DLL void VirtToPhy(const rect* Virtual, rect* Physical, const video* );
DLL void PhyToVirt(const rect* Physical, rect* Virtual, const video* );

DLL int UpperFourCC(int);
DLL bool_t AnyYUV( const pixel* );
DLL bool_t PlanarYUV( const pixel* );
DLL bool_t PackedYUV( const pixel* );
DLL void FillInfo( pixel* );
DLL int GetBPP( const pixel* );
DLL int BitMaskSize( unsigned int Mask );
DLL int BitMaskPos( unsigned int Mask );
DLL unsigned int RGBToFormat( unsigned int RGB, const pixel* );
DLL void DefaultPitch( video* );
DLL void DefaultRGB( pixel*, int BitCount, 
                     int RBits, int GBits, int BBits,
                     int RGaps, int GGaps, int BGaps );

DLL int IDCTCopy( idct* Src, idct* Dst, int MinBufferCount );

// <datadef table>, ... ,NULL
DLL int NodeEnumType( int* EnumNo, datadef* Out, ... );
// <class>, ... ,0
DLL int NodeClass( void* Data, int Size, ... );
DLL int NodeSetPin( pin* Dst, const pin* Src, void* Format, int FormatLength );

// variable names: Result, Data, Size

#define GETVALUE(Value,Type)								\
		if (Size == sizeof(Type))							\
		{													\
			*(Type*)Data = Value;							\
			Result = ERR_NONE;								\
		}

#define GETVALUECOND(Value,Type,Cond)						\
		if (Size == sizeof(Type) && (Cond))					\
		{													\
			*(Type*)Data = Value;							\
			Result = ERR_NONE;								\
		}

#define GETSTRING(Text)										\
		{													\
			tcsncpy((tchar_t*)Data,Text,Size/sizeof(tchar_t));\
			((tchar_t*)Data)[Size/sizeof(tchar_t)-1]=0;		\
			Result = ERR_NONE;								\
		}

#define SETVALUE(Value,Type,Update)							\
		if (Size == sizeof(Type))							\
		{													\
			Value = *(Type*)Data;							\
			Result = Update;								\
		}

#define SETVALUECMP(Value,Type,Update,EqFunc)				\
		if (Size == sizeof(Type))							\
		{													\
			Result = ERR_NONE;								\
			if (!EqFunc(&Value,(Type*)Data))				\
			{												\
				Value = *(Type*)Data;						\
				Result = Update;							\
			}												\
		}

#define SETSTRING(Text)										\
		{													\
			tcsncpy(Text,(tchar_t*)Data,sizeof(Text)/sizeof(tchar_t));\
			Text[sizeof(Text)/sizeof(tchar_t)-1]=0;			\
			Result = ERR_NONE;								\
		}

typedef struct merge
{
	uint8_t* Data;
	int Pos;
	int OverRead;
	int Allocated;
} merge;

DLL void MergeInit(merge*);
DLL void MergeDone(merge*);
DLL void MergeClear(merge*);
DLL void MergeAdd(merge*, const uint8_t* Ptr, int Length);
DLL void MergeEnd(merge*, int End );

DLL void SplitURL(const tchar_t* URL, tchar_t* Mime, tchar_t* Dir, tchar_t* Name, tchar_t* Ext );
DLL int CheckExts(const tchar_t* URL, const tchar_t* Exts );
DLL void AbsPath(tchar_t* Abs, const tchar_t* Any, const tchar_t* Base );

#define SWAP32(a) \
	( (((a) & 0xFF000000) >> 24) | (((a) & 0x00FF0000) >> 8) |\
	  (((a) & 0x0000FF00) << 8)  | (((a) & 0x000000FF) << 24) )

#define SWAP16(a) \
	( (((a) & 0xFF00) >> 8) | (((a) & 0x00FF) << 8) )

#define SWAP64(a) ((SWAP32(a) << 32) | SWAP32(a>>32))

#ifdef BIG_ENDIAN
#define INT64BE(a) (a)
#define INT64LE(a) SWAP64(a)
#define INT32BE(a) (a)
#define INT32LE(a) SWAP32(a)
#define INT16BE(a) (a)
#define INT16LE(a) SWAP16(a)
#else
#define INT64LE(a) (a)
#define INT64BE(a) SWAP64(a)
#define INT32LE(a) (a)
#define INT32BE(a) SWAP32(a)
#define INT16LE(a) (a)
#define INT16BE(a) SWAP16(a)
#endif

INLINE int Scale(int v,int Num,int Den)
{
	if (Den) 
		return (int)((int64_t)v * Num / Den);
	return 0;
}

#define OFS(name,item) ((int)&(((name*)NULL)->item))

#endif
