/*****************************************************************************
 *
 * This program is free software ; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 *
 * $Id$
 *
 ****************************************************************************/

#ifndef __VIDEO_H
#define __VIDEO_H

//---------------------------------------------------------------
//pixelformat flags
#define	PF_FOURCC		0x0001
#define	PF_PALETTE		0x0002
#define	PF_RGB			0x0004
#define	PF_YUV			0x0008
#define	PF_YUV420		0x0010
#define	PF_INVERTED		0x0020
#define PF_READONLY		0x0040
#define PF_FRAGMENTED	0x0080
#define PF_PIXELDOUBLE	0x0100

//---------------------------------------------------------------
//planar YUV formats
#define	FOURCC_YV12 	FOURCC('Y','V','1','2')
#define	FOURCC_IYUV 	FOURCC('I','Y','U','V')
#define	FOURCC_I420 	FOURCC('I','4','2','0')

//---------------------------------------------------------------
//packed YUV formats
#define	FOURCC_YUY2 	FOURCC('Y','U','Y','2')
#define	FOURCC_YUNV 	FOURCC('Y','U','N','V')
#define	FOURCC_V422 	FOURCC('V','4','2','2')
#define	FOURCC_YUYV 	FOURCC('Y','U','Y','V')
#define	FOURCC_YVYU 	FOURCC('Y','V','Y','U')
#define	FOURCC_UYVY 	FOURCC('U','Y','V','Y')
#define	FOURCC_Y422		FOURCC('Y','4','2','2')
#define	FOURCC_UYNV		FOURCC('U','Y','N','V')
#define	FOURCC_VYUY 	FOURCC('V','Y','U','Y')

//---------------------------------------------------------------
//direction flags
#define	DIR_SWAPXY			0x001
#define	DIR_MIRRORLEFTRIGHT	0x002
#define	DIR_MIRRORUPDOWN	0x004

//---------------------------------------------------------------

#define SCALE_ONE	(1<<16)
#define ASPECT_ONE	(1<<16)

typedef struct rgb
{
	unsigned char r,g,b,a;
} rgb;

typedef struct point
{
	int x;
	int y;
} point;

typedef struct rect
{
	int x;
	int y;
	int Width;
	int Height;
} rect;

typedef struct pixel
{
	int Flags;
	int FourCC;
	int BitCount;
	unsigned int RBitMask;
	unsigned int GBitMask;
	unsigned int BBitMask;
	rgb *Palette;
} pixel;

typedef struct video
{
	int Direction;	// direction flags
	int Aspect;		// 16.16 fixed point
	int ByteRate;
	fraction FrameRate; // frames/sec
	void* User;

	// surface (used by blit)
	int Width;		// phyisical width of surface
	int Height;		// phyisical height of surface
	int Pitch;
	pixel Pixel;

} video;

//---------------------------------------------------------------
// video output (abstract)

#define VOUT_CLASS			FOURCC('V','O','U','T')

// optional: overrides QueryOrientation (can be read without initiazliation) (int)
#define VOUT_ORIENTATION	0x7E
// primary display (bool_t readonly)
#define VOUT_PRIMARY		0x70
// input pin (TYPE_VIDEO)
#define VOUT_INPUT			0x71
// output format (TYPE_VIDEO readonly)
#define VOUT_OUTPUT			0x72
// hardware accelerated idct (idct* readonly)
#define VOUT_ACCEL_IDCT		0x74
// current idct (idct*) (possible idct rotation)
#define VOUT_IDCT			0x7D
// blit fx options (blitfx)
#define VOUT_FX				0x75
// window handle (int)
#define VOUT_WND			0x76
// overlay visible (bool_t)
#define VOUT_VISIBLE		0x77
// gui clipping needed (bool_t)
#define VOUT_CLIPPING		0x7F
// viewport rectangle (rect)
#define VOUT_VIEWPORT		0x78
// actuall output rectangle (rect readonly)
#define VOUT_OUTPUTRECT		0x79
// color key (rgb_t)
#define VOUT_COLORKEY		0x7A
// last frame is not visible anymore (bool_t)
#define VOUT_DIRTY			0x7B
// prerotate portrait input
#define VOUT_AUTOPREROTATE	0x7C
// aspect ratio
#define VOUT_ASPECT			0x81
// reset
#define VOUT_RESET			0x80

extern DLL const datadef VOutputParams[];

#define VOUT_ERROR_SIZE		0x100

//---------------------------------------------------------------
// null vide

#define NULLVIDEO_ID		FOURCC('N','U','L','V')

extern DLL void NullVideo_Init();
extern DLL void NullVideo_Done();

#endif
