/*****************************************************************************
 *
 * This program is free software ; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 *
 * $Id$
 *
 ****************************************************************************/

#include "stdafx.h"

#if defined(_WIN32)

#include <windows.h>

static tchar_t RegBase[256];
static const tchar_t* CmdLine = NULL; 
static const tchar_t* ProgramId = NULL;
static const tchar_t* Version = NULL;

#define NULL_ID			FOURCC('N','U','L','L')

#if defined(_WIN32_WCE)
#define ACCESS_LOAD		0
#define ACCESS_SAVE		0 
#else
#define ACCESS_LOAD		KEY_READ
#define ACCESS_SAVE		KEY_READ | KEY_WRITE | KEY_SET_VALUE 
#endif

bool_t NodeBase( node* p, tchar_t* Base )
{
	char Id[4];

	if (p->Get(p,NODE_ID,Id,sizeof(Id)) != ERR_NONE)
		return 0;

	stprintf(Base,T("%s\\%c%c%c%c"),RegBase,Id[0],Id[1],Id[2],Id[3]);
	return 1;
}

bool_t NodeLoadValue( int Id, void* Data, int Size, int* Type )
{
	bool_t Result = 0;
	HKEY Key;
	tchar_t Name[32];
	stprintf(Name,T("%d"),Id);

	if (RegOpenKeyEx( HKEY_LOCAL_MACHINE, RegBase, 0, ACCESS_LOAD, &Key ) == ERROR_SUCCESS)
	{
		DWORD RegSize = Size;
		DWORD RegType;

		if (RegQueryValueEx(Key, Name, 0, &RegType, (PBYTE)Data, &RegSize) == ERROR_SUCCESS)
			Result = 1;

		if (Type)
			switch (RegType)
			{
			case REG_SZ: *Type = TYPE_STRING; break;
			case REG_DWORD: *Type = TYPE_INT; break;
			default: *Type = 0; break;
			}

		RegCloseKey( Key );
	}

	return Result;
}

void NodeSaveValue( int Id, const void* Data, int Size, int Type )
{
	DWORD Disp;
	HKEY Key;
	tchar_t Name[32];
	stprintf(Name,T("%d"),Id);

	if (RegCreateKeyEx( HKEY_LOCAL_MACHINE, RegBase, 0, NULL, 0, ACCESS_SAVE, NULL, &Key, &Disp ) == ERROR_SUCCESS)
	{
		if (Size)
		{
			DWORD RegType;
			switch (Type)
			{
			case TYPE_STRING: RegType = REG_SZ; break;
			case TYPE_INT: RegType = REG_DWORD; break;
			default: RegType = REG_BINARY; break;
			}

			RegSetValueEx( Key, Name, 0, RegType, (PBYTE)Data, Size );
		}
		else
			RegDeleteValue(Key,Name);
		RegCloseKey( Key );
	}
}

void NodeLoadSetup( node* p )
{
	HKEY Key;
	tchar_t Base[256];
	tchar_t Name[64];

	if (!NodeBase(p,Base))
		return;
	
	if (RegOpenKeyEx( HKEY_LOCAL_MACHINE, Base, 0, ACCESS_LOAD, &Key ) == ERROR_SUCCESS)
	{
		uint8_t Buffer[512];
		DWORD RegSize;
		DWORD RegType;
		node* Ptr;

		int No;
		datadef Type;

		for (No=0;p->Enum(p,No,&Type)==ERR_NONE;++No)
			if ((Type.Flags & DF_SETUP) && !(Type.Flags & (DF_OWNREG|DF_RDONLY)))
			{
				RegSize = sizeof(Buffer);
				stprintf(Name,T("%d"),Type.No);
				if (RegQueryValueEx(Key, Name, 0, &RegType, Buffer, &RegSize) == ERROR_SUCCESS)
				{
					switch (Type.Type)
					{
					case TYPE_BOOL:
					case TYPE_INT:
					case TYPE_HOTKEY:
					case TYPE_FRACTION:
					case TYPE_STRING:
					case TYPE_RECT:
					case TYPE_POINT:
					case TYPE_RGB:
					case TYPE_EQUALIZER:
					case TYPE_TICK:
						p->Set(p,Type.No,Buffer,RegSize);
						break;

					case TYPE_NODE:
						if (RegType == REG_DWORD)
						{
							if (*(int*)Buffer == NULL_ID)
							{
								Ptr = NULL;
								p->Set(p,Type.No,&Ptr,sizeof(Ptr));
							}
							else
							{
								Ptr = NodeFind(*(int*)Buffer);
								if (Ptr)
									p->Set(p,Type.No,&Ptr,sizeof(Ptr));
							}
						}
						break;
					}
				}
			}

		RegCloseKey( Key );
	}
}

void NodeSaveSetup( node* p )
{
	DWORD Disp;
	HKEY Key;
	tchar_t Base[256];
	tchar_t Name[64];

	if (!NodeBase(p,Base))
		return;

	if (RegCreateKeyEx( HKEY_LOCAL_MACHINE, Base, 0, NULL, 0, ACCESS_SAVE, NULL, &Key, &Disp ) == ERROR_SUCCESS)
	{
		uint8_t Buffer[512];
		DWORD RegSize;
		DWORD RegType;
		node* Ptr;
		int Size;
		int No;
		datadef Type;

		RegSize = sizeof(Name);
		for (No=0;RegEnumValue(Key,No,Name,&RegSize,NULL,&RegType,NULL,NULL)!=ERROR_NO_MORE_ITEMS;++No)
		{
			RegDeleteValue(Key,Name);
			RegSize = sizeof(Name);
		}
		
		for (No=0;p->Enum(p,No,&Type)==ERR_NONE;++No)
			if ((Type.Flags & DF_SETUP) && !(Type.Flags & (DF_OWNREG|DF_RDONLY)))
			{
				RegType = RegSize = Size = 0;

				switch (Type.Type)
				{
				case TYPE_BOOL:
					RegType = REG_DWORD;
					RegSize = sizeof(DWORD);
					Size = sizeof(bool_t);
					break;
				case TYPE_TICK:
				case TYPE_INT:
				case TYPE_HOTKEY:
					RegType = REG_DWORD;
					RegSize = Size = sizeof(int);
					break;
				case TYPE_FRACTION:
					RegType = REG_BINARY;
					RegSize = Size = sizeof(fraction);
					break;
				case TYPE_STRING:
					RegType = REG_SZ;
					RegSize = Size = sizeof(Buffer);
					break;
				case TYPE_RECT:
					RegType = REG_BINARY;
					RegSize = Size = sizeof(rect);
					break;
				case TYPE_POINT:
					RegType = REG_BINARY;
					RegSize = Size = sizeof(point);
					break;
				case TYPE_RGB:
					RegType = REG_DWORD;
					RegSize = sizeof(DWORD);
					Size = sizeof(rgb);
					break;
				case TYPE_EQUALIZER:
					RegType = REG_BINARY;
					RegSize = Size = sizeof(equalizer);
					break;
				case TYPE_NODE:
					if (p->Get(p,Type.No,&Ptr,sizeof(Ptr)) == ERR_NONE)
					{
						RegType = REG_DWORD;
						RegSize = sizeof(DWORD);

						if (Ptr)
						{
							if (Ptr->Get(Ptr,NODE_ID,Buffer,sizeof(int)) != ERR_NONE)
								RegSize = 0;
						}
						else
							*(int*)Buffer = NULL_ID;
					}
					break;
				}

				if (Size)
				{
					memset(Buffer,0,RegSize);
					if (p->Get(p,Type.No,Buffer,Size) != ERR_NONE)
						continue;
				}

				if (RegSize)
				{
					stprintf(Name,T("%d"),Type.No);
					RegSetValueEx( Key, Name, 0, RegType, Buffer, RegSize );
				}
			}

		RegCloseKey( Key );
	}
}

typedef struct lib
{
	HMODULE Dll;
	FARPROC Func;
	uint8_t* Min;
	uint8_t* Max;
} lib;

static int NodeLibCount = 0;
static lib NodeLib[256];

const tchar_t* NodeProgramId() { return ProgramId; }
const tchar_t* NodeVersion() { return Version; }
const tchar_t* NodeCmdLine() { return CmdLine; }

void Node_Init( const tchar_t* NewProgramId, const tchar_t* NewVersion, funcany Any, const tchar_t* NewCmdLine )
{
	lib* Lib;

	ProgramId = NewProgramId;
	Version = NewVersion;
	CmdLine = NewCmdLine ? NewCmdLine : T("");

	tcscpy(RegBase,T("SOFTWARE\\"));
	tcscat(RegBase,NewProgramId);

	String_Init();
	Platform_Init();
	Blit_Init();
	PCM_Init();

	Lib = &NodeLib[0];
	Lib->Dll = NULL;
	Lib->Func = (FARPROC)Any;
	Lib->Min = Lib->Max = NULL;

	NodeLibCount = 1;
}

static void FilterString(tchar_t* i)
{
	tchar_t* j=i;
	for (;*i;++i,++j)
	{
		if (i[0]=='\\' && i[1]=='n')
		{
			*j=10;
			++i;
		}
		else
			*j=*i;
	}
	*j=0;
}

static void StringLoad(const tchar_t* Name)
{
	FILE* f = tfopen(Name,T("rb"));
	if (f)
	{
		char s8[512];
		unsigned short s16[512];
		tchar_t s[512];
		tchar_t* p;
		tchar_t* q;
		bool_t Unicode;
		int Lang = 0;
		int No;

		Unicode = fgetc(f)==0xFF;
		Unicode = Unicode && fgetc(f)==0xFE;
		if (!Unicode)
			fseek(f,0,SEEK_SET);

		for (;;)
		{
			if (Unicode)
			{
				fgetws(s16,sizeof(s16)/2,f);
				wcstotcs(s,s16,sizeof(s)/sizeof(tchar_t));
			}
			else
			{
				fgets(s8,sizeof(s8),f);
				strtotcs(s,s8,sizeof(s)/sizeof(tchar_t));
			}
			if (feof(f))
				break;

			p = s;
			while (*p==' ') ++p;
			if (*p==';' || *p==0 || *p==10 || *p==13)
				continue;

			if (*p=='[')
			{
				q = tcschr(p,']');
				if (!q)
					break; // not a valid language file

				*q = 0;

				Lang = StringToFourCC(p+1,1);

				if (Lang == FOURCC('D','E','F','A'))
					Lang = LANG_DEFAULT;
			}
			else
			{
				if (!Lang)
					break;

				if (Lang != LANG_DEFAULT && Lang != StringLang)
					continue;

				q = tcschr(p,'=');
				if (!q)
					break; // not a valid language file

				No = tcslen(s);
				if (No>0 && s[No-1]=='\n') s[--No]=0;
				if (No>0 && s[No-1]=='\r') s[--No]=0;

				*q=0;
				++q;

				if (tcslen(p)>4)
					if (tcslen(p)<8)
						No = StringToFourCC(p+4,1);
					else
					{
						No = StringToInt(p+4,1);
						if (No >= 32768)
							No -= 65536;
					}
				else
					No = 0;

				FilterString(q);
				StringAdd(Lang,StringToFourCC(p,1),No,q);
			}
		}

		fclose(f);
	}
}

static void NodePluginsFind(const tchar_t* Module, tchar_t* Path, const tchar_t* Skip )
{
	WIN32_FIND_DATA FindData;
	tchar_t FindPath[MAXPATH];
	HANDLE Find;
	tchar_t *s;
	int No;
	HMODULE Handle = NULL;

	if (Module)
		Handle = GetModuleHandle(Module);
	GetModuleFileName(Handle,Path,MAXPATH);
	s = tcsrchr(Path,'\\');
	if (s) *s = 0;
	tcscat(Path,T("\\"));

	if (Skip && tcsicmp(Path,Skip)==0)
		return;

	tcscpy(FindPath,Path);
	tcscat(FindPath,T("*.txt"));
	Find = FindFirstFile(FindPath,&FindData);

//DebugMessage(T("TIME:%d"),GetTime());

	if (Find != INVALID_HANDLE_VALUE)
	{
		do
		{
			tcscpy(FindPath,Path);
			tcscat(FindPath,FindData.cFileName);

			DEBUG_MSG1(T("LOAD TXT %s"),FindPath);
			StringLoad(FindPath);
		}
		while (FindNextFile(Find,&FindData));

		FindClose(Find);
	}

	tcscpy(FindPath,Path);
	tcscat(FindPath,T("*.plg"));
	Find = FindFirstFile(FindPath,&FindData);

	if (Find != INVALID_HANDLE_VALUE)
	{
		do
		{
			HMODULE Dll;
#if !defined(_WIN32_WCE)
			int OldMode = SetErrorMode(SEM_NOOPENFILEERRORBOX| SEM_FAILCRITICALERRORS);
#endif
			tcscpy(FindPath,Path);
			tcscat(FindPath,FindData.cFileName);
			Dll = LoadLibrary(FindPath);

			DEBUG_MSG2(T("LOAD PLUGIN %s %d"),FindPath,Dll);

//DebugMessage(T("TIME:%d %s"),GetTime(),FindPath);


#if !defined(_WIN32_WCE)
			SetErrorMode(OldMode);
#endif
			if (Dll)
			{
				FARPROC Register = GetProcAddress(Dll,T("DLLRegister"));
				if (!Register)
					Register = GetProcAddress(Dll,T("_DLLRegister@0"));

				for (No=0;No<NodeLibCount;++No)
					if (NodeLib[No].Dll == Dll)
						break;

				if (No<NodeLibCount)
					Register = NULL;

				if (Register && ((int(*)())Register)() != ERR_NONE)
					Register = NULL;

				if (Register)
				{
					lib* Lib = &NodeLib[NodeLibCount++];

					Lib->Min = Lib->Max = NULL;
					Lib->Dll = Dll;
					Lib->Func = Register;
				}
				else
					FreeLibrary(Dll);
			}

		}
		while (FindNextFile(Find,&FindData));

		FindClose(Find);
	}

//DebugMessage(T("TIME:%d"),GetTime());
}

void NodePlugins()
{
	tchar_t Path[MAXPATH];
	tchar_t Path2[MAXPATH];

	NodePluginsFind(NULL,Path,NULL);
	NodePluginsFind(T("common.plg"),Path2,Path); // just in case somebody using the emulator without setting up pathes
}

void Node_Done()
{
	int No;
	for (No=NodeLibCount-1;No>=1;--No)
	{
		lib* Lib = &NodeLib[No];
		FARPROC UnRegister = GetProcAddress(Lib->Dll,T("DLLUnRegister"));
		if (!UnRegister)
			UnRegister = GetProcAddress(Lib->Dll,T("_DLLUnRegister@0"));

		if (UnRegister)
			((void(*)())UnRegister)();

		FreeLibrary(Lib->Dll);
	}
	NodeLibCount = 0;

	PCM_Done();
	Blit_Done();
	Platform_Done();
	String_Done();
}

bool_t NodeLocate( void* Ptr, tchar_t* OutName, int* OutBase )
{
	lib* Lib;
	int No;
	uint8_t* v = (uint8_t*)Ptr;

	for (No=0;No<NodeLibCount;++No)
	{
		Lib = &NodeLib[No];

		if (!Lib->Min)
		{
			int PageSize;
			CodeFindPages(*(void**)&Lib->Func,&Lib->Min,&Lib->Max,&PageSize);
		}

		if (Lib->Min <= v && Lib->Max > v)
		{
			tchar_t Path[MAXPATH];
			GetModuleFileName(Lib->Dll,Path,MAXPATH);
			SplitURL(Path,NULL,NULL,OutName,OutName);
			*OutBase = v - Lib->Min;
			return 1;
		}
	}
	return 0;
}

#endif
