/*****************************************************************************
 *
 * This program is free software ; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 *
 * $Id$
 *
 ****************************************************************************/

#include "stdafx.h"

#if defined(_WIN32)

#include <windows.h>

#if defined(_MSC_VER) && !defined(_WIN32_WCE)
#include <crtdbg.h>
#endif

bool_t ShowMemoryLow = 1;

static const enumdef EnumProc[] =
{
	{ PLATFORM_ID, CPU_UNKNOWN },
	{ PLATFORM_ID, CPU_X86 },
	{ PLATFORM_ID, CPU_ARM },
	{ PLATFORM_ID, CPU_MIPS },
	{ PLATFORM_ID, CPU_SH3 },

	ENUMDEF_END
};

static const datadef PlatformParams[] = 
{
	{ PLATFORM_ID,			TYPE_CLASS },
	{ PLATFORM_LANG,		TYPE_INT, DF_SETUP|DF_ENUMCLASS|DF_RESTART, LANG_ID },
	{ PLATFORM_TYPE,		TYPE_STRING, DF_SETUP|DF_RDONLY|DF_GAP },
	{ PLATFORM_VER,			TYPE_INT, DF_SETUP|DF_RDONLY|DF_HIDDEN },
	{ PLATFORM_VERSION,		TYPE_STRING, DF_SETUP|DF_RDONLY },
	{ PLATFORM_OEMINFO,		TYPE_STRING, DF_SETUP|DF_RDONLY },
	{ PLATFORM_TYPENO,		TYPE_INT, DF_SETUP|DF_RDONLY|DF_HIDDEN },
	{ PLATFORM_MODEL,		TYPE_INT, DF_SETUP|DF_RDONLY|DF_HIDDEN },
	{ PLATFORM_CAPS,		TYPE_INT, DF_SETUP|DF_RDONLY|DF_HIDDEN|DF_HEX },
	{ PLATFORM_PROCESSOR,	TYPE_INT, DF_SETUP|DF_RDONLY|DF_ENUM, (int)&EnumProc },
	{ PLATFORM_ICACHE,		TYPE_INT, DF_SETUP|DF_RDONLY|DF_HIDDEN },
	{ PLATFORM_DCACHE,		TYPE_INT, DF_SETUP|DF_RDONLY|DF_HIDDEN },
	{ PLATFORM_NOBACKLIGHT, TYPE_BOOL, DF_SETUP|DF_CHECKLIST|DF_GAP },
	{ PLATFORM_OLDSHELL,	TYPE_BOOL, DF_SETUP|DF_RESTART|DF_CHECKLIST },

	DATADEF_END
};

typedef struct platform
{
	node VMT;
	int Processor;
	int Model;
	int Caps;
	int ICache;
	int DCache;
	int Type;
	int Ver;
	int OldShell;
	bool_t NoBackLight;
	tchar_t OemInfo[256];
	tchar_t PlatformType[256];
	int64_t Tick0;

} platform;

#if defined(_WIN32_WCE)

#ifndef DISP_CHANGE_SUCCESSFUL
#define DISP_CHANGE_SUCCESSFUL 0
#endif
#ifndef CDS_TEST
#define CDS_TEST            0x00000002
#endif
#ifndef DM_DISPLAYORIENTATION
#define DM_DISPLAYORIENTATION 0x00800000L
#endif
#ifndef DM_DISPLAYQUERYORIENTATION 
#define DM_DISPLAYQUERYORIENTATION 0x01000000L
#endif
#ifndef DMDO_0
#define DMDO_0      0
#endif
#ifndef DMDO_90
#define DMDO_90     1
#endif
#ifndef DMDO_180
#define DMDO_180    2
#endif
#ifndef DMDO_270
#define DMDO_270    4
#endif

#define SPI_GETOEMINFO				258
#define SPI_GETPLATFORMTYPE			257
#define CACHE_SYNC_INSTRUCTIONS		2

static BOOL (WINAPI* FuncSHGetDocumentsFolder)(LPCTSTR,LPTSTR) = NULL;
static BOOL (WINAPI* FuncSHGetSpecialFolderPath)(HWND,LPTSTR,int,BOOL) = NULL;
static LONG (WINAPI* FuncChangeDisplaySettingsEx)(LPCTSTR,LPDEVMODE,HWND,DWORD,LPVOID) = NULL;
static void (WINAPI* FuncCacheSync)(DWORD) = NULL;
static BOOL (WINAPI* FuncVirtualCopy)(LPVOID, LPVOID, DWORD, DWORD) = NULL;
static BOOL (WINAPI* FuncCeSetThreadQuantum)(HANDLE, DWORD ) = NULL;
static BOOL (WINAPI* FuncSetKMode)(BOOL) = NULL;
static BOOL (WINAPI* FuncRedrawWindow)(HWND,CONST RECT*,HRGN,UINT) = NULL;
static void (WINAPI* FuncSHIdleTimerReset)() = NULL;

static HMODULE CoreDLL = NULL;
static HMODULE CEShellDLL = NULL;

#define ACCESS_LOAD		0
#define ACCESS_SAVE		0 

#else

#define ACCESS_LOAD		KEY_READ
#define ACCESS_SAVE		KEY_READ | KEY_WRITE | KEY_SET_VALUE 

#endif

static HMODULE AygShellDLL = NULL;
static tchar_t DocPath[MAXPATH] = T("");
static tchar_t SysPath[MAXPATH] = T("");

static const tchar_t RegBacklight[] = T("ControlPanel\\Backlight");
static const tchar_t RegBatteryTimeout[] = T("BatteryTimeout");
static const tchar_t RegACTimeout[] = T("ACTimeout");
static const tchar_t RegScreenSaver[] = T("ControlPanel\\ScreenSaver");
static const tchar_t RegMode[] = T("Mode");
static const tchar_t RegPower[] = T("ControlPanel\\Power");
static const tchar_t RegDisplay[] = T("Display");

static const tchar_t RegCASIOBacklight[] = T("Drivers\\CASIO\\BackLight");
static const tchar_t RegTimeoutBattery[] = T("TimeoutBattery");
static const tchar_t RegTimeoutExPower[] = T("TimeoutExPower");

#define GETVFRAMEPHYSICAL			6144
#define GETVFRAMELEN				6145
#define DBGDRIVERSTAT				6146
#define SETPOWERMANAGEMENT			6147
#define GETPOWERMANAGEMENT			6148

typedef struct VIDEO_POWER_MANAGEMENT {
    ULONG Length;
    ULONG DPMSVersion;
    ULONG PowerState;
} VIDEO_POWER_MANAGEMENT;

extern int64_t STDCALL GetCpuId();

static rgb Palette[256];

static bool_t DisplayPower = 1;
static int Orientation = -1;
static void* VirtualBase = NULL;
static int PhyBase = 0;
static int PhySize = 0;

#if defined(MIPS)
int64_t STDCALL GetCpuId()
{
	int v;
	__asm(	".set noreorder;"
			"mfc0	$2, $15;"	//prid
			"nop;"
			"nop;"
			"nop;"
			"nop;"
			"sw		$2,0($4);"
			".set reorder;", &v );
	return v;
}
#elif !defined(ARM) && !defined(_M_IX86)
int64_t STDCALL GetCpuId() { return 0; }
#endif

#if defined( MIPS)
int GetTime()
{
	SYSTEMTIME t;
	GetSystemTime(&t);
	return (t.wDay*24+t.wHour)*60*60*1000+(t.wMinute*60+t.wSecond)*1000+t.wMilliseconds;
}
#elif defined(_WIN32_WCE)
int GetTime()
{
	return GetTickCount();
}
#else
int GetTime()
{
	return timeGetTime();
}
#endif

static int Enum( platform* p, int No, datadef* Param )
{
	int Result = NodeEnumType(&No,Param,NodeParams,PlatformParams,NULL);
	if (Result == ERR_NONE && Param->No == PLATFORM_OLDSHELL && (!AygShellDLL || p->Type == TYPE_SMARTPHONE))
		Param->Flags |= DF_HIDDEN;
	return Result;
}

static int Get( platform* p, int No, void* Data, int Size )
{
	int Result = ERR_INVALID_PARAM;
	switch (No)
	{
	case NODE_ID: GETVALUE(PLATFORM_ID,int); break;
	case PLATFORM_LANG: GETVALUE(StringLang,int); break;
	case PLATFORM_OLDSHELL: GETVALUE(p->OldShell,int); break;
	case PLATFORM_NOBACKLIGHT: GETVALUE(p->NoBackLight,bool_t); break;
	case PLATFORM_PROCESSOR: GETVALUE(p->Processor,int); break;
	case PLATFORM_TYPE:	GETSTRING(p->PlatformType[0] ? p->PlatformType : T("Windows")); break;
	case PLATFORM_OEMINFO: if (p->OemInfo[0]) GETSTRING(p->OemInfo); break;
	case PLATFORM_VER: GETVALUE(p->Ver,int); break;
	case PLATFORM_TYPENO: GETVALUE(p->Type,int); break;
	case PLATFORM_MODEL: GETVALUE(p->Model,int); break;
	case PLATFORM_CAPS: GETVALUE(p->Caps,int); break;
	case PLATFORM_ICACHE: GETVALUECOND(p->ICache,int,p->ICache>0); break;
	case PLATFORM_DCACHE: GETVALUECOND(p->DCache,int,p->DCache>0); break;
	case PLATFORM_VERSION: 
		stprintf((tchar_t*)Data,T("%d.%02d"),p->Ver/100,p->Ver%100);
		Result = ERR_NONE;
		break;
	}
	return Result;
}

static int Set( platform* p, int No, const void* Data, int Size )
{
	int Result = ERR_INVALID_PARAM;
	switch (No)
	{
	case PLATFORM_LANG: SETVALUE(StringLang,int,ERR_NONE); break;
	case PLATFORM_OLDSHELL: SETVALUE(p->OldShell,int,ERR_NONE); break;
	case PLATFORM_NOBACKLIGHT: SETVALUE(p->NoBackLight,bool_t,ERR_NONE); break;
	}
	return Result;
}

static platform Platform = 
{
	Enum,
	Get,
	Set,
};

int QueryPlatform( int Param )
{
	int Value = 0;
	Platform.VMT.Get(&Platform,Param,&Value,sizeof(Value));
	return Value;
}

void Platform_Init()
{
#if defined(_WIN32_WCE)
	tchar_t OemUpper[256];
	RECT WorkArea;
#endif

	char Lang8[16];
	tchar_t Lang[16];
	OSVERSIONINFO Ver;
	platform* p = &Platform;
	int64_t CpuId = 0;

#if defined(_DEBUG) && defined(_MSC_VER) && !defined(_WIN32_WCE)
	int Flag = _CrtSetDbgFlag( _CRTDBG_REPORT_FLAG );
	Flag |= _CRTDBG_LEAK_CHECK_DF;
	//Flag |= _CRTDBG_CHECK_ALWAYS_DF;
	_CrtSetDbgFlag( Flag );
#endif

	Ver.dwOSVersionInfoSize = sizeof(Ver);
	GetVersionEx(&Ver);

	p->Type = TYPE_UNKNOWN;
	p->Ver = Ver.dwMajorVersion*100 + Ver.dwMinorVersion;
	p->OemInfo[0] = 0;
	p->PlatformType[0] = 0;

#if defined(_WIN32_WCE)
	AygShellDLL = LoadLibrary(T("aygshell.dll"));
	if (AygShellDLL)
		*(FARPROC*)&FuncSHIdleTimerReset = GetProcAddress(AygShellDLL,MAKEINTRESOURCE(2006));

	p->OldShell = AygShellDLL==NULL && (Ver.dwMajorVersion<3);

	CoreDLL = LoadLibrary(T("coredll.dll"));
	if (CoreDLL)
	{
		*(FARPROC*)&FuncChangeDisplaySettingsEx = GetProcAddress(CoreDLL,T("ChangeDisplaySettingsEx"));
		*(FARPROC*)&FuncCacheSync = GetProcAddress(CoreDLL,T("CacheSync"));
		*(FARPROC*)&FuncVirtualCopy = GetProcAddress(CoreDLL,T("VirtualCopy"));
		*(FARPROC*)&FuncCeSetThreadQuantum = GetProcAddress(CoreDLL,T("CeSetThreadQuantum"));
		*(FARPROC*)&FuncSetKMode = GetProcAddress(CoreDLL,T("SetKMode"));
		*(FARPROC*)&FuncRedrawWindow = GetProcAddress(CoreDLL,T("RedrawWindow"));
		*(FARPROC*)&FuncSHGetSpecialFolderPath = GetProcAddress(CoreDLL,T("SHGetSpecialFolderPath"));
	}

	CEShellDLL = LoadLibrary(T("ceshell.dll"));
	if (CEShellDLL)
		*(FARPROC*)&FuncSHGetDocumentsFolder = GetProcAddress(CEShellDLL,MAKEINTRESOURCE(75));

	if (FuncSHGetSpecialFolderPath)
		FuncSHGetSpecialFolderPath(NULL,SysPath,0x24/*CSIDL_WINDOWS*/,FALSE);

	if (!FuncSHGetDocumentsFolder || !FuncSHGetDocumentsFolder(T("\\"),DocPath))
		DocPath[0] = 0;

	if (SystemParametersInfo(SPI_GETPLATFORMTYPE,sizeof(p->PlatformType),p->PlatformType,0)!=0)
	{
		if (tcsicmp(p->PlatformType,T("PocketPC"))==0)
			p->Type = TYPE_POCKETPC;
		else
		if (tcsicmp(p->PlatformType,T("Smartphone"))==0)
			p->Type = TYPE_SMARTPHONE;
	}
	else
	if (GetLastError()==ERROR_ACCESS_DENIED)
		p->Type = TYPE_SMARTPHONE;

	SystemParametersInfo(SPI_GETOEMINFO,sizeof(p->OemInfo),p->OemInfo,0);

	// detect fake navigator bar and use old style menu
	if (p->Type != TYPE_SMARTPHONE && AygShellDLL && 
		SystemParametersInfo(SPI_GETWORKAREA,0,&WorkArea,0) && WorkArea.top==0)
		p->OldShell = 1; 

	tcscpy(OemUpper,p->OemInfo);
	tcsupr(OemUpper);
#endif

	__try {	
		CpuId = GetCpuId(); 
	} __except (1) {}

	p->Processor = CPU_UNKNOWN;
	p->Model = MODEL_DEFAULT;

#if defined(ARM)
	p->Processor = CPU_ARM;

	p->ICache = 512 << (int)((CpuId >> (32+6)) & 7);
	p->DCache = 512 << (int)((CpuId >> (32+18)) & 7);

	if ((CpuId & 0xFFFF0000) == 0x69050000) //intel arm5e
		p->Caps |= CAPS_ARM_5E;

#if defined(_WIN32_WCE)

	if (tcsstr(OemUpper,T("PW10B1")) ||
		tcsstr(OemUpper,T("PW10A1")))
	{
		p->Model = MODEL_O2_XDA;
		p->Caps |= CAPS_ONLY12BITRGB;
	}
	else
	if (tcsstr(OemUpper,T("Compal P6C")))
		p->Model = MODEL_AMIGO_600C;
	else
	if (tcsstr(OemUpper,T("Pocket PC J710")))
		p->Model = MODEL_CASIO_E200;
	else
	if (tcsstr(OemUpper,T("JORNADA")))
	{
		p->Caps |= CAPS_LANDSCAPE;
		if (tcsstr(OemUpper,T("680")))
			p->Model = MODEL_JORNADA_680;
		else
		if (tcsstr(OemUpper,T("690")))
			p->Model = MODEL_JORNADA_690;
		else
		if (tcsstr(OemUpper,T("710")))
			p->Model = MODEL_JORNADA_710;
		else
		if (tcsstr(OemUpper,T("720")))
			p->Model = MODEL_JORNADA_720;
	}
	else
	if (tcsstr(OemUpper,T("IPAQ")))
	{
		p->Caps |= CAPS_PORTRAIT;
		if (tcsstr(OemUpper,T("H3600")))
		{
			p->Model = MODEL_IPAQ_H3600;
			p->Caps |= CAPS_ONLY12BITRGB;
		}
		else
		if (tcsstr(OemUpper,T("H3700")))
		{
			p->Model = MODEL_IPAQ_H3700;
			p->Caps |= CAPS_ONLY12BITRGB;
		}
		else
		if (tcsstr(OemUpper,T("H3800")))
		{
			p->Model = MODEL_IPAQ_H3800;
		}
		else
		if (tcsstr(OemUpper,T("H3900")))
			p->Model = MODEL_IPAQ_H3900;
	}
	else
	if (tcsstr(OemUpper,T("TOSHIBA")))
	{
		p->Caps |= CAPS_PORTRAIT;
		if (tcsstr(OemUpper,T("E740")))
			p->Model = MODEL_TOSHIBA_E740;
		else
		if (tcsstr(OemUpper,T("E750")))
			p->Model = MODEL_TOSHIBA_E750;
		else
		if (tcsstr(OemUpper,T("E80")))
			p->Model = MODEL_TOSHIBA_E800;
	}
#endif

#elif defined(MIPS)
	p->Processor = CPU_MIPS;

	if (((CpuId >> 8) & 255) == 0x0c)
		p->Caps |= CAPS_MIPS_VR41XX;

#if defined(_WIN32_WCE)
	if (tcsstr(OemUpper,T("POCKET PC")))
	{
		p->Caps |= CAPS_PORTRAIT;
		if (tcsstr(OemUpper,T(" J67")))
			p->Model = MODEL_CASIO_E125;
		else
		if (tcsstr(OemUpper,T(" J76")))
			p->Model = MODEL_CASIO_EM500;
		else
		if (tcsstr(OemUpper,T(" J58")))
			p->Model = MODEL_CASIO_E115;
		else
		if (tcsstr(OemUpper,T(" J53")) || tcsstr(OemUpper,T(" JX53")))
			p->Model = MODEL_CASIO_E105;
	}

	if (tcsstr(OemUpper,T("JVC")))
	{
		if (tcsstr(OemUpper,T("C33")))
			p->Model = MODEL_JVC_C33;
	}
#endif

#elif defined(SH3)
	p->Processor = CPU_SH3;

#if defined(_WIN32_WCE)
	if (tcsstr(OemUpper,T("JORNADA")))
	{
		if (tcsstr(OemUpper,T(" 43")))
		{
			p->Caps |= CAPS_ONLY12BITRGB;
		}
		if (tcsstr(OemUpper,T(" 54")))
		{
			p->Model = MODEL_JORNADA_540;
			p->Caps |= CAPS_ONLY12BITRGB | CAPS_PORTRAIT;
		}
	}
#endif

#elif defined(_M_IX86)
	p->Processor = CPU_X86;

	if (CpuId & 0x00800000)
		p->Caps |= CAPS_X86_MMX;
	if (CpuId & 0x02000000)
		p->Caps |= CAPS_X86_SSE;
#endif

	if (GetLocaleInfo(LOCALE_USER_DEFAULT,LOCALE_SABBREVLANGNAME,Lang,sizeof(Lang)/sizeof(tchar_t)) >= 2)
	{
		tcsupr(Lang);
		tcstostr(Lang8,Lang,sizeof(Lang8));
		StringLang = FOURCC(Lang8[0],Lang8[1],'_','_');
	}

	NodeRegister(&p->VMT,PRI_MAXIMUM+650);

	BacklightTimeout(0); // restore backlight settings (example if last time crashed)
}

void Platform_Done()
{
	NodeUnRegister(&Platform.VMT);

	UnlockFrameBuffer();
	SetDisplayPower(1);

#if defined(_WIN32_WCE)
	if (AygShellDLL) FreeLibrary(AygShellDLL);
	if (CoreDLL) FreeLibrary(CoreDLL);
	if (CEShellDLL) FreeLibrary(CEShellDLL);
#endif
}

bool_t IsOrientationChanged()
{
	int Old = Orientation;
	Orientation = -1;
	return QueryOrientation() != Old;
}

int QueryOrientation()
{
#if defined(_WIN32_WCE)
	if (Orientation < 0)
	{
		HKEY Key;
		char Buffer[256];
		DEVMODE* Mode = (DEVMODE*)Buffer;
		int i,No;
		node* Node;

		Mode->dmSize = 192;
		Mode->dmFields = DM_DISPLAYQUERYORIENTATION;

		if (Platform.Ver >= 421 && // we don't trust this method on pre wm2003se systems
			FuncChangeDisplaySettingsEx &&
			FuncChangeDisplaySettingsEx(NULL, Mode, NULL, CDS_TEST, NULL) == DISP_CHANGE_SUCCESSFUL)
		{
			Mode->dmFields = DM_DISPLAYORIENTATION;
			FuncChangeDisplaySettingsEx(NULL, Mode, NULL, CDS_TEST, NULL);

			switch ((&Mode->dmDisplayFrequency)[1]) //(Mode->dmDisplayOrientation)
			{
			case DMDO_0: Orientation = 0; break;
			case DMDO_90: Orientation = DIR_SWAPXY | DIR_MIRRORUPDOWN; break;
			case DMDO_270: Orientation = DIR_SWAPXY | DIR_MIRRORLEFTRIGHT; break;
			case DMDO_180: Orientation = DIR_MIRRORUPDOWN | DIR_MIRRORLEFTRIGHT; break;
			}
		}

		if (Orientation < 0)
			for (No=0;(Node = NodeEnum(VOUT_CLASS,No))!=NULL;++No)
				if (Node->Get(Node,VOUT_ORIENTATION,&i,sizeof(int))==ERR_NONE)
				{
					Orientation = i;
					break;
				}

		if (Orientation < 0 && RegOpenKeyEx( HKEY_LOCAL_MACHINE, T("System\\GDI\\ROTATION"), 0, ACCESS_LOAD, &Key ) == ERROR_SUCCESS)
		{
			DWORD Value;
			DWORD RegSize = sizeof(Value);
			DWORD RegType;

			if (RegQueryValueEx(Key, T("Angle"), 0, &RegType, (LPBYTE) &Value, &RegSize) == ERROR_SUCCESS)
				switch (Value)
				{
				case 0: Orientation = 0; break;
				case 90: Orientation = DIR_SWAPXY | DIR_MIRRORUPDOWN; break;
				case 270: Orientation = DIR_SWAPXY | DIR_MIRRORLEFTRIGHT; break;
				case 180: Orientation = DIR_MIRRORUPDOWN | DIR_MIRRORLEFTRIGHT; break;
				}

			RegCloseKey( Key );
		}

		if (Orientation < 0)
			Orientation = 0;
	}
#else
	Orientation = 0;
#endif
	return Orientation;
}

void QueryDesktop( video* p )
{
	HDC DC = GetDC(0);

	memset(p,0,sizeof(video));

	p->Width = GetDeviceCaps(DC,HORZRES);
	p->Height = GetDeviceCaps(DC,VERTRES);
	p->Aspect = ASPECT_ONE;
	p->Direction = 0;

	if (GetDeviceCaps(DC,BITSPIXEL)*GetDeviceCaps(DC,PLANES) <= 8)
	{
		int n,i;

		p->Pixel.Flags = PF_PALETTE;
		p->Pixel.BitCount = 1;

		n=GetSystemPaletteEntries(DC,0,256,NULL);
		for (i=2;i<n;i=i*i)
			p->Pixel.BitCount <<= 1;

		if (n>16) //4,2,1 bits are grayscale (default palette required)
		{
			GetSystemPaletteEntries(DC,0,256,(PALETTEENTRY*)Palette);
			p->Pixel.Palette = Palette;
		}
	}
	else
	if (QueryPlatform(PLATFORM_CAPS) & CAPS_ONLY12BITRGB)
		DefaultRGB(&p->Pixel,GetDeviceCaps(DC,BITSPIXEL),4,4,4,1,2,1);
	else
	{
		int i;
		int RBits = 4;
		int GBits = 4;
		int BBits = 4;
		int BitCount = GetDeviceCaps(DC,BITSPIXEL);

		COLORREF Old = GetPixel(DC,0,0);
		for (i=3;i>=0;--i)
		{
			COLORREF c = SetPixel(DC,0,0,0x010101<<i);
			if (c == (COLORREF)-1)
			{
				if (BitCount > 16)
					RBits = GBits = BBits = 8;
				else
				{
					RBits = BBits = 5; GBits = 6;
				}
				break;
			}
			if (c & 0xFF)	  RBits++;
			if (c & 0xFF00)	  GBits++;
			if (c & 0xFF0000) BBits++;
		}
		SetPixel(DC,0,0,Old);

		DefaultRGB(&p->Pixel,BitCount,RBits,GBits,BBits,0,0,0);
	}

	ReleaseDC(0,DC);
}

static void* Physical( int Addr, int Size )
{
#if defined(_WIN32_WCE)
	if (FuncVirtualCopy)
	{
		if (VirtualBase && PhyBase==Addr && PhySize==Size)
			return VirtualBase;

		PhyBase = Addr;
		PhySize = Size;
		Size = (Size + 0xFFFF) & ~0xFFFF;
		VirtualBase = VirtualAlloc(0, Size, MEM_RESERVE, PAGE_NOACCESS);
		if (VirtualBase && !FuncVirtualCopy(VirtualBase,(LPVOID)(Addr >> 8), Size,
			PAGE_READWRITE | PAGE_NOCACHE | PAGE_PHYSICAL))
		{
			VirtualFree(VirtualBase,0,MEM_RELEASE);
			VirtualBase = NULL;
		}
		return VirtualBase;
	}
#endif
	return NULL;
}

bool_t KernelMode(bool_t v)
{
#if defined(_WIN32_WCE)
	if (FuncSetKMode)
		return FuncSetKMode(v);
#endif
	return 0;
}

bool_t GetDisplayPower()
{
	return DisplayPower;
}

void SetDisplayPower(bool_t State)
{
	VIDEO_POWER_MANAGEMENT VPM;
	HDC DC;

	if (DisplayPower != State)
	{
		DisplayPower = State;

		DC = GetDC(NULL);
		VPM.Length = sizeof(VPM);
		VPM.DPMSVersion = 1;
		VPM.PowerState = State ? 1:4;

		ExtEscape(DC, SETPOWERMANAGEMENT, sizeof(VPM), (LPCSTR) &VPM, 0, NULL);
		ReleaseDC(NULL, DC);
	}
}

static void ChangeRegEntry( bool_t CurrentUser, const tchar_t* Reg, const tchar_t* Name, bool_t State, int NewValue, int BackupId ) 
{
	HKEY Key = 0;
	DWORD Size;
	DWORD Value;

	if (RegOpenKeyEx(CurrentUser ? HKEY_CURRENT_USER:HKEY_LOCAL_MACHINE,Reg, 0, ACCESS_SAVE, &Key ) == ERROR_SUCCESS) 
	{
		if (State)
		{
			// save old value 
			Size = sizeof(Value);
			if (RegQueryValueEx( Key, Name,NULL,NULL,(LPBYTE)&Value,&Size) == ERROR_SUCCESS)
				NodeSaveValue(BackupId,&Value,sizeof(Value),TYPE_INT);

			// set no screensaver
			Value = NewValue;
			Size = sizeof(Value);
			RegSetValueEx( Key, Name,0,REG_DWORD,(LPBYTE)&Value,Size);
		}
		else
		if (NodeLoadValue(BackupId,&Value,sizeof(Value),NULL))
		{
			// restore value (and delete old value)
			Size = sizeof(Value);
			RegSetValueEx( Key, Name,0,REG_DWORD,(LPBYTE)&Value,Size);

			NodeSaveValue(BackupId,NULL,0,TYPE_INT);
		}

		RegCloseKey( Key );
	}
}

void IdleTimerReset()
{
#if defined(_WIN32_WCE)
	SystemIdleTimerReset();
	if (FuncSHIdleTimerReset)
		FuncSHIdleTimerReset();
#endif
}

void BacklightTimeout(bool_t Disable)
{
	static bool_t State = -1;

	if (Platform.NoBackLight)
		Disable = 0;

	if (Disable != State)
	{
		HANDLE Handle;
		int Value;

		State = Disable;

#ifdef MIPS
		if (Platform.Model == MODEL_CASIO_E125 ||
			Platform.Model == MODEL_CASIO_EM500 ||
			Platform.Model == MODEL_CASIO_E115 ||
			Platform.Model == MODEL_CASIO_E105)
		{
			Value = 0;
			ChangeRegEntry(0,RegCASIOBacklight,RegTimeoutBattery,State,Value,PLATFORM_BATTERYTIMEOUT);
			ChangeRegEntry(0,RegCASIOBacklight,RegTimeoutExPower,State,Value,PLATFORM_ACTIMEOUT);
		}
		else
#endif
		{
			Value = 0x7FFFFFFF;
			ChangeRegEntry(1,RegScreenSaver,RegMode,State,1,PLATFORM_SCREENSAVER);
			ChangeRegEntry(1,RegBacklight,RegBatteryTimeout,State,Value,PLATFORM_BATTERYTIMEOUT);
			ChangeRegEntry(1,RegBacklight,RegACTimeout,State,Value,PLATFORM_ACTIMEOUT);
			ChangeRegEntry(1,RegPower,RegDisplay,State,-1,PLATFORM_DISPPOWER);
		}
	
		Handle = CreateEvent( NULL, FALSE, FALSE, T("BackLightChangeEvent") );
		if (Handle) 
		{
			SetEvent(Handle);
			CloseHandle(Handle);
		}
	}
}

void* LockFrameBuffer( video* p )
{
	void* Ptr = NULL;

	memset(p,0,sizeof(video));

	//set default values (for most of the models)
	p->Width = 240;
	p->Height = 320;
	p->Pitch = 480;
	p->Aspect = ASPECT_ONE;
	p->Direction = 0;

	if (QueryPlatform(PLATFORM_CAPS) & CAPS_ONLY12BITRGB)
		DefaultRGB(&p->Pixel,16,4,4,4,1,2,1);
	else
		DefaultRGB(&p->Pixel,16,5,6,5,0,0,0);

	switch (QueryPlatform(PLATFORM_MODEL))
	{
	case MODEL_IPAQ_H3600:
	case MODEL_IPAQ_H3700:
		p->Direction = DIR_SWAPXY | DIR_MIRRORLEFTRIGHT;
		Ptr = (void*)0xAC050020;
		p->Width = 320;
		p->Height = 240;
		p->Pitch = 640;
		break;

	case MODEL_IPAQ_H3800:
		p->Direction = DIR_SWAPXY | DIR_MIRRORUPDOWN;
		Ptr = (void*)0xAC050020;
		p->Width = 320;
		p->Height = 240;
		p->Pitch = 640;
		break;

	case MODEL_CASIO_E105:
	case MODEL_CASIO_E115:
		Ptr = (void*)0xAA200000;
		p->Pitch = 512;
		break;

	case MODEL_JORNADA_710:
	case MODEL_JORNADA_720:
		Ptr = Physical(0x48200000,240*1280);
		p->Width = 640;
		p->Height = 240;
		p->Pitch = 1280;
		break;

	case MODEL_JORNADA_680:
	case MODEL_JORNADA_690: 
		Ptr = (void*)0xB2000000; //only japan???
		p->Width = 640;
		p->Height = 240;
		p->Pitch = 1280;
		break;

	case MODEL_MOBILEGEAR2_550:
	case MODEL_MOBILEGEAR2_450:
	case MODEL_MOBILEGEAR2_530:
	case MODEL_MOBILEGEAR2_430:
	case MODEL_MOBILEPRO_780:
		Ptr = (void*)0xAA180100;
		p->Width = 640;
		p->Height = 240;
		p->Pitch = 1280;
		break;

	case MODEL_MOBILEPRO_770:
		Ptr = (void*)0xAA000000;
		p->Width = 640;
		p->Height = 240;
		p->Pitch = 1600;
		break;

	case MODEL_JVC_C33:
		Ptr = (void*)0x445C0000;
		p->Width = 1024;
		p->Height = 600;
		p->Pitch = 2048;
		break;

	case MODEL_SIGMARION:
		Ptr = (void*)0xAA000000;
		p->Width = 640;
		p->Height = 240;
		p->Pitch = 1280;
		break;

	case MODEL_SIGMARION2:
		Ptr = (void*)0xB0800000;
		p->Width = 640;
		p->Height = 240;
		p->Pitch = 1280;
		break;

	case MODEL_SIGMARION3:
		Ptr = Physical(0x14800000,480*1600);
		p->Width = 800;
		p->Height = 480;
		p->Pitch = 1600;
		break;

	case MODEL_INTERMEC_6651:
		Ptr = Physical(0x6C000000,480*1600);
		p->Width = 800;
		p->Height = 480;
		p->Pitch = 1600;
		break;
	}

	return Ptr;
}

void UnlockFrameBuffer()
{
	if (VirtualBase)
	{
		VirtualFree(VirtualBase,0,MEM_RELEASE);
		VirtualBase = NULL;
	}
}

void* CodeAlloc(int Size)
{
	return VirtualAlloc(NULL,Size,MEM_COMMIT,PAGE_EXECUTE_READ);
}

void CodeFree(void* Code,int Size)
{
	VirtualFree(Code,Size,MEM_DECOMMIT);
	VirtualFree(Code,0,MEM_RELEASE);
}

void CodeLock(void* Code,int Size)
{
	DWORD Protect;
	VirtualProtect(Code,Size,PAGE_READWRITE|PAGE_NOCACHE,&Protect);
}

void CodeUnlock(void* Code,int Size)
{
	DWORD Protect;
	VirtualProtect(Code,Size,PAGE_EXECUTE_READ,&Protect);
#if defined(_WIN32_WCE)
	if (FuncCacheSync)
		FuncCacheSync(CACHE_SYNC_INSTRUCTIONS);
#endif
}

void CodeFindPages(void* Ptr,uint8_t** Min,uint8_t** Max,int* PageSize)
{
#if defined(MIPS)
	*PageSize = 1024;
#else
	*PageSize = 4096;
#endif

	*Min = *Max = (uint8_t*) ((int)Ptr & ~(*PageSize-1));
	while (!IsBadCodePtr((FARPROC)(*Min-*PageSize)))
		*Min -= *PageSize;
	while (!IsBadCodePtr((FARPROC)*Max))
		*Max += *PageSize;
}

void Assert(const char* Exp,const char* File,int Line)
{
	TCHAR TExp[256];
	TCHAR TFile[MAXPATH];
	strtotcs(TExp,Exp,256);
	strtotcs(TFile,File,MAXPATH);
	MessageBox(NULL,TExp,TFile,MB_OK);
}

void* LockCreate()
{
	void* p = Alloc(sizeof(CRITICAL_SECTION));
	InitializeCriticalSection((LPCRITICAL_SECTION)p);
	return p;
}

void LockDelete(void* p)
{
	if (p)
	{
		DeleteCriticalSection((LPCRITICAL_SECTION)p);
		Free(p);
	}
}

void LockEnter(void* p)
{
	EnterCriticalSection((LPCRITICAL_SECTION)p);
}

void LockLeave(void* p)
{
	LeaveCriticalSection((LPCRITICAL_SECTION)p);
}

void* ThreadCreate(int(*Start)(void*),void* Parameter,int Quantum)
{
	DWORD Id;
	HANDLE Handle = CreateThread(NULL,0,(LPTHREAD_START_ROUTINE)Start,Parameter,0,&Id);

#if defined(_WIN32_WCE)
	if (FuncCeSetThreadQuantum)
		FuncCeSetThreadQuantum(Handle,Quantum);
#endif

	return Handle;
}

void CheckHeap()
{
#if defined(_DEBUG) && defined(_MSC_VER)
	assert(_CrtCheckMemory());
#endif
}

//#define FIND 0x009E64B0

/*INLINE void IsMemoryLow(void* p,int n)
{
	if (!p && n && ShowMemoryLow)
	{
		ShowMemoryLow = 0;
		NodeError((node*)&Platform,PLATFORM_ID,PLATFORM_MEMORY_LOW);
	}
}
*/
void* Alloc(int n)
{
	void* p = malloc(n);
	//IsMemoryLow(p,n);

#ifdef FIND
	if ((int)p == FIND) DebugBreak();
#endif

	return p;
}

void* CAlloc(int n,int m)
{
	void* p;
	
	n *= m;
	p = malloc(n*m);
	//IsMemoryLow(p,n);
	if (p)
		memset(p,0,n);

#ifdef FIND
	if ((int)p == FIND) DebugBreak();
#endif

	return p;
}

void* ReAlloc(void* p,int n)
{
	p = realloc(p,n);
	//IsMemoryLow(p,n);

#ifdef FIND
	if ((int)p == FIND) DebugBreak();
#endif

	return p;
}

void Free(void* p)
{
	free(p);
}

void ShowMessage(const tchar_t* Title,const tchar_t* Msg,...)
{
	tchar_t s[1024];
	vstprintf( s, Msg, (va_list)(&Msg+1) );
	MessageBox(NULL,s,Title,MB_OK);
}

static FILE* Debug = NULL;

void DebugMessage(const tchar_t* Msg,...)
{
/*	tchar_t s[1024];
	stprintf(s,T("%04x "),GetCurrentThreadId());
	vstprintf( s+tcslen(s), Msg, (va_list)(&Msg+1) );
	tcscat(s,T("\n"));
	OutputDebugString(s);
*/	
#if !defined(_WIN32_WCE)
	if (!Debug)
		Debug = fopen("\\debug.txt","w+");
	vftprintf(Debug,Msg,(va_list)(&Msg+1));
	putc(10,Debug);
	fflush(Debug);
#endif
}

typedef struct contextreg
{
	int Ofs;
	const tchar_t* Name;
} contextreg;

#if defined( _M_IX86 )
static contextreg Reg[] = 
{
	OFS(CONTEXT,Eax), T("eax"),
	OFS(CONTEXT,Ebx), T("ebx"),
	OFS(CONTEXT,Ecx), T("ecx"),
	OFS(CONTEXT,Edx), T("edx"),
	OFS(CONTEXT,Esi), T("esi"),
	OFS(CONTEXT,Edi), T("edi"),
	OFS(CONTEXT,Ebp), T("ebp"),
	OFS(CONTEXT,Esp), T("esp"),
	OFS(CONTEXT,Eip), T("eip"),
	OFS(CONTEXT,EFlags), T("flags"),

	OFS(CONTEXT,Esp), NULL, 
};
#elif defined( MIPS )
static contextreg Reg[] = 
{
	OFS(CONTEXT,IntZero), T("Zero"),
	OFS(CONTEXT,IntAt), T("At"),
	OFS(CONTEXT,IntV0), T("V0"),
	OFS(CONTEXT,IntV1), T("V1"),
	OFS(CONTEXT,IntA0), T("A0"),
	OFS(CONTEXT,IntA1), T("A1"),
	OFS(CONTEXT,IntA2), T("A2"),
	OFS(CONTEXT,IntA3), T("A3"),
	OFS(CONTEXT,IntT0), T("T0"),
	OFS(CONTEXT,IntT1), T("T1"),
	OFS(CONTEXT,IntT2), T("T2"),
	OFS(CONTEXT,IntT3), T("T3"),
	OFS(CONTEXT,IntT4), T("T4"),
	OFS(CONTEXT,IntT5), T("T5"),
	OFS(CONTEXT,IntT6), T("T6"),
	OFS(CONTEXT,IntT7), T("T7"),
	OFS(CONTEXT,IntS0), T("S0"),
	OFS(CONTEXT,IntS1), T("S1"),
	OFS(CONTEXT,IntS2), T("S2"),
	OFS(CONTEXT,IntS3), T("S3"),
	OFS(CONTEXT,IntS4), T("S4"),
	OFS(CONTEXT,IntS5), T("S5"),
	OFS(CONTEXT,IntS6), T("S6"),
	OFS(CONTEXT,IntS7), T("S7"),
	OFS(CONTEXT,IntT8), T("T8"),
	OFS(CONTEXT,IntT9), T("T9"),
	OFS(CONTEXT,IntK0), T("K0"),
	OFS(CONTEXT,IntK1), T("K1"),
	OFS(CONTEXT,IntGp), T("Gp"),
	OFS(CONTEXT,IntSp), T("Sp"),
	OFS(CONTEXT,IntS8), T("S8"),
	OFS(CONTEXT,IntRa), T("Ra"),
	OFS(CONTEXT,IntLo), T("Lo"),
	OFS(CONTEXT,IntHi), T("Hi"),
	OFS(CONTEXT,IntSp), NULL,
};
#elif defined( ARM )
static contextreg Reg[] = 
{
	OFS(CONTEXT,R0), T("R0"),
	OFS(CONTEXT,R1), T("R1"),
	OFS(CONTEXT,R2), T("R2"),
	OFS(CONTEXT,R3), T("R3"),
	OFS(CONTEXT,R4), T("R4"),
	OFS(CONTEXT,R5), T("R5"),
	OFS(CONTEXT,R6), T("R6"),
	OFS(CONTEXT,R7), T("R7"),
	OFS(CONTEXT,R8), T("R8"),
	OFS(CONTEXT,R9), T("R9"),
	OFS(CONTEXT,R10), T("R10"),
	OFS(CONTEXT,R11), T("R11"),
	OFS(CONTEXT,R12), T("R12"),
	OFS(CONTEXT,Sp), T("Sp"),
	OFS(CONTEXT,Lr), T("Lr"),
	OFS(CONTEXT,Pc), T("Pc"),
	OFS(CONTEXT,Psr), T("Psr"),
	OFS(CONTEXT,Sp), NULL,
};
#elif defined( SH3 )
static contextreg Reg[] = 
{
	OFS(CONTEXT,PR), T("PR"),
	OFS(CONTEXT,MACH), T("MACH"),
	OFS(CONTEXT,MACL), T("MACL"),
	OFS(CONTEXT,GBR), T("GBR"),
	OFS(CONTEXT,R0), T("R0"),
	OFS(CONTEXT,R1), T("R1"),
	OFS(CONTEXT,R2), T("R2"),
	OFS(CONTEXT,R3), T("R3"),
	OFS(CONTEXT,R4), T("R4"),
	OFS(CONTEXT,R5), T("R5"),
	OFS(CONTEXT,R6), T("R6"),
	OFS(CONTEXT,R7), T("R7"),
	OFS(CONTEXT,R8), T("R8"),
	OFS(CONTEXT,R9), T("R9"),
	OFS(CONTEXT,R10), T("R10"),
	OFS(CONTEXT,R11), T("R11"),
	OFS(CONTEXT,R12), T("R12"),
	OFS(CONTEXT,R13), T("R13"),
	OFS(CONTEXT,R14), T("R14"),
	OFS(CONTEXT,R15), T("R15"),
	OFS(CONTEXT,R15), NULL,
};
#else
static contextreg Reg[] = { -1, NULL };
#endif

void Dump()
{
	FILE* File;
	tchar_t Path[MAXPATH];
	stprintf(Path,T("%s\\dump.txt"),DocPath);

	File = tfopen(Path,T("w+"));
	if (File)
	{
		NodeDump(File);
		fclose(File);

		MessageBox(NULL,String(PLATFORM_ID,PLATFORM_DUMP_MESSAGE),
			String(PLATFORM_ID,PLATFORM_DUMP_TITLE),MB_OK|MB_SETFOREGROUND|MB_TOPMOST);
	}
}

int SafeException(void* p)
{
	EXCEPTION_POINTERS* Data = (EXCEPTION_POINTERS*)p;
	FILE* File;
	tchar_t Path[MAXPATH];
	stprintf(Path,T("%s\\crash.txt"),DocPath);

	File = tfopen(Path,T("w+"));
	if (File)
	{
		void** Stack;
		contextreg* r;
		int No;
		const uint8_t* Context = (const uint8_t*) Data->ContextRecord;
		EXCEPTION_RECORD* Record = Data->ExceptionRecord;
		tchar_t* Name;
		int DllBase;
		tchar_t DllName[MAXPATH];

		ftprintf(File,T("%s %s crash report\n----------------------------\n"),NodeProgramId(),NodeVersion());

		switch (Record->ExceptionCode)
		{
		case STATUS_ACCESS_VIOLATION:		Name = T("Access violation"); break;
		case STATUS_BREAKPOINT:				Name = T("Breakpoint"); break;
		case STATUS_DATATYPE_MISALIGNMENT:	Name = T("Datatype misalignment"); break;
		case STATUS_ILLEGAL_INSTRUCTION:	Name = T("Illegal instruction"); break;
		case STATUS_INTEGER_DIVIDE_BY_ZERO: Name = T("Int divide by zero"); break;
		case STATUS_INTEGER_OVERFLOW:		Name = T("Int overflow"); break;
		case STATUS_PRIVILEGED_INSTRUCTION: Name = T("Priv instruction"); break;
		case STATUS_STACK_OVERFLOW:			Name = T("Stack overflow"); break;
		default:							Name = T("Unknown"); break;
		}

		NodeLocate(Record->ExceptionAddress,DllName,&DllBase);
		ftprintf(File,T("%s(%08x) at %08x (%s:%08x)"),Name,Record->ExceptionCode,Record->ExceptionAddress,DllName,DllBase);

		if (Record->ExceptionCode == STATUS_ACCESS_VIOLATION)
		{
			if (Record->ExceptionInformation[0])
				Name = T("Write to");
			else
				Name = T("Read from");
			ftprintf(File,T("\n%s %08x"),Name,Record->ExceptionInformation[1]);
			if (NodeLocate((void*)Record->ExceptionInformation[1],DllName,&DllBase))
				ftprintf(File,T(" (%s:%08x)"),DllName,DllBase);
		}
		
		// context

		ftprintf(File,T("\n\ncpu dump:"));

		for (r=Reg;r->Name;++r)
		{
			void* Ptr = *(void**)(Context+r->Ofs);

			ftprintf(File,T("\n%-5s = %08x"),r->Name,Ptr);

			if (NodeLocate(Ptr,DllName,&DllBase))
				ftprintf(File,T(" (%s:%08x)"),DllName,DllBase);
		}

		if (r->Ofs >= 0)
		{
			ftprintf(File,T("\n\nstack dump:"));

			Stack = *(void***)(Context+r->Ofs);
			for (No=0;No<256;++No,++Stack)
			{
				if (!IsBadReadPtr(Stack,sizeof(void*)))
				{
					void* Ptr = *Stack;

					ftprintf(File,T("\n%08x %08x"),Stack,Ptr);

					if (NodeLocate(Ptr,DllName,&DllBase))
						ftprintf(File,T(" (%s:%08x)"),DllName,DllBase);
				}
				else
					ftprintf(File,T("\n%08x ????????"),Stack);
			}
		}

		ftprintf(File,T("\n\nnode dump:\n"));

		NodeDump(File);

		fclose(File);
	}

	MessageBox(NULL,String(PLATFORM_ID,PLATFORM_CRASH_MESSAGE),
		String(PLATFORM_ID,PLATFORM_CRASH_TITLE),MB_OK|MB_SETFOREGROUND|MB_TOPMOST|MB_ICONSTOP);

	TerminateProcess(GetCurrentProcess(),1);

	return 1;
}

void SetFileAssociation(const tchar_t* Ext,const tchar_t* Type,bool_t State)
{
	DWORD Disp=0;
	HKEY Key;
	DWORD RegSize;
	DWORD RegType;
	tchar_t Base[64];
	tchar_t Backup[256];

	stprintf(Base,T(".%s"),Ext);
	stprintf(Backup,T("%sBackup"),Type);

	if (RegOpenKeyEx( HKEY_CLASSES_ROOT, Base, 0, ACCESS_SAVE, &Key ) == ERROR_SUCCESS)
	{
		RegSize=sizeof(Base);
		if (RegQueryValueEx( Key, NULL, 0, &RegType, (LPBYTE)Base, &RegSize ) == ERROR_SUCCESS)
		{
			if (tcsicmp(Base,Type)!=0)
			{
				if (State)
				{
					RegSetValueEx( Key, NULL, 0, REG_SZ, (LPBYTE)Type, (tcslen(Type)+1)*sizeof(tchar_t) );
					RegSetValueEx( Key, Backup, 0, REG_SZ, (LPBYTE)Base, RegSize );
				}
			}
			else 
			if (!State)
			{
				RegSize = sizeof(Base);
				if (RegQueryValueEx( Key, Backup, 0, &RegType, (LPBYTE)Base, &RegSize ) != ERROR_SUCCESS)
				{
					Base[0] = 0;
					RegSize = sizeof(tchar_t);
				}

				RegSetValueEx( Key, NULL, 0, REG_SZ, (LPBYTE)Base, RegSize);
			}
		}
		RegCloseKey( Key );
	}
	else 
	if (State)
	{
		if (RegCreateKeyEx( HKEY_CLASSES_ROOT, Base, 0, NULL, 0, ACCESS_SAVE, NULL, &Key, &Disp ) == ERROR_SUCCESS)
		{
			RegSetValueEx( Key, NULL, 0, REG_SZ, (LPBYTE)Type, (tcslen(Type)+1)*sizeof(tchar_t) );
			RegCloseKey( Key );
		}
	}
}

bool_t GetFileAssociation(const tchar_t* Ext,const tchar_t* Type)
{
	bool_t Result = 0;
	HKEY Key;
	tchar_t Base[64];
	stprintf(Base,T(".%s"),Ext);

	if (RegOpenKeyEx( HKEY_CLASSES_ROOT, Base, 0, ACCESS_LOAD, &Key ) == ERROR_SUCCESS)
	{
		tchar_t s[64];
		DWORD RegSize = sizeof(s);
		DWORD RegType;

		if (RegQueryValueEx( Key, NULL, 0, &RegType, (LPBYTE)s, &RegSize ) == ERROR_SUCCESS && RegType == REG_SZ)
			Result = tcsicmp(s,Type)==0;

		RegCloseKey( Key );
	}

	return Result;
}

void SetFileType(const tchar_t* Type, const tchar_t* Open, const tchar_t* Icon )
{
	DWORD Disp=0;
	HKEY Key;
	tchar_t Base[256];

	if (RegCreateKeyEx( HKEY_CLASSES_ROOT, Type, 0, NULL, 0, ACCESS_SAVE, NULL, &Key, &Disp ) == ERROR_SUCCESS)
	{
		RegSetValueEx(Key, NULL, 0, REG_SZ, (LPBYTE)Type, (tcslen(Type)+1)*sizeof(tchar_t));
		RegCloseKey(Key);
	}

	stprintf(Base,T("%s\\DefaultIcon"),Type);
	if (RegCreateKeyEx( HKEY_CLASSES_ROOT, Base, 0, NULL, 0, ACCESS_SAVE, NULL, &Key, &Disp ) == ERROR_SUCCESS)
	{
		RegSetValueEx(Key, NULL, 0, REG_SZ, (LPBYTE)Icon, (tcslen(Icon)+1)*sizeof(tchar_t));
		RegCloseKey(Key);
	}

	stprintf(Base,T("%s\\Shell\\Open\\Command"),Type);
	if (RegCreateKeyEx( HKEY_CLASSES_ROOT, Base, 0, NULL, 0, ACCESS_SAVE, NULL, &Key, &Disp ) == ERROR_SUCCESS)
	{
		RegSetValueEx(Key, NULL, 0, REG_SZ, (LPBYTE)Open, (tcslen(Open)+1)*sizeof(tchar_t));
		RegCloseKey(Key);
	}
}

void TryInvalidate( HWND Wnd, RECT* r0 )
{
	RECT r;

	if (IsWindowVisible(Wnd))
	{
		GetWindowRect(Wnd,&r);
		if (r0->left < r.right && r0->right > r.left &&
			r0->top < r.bottom && r0->bottom > r.top)
		{
			// window is visible and intersects with r0

			HWND Child;
			POINT o;
			o.x = o.y = 0;
			ClientToScreen(Wnd,&o);

			r.left = r0->left - o.x;
			r.right = r0->right - o.x;
			r.top = r0->top - o.y;
			r.bottom = r0->bottom - o.y;
			InvalidateRect(Wnd,&r,0);

			Child = GetWindow(Wnd,GW_CHILD);
			while (Child && IsWindow(Child))
			{
				TryInvalidate(Child,r0);
				Child = GetWindow(Child,GW_HWNDNEXT);
			}
		}
	}
}

BOOL CALLBACK EnumInvalidate( HWND Wnd, LPARAM lParam )
{
	TryInvalidate(Wnd,(RECT*)lParam);
	return 1;
}

void GlobalInvalidate( const rect* Rect )
{
	RECT r;
	r.left = Rect->x;
	r.top = Rect->y;
	r.right = r.left + Rect->Width;
	r.bottom = r.top + Rect->Height;

#if defined(_WIN32_WCE)
	if (FuncRedrawWindow) // doesn't work for mio558???
		FuncRedrawWindow(NULL,&r,NULL,0x0001|0x0080); //RDW_INVALIDATE|RDW_ALLCHILDREN
	else
		EnumWindows(EnumInvalidate,(LPARAM)&r);
#else
	RedrawWindow(NULL,&r,NULL,RDW_INVALIDATE|RDW_ALLCHILDREN);
#endif
}

/*
#include <winioctl.h>

BOOL WINAPI KernelIoControl(
   DWORD dwIoControlCode,
   LPVOID lpInBuf,
   DWORD nInBufSize,
   LPVOID lpOutBuf,
   DWORD nOutBufSize,3
   LPDWORD lpBytesReturned);

#define IOCTL_HAL_REBOOT CTL_CODE(FILE_DEVICE_HAL, 15, METHOD_BUFFERED, FILE_ANY_ACCESS)

void SoftReset()
{
    KernelIoControl(IOCTL_HAL_REBOOT, NULL, 0, NULL, 0, NULL);
}
*/

void HotKeyToString( tchar_t* Out, int HotKey, int OutSize )
{
	int Id = 0;
	
	Out[0] = 0;
	if (HotKey)
	{
		if ((HotKey & HOTKEY_MASK) < 0xC1 || (HotKey & HOTKEY_MASK) > 0xCF)
		{
			if (HotKey & HOTKEY_WIN) { tcscat(Out,String(PLATFORM_ID,PLATFORM_KEY_WIN)); tcscat(Out,T("+")); }
			if (HotKey & HOTKEY_SHIFT) { tcscat(Out,String(PLATFORM_ID,PLATFORM_KEY_SHIFT)); tcscat(Out,T("+")); }
			if (HotKey & HOTKEY_CTRL) { tcscat(Out,String(PLATFORM_ID,PLATFORM_KEY_CTRL)); tcscat(Out,T("+")); }
			if (HotKey & HOTKEY_ALT) { tcscat(Out,String(PLATFORM_ID,PLATFORM_KEY_ALT)); tcscat(Out,T("+")); }
		}
		HotKey &= HOTKEY_MASK;

		if ((HotKey >= '0' && HotKey <= '9') ||
			(HotKey >= 'A' && HotKey <= 'Z'))
			stprintf(Out+tcslen(Out),T("%c"),HotKey);
		else
		if (HotKey >= 0x70 && HotKey < 0x80)
			stprintf(Out+tcslen(Out),T("F%d"),HotKey-0x70+1);
		else
		switch (HotKey)
		{
		case 0x86:
			Id = PLATFORM_KEY_ACTION;
			break;
		case VK_RETURN:
			Id = PLATFORM_KEY_ENTER;
			break;
		case VK_LEFT:
			Id = PLATFORM_KEY_LEFT;
			break;
		case VK_RIGHT:
			Id = PLATFORM_KEY_RIGHT;
			break;
		case VK_UP:
			Id = PLATFORM_KEY_UP;
			break;
		case VK_DOWN:
			Id = PLATFORM_KEY_DOWN;
			break;
		case VK_SPACE:
			Id = PLATFORM_KEY_SPACE;
			break;
		case 0xC1:
		case 0xC2:
		case 0xC3:
		case 0xC4:
		case 0xC5:
		case 0xC6:
		case 0xC7:
		case 0xC8:
		case 0xC9:
		case 0xCA:
		case 0xCB:
		case 0xCC:
		case 0xCD:
		case 0xCE:
		case 0xCF:
			stprintf(Out+tcslen(Out),String(PLATFORM_ID,PLATFORM_KEY_APP),HotKey-0xC0);
			break;
		default:
			stprintf(Out+tcslen(Out),T("#%02X"),HotKey);
			break;
		}
		if (Id)
			tcscat(Out,String(PLATFORM_ID,Id));
	}
}

void WaitBegin()
{
	DEBUG_MSG(T("WaitBegin"));
	SetCursor(LoadCursor (NULL, IDC_WAIT));
}

void WaitEnd()
{
	DEBUG_MSG(T("WaitEnd"));
	SetCursor(LoadCursor (NULL, IDC_ARROW));
}

const tchar_t* GetDocumentPath() { return DocPath; }
const tchar_t* GetSystemPath() { return SysPath; }

#endif
