#include <windows.h>
#include <tchar.h>

#include "Parser.h"

#include "AVInterface.h"

#define	CACHE_SIZE	1048576

#define	MVP_CHAPTER_COUNT   50 // XXX
#define	MVP_CHAPTER_DISPLAY 100	// XXX

MatroskaFile	*g_mf;
videoinfo	g_vi;
int		g_audio;
int		g_video;
ulonglong	g_chapters[MVP_CHAPTER_COUNT];
ulonglong	g_lasttime;

CRITICAL_SECTION  g_cs;

static void VideoTrack(int num,TrackInfo *ti) {
  BITMAPINFOHEADER  *bmh = (BITMAPINFOHEADER*)ti->CodecPrivate;

  if (g_video>=0 || !bmh || ti->CodecPrivateSize<sizeof(*bmh))
    return;
  g_video = num;

  g_vi.width = ti->Video.PixelWidth;
  g_vi.height = ti->Video.PixelHeight;
  g_vi.fps = (ti->DefaultDuration ? 1e9/(longlong)ti->DefaultDuration : 25.0);
  memcpy(g_vi.compressor,&bmh->biCompression,4);
  g_vi.compressor[4] = 0;
  g_vi.video_frames = (long)(mkv_GetFileInfo(g_mf)->Duration * (longlong)mkv_GetFileInfo(g_mf)->TimecodeScale * g_vi.fps / 1e9);
}

static void AudioTrack(int num,TrackInfo *ti,unsigned int tag) {
  if (g_audio>=0)
    return;
  g_audio = num;

  g_vi.a_fmt = tag;
  g_vi.a_chans = ti->Audio.Channels;
  g_vi.a_rate = (long)ti->Audio.OutputSamplingFreq;
  g_vi.a_bits = ti->Audio.BitDepth;
  g_vi.audio_bytes = 1048576; // XXX
}

static void ProcessChapter(Chapter *ch) {
  unsigned int	  i,j;

  if (g_vi.ChapterCount >= MVP_CHAPTER_COUNT)
    return;

  for (i=0;i<ch->nChildren;++i) {
    for (j=0;j<ch->Children[i].nDisplay;++i)
      if (ch->Children[i].Display[j].String) {
	g_chapters[g_vi.ChapterCount] = ch->Children[i].Start;
	MultiByteToWideChar(CP_UTF8,0,ch->Children[i].Display[j].String,-1,
	  g_vi.ChapterDisplay[g_vi.ChapterCount],MVP_CHAPTER_DISPLAY);
	g_vi.ChapterCount++;
	break;
      }
    ProcessChapter(&ch->Children[i]);
  }
}

int CALLBACK	MatroskaOpen(LPTSTR lpFilename, int Cache,
			     int CacheSize,videoinfo** vidinfo)
{
  char	      errmsg[128];
  char	      filename[MAX_PATH];
  InputFile   *inf;
  unsigned    i;
  Chapter     *ch;

  if (g_mf) {
    mkv_Close(g_mf);
    g_mf = NULL;
  }
 
  WideCharToMultiByte(CP_ACP,0,lpFilename,-1,filename,sizeof(filename),NULL,NULL);
  inf = fileio_open(filename, errmsg, sizeof(errmsg));
  if (inf == NULL)
    return 0;

  g_mf = mkv_Open(inf, CACHE_SIZE, errmsg, sizeof(errmsg));
  if (g_mf == NULL) {
    TCHAR   tmp[128];
    MultiByteToWideChar(CP_ACP,0,errmsg,-1,tmp,sizeof(tmp)/sizeof(tmp[0]));
    MessageBox(GetActiveWindow(),tmp,_T("MKV Open Error"),MB_OK);
    return 0;
  }

  memset(&g_vi, 0, sizeof(g_vi));

  g_audio = g_video = -1;

  // find video tracks
  for (i=0;i<mkv_GetNumTracks(g_mf);++i) {
    TrackInfo *ti = mkv_GetTrackInfo(g_mf, i);
    if (ti) {
      if (strcmp(ti->CodecID, "V_MS/VFW/FOURCC")==0)
	VideoTrack(i,ti);
      else if (strcmp(ti->CodecID, "A_MPEG/L3")==0)
	AudioTrack(i,ti,0x0055);
      else if (strcmp(ti->CodecID, "A_MPEG/L2")==0 || strcmp(ti->CodecID, "A_MPEG/L1")==0)
	AudioTrack(i,ti,0x0050);
      else if (strcmp(ti->CodecID, "A_PCM/INT/LIT")==0)
	AudioTrack(i,ti,0x0001);
      else if (strcmp(ti->CodecID, "A_PCM/FLOAT/IEEE")==0)
	AudioTrack(i,ti,0x0003);
      else if (strcmp(ti->CodecID, "A_AC3")==0 || strcmp(ti->CodecID, "A_DTS")==0)
	AudioTrack(i,ti,0x2000);
    }
  }

  if (g_audio<0 && g_video<0)
    return 0;

  i = 0;
  if (g_audio>=0)
    i |= 1<<g_audio;
  if (g_video>=0)
    i |= 1<<g_video;
  mkv_SetTrackMask(g_mf, ~i);

  mkv_GetChapters(g_mf,&ch,&i);
  if (i>0)
    ProcessChapter(&ch[0]);

  *vidinfo = &g_vi;

  return 1;
}

ulonglong   g_apos;
unsigned    g_abytes;

int CALLBACK	MatroskaReadAudio(char *audbuf, int bytes)
{
  ulonglong EndTime;
  unsigned  track,FrameFlags,nb,todo;

  if (!g_mf || g_audio<0)
    return 0;

  EnterCriticalSection(&g_cs);

  nb = bytes;
  while (nb>0) {
    if (g_abytes == 0)
      if (mkv_ReadFrame(g_mf,~(1<<g_audio),&track,&g_lasttime,&EndTime,&g_apos,&g_abytes,&FrameFlags)<0)
	break;

    todo = nb;
    if (todo > g_abytes)
      todo = g_abytes;

    if (mkv_ReadData(g_mf,g_apos,audbuf,todo)<0) {
      g_abytes = 0;
      break;
    }

    g_apos += todo;
    g_abytes -= todo;
    audbuf += todo;
    nb -= todo;
  }

  LeaveCriticalSection(&g_cs);

  return bytes - nb;
}

int CALLBACK	MatroskaNextVideoFrame(char *buffer, int drop)
{
  ulonglong EndTime,FilePos;
  unsigned  track,FrameSize,FrameFlags;

  if (!g_mf || g_video<0)
    return 0;

  EnterCriticalSection(&g_cs);

  if (mkv_ReadFrame(g_mf,~(1<<g_video),&track,&g_lasttime,&EndTime,&FilePos,&FrameSize,&FrameFlags)) {
    FrameSize = 0;
    goto out;
  }

  if (mkv_ReadData(g_mf,FilePos,buffer,FrameSize)<0)
    FrameSize = 0;

out:
  LeaveCriticalSection(&g_cs);

  return FrameSize;
}

BOOL CALLBACK	MatroskaIsKeyFrame(long frame)
{
  return TRUE;
}

int CALLBACK	MatroskaSeek(int percent, int isFrame)
{
  ulonglong tc;

  if (!g_mf)
    return 0;

  if (isFrame)
    tc = (ulonglong)(percent * mkv_GetFileInfo(g_mf)->Duration / g_vi.video_frames);
  else
    tc = (ulonglong)(percent * mkv_GetFileInfo(g_mf)->Duration / 100);

  EnterCriticalSection(&g_cs);
  mkv_Seek(g_mf,(ulonglong)(tc * mkv_GetFileInfo(g_mf)->TimecodeScale));
  LeaveCriticalSection(&g_cs);

  return 0;
}

int CALLBACK	MatroskaSeekChapter(int chapter)
{
  if (!g_mf)
    return 0;

  EnterCriticalSection(&g_cs);
  mkv_Seek(g_mf,g_chapters[chapter] * mkv_GetFileInfo(g_mf)->TimecodeScale);
  LeaveCriticalSection(&g_cs);
  return 0;
}

int CALLBACK	MatroskaReSeekAudio()
{
  return 1;
}

double CALLBACK	MatroskaGetProgress()
{
  if (!g_mf)
    return 0;

  return (longlong)g_lasttime * 100.0 / (mkv_GetFileInfo(g_mf)->Duration * (longlong)mkv_GetFileInfo(g_mf)->TimecodeScale);
}

int CALLBACK	MatroskaClose()
{
  if (g_mf)
    mkv_Close(g_mf);

  g_mf = NULL;

  return 1;
}

void CALLBACK	MatroskaPause(int pause)
{
}

int CALLBACK	MatroskaStop()
{
  EnterCriticalSection(&g_cs);
  if (g_mf) {
    mkv_Close(g_mf);
    g_mf = NULL;
  }
  LeaveCriticalSection(&g_cs);
  return 1;
}

void WINAPI InitializeReader(reader* rd) {
  rd->READER_Open=MatroskaOpen;
  rd->READER_ReadAudio=MatroskaReadAudio;
  rd->READER_NextVideoFrame=MatroskaNextVideoFrame;
  rd->READER_Seek=MatroskaSeek;
  rd->READER_ReSeekAudio=MatroskaReSeekAudio;
  rd->READER_GetProgress=MatroskaGetProgress;
  rd->READER_Close=MatroskaClose;
  rd->READER_isKeyFrame=MatroskaIsKeyFrame;
  rd->READER_SeekChapter=MatroskaSeekChapter;
  rd->READER_Stop=MatroskaStop;
}

BOOL WINAPI DllMain(HANDLE hDll, DWORD dwReason, LPVOID lpvReserved) {
  if (dwReason == DLL_PROCESS_ATTACH)
    InitializeCriticalSection(&g_cs);
  else if (dwReason == DLL_PROCESS_DETACH)
    DeleteCriticalSection(&g_cs);
  return TRUE;
}