#ifdef _WIN32
#include <windows.h>
#include <tchar.h>
#include <io.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <fcntl.h>

#include "Parser.h"

int main(int argc,char **argv) {
  char		errmsg[128];
  InputFile	*io;
  SegmentInfo	*si;
  TrackInfo	*ti;
  unsigned	track,flags;
  ulonglong	StartTime,EndTime;
#ifdef _WIN32
  DWORD		start,end;
#endif
  char		zcomp = 0;
  MatroskaFile	*mf;
  CompressedStream  *cs;
  unsigned	FrameSize;
  ulonglong	FilePos;
  unsigned char	ibuffer[4096];


#ifdef _WIN32
  start = GetTickCount();

  // windows madness
  _setmode( _fileno( stdout ), _O_BINARY );
#endif

  if (argc<2) {
    fprintf(stderr,"Usage: mp filename\n");
    return 1;
  }

  io = fileio_open(argv[1],errmsg,sizeof(errmsg));
  if (io == NULL) {
    fprintf(stderr,"Can't open '%s': %s\n",argv[1],errmsg);
    return 1;
  }

  mf = mkv_Open(io,1048576*2,errmsg,sizeof(errmsg));
  if (mf == NULL) {
    fprintf(stderr,"Can't open Matroska file: %s\n",errmsg);
    return 1;
  }

  si = mkv_GetFileInfo(mf);

  fprintf(stderr,"Muxing App: %s\nWritingApp: %s\nTitle: %s\n",
    si->MuxingApp,
    si->WritingApp,
    si->Title);

  // find some S_VOBSUB track
  for (track=0;track<mkv_GetNumTracks(mf);++track) {
    ti = mkv_GetTrackInfo(mf,track);
    if (ti && strcmp(ti->CodecID,"S_VOBSUB")==0) {
      if (ti->CompEnabled && ti->CompMethod==COMP_ZLIB)
	zcomp = 1;
      goto found;
    }
  }

  fprintf(stderr,"No VobSub tracks found\n");
  mkv_Close(mf);
  return 1;

found:
  mkv_SetTrackMask(mf, ~(1 << track));
  if (zcomp) {
    cs = cs_Create(mf, track, errmsg, sizeof(errmsg));
    if (cs == NULL) {
      fprintf(stderr,"Can't create decompressor: %s\n",errmsg);
      goto out;
    }
  }
  for (;;) {
    if (mkv_ReadFrame(mf,0,&track,&StartTime,&EndTime,&FilePos,&FrameSize,&flags)==EOF)
      break;

#ifdef _WIN32
    fprintf(stderr,"Frame %u@%I64u, Track %u\n",FrameSize,FilePos,track);
#else
    fprintf(stderr,"Frame %u@%llu, Track %u\n",FrameSize,FilePos,track);
#endif

    if (zcomp) {
      int   rd;

      cs_NextFrame(cs, FilePos, FrameSize);

      do {
	rd = cs_ReadData(cs, ibuffer, sizeof(ibuffer));
	if (rd<0) {
	  fprintf(stderr,"Decompressor error: %s\n",cs_GetLastError(cs));
	  goto out;
	}
	if (rd>0 && fwrite(ibuffer,1,rd,stdout)!=rd) {
	  fprintf(stderr,"Short write while writing a frame to stdout\n");
	  goto out;
	}
      } while (rd>0);
    } else
      while (FrameSize>0) {
	unsigned	rd = FrameSize > sizeof(ibuffer) ? sizeof(ibuffer) : FrameSize;
	if (mkv_ReadData(mf,FilePos,ibuffer,rd)<0) {
	  fprintf(stderr,"MKV Error: %s\n",mkv_GetLastError(mf));
	  goto out;
	}

	if (fwrite(ibuffer,1,rd,stdout)!=rd) {
	  fprintf(stderr,"Short write while writing a frame to stdout\n");
	  goto out;
	}
	FrameSize -= rd;
	FilePos += rd;
      }
  }

out:
  if (mkv_GetLastError(mf))
    fprintf(stderr,"MKV Error: %s\n",mkv_GetLastError(mf));

  if (zcomp)
    cs_Destroy(cs);

  mkv_Close(mf);

#ifdef _WIN32
  end = GetTickCount();

  fprintf(stderr,"Time: %ums\n",end - start);
#endif

  return 0;
}
