/*
 * Copyright (c) 2004 Mike Matsnev.  All Rights Reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice immediately at the beginning of the file, without modification,
 *    this list of conditions, and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Absolutely no warranty of function or purpose is made by the author
 *    Mike Matsnev.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * $Id: main.cpp,v 1.2 2004/04/28 07:49:13 mike Exp $
 * 
 */

#include <windows.h>
#include <tchar.h>
#include <vfw.h>

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>

#include "Parser.h"

int main(int argc,char **argv) {
  char		errmsg[128];
  InputFile	*io;
  SegmentInfo	*si;
  TrackInfo	*ti;
  unsigned	track,flags;
  ulonglong	StartTime,EndTime;
  DWORD		start,end;
  MatroskaFile	*mf;
  unsigned	FrameSize;
  ulonglong	FilePos;
  unsigned char	*framebuf = NULL;
  unsigned	framebufsize = 0;
  PAVIFILE	avi;
  PAVISTREAM	stream;
  AVISTREAMINFO	asi;
  BITMAPINFOHEADER  *bmh;
  FILE		*vfr;
  longlong	prev_duration = 0;
  LONG		range_start = 0;
  ulonglong	range_start_time = 0;
  ulonglong	tolerance;

  start = GetTickCount();

  if (argc<4) {
    fprintf(stderr,"Usage: mp matroska_file avi_file vfr_file\n");
    return 1;
  }

  io = fileio_open(argv[1],errmsg,sizeof(errmsg));
  if (io == NULL) {
    fprintf(stderr,"Can't open '%s': %s\n",argv[1],errmsg);
    return 1;
  }

  mf = mkv_Open(io,1048576*2,errmsg,sizeof(errmsg));
  if (mf == NULL) {
    fprintf(stderr,"Can't open Matroska file: %s\n",errmsg);
    return 1;
  }

  si = mkv_GetFileInfo(mf);

  fprintf(stderr,"Muxing App: %s\nWritingApp: %s\nTitle: %s\n",
    si->MuxingApp,
    si->WritingApp,
    si->Title);

  // find some V_MS/VFW/FOURCC track
  for (track=0;track<mkv_GetNumTracks(mf);++track) {
    ti = mkv_GetTrackInfo(mf,track);
    if (ti && strcmp(ti->CodecID,"V_MS/VFW/FOURCC")==0) {
      if (ti->CodecPrivateSize < sizeof(BITMAPINFOHEADER)) {
	fprintf(stderr, "CodecPrivateSize is too small.\n");
	mkv_Close(mf);
	return 1;
      }
      goto found;
    }
  }

  fprintf(stderr,"No Video tracks found\n");
  mkv_Close(mf);
  return 1;

found:
  mkv_SetTrackMask(mf, ~(1 << track));

  AVIFileInit();

  // create and open an avi file
  if (AVIFileOpen(&avi,argv[2],OF_CREATE|OF_WRITE|OF_SHARE_DENY_WRITE,NULL)) {
    fprintf(stderr,"AVIFileOpen failed.\n");
    goto out2;
  }

  // fill in stream info
  memset(&asi,0,sizeof(asi));

  bmh = (BITMAPINFOHEADER *)ti->CodecPrivate;

  asi.fccType = streamtypeVIDEO;
  asi.fccHandler = bmh->biCompression;
  asi.dwScale = 1000;
  asi.dwRate = 23976; // FIXME use defaultduration
  asi.dwQuality = -1;
  asi.rcFrame.right = bmh->biWidth;
  asi.rcFrame.bottom = abs(bmh->biHeight);

  if (AVIFileCreateStream(avi,&stream,&asi)) {
    fprintf(stderr,"AVIFileCreateStream failed.\n");
    goto out3;
  }

  if (AVIStreamSetFormat(stream,0,ti->CodecPrivate,ti->CodecPrivateSize)) {
    fprintf(stderr,"AVIStreamSetFormat failed.\n");
    goto out4;
  }

  // create/open the vfr file
  vfr = fopen(argv[3],"w");
  if (vfr == NULL) {
    fprintf(stderr,"Can't open vfr file.\n");
    goto out4;
  }

  // write the required header
  fprintf(vfr,"Assume 23.976\n");

  // compure the maximum delta
  tolerance = si->TimecodeScale + si->TimecodeScale/100;

  for (LONG framenum = 0;;++framenum) {
    if (mkv_ReadFrame(mf,0,&track,&StartTime,&EndTime,&FilePos,&FrameSize,&flags)==EOF)
      break;

#ifdef DEBUG
    fprintf(stderr,"Frame %u@%I64u, %I64d %I64d, Track %u\n",FrameSize,FilePos,StartTime,EndTime-StartTime,track);
#endif

    // allocate a bigger buffer
    if (framebufsize < FrameSize) {
      free(framebuf);
      framebufsize = 0;
      framebuf = (unsigned char *)malloc(FrameSize);
      if (framebuf == NULL) {
	fprintf(stderr, "Ouf of memory.\n");
	goto out;
      }
      framebufsize = FrameSize;
    }

    // read frame data
    if (mkv_ReadData(mf,FilePos,framebuf,FrameSize)<0) {
      fprintf(stderr,"MKV Error: %s\n",mkv_GetLastError(mf));
      goto out;
    }

    // write frame to avi file
    LONG  nsamp,nbytes;
    if (AVIStreamWrite(stream,framenum,1,framebuf,FrameSize,
      flags & FRAME_NOTKF ? 0 : AVIIF_KEYFRAME,&nsamp,&nbytes))
    {
      fprintf(stderr,"AVIStreamWrite failed.\n");
      goto out;
    }

    // handle vfr stuff
    if (framenum > 0) {
      longlong	  d = EndTime - StartTime - prev_duration;
      if (d < 0)
	d = -d;

      // matroska files are usually written with low timecode resolution,
      // so frame duration will vary due to rounding. We tolerate duration
      // differences of TimeCodeScale and less.
      if ((ulonglong)d > tolerance) { // rate change
	if (StartTime == range_start_time)
	  fprintf(stderr, "Frame duration is 0!\n");
	else
	  fprintf(vfr,"%d,%d,%.4f\n",range_start,framenum-1,1000000000.0 * (double)(framenum - range_start) / (StartTime - range_start_time));
	range_start = framenum;
	range_start_time = StartTime;
	prev_duration = EndTime - StartTime;
      }
    } else {
      prev_duration = EndTime - StartTime;
      range_start_time = StartTime;
    }
  }

  // last segment
  if (EndTime == range_start_time)
    fprintf(stderr, "Frame duration is 0!\n");
  else
    fprintf(vfr,"%d,%d,%.4f\n",range_start,framenum-1,1000000000.0 * (double)(framenum - range_start) / (EndTime - range_start_time));

out:
  fclose(vfr);

out4:
  AVIStreamRelease(stream);

out3:
  AVIFileRelease(avi);

out2:
  AVIFileExit();

  if (mkv_GetLastError(mf))
    fprintf(stderr,"MKV Error: %s\n",mkv_GetLastError(mf));

  mkv_Close(mf);

  end = GetTickCount();

  fprintf(stderr,"Time: %ums\n",end - start);

  return 0;
}
