/*****************************************************************************
 *
 * This program is free software ; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 *
 * $Id$
 *
 ****************************************************************************/
 
#include "stdafx.h"

static mjpeg_decode MJPEG;

const datadef MJPEGParams[] = 
{
	{ MJPEG_ID,		TYPE_CLASS },
	{ MJPEG_INPUT,	TYPE_PACKET, DF_INPUT, PACKET_VIDEO },
	{ MJPEG_IDCT,	TYPE_NODE, DF_OUTPUT, IDCT_CLASS },

	DATADEF_END
};

static int MJPEG_Packet( mjpeg_decode* p, const packet* Packet )
{
	int Result;

	if (!p->IDCT)
		return ERR_NOT_SUPPORTED;

	p->Packet.CurrTime = Packet->CurrTime; 

	if (Packet->DropLevel > 1)
	{
		p->Show = -1;
		p->IDCT->Drop(p->IDCT);
		return ERR_NONE;
	}

	if (p->Show>=0)
	{
		// pending frame
		
		Result = p->IDCT->Send(p->IDCT,&p->Packet,&p->Dropped);
		if (Result == ERR_BUFFER_FULL)
			return Result;

		p->Show = -1;
	}

	if (Packet->Length == 0)
		return ERR_INVALID_DATA;

	initbits(p,Packet->Data[0],Packet->Length);

	Result = ERR_NONE;

	if (p->Show>=0)
	{
		if (p->IDCT->Send(p->IDCT,&p->Packet,&p->Dropped) != ERR_BUFFER_FULL)
			p->Show = -1;
	}

	return Result;
}

static int MJPEG_SetIDCT( mjpeg_decode* p, idct* IDCT )
{
	int Format = IDCTFORMAT_BLOCK;
	if (IDCT && IDCT->Set(IDCT,IDCT_FORMAT,&Format,sizeof(Format)) != ERR_NONE)
		return ERR_NOT_SUPPORTED;

	if (p->IDCT)
	{
		if (IDCTCopy(p->IDCT,IDCT,1) != ERR_NONE)
			return ERR_NOT_SUPPORTED;
	}
	else
	{
		video Output;
		int Count;

		Output = p->Format;
		Output.Pixel.Flags = PF_YUV420;

		if (IDCT)
		{
			if (IDCT->Set(IDCT,IDCT_OUTPUT|PACKET_FORMAT,&Output,sizeof(Output)) != ERR_NONE)
				return ERR_NOT_SUPPORTED;

			Count = 1;
			IDCT->Set(IDCT,IDCT_BUFFERCOUNT,&Count,sizeof(int));
			IDCT->Drop(IDCT);
		}
	}

	p->IDCT = IDCT;
	return ERR_NONE;
}

static int MJPEG_SetFormat( mjpeg_decode* p, const video* Format )
{
	memset(&p->Format,0,sizeof(p->Format));

	p->mb_xsize = 0;
	p->mb_ysize = 0;

	p->Total = 0;
	p->Dropped = 0;
	p->Show = -1;

	if (Format)
	{
		int FourCC = Format->Pixel.FourCC;

		if (!(Format->Pixel.Flags & PF_FOURCC))
			return ERR_INVALID_DATA;

		if ((FourCC != FOURCC('M','J','P','G')))
			return ERR_INVALID_DATA;

		p->Format = *Format;
		p->mb_xsize = ( p->Format.Width + 15 ) / 16; 
		p->mb_ysize = ( p->Format.Height + 15 ) / 16;
	}

	return ERR_NONE;
}

int MJPEG_Get( mjpeg_decode* p, int No, void* Data, int Size )
{
	int Result = ERR_INVALID_PARAM;
	switch (No)
	{
	case NODE_ID: GETVALUE(MJPEG_ID,int); break;
	case FLOW_TOTAL:GETVALUE(p->Total,int); break;
	case FLOW_DROPPED: GETVALUE(p->Dropped,int); break;
	case MJPEG_INPUT|PACKET_FORMAT: GETVALUE(p->Format,video); break;
	case MJPEG_IDCT: GETVALUE(p->IDCT,idct*); break;
	}
	return Result;
}

int MJPEG_Set( mjpeg_decode* p, int No, const void* Data, int Size )
{
	int Result = ERR_INVALID_PARAM;
	switch (No)
	{
	case MJPEG_IDCT:
		if (Size == sizeof(idct*) && NodeIsClass(*(node**)Data,IDCT_CLASS))
			Result = MJPEG_SetIDCT(p,*(idct**)Data);
		break;

	case MJPEG_INPUT | PACKET_FORMAT:
		if (!Size || Size == sizeof(video))
		{
			if (!Size) Data = NULL;
			Result = MJPEG_SetFormat(p,(const video*)Data);
		}
		break;

	case MJPEG_INPUT:
		if (Size == sizeof(packet))
			Result = MJPEG_Packet(p,(const packet*)Data);
		break;

	case FLOW_TOTAL: SETVALUE(p->Total,int,ERR_NONE); break;
	case FLOW_DROPPED: SETVALUE(p->Dropped,int,ERR_NONE); break;

	case FLOW_EMPTY:
		p->Show = -1;
		if (p->IDCT)
			p->IDCT->Drop(p->IDCT);
		Result = ERR_NONE;
		break;
	}
	return Result;
}

int MJPEG_Enum( mjpeg_decode* p, int No, datadef* Param )
{
	return NodeEnumType(&No,Param,NodeParams,FlowParams,MJPEGParams,NULL);
}

void MJPEG_Init()
{
	mjpeg_decode* p = &MJPEG;

	memset(p,0,sizeof(mjpeg_decode));

	p->VMT.Enum = MJPEG_Enum;
	p->VMT.Set = MJPEG_Set;
	p->VMT.Get = MJPEG_Get;

	NodeRegister(&p->VMT,PRI_DEFAULT);
}

void MJPEG_Done()
{
	mjpeg_decode* p = &MJPEG;
	
	MJPEG_SetIDCT(p,NULL);
	MJPEG_SetFormat(p,NULL);

	NodeUnRegister(&p->VMT);
}
