/*****************************************************************************
 *
 * This code has been developed by Project Mayo. This software is an
 * implementation of a part of one or more MPEG-4 Video tools as
 * specified in ISO/IEC 14496-2 standard.  Those intending to use this
 * software module in hardware or software products are advised that its
 * use may infringe existing patents or copyrights, and any such use
 * would be at such party's own risk.  The original developer of this
 * software module and his/her company, and subsequent editors and their
 * companies (including Project Mayo), will have no liability for use of
 * this software or modifications or derivatives thereof.
 *
 *****************************************************************************
 *																				*	
 * This program is free software ; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 *
 * $Id$
 *
 *****************************************************************************
 *
 * Authors:
 *
 *	Andrea	Graziani  (Ag): Original source code (Open Divx Decoder 0.4a).
 *	Marc	Dukette   (Md) and
 *	Pedro	Mateu     (Pm) and
 *	Gabor	Kovacs    (Kg) Heavily modified and optimized code
 *
 ****************************************************************************/

#include "stdafx.h"

mp4_decode mp4 = { 0 };

extern idct* SoftIDCT_Init(int Id);
extern void SoftIDCT_Done();

_CONST datadef MP4Params[] = 
{
	{ MPEG4_ID,		TYPE_CLASS },
	{ MPEG4_INPUT,	TYPE_PACKET, DF_INPUT, PACKET_VIDEO },
	{ MPEG4_IDCT,	TYPE_NODE, DF_OUTPUT, IDCT_CLASS },

	DATADEF_END
};

_CONST datadef MSMP4Params[] = 
{
	{ MSMPEG4_ID,	TYPE_CLASS },
	{ MPEG4_INPUT,	TYPE_PACKET, DF_INPUT, PACKET_VIDEO },
	{ MPEG4_IDCT,	TYPE_NODE, DF_OUTPUT, IDCT_CLASS },

	DATADEF_END
};

int mp4_updatecount( mp4_decode* dec )
{
	if (!dec->idct)
	{
		dec->bufcount = 0;
		return ERR_NONE;
	}

	dec->idct->Get(dec->idct,IDCT_BUFFERCOUNT,&dec->bufcount,sizeof(int));

	if (dec->bufcount >= 3)
	{
		dec->mv_bufmask = ~0; // full size mv buffer needed
		dec->low_delay = 0;
	}
	else
	{
		dec->mv_bufmask = MB_X-1; //only one row of mv buffer needed
		dec->low_delay = 1;
	}

	if (dec->format.Width > 0 && dec->bufcount < 2)
		return ERR_OUT_OF_MEMORY;

	return ERR_NONE;
}

bool_t mp4_initidct( mp4_decode* dec, idct* p )
{
	video idctformat;
	int count;

	idctformat = dec->format;
	idctformat.Pixel.Flags = PF_YUV420;

	if (p)
	{
		if (p->Set(p,IDCT_OUTPUT|PACKET_FORMAT,&idctformat,sizeof(idctformat)) != ERR_NONE)
			return 0;

		count = 2;
		p->Set(p,IDCT_BUFFERCOUNT,&count,sizeof(int));
		p->Drop(p);
	}

	return 1;
}

int mp4_updatesize( mp4_decode* dec, int width, int height)
{
	int result = ERR_NONE;

	if (dec->format.Width != width || dec->format.Height != height)
	{
		if (width > MB_X*16 || height > MB_Y*16)
		{
			NodeError(&dec->vmt,VOUT_CLASS,VOUT_ERROR_SIZE);
			result = ERR_NOT_SUPPORTED;
		}
		else
		{
			Free(dec->_mv_buf);

			dec->format.Width = width;
			dec->format.Height = height;
			dec->mb_xsize = ( dec->format.Width + 15 ) / 16; 
			dec->mb_ysize = ( dec->format.Height + 15 ) / 16;
			dec->pos_end = dec->mb_ysize * MB_X;
			dec->_mv_buf = (int*) Alloc(32 + 4*sizeof(int) * dec->pos_end);
			if (!dec->_mv_buf)
				return ERR_INVALID_DATA;
			dec->mv_buf = (int*)(((int)dec->_mv_buf + 31) & ~31);

			if (dec->idct)
			{
				if (!mp4_initidct(dec,dec->idct))
					result = ERR_NOT_SUPPORTED;
				else
					result = mp4_updatecount(dec);
			}
		}
	}
	return result;
}

static int mp4_setidct( mp4_decode* dec, idct* p )
{
	int format = IDCTFORMAT_BLOCK;
	if (p && p->Set(p,IDCT_FORMAT,&format,sizeof(format)) != ERR_NONE)
		return ERR_NOT_SUPPORTED;

	if (dec->idct)
	{
		if (IDCTCopy(dec->idct,p,2) != ERR_NONE)
			return ERR_NOT_SUPPORTED;
	}
	else
	{
		if (!mp4_initidct(dec,p))
			return ERR_NOT_SUPPORTED;
	}

	dec->idct = p;
	return mp4_updatecount(dec);
}

static void mp4_setempty( mp4_decode* dec )
{
	dec->frame = 0;
	dec->show = -1;
	dec->packed_length = 0;
	dec->starttime = 0;
	dec->startstate = -1;
	dec->startpos = 0;
	dec->startfound = 0;
	MergeClear(&dec->buffer);
	if (dec->idct)
		dec->idct->Drop(dec->idct);
}

static int mp4_setformat( mp4_decode* dec, const video* format )
{
	memset(&dec->format,0,sizeof(dec->format));

	Free(dec->packed_data);
	dec->packed_data = NULL;
	dec->packed_allocated = 0;
	dec->packed_length = 0;

	Free(dec->_mv_buf);
	dec->_mv_buf = NULL;
	dec->mv_buf = NULL;

	dec->mb_xsize = 0; 
	dec->mb_ysize = 0;
	dec->pos_end = 0;

	dec->quant_precision = 5;
	dec->time_increment_resolution = 0;
	dec->last_reftime = 0;
	dec->time_pp = 0;
	dec->time_bp = 0;
	dec->low_delay = 0;
	dec->tried_bvop = 0;
	dec->total = 0;
	dec->dropped = 0;
	dec->frame = 0;
	dec->show = -1;
	dec->valid = 0;
	MergeDone(&dec->buffer);

	if (format)
	{
		int fourcc = format->Pixel.FourCC;

		if (!(format->Pixel.Flags & PF_FOURCC))
			return ERR_INVALID_DATA;

		if (dec->params == MSMP4Params)
		{
			if ((fourcc != FOURCC('M','P','4','3')) &&
				(fourcc != FOURCC('M','P','G','3')) &&
				(fourcc != FOURCC('D','I','V','3')) &&
				(fourcc != FOURCC('D','I','V','4')) &&
				(fourcc != FOURCC('D','I','V','5')) &&
				(fourcc != FOURCC('3','I','V','D')))
			return ERR_INVALID_DATA;

			dec->valid = 1;
		}
		else
		{
			if ((fourcc != FOURCC('D','I','V','X')) &&
				(fourcc != FOURCC('D','X','5','0')) &&
				(fourcc != FOURCC('X','V','I','D')) &&
				(fourcc != FOURCC('3','I','V','2')) &&
				(fourcc != FOURCC('M','P','4','S')) &&
				(fourcc != FOURCC('U','M','P','4')) &&
				(fourcc != FOURCC('M','P','4','V')))

			return ERR_INVALID_DATA;
		}

		dec->format = *format;
		dec->format.Width = -1;

		if (mp4_updatesize(dec,format->Width,format->Height) != ERR_NONE)
			return ERR_INVALID_DATA;

		dec->showerror = 1;
		dec->bufcount = 0;
		dec->frametime = Scale(TICKSPERSEC,format->FrameRate.Den,format->FrameRate.Num);
		MergeInit(&dec->buffer);

		mp4_setempty(dec);

		if (dec->ownidct)
			mp4_setidct(dec,dec->ownidct);
	}
	else
	if (dec->ownidct)
		mp4_setidct(dec,NULL);

	return ERR_NONE;
}

#define DEC(p) ((mp4_decode*)((char*)(p)-(int)&(((mp4_decode*)0)->vmt)))

bool_t mp4_isempty( mp4_decode* dec );

int mp4_get( node* n, int No, void* Data, int Size )
{
	int Result = ERR_INVALID_PARAM;
	mp4_decode* dec = DEC(n);
	switch (No)
	{
	case NODE_ID: GETVALUE(dec->params[0].No,int); break;
	case FLOW_TOTAL:GETVALUE(dec->total,int); break;
	case FLOW_DROPPED: GETVALUE(dec->dropped,int); break;
	case MPEG4_INPUT|PACKET_FORMAT: GETVALUE(dec->format,video); break;
	case MPEG4_IDCT: GETVALUE(dec->idct,idct*); break;
	case FLOW_EMPTY: GETVALUE(mp4_isempty(dec),bool_t); break;
	}
	return Result;
}

int mp4_vop( mp4_decode *dec, bool_t packed )
{
	int i;
	int64_t TRB,TRD;

	switch (dec->prediction_type) 
	{
	case N_VOP:
		if (dec->frame < 1)
		{
			dec->total++;
			dec->dropped++;
			return ERR_INVALID_DATA;
		}
		// nothing to do here
		if (!packed)
		{
			dec->total++;
			dec->show = -1;
		}
		return ERR_NONE;

	case P_VOP:
		if (dec->frame < 1)
		{
			dec->total++;
			dec->dropped++;
			return ERR_INVALID_DATA;
		}

	case I_VOP:

		// I/P-frames are the "reference" frames
		// they are decode into refframe[0] 
		// P-frames depend on last refenrence frame: refframe[1]

		i = dec->refframe[0];
		dec->refframe[0] = dec->refframe[1];
		dec->refframe[1] = i;

		if (dec->low_delay || dec->frame==0)
			dec->show = dec->refframe[0]; // show this refframe
		else
			dec->show = dec->refframe[1]; // show last refframe 

		if (dec->prediction_type == I_VOP)
		{
			dec->bufframe = dec->idct->FrameStart(dec->idct,dec->frame,dec->refframe[0],-1,-1,dec->show,0);
			dec->IVOP_frame( dec );
		}
		else 
		{
			dec->bufframe = dec->idct->FrameStart(dec->idct,dec->frame,dec->refframe[0],dec->refframe[1],-1,dec->show,dec->pck.DropLevel);
			dec->PVOP_frame( dec );
		}
		break;

	case B_VOP:
		if (dec->low_delay && !dec->tried_bvop)
		{
			dec->tried_bvop = 1;
			dec->bufcount = 3;
			dec->idct->Set(dec->idct,IDCT_BUFFERCOUNT,&dec->bufcount,sizeof(int));
			mp4_updatecount(dec);
		}

		if (dec->low_delay || dec->frame < 2)
		{
			dec->total++;
			dec->dropped++;
			return ERR_INVALID_DATA;
		}

		if (dec->pck.DropLevel)
		{
			dec->total++;
			dec->dropped++;
			dec->show = -1;
			return ERR_NONE;
		}

		// save this b-frame if there was already a packed one
		if (packed) 
			return ERR_NONE;

		dec->show = 2;
		dec->bufframe = dec->idct->FrameStart(dec->idct,-dec->frame,2,dec->refframe[0],dec->refframe[1],2,dec->pck.DropLevel);
			
		TRB = dec->time_pp - dec->time_bp;
		TRD = dec->time_pp;

		if (TRD)
		{
			dec->TRB = (TRB << 32) / TRD;
			dec->TRB_TRD = ((TRB - TRD) << 32) / TRD;
		}

		dec->BVOP_frame( dec );
		break;

	default:
		return ERR_INVALID_DATA;
	}

	dec->idct->FrameEnd(dec->idct);

	dec->frame++;
	dec->total++;

	// is there a possible (uint8) framemap overflow?
	if ((dec->frame - dec->mapofs) >= 128) 
	{
		int pos;
		for (pos=0;pos<dec->pos_end;pos+=MB_X-dec->mb_xsize)
			for (;POSX(pos)<dec->mb_xsize;++pos)
			{
				int i = dec->framemap[pos];
				if (i >= (120<<1))
					i -= 120<<1;
				else
					i &= 1;
				dec->framemap[pos] = (uint8_t)i;
			}

		dec->mapofs += 120;
	}
	return ERR_NONE;
}

int mp4_frame( mp4_decode* dec, const uint8_t* ptr, int len )
{
	const uint8_t* pos;
	bool_t packed = 0;
	int result;

	if (dec->packed_length)
	{
		initbits(dec,dec->packed_data,dec->packed_length);

		packed = dec->gethdr(dec) && mp4_vop(dec,0)==ERR_NONE;

		dec->packed_length = 0;
	}

	if (len == 0)
		return ERR_INVALID_DATA;

	initbits(dec,ptr,len);

	if (!dec->gethdr(dec) || !dec->valid)
		return ERR_INVALID_DATA; // invalid header or no vol information

	result = mp4_vop(dec,packed);

	if (result == ERR_NONE && dec->show>=0) 
	{
		// check for packed b-frame
		pos = bytepos(dec);
		len -= (pos - ptr);

		if (len > 4) 
		{
			if (dec->low_delay && !dec->tried_bvop)
			{
				dec->tried_bvop = 1;
				dec->bufcount = 3;
				dec->idct->Set(dec->idct,IDCT_BUFFERCOUNT,&dec->bufcount,sizeof(int));
				mp4_updatecount(dec);
			}

			if (!dec->low_delay)
			{
				// save packed b-frame

				if (dec->packed_allocated < len)
				{
					if (dec->packed_data)
						Free(dec->packed_data);
					dec->packed_allocated = (len+8 + 0x1FFF) & ~0x1FFF;
					dec->packed_data = Alloc(dec->packed_allocated);
					if (!dec->packed_data)
						dec->packed_allocated = 0;
				}

				if (dec->packed_data) 
				{
					dec->packed_length = len;
					memcpy(dec->packed_data,pos,len);
				}
			}
		}
	}

	if (dec->show>=0)
	{
		if (dec->idct->Send(dec->idct,&dec->pck,&dec->dropped) != ERR_BUFFER_FULL)
			dec->show = -1;
	}

	return result;
}

int mp4_packet( mp4_decode* dec, const packet* p )
{
	int result;
	
	if (!dec->idct)
		return ERR_NOT_SUPPORTED;

	if (p->DropLevel > 1)
	{
		mp4_setempty(dec);
		return ERR_NONE;
	}

	dec->pck.CurrTime = p->CurrTime; 

	if (dec->show>=0)
	{
		// pending frame
		
		result = dec->idct->Send(dec->idct,&dec->pck,&dec->dropped);
		if (result == ERR_BUFFER_FULL)
			return result;

		dec->show = -1;
	}

	if (dec->format.Pixel.Flags & PF_FRAGMENTED)
	{
		int processed = 0;
		bool_t added = 0;
		int drop = p->DropLevel;

		result = ERR_NEED_MORE_DATA;

		for (;;)
		{
			if (!getnextframe(dec))
				if (!added)
				{
					added = 1;
					MergeAdd(&dec->buffer,p->Data[0],p->Length);
					if (p->RefTime >= 0)
						dec->starttime = p->RefTime;
					continue;
				}
				else
					break;

			dec->pck.RefTime = dec->starttime;
			dec->pck.DropLevel = drop;

			dec->starttime += dec->frametime;
			drop = 0;

			result = mp4_frame(dec,dec->buffer.Data+processed,dec->startpos-processed);
			
			processed = dec->startpos;

			if (result == ERR_NONE && dec->show>=0) // pending
			{
				if (!added)
					result = ERR_BUFFER_FULL; // resend packet next time
				break;
			}
		}

		dec->startpos -= processed;
		MergeEnd(&dec->buffer,processed);
	}
	else
	{
		dec->pck.RefTime = p->RefTime;
		dec->pck.DropLevel = p->DropLevel;

		result = mp4_frame(dec,p->Data[0],p->Length);
	}

	return result;
}

bool_t mp4_isempty( mp4_decode* dec )
{
	if (!dec->idct)
		return 1;

	if (dec->show>=0)
	{
		packet zero;

		zero.CurrTime = -1;
		zero.DropLevel = dec->pck.DropLevel;
		zero.Length = 0;
		zero.RefTime = -1;

		mp4_packet(dec,&zero);
	}
	return dec->show<0;
}

int mp4_set( node* n, int no, const void* ptr, int size )
{
	int Result = ERR_INVALID_PARAM;
	mp4_decode* dec = DEC(n);
	switch (no)
	{
	case MPEG4_IDCT:
		if (size == sizeof(idct*) && NodeIsClass(*(node**)ptr,IDCT_CLASS))
			Result = mp4_setidct(dec,*(idct**)ptr);
		break;

	case MPEG4_INPUT | PACKET_FORMAT:
		if (!size || size == sizeof(video))
		{
			if (!size) ptr = NULL;
			Result = mp4_setformat(dec,(const video*)ptr);
		}
		break;

	case MPEG4_INPUT:
		if (size == sizeof(packet))
			Result = mp4_packet(dec,(const packet*)ptr);
		break;

	case FLOW_TOTAL:
		if (size == sizeof(int))
		{
			dec->total = *(int*)ptr;
			Result = ERR_NONE;
		}
		break;

	case FLOW_DROPPED:
		if (size == sizeof(int))
		{
			dec->dropped = *(int*)ptr;
			Result = ERR_NONE;
		}
		break;

	case FLOW_EMPTY:
		mp4_setempty(dec);
		Result = ERR_NONE;
		break;
	}
	return Result;
}

int mp4_enum( node* n, int no, datadef* param )
{
	mp4_decode* dec = DEC(n);
	int Result = NodeEnumType(&no,param,NodeParams,FlowParams,dec->params,NULL);
	if (dec->ownidct && Result == ERR_NONE && param->No == MPEG4_IDCT)
		param->Flags |= DF_RDONLY;
	return Result;
}

void mp4_initdec( mp4_decode* dec )
{
	int i;
	dec->vmt.Enum = mp4_enum;
	dec->vmt.Set = mp4_set;
	dec->vmt.Get = mp4_get;

	dec->frame = 0;
	dec->refframe[0] = 0;
	dec->refframe[1] = 1;

	for (i = 0; i < 4*MB_X*2; i++)
		dec->dc_lum[i] = 1024;
	for (i = 0; i < 2*MB_X; i++) 
	{
		dec->dc_chr[0][i] = 1024;
		dec->dc_chr[1][i] = 1024;
	}
}

void mp4_init()
{
	mp4_decode* dec;
	
	dec = &mp4;
	mp4_initdec(dec);
	dec->params = MP4Params;
	dec->gethdr = gethdr;
	dec->IVOP_frame = IVOP_frame;
	dec->PVOP_frame = PVOP_frame;
	dec->BVOP_frame = BVOP_frame;

	NodeRegister(&dec->vmt,PRI_DEFAULT);

	dec = &msmp4;
	mp4_initdec(dec);
	dec->vlctable = vlc;
	dec->params = MSMP4Params;
	dec->gethdr = gethdr_ms;
	dec->IVOP_frame = IVOP_frame_ms;
	dec->PVOP_frame = PVOP_frame_ms;
	dec->BVOP_frame = BVOP_frame_ms;
	msmpeg4_init(dec);

	NodeRegister(&dec->vmt,PRI_DEFAULT);

#if defined(MIPS64)
	msmp4.ownidct = mp4.ownidct = SoftIDCT_Init(MPEG4IDCT_ID);
#endif
}

void mp4_done()
{
	mp4_decode* dec;
	
	dec = &mp4;
	mp4_setidct(dec,NULL);
	mp4_setformat(dec,NULL);
	NodeUnRegister(&dec->vmt);

	dec = &msmp4;
	mp4_setidct(dec,NULL);
	mp4_setformat(dec,NULL);
	NodeUnRegister(&dec->vmt);

#if defined(MIPS64)
	SoftIDCT_Done();
#endif
}
