/*****************************************************************************
 *
 * This code has been developed by Project Mayo. This software is an
 * implementation of a part of one or more MPEG-4 Video tools as
 * specified in ISO/IEC 14496-2 standard.  Those intending to use this
 * software module in hardware or software products are advised that its
 * use may infringe existing patents or copyrights, and any such use
 * would be at such party's own risk.  The original developer of this
 * software module and his/her company, and subsequent editors and their
 * companies (including Project Mayo), will have no liability for use of
 * this software or modifications or derivatives thereof.
 *
 *****************************************************************************
 *																				*	
 * This program is free software ; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 *
 * $Id$
 *
 *****************************************************************************
 *
 * Authors:
 *
 *	Andrea	Graziani  (Ag): Original source code (Open Divx Decoder 0.4a).
 *	Marc	Dukette   (Md) and
 *	Pedro	Mateu     (Pm) and
 *	Gabor	Kovacs    (Kg) Heavily modified and optimized code
 *
 ****************************************************************************/

#include "stdafx.h"

_CONST uint8_t def_quant_intra[64] = 
{
	 8, 17, 18, 19, 21, 23, 25, 27,
	17, 18, 19, 21, 23, 25, 27, 28,
	20, 21, 22, 23, 24, 26, 28, 30,
	21, 22, 23, 24, 26, 28, 30, 32,
	22, 23, 24, 26, 28, 30, 32, 35,
	23, 24, 26, 28, 30, 32, 35, 38,
	25, 26, 28, 30, 32, 35, 38, 41,
	27, 28, 30, 32, 35, 38, 41, 45
};

_CONST uint8_t def_quant_inter[64] = 
{
	16, 17, 18, 19, 20, 21, 22, 23,
	17, 18, 19, 20, 21, 22, 23, 24,
	18, 19, 20, 21, 22, 23, 24, 25,
	19, 20, 21, 22, 23, 24, 26, 27,
	20, 21, 22, 23, 25, 26, 27, 28,
	21, 22, 23, 24, 26, 27, 28, 30,
	22, 23, 24, 26, 27, 28, 30, 31,
	23, 24, 25, 27, 28, 30, 31, 33
};

INLINE int _log2(uint32_t data)
{
	int i;
	for (i=0;data;++i)
		data >>= 1;
    return i;
}

bool_t getnextframe( mp4_decode* dec )
{
	uint8_t* ptr = dec->buffer.Data;
	int len = dec->buffer.Pos;
	int pos = dec->startpos;
	int32_t v = dec->startstate;

	if (!dec->startfound)
		for (;pos<len;++pos)
		{
			v = (v<<8) | ptr[pos];
			if (v == VOP_START_CODE)
			{
				++pos;
				dec->startfound = 1;
				break;
			}
		}

	if (dec->startfound)
		for (;pos<len;++pos)
		{
			v = (v<<8) | ptr[pos];
			if (v == VOP_START_CODE)
			{
				dec->startpos = pos-3;
				dec->startfound = 0;
				dec->startstate = -1;
				return 1;
			}
		}

	dec->startpos = pos;
	dec->startstate = v;
	return 0;
}

void getmatrix( mp4_decode* dec, uint8_t* m )
{
	int i = 0;
	int last,value = 0;

	do 
	{
		last = value;
		loadbits(dec);
		value = getbits(dec,8);
		m[scan[0][i++]] = (uint8_t)value;
	}
	while (value!=0 && i<64);

    i--;
	while (i < 64) 
		m[scan[0][i++]] = (uint8_t)last;
}

int gethdr( mp4_decode *dec )
{
	int code;
	for (;;)
	{
		bytealign(dec);
		loadbits(dec);
		if (eofbits(dec))
			break;
		code = showbits(dec,32);

		if ((code & ~VOL_START_CODE_MASK) == VOL_START_CODE)
		{
			int width,height;
			int visual_object_layer_verid;
			int is_object_layer_identifier;
			flushbits(dec,32+1+8); // start_code + random_accessible_vol + video_object_type_indication
			loadbits(dec);
			 
			is_object_layer_identifier = getbits(dec,1);

			if (is_object_layer_identifier) {
				visual_object_layer_verid = getbits(dec,4);
				flushbits(dec,3); // video_object_layer_priority
			} 
			else {
				visual_object_layer_verid = 1;
			}

			dec->aspect = getbits(dec,4); // aspect ratio
			if (dec->aspect == ASPECT_CUSTOM)
			{
				loadbits(dec);
				dec->aspect_width = getbits(dec,8);
				dec->aspect_height = getbits(dec,8);
			}

			if (getbits1(dec)) // vol control parameters
			{
				flushbits(dec,2); // chroma_format

				dec->tried_bvop = 1;
				dec->bufcount = getbits(dec,1) ? 2:3; // low_delay
				dec->idct->Set(dec->idct,IDCT_BUFFERCOUNT,&dec->bufcount,sizeof(int));
				mp4_updatecount(dec);

				if (getbits1(dec)) // vbv parameters
				{
					flushbits(dec,15+1+15+1+15+1+3+11+1+15+1);
				}
			}
			else
				dec->tried_bvop = 0;

			flushbits(dec,2+1); // shape + marker
			loadbits(dec);

			dec->time_increment_resolution = getbits(dec,16);
			if (dec->time_increment_resolution <= 0)
				dec->time_increment_resolution = 1;
			dec->time_increment_bits = _log2(dec->time_increment_resolution-1);

			flushbits(dec,1); // marker
			if (getbits1(dec)) //fixed_vop_rate
				flushbits(dec,dec->time_increment_bits);

			flushbits(dec,1); // marker
			loadbits(dec);
			width = getbits(dec,13); //width
			flushbits(dec,1);
			height = getbits(dec,13); //height
			flushbits(dec,1+1+1); // marker + interlacing + obmc_disable

			if (mp4_updatesize(dec,width,height) != ERR_NONE)
			{
				dec->valid = 0;
				continue;
			}

			loadbits(dec);
			dec->sprite = getbits(dec,(visual_object_layer_verid==1)?1:2);

			if (dec->sprite == SPRITE_STATIC || dec->sprite == SPRITE_GMC)
			{
				if (dec->showerror)
					NodeError(&dec->vmt,MPEG4_ID,MPEG4_ERROR_GMC);

				if (dec->sprite != SPRITE_GMC)
				{
					flushbits(dec,13+1+13+1+13+1+13+1); //width+marker+height+marker+left+marker+top+marker
					loadbits(dec);
				}
				dec->sprite_warping_points = getbits(dec,6);	
				dec->sprite_warping_accuracy = getbits(dec,2);
				dec->sprite_brightness_change = getbits(dec,1);

				if (dec->sprite != SPRITE_GMC)
					flushbits(dec,1); // low_latency_sprite_enable
			}
				
			loadbits(dec);
			if (getbits1(dec))  // not 8 bit
			{
				dec->quant_precision = getbits(dec,4);
				flushbits(dec,4); // bit per pixel
			}
			else 
				dec->quant_precision = 5;

			if (getbits1(dec)) // quant type
			{
				//if (dec->showerror)
				//	NodeError(&dec->vmt,MPEG4_ID,MPEG4_ERROR_QUANT);

				if (getbits1(dec))	// load intra
					getmatrix(dec,dec->quant_intra);
				else
					memcpy(dec->quant_intra,def_quant_intra,sizeof(def_quant_intra));

				if (getbits1(dec))	// load inter
					getmatrix(dec,dec->quant_inter);
				else
					memcpy(dec->quant_inter,def_quant_inter,sizeof(def_quant_inter));
			}

			loadbits(dec);
			if (visual_object_layer_verid != 1) 
			{
				if (getbits1(dec) && dec->showerror)
					NodeError(&dec->vmt,MPEG4_ID,MPEG4_ERROR_QPEL);
			} 

			flushbits(dec,4);

			dec->valid = 1;
			dec->showerror = 0;
		}
		else if (code == VOP_START_CODE) {

			int time_increment;
			bool_t rounding;

			flushbits(dec,32);

			loadbits(dec);
			dec->prediction_type = getbits(dec,2);

			while (getbits1(dec)) 
				loadbits(dec);

			flushbits(dec,1); //marker
			loadbits(dec);
			time_increment = getbits(dec,dec->time_increment_bits);

			if (!dec->low_delay)
			{
				if (dec->prediction_type != B_VOP) {
					dec->time_pp = (dec->time_increment_resolution + 
						time_increment - dec->last_reftime) % dec->time_increment_resolution;
					dec->last_reftime = time_increment;
				} else {
					dec->time_bp = (dec->time_increment_resolution + 
						dec->last_reftime - time_increment) % dec->time_increment_resolution;
				}
			}

			flushbits(dec,1); //marker

			loadbits(dec);
			if (!getbits1(dec)) 
			{
				dec->prediction_type = N_VOP;
				return 1;
			}  

			rounding = 0;
			if (dec->prediction_type == P_VOP || dec->prediction_type == S_VOP)
				rounding = getbits(dec,1);

			dec->idct->Set(dec->idct,IDCT_ROUNDING,&rounding,sizeof(bool_t));

			flushbits(dec,3);// intra_dc_vlc_thr

			dec->quantizer = getbits(dec,dec->quant_precision); // vop quant

			loadbits(dec);
			if (dec->prediction_type != I_VOP) 
				dec->fcode_for = getbits(dec,3); 

			if (dec->prediction_type == B_VOP) 
				dec->fcode_back = getbits(dec,3); 

			return 1;
		}
		else
			flushbits(dec,8);
	}
	return 0;
}

