/*****************************************************************************
 *
 * This program is free software ; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 *
 * $Id$
 *
 ****************************************************************************/

#include "..\common\common.h"
#include "mpegaudio.h"
#include "libid3tag\id3tag.h"

#define ID3TAG1SIZE	(3+30+30+30+4+30+1)

typedef struct mp3
{
	format_base p;
	bool_t Tag;

} mp3;

static int Init(mp3* p)
{
	format_stream* s = Format_AddStream(&p->p,sizeof(format_stream),sizeof(audio));
	if (s)
	{
		s->FormatType = PACKET_AUDIO;
		s->Fragmented = 1;
		s->DisableDrop = 1; // VBR MP3 could drop because timestamp and duration calculation is wrong
		((audio*)s->Format)->Format = AUDIOFMT_MP3;
	}
	p->Tag = 0;
	return ERR_NONE;
}

static int Probe(mp3* p, const void* Ptr, int Length, const tchar_t* URL )
{
	const uint8_t* Data = (const uint8_t*)Ptr;

	if (id3_tag_query(Ptr,Length)>0)
		return ERR_NONE;

	if (Data[0] == 0xFF && (Data[1] & 0xE0) == 0xE0)
		return ERR_NONE;

	if (CheckExts(URL,p->p.Exts))
		return ERR_NONE;

	return ERR_INVALID_DATA;
}

int Seek(mp3* p, tick_t Time, int FilePos )
{
	format_stream* Stream = p->p.Streams[0];

	if (!p->Tag && Stream->Comment.Node)
	{
		char Buffer[ID3TAG1SIZE];

		p->Tag = 1;

		if (p->p.Input->Seek(p->p.Input,-(int)sizeof(Buffer),SEEK_END)>=0 &&
		    p->p.Input->Read(p->p.Input,Buffer,sizeof(Buffer)) == sizeof(Buffer))

			TagParse(Buffer,sizeof(Buffer),&Stream->Comment);
	}

	if (FilePos < 0)
	{
		if (Time > 0)
		{
			audio* Format = (audio*)Stream->Format;
			if (Format->ByteRate)
				FilePos = Scale(Time,Format->ByteRate,TICKSPERSEC);
			else
				return ERR_NOT_SUPPORTED;
		}
		else
			FilePos = 0;
	}

	return Format_InputSeek(&p->p,FilePos);
}

int ReadPacket(mp3* p, format_packet* Packet)
{
	format_stream* Stream = p->p.Streams[0];
	audio* Format = (audio*)Stream->Format;

	if (p->p.FilePos==0)
		Packet->RefTime = 0;
	else
	if (Format->ByteRate)
		Packet->RefTime = Scale(p->p.FilePos,TICKSPERSEC,Format->ByteRate);
	else
		Packet->RefTime = -1;

	Packet->Data = Format_ReadAsRef(&p->p,-CHUNKSIZE);
	Packet->Stream = Stream;
	return ERR_NONE;
}

static mp3 MP3;

void MP3_Init()
{
	MP3.p = FormatBase;
	MP3.p.VMT.Probe = Probe;
	MP3.p.Init = Init;
	MP3.p.Seek = Seek;
	MP3.p.ReadPacket = ReadPacket;
	MP3.p.Id = MP3_ID;
	MP3.p.Exts = T("mp1:A;mp2:A;mp3:A;mpa:A");

	NodeRegister((node*)&MP3.p.VMT,PRI_DEFAULT);
}

void MP3_Done()
{
	NodeUnRegister((node*)&MP3.p.VMT);
}
