/*****************************************************************************
 *
 * This program is free software ; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 *
 * $Id$
 *
 ****************************************************************************/

#include "stdafx.h"

const datadef MPEG1Params[] = 
{
	{ MPEG1_ID,		TYPE_CLASS },
	{ MPEG1_INPUT,	TYPE_PACKET, DF_INPUT, PACKET_VIDEO },
	{ MPEG1_IDCT,	TYPE_NODE, DF_OUTPUT, IDCT_CLASS },

	DATADEF_END
};

int UpdateCount( mpeg_decode* dec )
{
	if (!dec->IDCT)
	{
		dec->BufCount = 0;
		return ERR_NONE;
	}

	dec->IDCT->Get(dec->IDCT,IDCT_BUFFERCOUNT,&dec->BufCount,sizeof(int));

	if (dec->BufCount >= 3)
		dec->LowDelay = 0;
	else
		dec->LowDelay = 1;

	if (dec->Format.Width > 0 && dec->BufCount < 2)
		return ERR_OUT_OF_MEMORY;

	return ERR_NONE;
}

bool_t InitIDCT( mpeg_decode* dec, idct* IDCT )
{
	video Format;
	int Count;

	Format = dec->Format;
	Format.Pixel.Flags = PF_YUV420;

	if (IDCT)
	{
		if (IDCT->Set(IDCT,IDCT_OUTPUT|PACKET_FORMAT,&Format,sizeof(Format)) != ERR_NONE)
			return 0;

		Count = 2;
		IDCT->Set(IDCT,IDCT_BUFFERCOUNT,&Count,sizeof(int));
		IDCT->Drop(IDCT);
	}

	return 1;
}

int UpdateSize( mpeg_decode* dec, int Width, int Height)
{
	int Result = ERR_NONE;
	if (dec->Format.Width != Width || dec->Format.Height != Height)
	{
		if (Width > MB_X*16 || Height > MB_Y*16)
		{
			NodeError(&dec->VMT,VOUT_CLASS,VOUT_ERROR_SIZE);
			Result = ERR_NOT_SUPPORTED;
		}
		else
		{
			Free(dec->_mv_buf);

			dec->Format.Width = Width;
			dec->Format.Height = Height;
			dec->mb_xsize = ( dec->Format.Width + 15 ) / 16; 
			dec->mb_ysize = ( dec->Format.Height + 15 ) / 16;
			dec->pos_end = dec->mb_ysize * MB_X;
			dec->_mv_buf = (int*) Alloc(32 + 4*sizeof(int) * dec->pos_end);
			if (!dec->_mv_buf)
				return ERR_INVALID_DATA;
			dec->mv_buf = (int*)(((int)dec->_mv_buf + 31) & ~31);

			if (dec->IDCT)
			{
				if (!InitIDCT(dec,dec->IDCT))
					Result = ERR_NOT_SUPPORTED;
				else
					Result = UpdateCount(dec);
			}
		}
	}
	return Result;
}

static int SetIDCT( mpeg_decode* dec, idct* New )
{
	//!!! support for IDCTFORMAT_ATI

	int Format = IDCTFORMAT_BLOCK;
	if (New && New->Set(New,IDCT_FORMAT,&Format,sizeof(Format)) != ERR_NONE)
		return ERR_NOT_SUPPORTED;

	if (dec->IDCT)
	{
		if (IDCTCopy(dec->IDCT,New,2) != ERR_NONE)
			return ERR_NOT_SUPPORTED;
	}
	else
		if (!InitIDCT(dec,New))
			return ERR_NOT_SUPPORTED;

	dec->IDCT = New;
	return UpdateCount(dec);
}

static void Empty( mpeg_decode* dec )
{
	dec->Frame = 0;
	dec->Show = -1;
	dec->StartTime = -1;
	dec->StartState = -1;
	dec->StartPos = 0;
	dec->StartFound = 0;
	MergeClear(&dec->Merge);
	if (dec->IDCT)
		dec->IDCT->Drop(dec->IDCT);
}

static int SetFormat( mpeg_decode* dec, const video* Format )
{
	memset(&dec->Format,0,sizeof(dec->Format));

	Free(dec->_mv_buf);
	dec->_mv_buf = NULL;
	dec->mv_buf = NULL;

	dec->mb_xsize = 0; 
	dec->mb_ysize = 0;
	dec->pos_end = 0;

	dec->Total = 0;
	dec->Dropped = 0;
	dec->Frame = 0;
	dec->Show = -1;
	dec->BufCount = 0;
	dec->LowDelay = 0;
	MergeDone(&dec->Merge);

	if (Format)
	{
		int FourCC = Format->Pixel.FourCC;

		if (!(Format->Pixel.Flags & PF_FOURCC))
			return ERR_INVALID_DATA;

		if (FourCC != FOURCC('M','P','E','G') &&
			FourCC != FOURCC('M','P','G','1'))
			return ERR_INVALID_DATA;

		dec->Format = *Format;
		dec->Format.Width = -1;

		if (UpdateSize(dec,Format->Width,Format->Height) != ERR_NONE)
			return ERR_INVALID_DATA;

		MergeInit(&dec->Merge);

		Empty(dec);
	}

	return ERR_NONE;
}

bool_t GetNextFrame( mpeg_decode* dec )
{
	uint8_t* Ptr = dec->Merge.Data;
	int Len = dec->Merge.Pos;
	int Pos = dec->StartPos;
	int32_t v = dec->StartState;

	if (!dec->StartFound)
		for (;Pos<Len;++Pos)
		{
			v = (v<<8) | Ptr[Pos];
            if (v >= SLICE_MIN_START_CODE && v <= SLICE_MAX_START_CODE)
			{
                Pos++;
                dec->StartFound=1;
                break;
            }
		}

	if (dec->StartFound)
		for (;Pos<Len;++Pos)
		{
			v = (v<<8) | Ptr[Pos];
			if ((v & ~0xFF) == 0x100 &&
				(v < SLICE_MIN_START_CODE || v > SLICE_MAX_START_CODE))
			{
				dec->StartPos = Pos-3;
				dec->StartFound = 0;
				dec->StartState = -1;
				return 1;
			}
		}

	dec->StartPos = Pos;
	dec->StartState = v;
	return 0;
}

#define MPEG1_FRAME_RATE_BASE 1001

_CONST int FrameRate[16] = 
{
        0,
    24000,
    24024,
    25025,
    30000,
    30030,
    50050,
    60000,
    60060,
    15015,
     5005,
    10010,
    12012,
    15015,
    25025,
    25025,
};

_CONST int16_t MPEG1_IntraMatrix[64] = 
{
	8,  16, 19, 22, 26, 27, 29, 34,
	16, 16, 22, 24, 27, 29, 34, 37,
	19, 22, 26, 27, 29, 34, 34, 38,
	22, 22, 26, 27, 29, 34, 37, 40,
	22, 26, 27, 29, 32, 35, 40, 48,
	26, 27, 29, 32, 35, 40, 48, 58,
	26, 27, 29, 34, 38, 46, 56, 69,
	27, 29, 35, 38, 46, 56, 69, 83
};

_CONST int16_t MPEG1_InterMatrix[64] = 
{
    16, 16, 16, 16, 16, 16, 16, 16,
    16, 16, 16, 16, 16, 16, 16, 16,
    16, 16, 16, 16, 16, 16, 16, 16,
    16, 16, 16, 16, 16, 16, 16, 16,
    16, 16, 16, 16, 16, 16, 16, 16,
    16, 16, 16, 16, 16, 16, 16, 16,
    16, 16, 16, 16, 16, 16, 16, 16,
    16, 16, 16, 16, 16, 16, 16, 16,
};

static void Sequence( mpeg_decode* dec )
{
	int i,v;
	int Width,Height;

	loadbits(dec);
	Width = getbits(dec,12);
	Height = getbits(dec,12);

	loadbits(dec);
    getbits(dec,4); // aspect

	dec->Format.FrameRate.Den = 1001;
	dec->Format.FrameRate.Num = FrameRate[getbits(dec,4)];
	dec->FrameTime = Scale(TICKSPERSEC,dec->Format.FrameRate.Den,dec->Format.FrameRate.Num);

	loadbits(dec);
    dec->Format.ByteRate = getbits(dec,18) * 50;

    flushbits(dec,1); // marker

	UpdateSize(dec,Width,Height);

    flushbits(dec,10); // vbv_buffer_size
    flushbits(dec,1);

	loadbits(dec);
    if (getbits1(dec)) 
	{
        for (i=0;i<64;i++) 
		{
			loadbits(dec);
            v = getbits(dec,8);
            dec->IntraMatrix[i] = (uint8_t)v;
            dec->ChromaIntraMatrix[i] = (uint8_t)v;
        }
    } 
	else 
	{
        for (i=0;i<64;i++) 
		{
            v = MPEG1_IntraMatrix[i];
            dec->IntraMatrix[i] = (uint8_t)v;
            dec->ChromaIntraMatrix[i] = (uint8_t)v;
        }
    }

    if (getbits1(dec)) 
	{
        for (i=0;i<64;i++) 
		{
			loadbits(dec);
            v = getbits(dec, 8);
            dec->InterMatrix[i] = (uint8_t)v;
            dec->ChromaInterMatrix[i] = (uint8_t)v;
        }
    } 
	else 
	{
        for (i=0;i<64;i++) 
		{
            v = MPEG1_InterMatrix[i];
            dec->InterMatrix[i] = (uint8_t)v;
            dec->ChromaInterMatrix[i] = (uint8_t)v;
        }
    }
}

static void Picture( mpeg_decode* dec )
{
	int f_code;

    flushbits(dec,10);

    loadbits(dec);
	dec->prediction_type = getbits(dec,3);

    flushbits(dec,16);

	loadbits(dec);
    if (dec->prediction_type == P_VOP || dec->prediction_type == B_VOP)
	{
        dec->full_pel[0] = getbits(dec,1);
        f_code = getbits(dec,3);
        dec->f_code[0][0] = f_code;
        dec->f_code[0][1] = f_code;
    }

    if (dec->prediction_type == B_VOP) 
	{
        dec->full_pel[1] = getbits(dec,1);
        f_code = getbits(dec,3);
        dec->f_code[1][0] = f_code;
        dec->f_code[1][1] = f_code;
    }

    dec->dc_lum_scaler = 8;
    dec->dc_chr_scaler = 8;
    dec->first_slice = 1;
}

static void ExtSequence( mpeg_decode* dec)
{
	//!!!
}

static void ExtQuantMatrix( mpeg_decode* dec)
{
	//!!!
}

static void ExtPictureCoding( mpeg_decode* dec)
{
	//!!!
}

static int Frame( mpeg_decode* dec, const uint8_t* Ptr, int Len, bool_t Drop, int RefTime )
{
	int Result;
	int Code;
	int Type;

	if (Len == 0)
		return ERR_INVALID_DATA;

	initbits(dec,Ptr,Len);

	Result = ERR_NONE;

	DEBUG_MSG1(T("MPEG Frame Length:%d"),Len);

	for (;;)
	{
		bytealign(dec);
		loadbits(dec);
		if (eofbits(dec))
			break;

		Code = showbits(dec,32);
		if ((Code & ~0xFF)==0x100)
		{
			DEBUG_MSG1(T("MPEG StartCode:%04x"),Code);

			flushbits(dec,32);
			switch (Code)
			{
			case SEQ_START_CODE:
				Sequence(dec);
				break;
			case PICTURE_START_CODE:
				Picture(dec);
				break;
            case EXT_START_CODE:
				Type = getbits(dec,4);
				switch(Type) 
				{
				case 0x1:
					ExtSequence(dec);
					break;
				case 0x3:
					ExtQuantMatrix(dec);
					break;
				case 0x8:
					ExtPictureCoding(dec);
					break;
				}
				break;
	        case GOP_START_CODE:
                break;
			default:
				if (Code >= SLICE_MIN_START_CODE && Code <= SLICE_MAX_START_CODE)
				{


				}
				break;
			}
		}
		else
			flushbits(dec,8);
	}

	if (dec->Show>=0)
	{
		if (dec->IDCT->Send(dec->IDCT,&dec->Packet,&dec->Dropped) != ERR_BUFFER_FULL)
			dec->Show = -1;
	}

	return Result;
}

static int Packet( mpeg_decode* dec, const packet* p )
{
	int Result;

	if (!dec->IDCT)
		return ERR_NOT_SUPPORTED;

	if (p->DropLevel > 1)
	{
		Empty(dec);
		return ERR_NONE;
	}

	dec->Packet.CurrTime = p->CurrTime; 

	if (dec->Show>=0)
	{
		// pending frame
		
		Result = dec->IDCT->Send(dec->IDCT,&dec->Packet,&dec->Dropped);
		if (Result == ERR_BUFFER_FULL)
			return Result;

		dec->Show = -1;
	}

	if (dec->Format.Pixel.Flags & PF_FRAGMENTED)
	{
		int Processed = 0;
		bool_t Added = 0;
		bool_t Drop = p->DropLevel > 0;

		Result = ERR_NEED_MORE_DATA;

		for (;;)
		{
			if (!GetNextFrame(dec))
				if (!Added)
				{
					Added = 1;
					MergeAdd(&dec->Merge,p->Data[0],p->Length);
					if (p->RefTime >= 0)
						dec->StartTime = p->RefTime;
					continue;
				}
				else
					break;

			Result = Frame(dec,dec->Merge.Data+Processed,dec->StartPos-Processed,Drop,dec->StartTime);
			
			dec->StartTime += dec->FrameTime;
			Processed = dec->StartPos;
			Drop = 0;

			if (Result == ERR_NONE && dec->Show>=0) // pending
			{
				if (!Added)
					Result = ERR_BUFFER_FULL; // resend packet next time
				break;
			}
		}

		dec->StartPos -= Processed;
		MergeEnd(&dec->Merge,Processed);
	}
	else
		Result = Frame(dec,p->Data[0],p->Length,p->DropLevel>0,p->RefTime);

	return Result;
}

static int Enum( mpeg_decode* dec, int No, datadef* Param )
{
	return NodeEnumType(&No,Param,NodeParams,FlowParams,MPEG1Params,NULL);
}

static int Get( mpeg_decode* dec, int No, void* Data, int Size )
{
	int Result = ERR_INVALID_PARAM;
	switch (No)
	{
	case NODE_ID: GETVALUE(MPEG1_ID,int); break;
	case FLOW_TOTAL:GETVALUE(dec->Total,int); break;
	case FLOW_DROPPED: GETVALUE(dec->Dropped,int); break;
	case MPEG1_INPUT|PACKET_FORMAT: GETVALUE(dec->Format,video); break;
	case MPEG1_IDCT: GETVALUE(dec->IDCT,idct*); break;
	}
	return Result;
}

static int Set( mpeg_decode* dec, int No, const void* Data, int Size )
{
	int Result = ERR_INVALID_PARAM;
	switch (No)
	{
	case MPEG1_IDCT:
		if (Size == sizeof(idct*) && NodeIsClass(*(node**)Data,IDCT_CLASS))
			Result = SetIDCT(dec,*(idct**)Data);
		break;

	case MPEG1_INPUT | PACKET_FORMAT:
		if (!Size || Size == sizeof(video))
		{
			if (!Size) Data = NULL;
			Result = SetFormat(dec,(const video*)Data);
		}
		break;

	case MPEG1_INPUT:
		if (Size == sizeof(packet))
			Result = Packet(dec,(const packet*)Data);
		break;

	case FLOW_TOTAL: SETVALUE(dec->Total,int,ERR_NONE); break;
	case FLOW_DROPPED: SETVALUE(dec->Dropped,int,ERR_NONE); break;

	case FLOW_EMPTY:
		dec->Frame = 0;
		dec->Show = -1;
		dec->StartState = -1;
		dec->StartFound = 0;
		MergeClear(&dec->Merge);
		if (dec->IDCT)
			dec->IDCT->Drop(dec->IDCT);
		Result = ERR_NONE;
		break;
	}
	return Result;
}

static mpeg_decode MPEG1 = 
{
	Enum,
	Get,
	Set,
};

void MPEG1_Init()
{
	NodeRegister(&MPEG1.VMT,PRI_DEFAULT);
}

void MPEG1_Done()
{
	NodeUnRegister(&MPEG1.VMT);
}