/*****************************************************************************
 *
 * This program is free software ; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 *
 * $Id$
 *
 ****************************************************************************/

#include "stdafx.h"

typedef struct mves
{
	format_base p;
	int LastSeek;
	bool_t MP4;

} mves;

static int Init(mves* p)
{
	format_stream* s = Format_AddStream(&p->p,sizeof(format_stream),sizeof(video));
	if (s)
	{
		video* Video = (video*)s->Format;
		s->FormatType = PACKET_VIDEO;
		s->Fragmented = 1;
		Video->Pixel.Flags = PF_FOURCC | PF_FRAGMENTED;
		Video->Pixel.FourCC = p->MP4 ? FOURCC('M','P','4','V') : FOURCC('M','P','E','G');
		Video->FrameRate.Num = 25; //fixed frame rate
		Video->FrameRate.Den = 1;
	}
	p->LastSeek = 0;

	return ERR_NONE;
}

static int Probe(mves* p, const void* Ptr, int Length, const tchar_t* URL )
{
	if (CheckExts(URL,p->p.Exts))
	{
		p->MP4 = CheckExts(URL,T("m4v:V")) != 0;
		return ERR_NONE;
	}

	return ERR_INVALID_DATA;
}

int Seek(mves* p, tick_t Time, int FilePos )
{
	if (FilePos < 0)
	{
		if (Time > 0)
			return ERR_NOT_SUPPORTED;
		else
			FilePos = 0;
	}

	p->LastSeek = FilePos;
	return Format_InputSeek(&p->p,FilePos);
}

int ReadPacket(mves* p, format_packet* Packet)
{
	format_stream* Stream = p->p.Streams[0];

	if (p->p.FilePos==0)
		Packet->RefTime = 0;
	else
	if (p->p.FilePos==p->LastSeek)
		Packet->RefTime = 1;
	else
		Packet->RefTime = -1;

	Packet->Data = Format_ReadAsRef(&p->p,-CHUNKSIZE);
	Packet->Stream = Stream;
	return ERR_NONE;
}

static mves MVES;

void MVES_Init()
{
	MVES.p = FormatBase;
	MVES.p.VMT.Probe = Probe;
	MVES.p.Init = Init;
	MVES.p.Seek = Seek;
	MVES.p.ReadPacket = ReadPacket;
	MVES.p.Id = MVES_ID;
	MVES.p.Exts = T("m1v:V;m2v:V;m4v:V");

	NodeRegister((node*)&MVES.p.VMT,PRI_DEFAULT);
}

void MVES_Done()
{
	NodeUnRegister((node*)&MVES.p.VMT);
}
