/*****************************************************************************
 *
 * This program is free software ; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 *
 * $Id$
 *
 ****************************************************************************/

#include "stdafx.h"

#ifdef ARM

#define CLIENTVERSION "1.29.031201.07700" 
#define MAXBUF 6
#define SCALE	MAXBUF

#define ALIGN16(x) (((x) + 15) & ~15)
#define DCFACTOR 8

enum
{
	PLANE_Y			= 0x10000000,
	PLANE_U			= 0x20000000,
	PLANE_V			= 0x40000000,
	SUCCESS			= 0,
	FAIL			= 1,
	ENDENUM			= 5,

	FORMAT_PAL4		= 1,
	FORMAT_PAL8		= 2,
	FORMAT_RGB444	= 3,
	FORMAT_RGB555	= 4,
	FORMAT_RGB565	= 5,
	FORMAT_YUYV		= 6,
	FORMAT_YVYU		= 7,
	FORMAT_YUV422	= 8,
	FORMAT_YUV420	= 9,
	FORMAT_UYVY		= 10,
	FORMAT_VYUY		= 11,
	FORMAT_PAL2		= 12,

	MIRROR_UPDOWN	= 1,
	MIRROR_LEFTRIGHT= 2,

	MEM_EXTERNAL	= 0x20,
	MEM_INTERNAL	= 0x40,

	MEM_ANY_INT		= -1, // prefer internal
	MEM_ANY_EXT		= -2, // prefer external

	PROP_SCANTYPE	= 0,
	PROP_PADDING	= 1,
	PROP_BIAS		= 2,
	PROP_ROUNDING	= 3,
	PROP_FRAMENO	= 4,
	PROP_CODING		= 5,
	PROP_DEBUG		= 6,
};

typedef struct devinfo
{
	char Name[0x50];
	char Version[0x50];
	int Revision;
	int ChipId;
	int RevisionId;
	int TotalMemory;
	int BusInterfaceMode;
	int InternalMemSize;
	int ExternalMemSize;
	int Caps1;
	int Caps2;
	int Caps3;
	int Caps4;
	char Pad[128];
} devinfo;

typedef struct modeinfo
{
	int Width;
	int Height;
	int PixelFormat;
	int Freq;
	int Rotation;
	int Mirror;
} modeinfo;

typedef struct surfinfo
{
	int Width;
	int Height;
	void* Addr;
	int Pitch;
	int Size;
	int Class;
	int One;
} surfinfo;

typedef struct planeinfo
{
	int Width;
	int Height;
	void* Addr;
	int Pitch;
	int One;
} planeinfo;

typedef struct ahirect
{
	int Left;
	int Top;
	int Right;
	int Bottom;
} ahirect;

typedef struct ahipoint
{
	int x;
	int y;
} ahipoint;

typedef void* ahidevice;
typedef void* ahicontext;
typedef void* ahisurface;
typedef void* ahiidct;

typedef struct ahioverlay
{
	ahirect DstRect;
	ahirect SrcRect;
	int ColorKey;
	int ColorMask;
	int Rotate;
	int Mirror;
	int Zero;
} ahioverlay;

typedef struct ahi
{
	overlay_base p;
	idct IdctVMT;

	bool_t Shrink;
	bool_t Stretch;
	bool_t StretchOnlyInternal;
	rect OverlayRect;
	video Overlay;
	blitfx SoftFX;

	bool_t FullScreen;	  // overlay covers fullscreen	
	bool_t Internal;	  // IDCT is using internal memory (it's faster, but less memory)
	bool_t Landscape;	  // avoid external memory overlays in landscape mode
	bool_t IDCTBug;
	bool_t IDCTInit;
	bool_t ErrorMemory;

	int OverlayRPitch;
	int ScaleWidth;
	int ScaleHeight;
	devinfo DevInfo;
	modeinfo ModeInfo;
	ahidevice* Device;
	ahicontext* Context;
	int MaxCount;
	int AllCount; //BufCount + TempCount
	int BufCount;
	int TempCount;
	int ShowNext;
	int ShowCurr;
	int ShowLast;
	ahisurface* Buf[MAXBUF+1];  
	ahisurface* BufR[MAXBUF+1]; 
	int BufMode[MAXBUF+1]; 
	int BufFrameNo[MAXBUF+1];
	ahioverlay Param;
	tchar_t Version[128];

	//idct
	tick_t LastTime;
	pin IDCTOutput;
	int FrameNo;
	int Zero;
	bool_t IDCTRounding;
	int IDCTFormat;
	ahiidct* IDCT;
	int DCAdd;
	//int DCAdj;
	int MacroWidth;
	int MacroNo;
	int MacroSubNo;

	int MVBack;
	int MVFwd;
	idct_ati Data[16];

	int SaveCount;
	int SaveScanType;
	int SaveMVBack;
	int SaveMVFwd;
	idct_ati SaveData[16];

	idct_block_t Block[8*8];

	int (*AhiInit)( int Version );
	int (*AhiTerm)();

	int (*AhiDevEnum)( ahidevice** Device, devinfo* Info, int No );
	int (*AhiDevOpen)( ahicontext** ContextOut, ahidevice* Device, const char* Name, int v );
	int (*AhiDevClose)( ahicontext* Context );
	int (*AhiDevClientVersion)( ahicontext* Context, const char* s );

	int (*AhiDispModeGet)( ahicontext* Context, modeinfo* Info );
	int (*AhiDispModeEnum)( ahicontext* Context, modeinfo* Info, int No );
	int (*AhiDispModeClip)( ahicontext* Context, const ahirect* ); //not sure
	int (*AhiDispSurfGet)(ahicontext* Context, ahisurface** SurfaceOut );

	int (*AhiSurfGetLargestFreeBlockSize)( ahicontext* Context, int Format, int* a, int* b, int Mem );
	int (*AhiSurfAlloc)(ahicontext* Context, ahisurface** SurfaceOut, int* Size, int Format, int Mem );
	int (*AhiSurfReuse)(ahicontext* Context, ahisurface** SurfaceOut, ahisurface* SurfaceReuse, int* Size, int Format, int Zero );
	int (*AhiSurfFree)(ahicontext* Context, ahisurface* s );
	int (*AhiSurfLock)(ahicontext* Context, ahisurface* s, void** Ptr, int Plane );
	int (*AhiSurfUnlock)( ahicontext* Context, ahisurface* s, int Plane );
	int (*AhiSurfInfo)(ahicontext* Context, ahisurface* s, surfinfo* Info );
	int (*AhiSurfPlaneInfo)(ahicontext* Context, ahisurface* s, planeinfo* Info, int Plane);

	int (*AhiDrawRopSet)(ahicontext* Context, int ROP);
	int (*AhiDrawSurfSrcSet)(ahicontext* Context, ahisurface* Src, int Plane);
	int (*AhiDrawSurfDstSet)(ahicontext* Context, ahisurface* Dst, int Plane);
	int (*AhiDrawIdle)(ahicontext* Context, int v);
	int (*AhiDispWaitVBlank)(ahicontext* Context, int v);
	int (*AhiDrawBitBlt)(ahicontext* Context, const ahirect* DstRect, const ahipoint* SrcPos );
	int (*AhiDrawStretchBlt)(ahicontext* Context, const ahirect* DstRect, const ahirect* SrcPos, int Mode );
	int (*AhiDrawRotateBlt)(ahicontext* Context, const ahirect* DstRect, const ahipoint* SrcPos, int Rotate, int Mirror, int Zero);

	int (*AhiDispOverlayCaps)(ahicontext* Context, int Mode, int* Caps );
	int (*AhiDispOverlayState)(ahicontext* Context, int State, int Zero );
	int (*AhiDispOverlayPos)(ahicontext* Context, const ahirect* Rect, int Zero );
	int (*AhiDispOverlaySurf)(ahicontext* Context, ahisurface* Surface, ahioverlay* Info, int Mode);

	int (*AhiIDCTOpen)(ahicontext* Context, ahiidct** IDCTOut, int Zero );
	int (*AhiIDCTClose)( ahiidct* IDCT );
	int (*AhiIDCTPropGet)(ahiidct* IDCT, int No, int* Data, int Size);
	int (*AhiIDCTPropSet)(ahiidct* IDCT, int No, const int* Data, int Size);
	int (*AhiIDCTFrameStart)( ahiidct* IDCT, ahisurface* Dst, ahisurface* Back, ahisurface* Fwd, int Frame, int Zero );
	int (*AhiIDCTFrameEnd)( ahiidct* IDCT, int Frame );
	int (*AhiIDCTMComp8x8)(ahiidct* IDCT, const idct_ati* Data, int Len, int* Back, int* Fwd, int No, int ZigZagType);
	int (*AhiIDCTMComp)(ahiidct* IDCT, int* Back, int* Fwd, int No );
	int (*AhiIDCT8x8)(ahiidct* IDCT, const idct_ati* Data, int Len, int No, int ZigZagType);

} ahi;

const datadef AHIParams[] = 
{
	{ AHI_ID,			TYPE_CLASS },
	{ AHI_DEVICEBITMAP,	TYPE_BOOL, DF_SETUP|DF_SOFTRESET|DF_OWNREG|DF_CHECKLIST },
	{ AHI_IDCTBUG,		TYPE_BOOL, DF_SETUP|DF_CHECKLIST },
	{ AHI_NAME,			TYPE_STRING, DF_SETUP | DF_RDONLY|DF_GAP },
	{ AHI_VERSION,		TYPE_STRING, DF_SETUP | DF_RDONLY },
	{ AHI_REVISION,		TYPE_INT, DF_SETUP | DF_RDONLY },
	{ AHI_CHIPID,		TYPE_INT, DF_SETUP | DF_RDONLY|DF_HEX },
	{ AHI_REVISIONID,	TYPE_INT, DF_SETUP | DF_RDONLY },
	{ AHI_TOTALMEMORY,	TYPE_INT, DF_SETUP | DF_RDONLY },
	{ AHI_INTERNALTOTALMEM,	TYPE_INT, DF_SETUP | DF_RDONLY },
	{ AHI_INTERNALFREEMEM,	TYPE_INT, DF_SETUP | DF_RDONLY },
	{ AHI_EXTERNALTOTALMEM,	TYPE_INT, DF_SETUP | DF_RDONLY },
	{ AHI_EXTERNALFREEMEM,	TYPE_INT, DF_SETUP | DF_RDONLY },
	{ AHI_CAPS1,		TYPE_INT, DF_SETUP | DF_RDONLY|DF_HEX },
	{ AHI_CAPS2,		TYPE_INT, DF_SETUP | DF_RDONLY|DF_HEX },
	{ AHI_CAPS3,		TYPE_INT, DF_SETUP | DF_RDONLY|DF_HEX },
	{ AHI_CAPS4,		TYPE_INT, DF_SETUP | DF_RDONLY|DF_HEX },

	DATADEF_END
};

static int IDCTSetFormat(ahi* p,int Format);

const tchar_t* ATIBase = T("System\\GDI\\Drivers\\ATI");
const tchar_t* DisableDevice = T("DisableDeviceBitmap");

static int Set(ahi* p,int No,const void* Data,int Size)
{
	int Result = OvlSet(&p->p,No,Data,Size);
	switch (No)
	{
	case AHI_DEVICEBITMAP:
		if (Size==sizeof(bool_t))
		{
			DWORD Disp;
			DWORD Value;
			HKEY Key;
			if (RegCreateKeyEx( HKEY_LOCAL_MACHINE, ATIBase, 0, NULL, 0, 0, NULL, &Key, &Disp ) == ERROR_SUCCESS)
			{
				Value = *(bool_t*)Data;
				RegSetValueEx( Key, DisableDevice, 0, REG_DWORD, (LPBYTE)&Value, sizeof(Value));
				RegCloseKey( Key );
				Result = ERR_NONE;
			}

		}
		break;

	case AHI_IDCTBUG: 
		if (Size==sizeof(bool_t))
		{
			p->IDCTBug = *(bool_t*)Data;
			IDCTSetFormat(p,p->IDCTFormat);
		}
		break;
	}

	return Result;
}

static bool_t CheckVersion(ahi* p)
{
	int a,b;

	strtotcs(p->Version,p->DevInfo.Version,sizeof(p->Version)/sizeof(tchar_t));
	if (stscanf(p->Version,T("%d.%d"),&a,&b)!=2)
		a = 2; // force true 
	return a>1 || (a==1 && b>=27);
}

static int GetFreeMemory(ahi* p,int Mem)
{
	int a=0;
	int b;
	if (p->AhiSurfGetLargestFreeBlockSize)
	{
		if (p->Context)
			p->AhiSurfGetLargestFreeBlockSize(p->Context,FORMAT_YUV420,&a,&b,Mem);
		else
		if (p->AhiInit(0)==SUCCESS)
		{
			int No;
			for (No=0;;No++)
			{
				int Result = p->AhiDevEnum(&p->Device,&p->DevInfo,No);
				if (Result == SUCCESS && CheckVersion(p))
				{
					if (p->AhiDevOpen(&p->Context,p->Device,"BetaPlayer",2)==SUCCESS)
					{
						p->AhiDevClientVersion(p->Context,CLIENTVERSION);
						p->AhiSurfGetLargestFreeBlockSize(p->Context,FORMAT_YUV420,&a,&b,Mem);
						p->AhiDevClose(p->Context);

						a = (a * 3) >> 1; // 12bit/pixel
					}
					p->Context = NULL;
					break;
				}
				if (Result == ENDENUM)
					break;		
			}

			p->AhiTerm();
		}
	}
	return a;
}

static int Get(ahi* p,int No,void* Data,int Size)
{
	int Result = OvlGet(&p->p,No,Data,Size);
	switch (No)
	{
	case AHI_NAME: 
		strtotcs((tchar_t*)Data,p->DevInfo.Name,Size/sizeof(tchar_t));
		Result = ERR_NONE;
		break;

	case AHI_VERSION: 
		strtotcs((tchar_t*)Data,p->DevInfo.Version,Size/sizeof(tchar_t));
		Result = ERR_NONE;
		break;

	case AHI_DEVICEBITMAP:
		if (Size==sizeof(bool_t))
		{
			HKEY Key;

			Result = ERR_NONE;
			*(bool_t*)Data = 0;

			if (RegOpenKeyEx( HKEY_LOCAL_MACHINE, ATIBase, 0, 0, &Key ) == ERROR_SUCCESS)
			{
				DWORD Value;
				DWORD RegSize = sizeof(Value);
				DWORD RegType;

				if (RegQueryValueEx( Key, DisableDevice, 0, &RegType, (LPBYTE)&Value, &RegSize ) == ERROR_SUCCESS && 
					RegType == REG_DWORD && Value!=0)
					*(bool_t*)Data = 1;

				RegCloseKey( Key );
			}
		}
		break;

	case AHI_IDCTBUG: GETVALUE(p->IDCTBug,bool_t); break;
	case AHI_REVISION: GETVALUE(p->DevInfo.Revision,int); break;
	case AHI_CHIPID: GETVALUE(p->DevInfo.ChipId,int); break;
	case AHI_REVISIONID: GETVALUE(p->DevInfo.RevisionId,int); break;
	case AHI_TOTALMEMORY: GETVALUE(p->DevInfo.TotalMemory,int); break;
	case AHI_INTERNALTOTALMEM: GETVALUE(p->DevInfo.InternalMemSize,int); break;
	case AHI_EXTERNALTOTALMEM: GETVALUE(p->DevInfo.ExternalMemSize,int); break;
	case AHI_INTERNALFREEMEM: GETVALUE(GetFreeMemory(p,MEM_INTERNAL),int); break;
	case AHI_EXTERNALFREEMEM: GETVALUE(GetFreeMemory(p,MEM_EXTERNAL),int); break;
	case AHI_CAPS1: GETVALUE(p->DevInfo.Caps1,int); break;
	case AHI_CAPS2: GETVALUE(p->DevInfo.Caps2,int); break;
	case AHI_CAPS3: GETVALUE(p->DevInfo.Caps3,int); break;
	case AHI_CAPS4: GETVALUE(p->DevInfo.Caps4,int); break;
	}
	return Result;
}

static bool_t Ahi2PixelFormat(int Format, pixel* p )
{
	switch (Format)
	{
	case FORMAT_PAL2:
		p->Flags = PF_PALETTE;
		p->BitCount = 2;
		break;
	case FORMAT_PAL4:
		p->Flags = PF_PALETTE;
		p->BitCount = 2;
		break;
	case FORMAT_PAL8:
		p->Flags = PF_PALETTE;
		p->BitCount = 2;
		break;
	case FORMAT_RGB444:
		DefaultRGB(p,16,4,4,4,0,0,0);
		break;
	case FORMAT_RGB555:
		DefaultRGB(p,16,5,5,5,0,0,0);
		break;
	case FORMAT_RGB565:
		DefaultRGB(p,16,5,6,5,0,0,0);
		break;
	case FORMAT_YUYV:
		p->Flags = PF_FOURCC;
		p->FourCC = FOURCC_YUYV;
		break;
	case FORMAT_YVYU:
		p->Flags = PF_FOURCC;
		p->FourCC = FOURCC_YVYU;
		break;
	case FORMAT_YUV420:
		p->Flags = PF_FOURCC;
		p->FourCC = FOURCC_I420;
		break;
	case FORMAT_UYVY:
		p->Flags = PF_FOURCC;
		p->FourCC = FOURCC_UYVY;
		break;
	case FORMAT_VYUY:
		p->Flags = PF_FOURCC;
		p->FourCC = FOURCC_VYUY;
		break;

	case FORMAT_YUV422:
	default:
		return 0;
	}
	return 1;
}

static bool_t Hide(ahi* p)
{
	if (p->p.Show && p->AllCount)
	{
		p->AhiDispOverlayState(p->Context,0,0);
		p->AhiDispWaitVBlank(p->Context,0);
	}
	return 1;
}

static void FreeScale(ahi* p,bool_t* Hidden)
{
	if (p->Buf[SCALE])
	{
		*Hidden = Hide(p);
		p->AhiSurfFree(p->Context,p->Buf[SCALE]);
		p->Buf[SCALE] = NULL;
		p->BufMode[SCALE] = 0;
	}
	p->ScaleWidth = 0;
	p->ScaleHeight = 0;
}

static void FreeBuffer(ahi* p,int No,bool_t* Hidden)
{
	if (Hidden && p->ShowCurr == No)
		*Hidden = Hide(p);

	if (p->Buf[No])
		p->AhiSurfFree(p->Context,p->Buf[No]);
	p->Buf[No] = NULL;

	if (p->BufR[No])
		p->AhiSurfFree(p->Context,p->BufR[No]);
	p->BufR[No] = NULL;

	p->BufMode[No] = 0;
}

static bool_t AllocBuffer(ahi* p,int No,int Width,int Height,int Mode,int Clear)
{
	int FirstMode;
	int SecondMode;
	ahisurface* Surf = NULL;
	ahisurface* SurfR = NULL;
	int FinalMode = 0;

	int Size[2];
	Size[0] = Width;
	Size[1] = Height;

	if (Mode == MEM_ANY_INT)
	{
		FirstMode = MEM_INTERNAL;
		SecondMode = MEM_EXTERNAL;
	}
	else
	if (Mode == MEM_ANY_EXT)
	{
		FirstMode = MEM_EXTERNAL;
		SecondMode = MEM_INTERNAL;
	}
	else
	{
		FirstMode = Mode;
		SecondMode = 0;
	}

	p->AhiSurfAlloc(p->Context,&Surf,Size,FORMAT_YUV420,FirstMode);
	if (Surf)
		FinalMode = FirstMode;

	if (!Surf && SecondMode)
	{
		p->AhiSurfAlloc(p->Context,&Surf,Size,FORMAT_YUV420,SecondMode);
		if (Surf)
			FinalMode = SecondMode;
	}

	DEBUG_MSG4(T("AHIALLOC %d: %d,%d %d"),No,Size[0],Size[1],FinalMode);

	if (Surf)
	{
		planeinfo Info;

		p->AhiSurfPlaneInfo(p->Context,Surf,&Info,PLANE_Y);
		if (No != SCALE)
		{
			p->Overlay.Pitch = Info.Pitch;
			Size[0] = Height;
			Size[1] = Width;

			p->AhiSurfReuse(p->Context,&SurfR,Surf,Size,FORMAT_YUV420,0);

			if (!SurfR)
			{
				// this shouldn't happen
				p->AhiSurfFree(p->Context,Surf);
				return 0;
			}

			p->AhiSurfPlaneInfo(p->Context,SurfR,&Info,PLANE_Y);
			p->OverlayRPitch = Info.Pitch;
		}

		p->Buf[No] = Surf;
		p->BufR[No] = SurfR;
		p->BufMode[No] = FinalMode;
		p->BufFrameNo[No] = -1;

		if (Clear)
		{
			void* Ptr;

			p->AhiSurfPlaneInfo(p->Context,Surf,&Info,PLANE_Y);
			if (p->AhiSurfLock(p->Context,Surf,&Ptr,PLANE_Y)==SUCCESS)
			{
				FillColor(Ptr,Info.Pitch,Info.Width,Info.Height,8,0);
				p->AhiSurfUnlock(p->Context,Surf,PLANE_Y);
			}

			p->AhiSurfPlaneInfo(p->Context,Surf,&Info,PLANE_U);
			if (p->AhiSurfLock(p->Context,Surf,&Ptr,PLANE_U)==SUCCESS)
			{
				FillColor(Ptr,Info.Pitch,Info.Width,Info.Height,8,128);
				p->AhiSurfUnlock(p->Context,Surf,PLANE_U);
			}

			p->AhiSurfPlaneInfo(p->Context,Surf,&Info,PLANE_V);
			if (p->AhiSurfLock(p->Context,Surf,&Ptr,PLANE_V)==SUCCESS)
			{
				FillColor(Ptr,Info.Pitch,Info.Width,Info.Height,8,128);
				p->AhiSurfUnlock(p->Context,Surf,PLANE_V);
			}
		}
		return 1;
	}
	return 0;
}

static void Stretch(ahi* p)
{
	ahirect DstRect;
	ahirect SrcRect;
	int ScaleMode = (p->p.FX.Flags & BLITFX_ARITHSTRETCH) ? 5:1;

	DstRect.Left = 0;
	DstRect.Top = 0;
	DstRect.Right = p->ScaleWidth >> 1;
	DstRect.Bottom = p->ScaleHeight >> 1;

	SrcRect.Left = p->OverlayRect.x >> 1;
	SrcRect.Top = p->OverlayRect.y >> 1;
	SrcRect.Right = (p->OverlayRect.x + p->OverlayRect.Width) >> 1;
	SrcRect.Bottom = (p->OverlayRect.y + p->OverlayRect.Height) >> 1;

	if (p->Shrink) p->AhiDrawIdle(p->Context,0);
	p->AhiDrawSurfSrcSet(p->Context,p->Buf[p->ShowCurr],PLANE_U);
	p->AhiDrawSurfDstSet(p->Context,p->Buf[SCALE],PLANE_U);
	if (p->Stretch)
		p->AhiDrawStretchBlt(p->Context,&DstRect,&SrcRect,ScaleMode);
	else
		p->AhiDrawBitBlt(p->Context,&DstRect,(ahipoint*)&SrcRect);

	if (p->Shrink) p->AhiDrawIdle(p->Context,0);
	p->AhiDrawSurfSrcSet(p->Context,p->Buf[p->ShowCurr],PLANE_V);
	p->AhiDrawSurfDstSet(p->Context,p->Buf[SCALE],PLANE_V);
	if (p->Stretch)
		p->AhiDrawStretchBlt(p->Context,&DstRect,&SrcRect,ScaleMode);
	else
		p->AhiDrawBitBlt(p->Context,&DstRect,(ahipoint*)&SrcRect);

	DstRect.Left <<= 1;
	DstRect.Top <<= 1;
	DstRect.Right <<= 1;
	DstRect.Bottom <<= 1;

	SrcRect.Left <<= 1;
	SrcRect.Top <<= 1;
	SrcRect.Right <<= 1;
	SrcRect.Bottom <<= 1;

	if (p->Shrink) p->AhiDrawIdle(p->Context,0);
	p->AhiDrawSurfSrcSet(p->Context,p->Buf[p->ShowCurr],PLANE_Y);
	p->AhiDrawSurfDstSet(p->Context,p->Buf[SCALE],PLANE_Y);
	if (p->Stretch)
		p->AhiDrawStretchBlt(p->Context,&DstRect,&SrcRect,ScaleMode);
	else
		p->AhiDrawBitBlt(p->Context,&DstRect,(ahipoint*)&SrcRect);
}

static void UpdateOverlaySurf(ahi* p)
{
	int Show = p->ShowCurr;
	if (p->Buf[SCALE])
		Show = SCALE;
	if (p->Buf[Show])
	{
		p->AhiDrawIdle(p->Context,0);
		p->AhiDispOverlaySurf(p->Context,p->Buf[Show],&p->Param,0);
		DEBUG_MSG1(T("ATI %08x"),p->Buf[Show]);
	}
}

static int Lock(ahi* p,int No,planes Planes)
{
	if (No<0 || No>=p->BufCount)
	{
		Planes[0] = NULL;
		Planes[1] = NULL;
		Planes[2] = NULL;
		return ERR_INVALID_PARAM;
	}

	p->AhiDrawIdle(p->Context,0);

	if (p->AhiSurfLock(p->Context,p->Buf[No],&Planes[0],PLANE_Y)==SUCCESS)
	{
		if (p->AhiSurfLock(p->Context,p->Buf[No],&Planes[1],PLANE_U)==SUCCESS)
		{
			if (p->AhiSurfLock(p->Context,p->Buf[No],&Planes[2],PLANE_V)==SUCCESS)
				return ERR_NONE;
			p->AhiSurfUnlock(p->Context,p->Buf[No],PLANE_U);
		}
		p->AhiSurfUnlock(p->Context,p->Buf[No],PLANE_Y);
	}
	return ERR_DEVICE_ERROR;
}

static void Unlock(ahi* p,int No)
{
	p->AhiSurfUnlock(p->Context,p->Buf[No],PLANE_Y);
	p->AhiSurfUnlock(p->Context,p->Buf[No],PLANE_U);
	p->AhiSurfUnlock(p->Context,p->Buf[No],PLANE_V);
}

static void SwapPtr(ahi* p,int a,int b,int SwapShow)
{
	ahisurface* t;
	
	t = p->Buf[a];
	p->Buf[a] = p->Buf[b];
	p->Buf[b] = t;

	t = p->BufR[a];
	p->BufR[a] = p->BufR[b];
	p->BufR[b] = t;

	if (SwapShow)
	{
		SwapInt(&p->BufFrameNo[a],&p->BufFrameNo[b]);
	
		if (p->ShowNext == a)
			p->ShowNext = b;
		else
		if (p->ShowNext == b)
			p->ShowNext = a;

		if (p->ShowLast == a)
			p->ShowLast = b;
		else
		if (p->ShowLast == b)
			p->ShowLast = a;

		if (p->ShowCurr == a)
			p->ShowCurr = b;
		else
		if (p->ShowCurr == b)
			p->ShowCurr = a;
	}
}

static void UpdateBrightness(ahi* p,int Brightness)
{
	int Delta,No;
	int DCAdd = Brightness * DCFACTOR;

	if (p->DCAdd != DCAdd)
	{
		Delta = DCAdd - p->DCAdd;
		p->DCAdd = DCAdd;
		
		if (p->IDCT)
		{
			planes s,Tmp;
			int ValidTmp = 0;
			int SwapNo = p->TempCount ? p->BufCount : 0;
			blitfx FX;
			int MacroNo;
			int MacroCount = (p->Overlay.Width >> 4)*(p->Overlay.Height >> 4);

			// we need to save one buffer and adjust manually (if there are no temp buffers)
			if (p->BufCount && !p->TempCount && Lock(p,SwapNo,s)==ERR_NONE)
			{
				if (SurfaceAlloc(Tmp,&p->Overlay)==ERR_NONE)
				{
					ValidTmp = 1;
					memset(&FX,0,sizeof(FX));
					FX.ScaleX = SCALE_ONE;
					FX.ScaleY = SCALE_ONE;
					SurfaceCopy(&p->Overlay,&p->Overlay,s,Tmp,&FX);
				}
				Unlock(p,SwapNo);
			}

			*(int32_t*)&p->Data[0] = -1;
			p->Data[0].skip0 = 0;
			p->Data[0].data0 = (idct_block_t)Delta;

			for (No=p->TempCount ? 0:1;No<p->BufCount;++No)
			{
				// swap context
				p->AhiIDCTFrameStart(p->IDCT,p->Buf[SwapNo],p->Buf[No],NULL,p->FrameNo,0);
				for (MacroNo=0;MacroNo<MacroCount;++MacroNo)
				{
					p->AhiIDCTMComp8x8(p->IDCT,p->Data,1,&p->Zero,NULL,MacroNo,1);
					p->AhiIDCTMComp8x8(p->IDCT,p->Data,1,&p->Zero,NULL,MacroNo,1);
					p->AhiIDCTMComp8x8(p->IDCT,p->Data,1,&p->Zero,NULL,MacroNo,1);
					p->AhiIDCTMComp8x8(p->IDCT,p->Data,1,&p->Zero,NULL,MacroNo,1);
					p->AhiIDCTMComp8x8(p->IDCT,NULL,0,&p->Zero,NULL,MacroNo,0);
					p->AhiIDCTMComp8x8(p->IDCT,NULL,0,&p->Zero,NULL,MacroNo,0);
				}
				p->AhiIDCTFrameEnd(p->IDCT,p->FrameNo++);

				// swap pointer
				SwapPtr(p,No,SwapNo,0);
			}
			
			if (ValidTmp)
			{
				if (Lock(p,SwapNo,s)==ERR_NONE)
				{
					FX.Brightness = Delta / DCFACTOR;
					SurfaceCopy(&p->Overlay,&p->Overlay,Tmp,s,&FX);
					Unlock(p,SwapNo);
				}
				SurfaceFree(Tmp);
			}
		}
	}
}

static void UpdateDirection(ahi* p,int Dir)
{
	int No;
	if (p->Overlay.Direction != Dir)
	{
		planes s,Tmp;
		video TmpSurface;
		int ValidTmp = 0;
		int SwapNo = p->TempCount ? p->BufCount : 0;
		blitfx FX;

		// hide overlay
		if (p->p.Show && p->AllCount)
		{
			p->AhiDispOverlayState(p->Context,0,0);
			p->AhiDispWaitVBlank(p->Context,0);
		}

		// we need to save one buffer and swap manually (if there are no temp buffers)
		if (p->BufCount && !p->TempCount && Lock(p,SwapNo,s)==ERR_NONE)
		{
			if (SurfaceAlloc(Tmp,&p->Overlay)==ERR_NONE)
			{
				ValidTmp = 1;
				TmpSurface = p->Overlay;
				memset(&FX,0,sizeof(FX));
				FX.ScaleX = SCALE_ONE;
				FX.ScaleY = SCALE_ONE;
				SurfaceCopy(&p->Overlay,&p->Overlay,s,Tmp,&FX);
			}
			Unlock(p,SwapNo);
		}

		p->Overlay.Direction = Dir;
		SwapInt(&p->Overlay.Width,&p->Overlay.Height);
		SwapInt(&p->Overlay.Pitch,&p->OverlayRPitch);
		IDCTSetFormat(p,p->IDCTFormat);

		// swap buffer pointers

		for (No=0;No<p->AllCount;++No)
		{
			ahisurface* t = p->Buf[No];
			p->Buf[No] = p->BufR[No];
			p->BufR[No] = t;
		}

		// swapxy content of buffers

		for (No=p->TempCount ? 0:1;No<p->BufCount;++No)
		{
			ahirect DstRect;
			ahipoint SrcPos;

			DstRect.Left = DstRect.Top = 0;
			DstRect.Right = p->Overlay.Width;
			DstRect.Bottom = p->Overlay.Height;
			SrcPos.x = SrcPos.y = 0;

			p->AhiDrawSurfSrcSet(p->Context,p->BufR[No],PLANE_Y);
			p->AhiDrawSurfDstSet(p->Context,p->Buf[SwapNo],PLANE_Y);
			p->AhiDrawRotateBlt(p->Context,&DstRect,&SrcPos,1,1,0);

			DstRect.Right >>= 1;
			DstRect.Bottom >>= 1;

			p->AhiDrawSurfSrcSet(p->Context,p->BufR[No],PLANE_U);
			p->AhiDrawSurfDstSet(p->Context,p->Buf[SwapNo],PLANE_U);
			p->AhiDrawRotateBlt(p->Context,&DstRect,&SrcPos,1,1,0);

			p->AhiDrawSurfSrcSet(p->Context,p->BufR[No],PLANE_V);
			p->AhiDrawSurfDstSet(p->Context,p->Buf[SwapNo],PLANE_V);
			p->AhiDrawRotateBlt(p->Context,&DstRect,&SrcPos,1,1,0);

			SwapPtr(p,No,SwapNo,0);
		}
		
		if (ValidTmp)
		{
			if (Lock(p,SwapNo,s)==ERR_NONE)
			{
				FX.Direction = DIR_SWAPXY;
				SurfaceCopy(&TmpSurface,&p->Overlay,Tmp,s,&FX);
				Unlock(p,SwapNo);
			}
			SurfaceFree(Tmp);
		}
	}
}

static void UpdateAlign(ahi* p,int ScaleMode)
{
	blitfx OvlFX = p->p.FX;
	rect FullScreen;

	PhyToVirt(NULL,&FullScreen,&p->p.Output);

	VirtToPhy(&p->p.Viewport,&p->p.DstAlignedRect,&p->p.Output);
	VirtToPhy(NULL,&p->OverlayRect,&p->Overlay);

	OvlFX.Direction &= ~DIR_SWAPXY;
	if (!ScaleMode ||
		(OvlFX.ScaleX == OvlFX.ScaleY &&
		 OvlFX.ScaleX >= (SCALE_ONE*98)/100 && OvlFX.ScaleX < (SCALE_ONE*103)/100))
		OvlFX.ScaleX = OvlFX.ScaleY = SCALE_ONE;

	p->Stretch = OvlFX.ScaleX != SCALE_ONE || OvlFX.ScaleY != SCALE_ONE;
	p->StretchOnlyInternal = (p->Stretch && ScaleMode==2);
	p->Shrink = OvlFX.ScaleX < SCALE_ONE || OvlFX.ScaleY < SCALE_ONE;

	AnyAlign(&p->p.DstAlignedRect, &p->OverlayRect, &OvlFX,
		16,1,SCALE_ONE>>1,SCALE_ONE*16);

	PhyToVirt(&p->p.DstAlignedRect,&p->p.GUIAlignedRect,&p->p.Output);

	p->FullScreen = EqRect(&FullScreen,&p->p.GUIAlignedRect);

	VirtToPhy(NULL,&p->p.SrcAlignedRect,&p->p.Input);

/*	if (ScaleMode>0 && p->IDCT && !p->Internal && p->Stretch && OvlFX.ScaleX < (SCALE_ONE*13/10))
	{
		// external idct buffer with external scale buffer would be very slow
		// not too much scale factor -> fall back to 100% if internal scale buffer not available
		p->StretchOnlyInternal = 1;
	}
*/
	if (ScaleMode>0 && p->IDCT && p->Landscape && !p->Stretch && !p->FullScreen)
	{
		// using idct with external memory and 100% zoom in landscape mode
		// this means external overlay in landscape mode -> try internal scale buffer (100% scale)
		// (except when the overlay covers the full screen)
		p->Stretch = 1;
		p->StretchOnlyInternal = 1; // but don't use external memory (it would make no sense)
	}

	if (p->Stretch)
		p->SoftFX.Flags &= ~(BLITFX_AVOIDTEARING|BLITFX_VMEMROTATED);
	else
	{
		// no BLITFX_VMEMROTATED needed because overlay is always in lcd space
		// (just the primary RGB surface is rotated)
		p->SoftFX.Flags |= BLITFX_AVOIDTEARING;
	}

	BlitRelease(p->p.Soft);
	p->p.Soft = BlitCreate(&p->Overlay,&p->p.Input,&p->SoftFX);

	BlitAlign(p->p.Soft, &p->OverlayRect, &p->p.SrcAlignedRect);
}

// build ahi overlay param
static void UpdateAHI(ahi* p)
{
	if (p->Buf[SCALE])
	{
		p->Param.SrcRect.Left = 0;
		p->Param.SrcRect.Top = 0;
	}
	else
	{
		p->Param.SrcRect.Left = p->OverlayRect.x;
		p->Param.SrcRect.Top = p->OverlayRect.y;
	}
	p->Param.SrcRect.Right = p->Param.SrcRect.Left + p->p.DstAlignedRect.Width;
	p->Param.SrcRect.Bottom = p->Param.SrcRect.Top + p->p.DstAlignedRect.Height;

	p->Param.DstRect.Left = p->p.DstAlignedRect.x;
	p->Param.DstRect.Top = p->p.DstAlignedRect.y;
	p->Param.DstRect.Right = p->p.DstAlignedRect.x + p->p.DstAlignedRect.Width;
	p->Param.DstRect.Bottom = p->p.DstAlignedRect.y + p->p.DstAlignedRect.Height;

	DEBUG_MSG4(T("UpdateAHI %d,%d,%d,%d"),
		p->Param.SrcRect.Left,
		p->Param.SrcRect.Top,
		p->Param.SrcRect.Right,
		p->Param.SrcRect.Bottom);

	p->Param.Mirror = 0;
	if (p->p.FX.Direction & DIR_MIRRORLEFTRIGHT)
		p->Param.Mirror |= MIRROR_LEFTRIGHT;
	if (p->p.FX.Direction & DIR_MIRRORUPDOWN)
		p->Param.Mirror |= MIRROR_UPDOWN;

	if (p->p.ColorKey != -1) {
		p->Param.ColorMask = 
			p->p.Output.Pixel.RBitMask | 
			p->p.Output.Pixel.GBitMask |
			p->p.Output.Pixel.BBitMask;
		p->Param.ColorKey = RGBToFormat(p->p.ColorKey,&p->p.Output.Pixel);
	}
	else
	{
		p->Param.ColorMask = 0;
		p->Param.ColorKey = 0;
	}
}

static void UpdateAlloc(ahi* p,int BufCount,int AllCount,int ScaleMode,bool_t* Hidden)
{
	int No;
	int Mode;

	if (p->IDCT || ScaleMode==2)
		Mode = MEM_EXTERNAL; // only accept external
	else
		Mode = MEM_ANY_INT; // prefer internal memory transfer

	if (!p->IDCT && p->Landscape && p->AllCount)
	{
		// try to reallocate buffer if it's not in the right memory

		if ((Mode == MEM_EXTERNAL && p->BufMode[0] != MEM_EXTERNAL) ||
			(Mode == MEM_ANY_INT && p->BufMode[0] != MEM_INTERNAL))
		{
			// free surfaces

			FreeScale(p,Hidden);
			for (No=p->AllCount-1;No>=0;--No)
				FreeBuffer(p,No,Hidden);
			p->AllCount = 0;
		}
	}

	if (AllCount != p->AllCount)
	{
		FreeScale(p,Hidden);

		//free unneccessary buffers
		for (No=p->AllCount-1;No>=AllCount;--No)
			FreeBuffer(p,No,Hidden);

		if (p->AllCount > AllCount)
			p->AllCount = AllCount;

		if (p->IDCT && p->Internal)
		{
			//allocate neccessary buffers in internal memory (if possible)
			for (No=p->AllCount;No<AllCount;++No)
			{
				if (!AllocBuffer(p,No,p->Overlay.Width,p->Overlay.Height,MEM_INTERNAL,1))
				{
					// failed? need use to external? (temp buffers doesn't count)
					if (p->AllCount < BufCount && p->DevInfo.ExternalMemSize) 
					{
						p->Internal = 0;

						// free all buffers, so they can be realloced as external
						for (No=0;No<p->AllCount;++No)
							FreeBuffer(p,No,Hidden);
						p->AllCount = 0;
					}
					break;
				}
				p->AllCount = No+1;
			}
		}

		//allocate neccessary buffers
		for (No=p->AllCount;No<AllCount;++No)
		{
			if (!AllocBuffer(p,No,p->Overlay.Width,p->Overlay.Height,Mode,1))
			{
				p->MaxCount = max(1,p->AllCount);
				break;
			}
			p->AllCount = No+1;
		}

		if (p->ShowCurr >= p->AllCount)
			p->ShowCurr = 0;
		if (p->ShowNext >= p->AllCount)
			p->ShowNext = -1;
		if (p->ShowLast >= p->AllCount)
			p->ShowLast = -1;
		p->BufCount = min(p->AllCount,BufCount);
		p->TempCount = p->AllCount - p->BufCount;
	}
}

static bool_t UpdateOverlay(ahi* p,int BufCount,int TempCount)
{
	int No;
	int ScaleMode;
	bool_t Hidden = 0;
	int AllCount = BufCount + TempCount;
	if (AllCount>p->MaxCount) return 0;

	p->SoftFX.Flags = 0;
	p->SoftFX.Brightness = p->p.FX.Brightness;
	p->SoftFX.Contrast = p->p.FX.Contrast;
	p->SoftFX.Saturation = p->p.FX.Saturation;
	p->SoftFX.Direction = p->p.FX.Direction & DIR_SWAPXY;

	ScaleMode = 1;
	if (!p->IDCT && p->Landscape && (p->p.FX.ScaleX != SCALE_ONE || p->p.FX.ScaleY != SCALE_ONE))
		ScaleMode = 2; // first try normal=external scale=internal

scaletry:

	// alloc buffers
	UpdateAlloc(p,BufCount,AllCount,ScaleMode,&Hidden);

	// align
	UpdateAlign(p,ScaleMode);

	// allocate/free scale surface
	if (p->AllCount && p->Stretch)
	{
		if (p->ScaleWidth != p->p.DstAlignedRect.Width ||
			p->ScaleHeight != p->p.DstAlignedRect.Height)
		{
			int Mode;

			FreeScale(p,&Hidden);

			if (p->StretchOnlyInternal)
				Mode = MEM_INTERNAL; // only accept internal
			else
			if (p->BufMode[0] == MEM_INTERNAL) // try opposite of normal buffer
				Mode = MEM_ANY_EXT;  
			else
				Mode = MEM_ANY_INT;

			AllocBuffer(p,SCALE,p->p.DstAlignedRect.Width,p->p.DstAlignedRect.Height,Mode,0);

			if (p->Buf[SCALE])
			{
				p->ScaleWidth = p->p.DstAlignedRect.Width;
				p->ScaleHeight = p->p.DstAlignedRect.Height;
			
				Stretch(p);
			}
		}
	}
	else
		FreeScale(p,&Hidden);

	if (p->Stretch && !p->Buf[SCALE])
	{
		--ScaleMode;
		goto scaletry; // no memory for scale buffer -> try again
	}

	p->p.ColorKey = COLORKEY; // use colorkey by default
	if (p->Landscape)
	{
		// disable colorkey for external overlays in landscape mode
		if (p->Buf[SCALE])
		{
			if (p->BufMode[SCALE] == MEM_EXTERNAL)
				p->p.ColorKey = -1;
		}
		else
		{
			for (No=0;No<p->AllCount;++No)
				if (p->BufMode[No] == MEM_EXTERNAL)
				{
					p->p.ColorKey = -1;
					break;
				}
		}
	}

	if (p->p.ColorKey != -1)
		OvlInvalidate(&p->p,&p->p.Viewport,1);

	UpdateAHI(p);
	UpdateOverlaySurf(p);

	if (p->p.Show && p->BufCount)
		p->AhiDispOverlayState(p->Context,p->p.Show,0);

	if (!TempCount && p->AllCount != AllCount && p->BufCount<2 && !p->ErrorMemory)
	{
		p->ErrorMemory = 1;
		NodeError((node*)p,AHI_ID,AHI_OUT_OF_MEMORY);
	}

	return p->AllCount == AllCount;
}

static void GetMode(ahi* p)
{
	// get current mode parameters
	p->AhiDispModeGet(p->Context,&p->ModeInfo);
	memset(&p->p.Output,0,sizeof(video));
	p->p.Output.Width = p->ModeInfo.Width;
	p->p.Output.Height = p->ModeInfo.Height;
	p->p.Output.Direction = 0;
	p->p.Output.Aspect = ASPECT_ONE;

	if (p->ModeInfo.Rotation!=0 && p->ModeInfo.Rotation!=2)
		p->p.Output.Direction ^= DIR_SWAPXY;

	switch (p->ModeInfo.Rotation)
	{
	case 1:
		p->p.Output.Direction ^= DIR_MIRRORLEFTRIGHT;
		break;
	case 2:
		p->p.Output.Direction ^= DIR_MIRRORLEFTRIGHT|DIR_MIRRORUPDOWN;
		break;
	case 3:
		p->p.Output.Direction ^= DIR_MIRRORUPDOWN;
		break;
	}

	if (p->ModeInfo.Mirror & MIRROR_LEFTRIGHT)
		p->p.Output.Direction ^= DIR_MIRRORLEFTRIGHT;

	if (p->ModeInfo.Mirror & MIRROR_UPDOWN)
		p->p.Output.Direction ^= DIR_MIRRORUPDOWN;

	Ahi2PixelFormat(p->ModeInfo.PixelFormat,&p->p.Output.Pixel);
	FillInfo(&p->p.Output.Pixel);

	OvlAdjustOrientation(&p->p);
}

static int Reset(ahi* p)
{
	if (p->Context)
		GetMode(p);
	return ERR_NONE;
}

static int Orientation(ahi* p)
{
	int No;
	if (!p->Context)
	{
		p->p.Output.Direction = -1;

		if (p->AhiInit(0)==SUCCESS)
		{
			for (No=0;;No++)
			{
				int Result = p->AhiDevEnum(&p->Device,&p->DevInfo,No);
				if (Result == SUCCESS && CheckVersion(p))
				{
					if (p->AhiDevOpen(&p->Context,p->Device,"BetaPlayer",2)==SUCCESS)
					{
						p->AhiDevClientVersion(p->Context,CLIENTVERSION);
						GetMode(p);
						p->AhiDevClose(p->Context);
					}
					p->Context = NULL;
					break;
				}
				if (Result == ENDENUM)
					break;		
			}

			p->AhiTerm();
		}
	}

	return p->p.Output.Direction;
}

static int Init(ahi* p)
{
	int No;
	int Result = ERR_DEVICE_ERROR;

	p->Device = NULL;
	p->Context = NULL;
	memset(&p->Param,0,sizeof(p->Param));

	for (No=0;No<MAXBUF+1;++No)
	{
		p->Buf[No] = NULL;
		p->BufR[No] = NULL;
		p->BufMode[No] = 0;
		p->BufFrameNo[No] = -1;
	}

	if (p->AhiInit(0)==SUCCESS)
	{
		p->SoftFX = p->p.FX;
		p->SoftFX.ScaleX = SCALE_ONE;
		p->SoftFX.ScaleY = SCALE_ONE;
		p->SoftFX.Direction = 0;

		memset(&p->Overlay,0,sizeof(video));
		p->Overlay.Width = ALIGN16(p->p.Input.Width);
		p->Overlay.Height = ALIGN16(p->p.Input.Height);
		p->Overlay.Aspect = ASPECT_ONE;
		p->Overlay.Direction = 0;
		p->Overlay.Pixel.Flags = PF_YUV420;

		p->ErrorMemory = 0;
		p->Landscape = 0;
		p->Internal = 1;
		p->ScaleWidth = 0;
		p->ScaleHeight = 0;
		p->ShowCurr = 0;
		p->ShowNext = -1;
		p->ShowLast = -1;
		p->MaxCount = MAXBUF;
		p->AllCount = 0;
		p->BufCount = 0;
		p->TempCount = 0;
		p->LastTime = -1;
		p->Zero = 0;
		p->FrameNo = 0;

		for (No=0;;No++)
		{
			int AtiResult = p->AhiDevEnum(&p->Device,&p->DevInfo,No);
			if (AtiResult == SUCCESS)
			{
				if (!CheckVersion(p))
				{
					NodeError((node*)p,AHI_ID,AHI_OLDVERSION,p->Version);
					p->Context = NULL;
					Result = ERR_NOT_SUPPORTED;
				}
				else
				if (p->AhiDevOpen(&p->Context,p->Device,"BetaPlayer",2)==SUCCESS)
				{
					p->AhiDevClientVersion(p->Context,CLIENTVERSION);

					p->AhiDrawRopSet(p->Context,0xCCCC);
					p->AhiDispOverlayState(p->Context,0,0);

					GetMode(p);

					// landscape mode doesn't like external memory overlays, 
					// probably reading the external primary framebuffer conflicts with overlay readings

					if ((p->ModeInfo.Rotation!=0 && p->ModeInfo.Rotation!=2) && p->DevInfo.ExternalMemSize>0)
					{	
						p->Landscape = 1;
						p->Internal = 0;
					}

					if (p->IDCTInit)
					{
						p->AhiIDCTOpen(p->Context,&p->IDCT,0);

						if (!p->IDCT)
						{
							p->AhiDevClose(p->Context);
							p->Context = NULL;
							NodeError((node*)p,AHI_ID,AHI_OPEN_IDCT);
							Result = ERR_NOT_SUPPORTED; // don't show device error message afterwards
							break;
						}

						p->IDCTRounding = 1;
						//p->DCAdj = 3;
						IDCTSetFormat(p,p->IDCTFormat);

						p->AhiIDCTPropSet(p->IDCT,PROP_ROUNDING,&p->IDCTRounding,sizeof(int));
						p->AhiIDCTPropSet(p->IDCT,PROP_BIAS,&p->Zero,sizeof(int));

					}
					Result = ERR_NONE;
				}
				else
					p->Context = NULL;
				break;
			}
			if (AtiResult == ENDENUM)
				break;
		}

		if (!p->Context)
			p->AhiTerm();
	}

	return Result;
}

static void Done(ahi* p)
{
	int No;

	if (p->Context && p->IDCT)
	{
		p->AhiIDCTClose(p->IDCT);
		p->IDCT = NULL;
	}

	for (No=MAXBUF;No>=0;--No)
		FreeBuffer(p,No,NULL);

	if (p->Context)
	{
		p->AhiDispOverlayState(p->Context,0,0);
		p->AhiDevClose(p->Context);
		p->AhiTerm();
		p->Context = NULL;
	}
}

static int Update(ahi* p)
{	
	UpdateDirection(p,p->p.FX.Direction & DIR_SWAPXY);
	UpdateBrightness(p,p->p.FX.Brightness);
	// temp buffers are freed!
	if (!UpdateOverlay(p,p->IDCT?p->BufCount:1,0))
		return ERR_OUT_OF_MEMORY;
	return ERR_NONE;
}

static int UpdateShow(ahi* p)
{
	p->AhiDispOverlayState(p->Context,p->p.Show,0);
	return ERR_NONE;
}

static int Blit(ahi* p, const planes Data, const planes DataLast )
{
	planes Planes;

	int Result = Lock(p,p->ShowCurr,Planes);

	if (Result == ERR_NONE)
	{
		BlitImage(p->p.Soft,Planes,Data,DataLast,-1,-1);

		Unlock(p,p->ShowCurr);

		if (p->Buf[SCALE])
			Stretch(p);
	}
	return Result;
}

//**********************
//* IDCT               *
//**********************

const uint8_t ScanTable[3][64] = {
{
	 0,  1,  8, 16,  9,  2,  3, 10, 
	17,	24, 32, 25, 18, 11,  4,  5,
	12, 19, 26, 33, 40, 48, 41, 34, 
	27,	20, 13,  6,  7, 14, 21, 28, 
	35, 42,	49, 56, 57, 50, 43, 36, 
	29, 22,	15, 23, 30, 37, 44, 51, 
	58, 59, 52, 45, 38, 31, 39, 46, 
	53, 60, 61,	54, 47, 55, 62, 63
},
{
 	 0,  1,  2,  3,  8,  9, 16, 17, 
	10, 11,  4,  5,  6,  7, 15, 14,
	13, 12, 19, 18, 24, 25, 32, 33, 
	26, 27, 20, 21, 22, 23, 28, 29,
	30, 31, 34, 35, 40, 41, 48, 49, 
	42, 43, 36, 37, 38, 39, 44, 45,
	46, 47, 50, 51, 56, 57, 58, 59, 
	52, 53, 54, 55, 60, 61, 62, 63
},
{
	 0,  8, 16, 24,  1,  9,  2, 10, 
	17, 25, 32, 40, 48, 56, 57, 49,
	41, 33, 26, 18,  3, 11,  4, 12, 
	19, 27, 34, 42, 50, 58, 35, 43,
	51, 59, 20, 28,  5, 13,  6, 14, 
	21, 29, 36, 44, 52, 60, 37, 45,
	53, 61, 22, 30,  7, 15, 23, 31, 
	38, 46, 54, 62, 39, 47, 55, 63
}};
const uint8_t ZigZagSwap[64] = {
	 0,  8,  1,  2,  9, 16, 24, 17, 
	10,	 3,  4, 11, 18, 25, 32, 40,
	33, 26, 19, 12,  5,  6, 13, 20, 
	27,	34, 41, 48, 56, 49, 42, 35, 
	28, 21,	14,  7, 15, 22, 29, 36, 
	43, 50,	57, 58, 51, 44, 37, 30, 
	23, 31, 38, 45, 52, 59, 60, 53, 
	46, 39, 47,	54, 61, 62, 55, 63
};
const uint8_t ZigZagSwapLength[64] = {
	1,   3,  3,  6,  6,  6, 10,	10,	
	10, 10, 15, 15, 15, 15, 15,	21, 
	21, 21, 21, 21, 21, 28, 28, 28,	
	28, 28, 28, 28, 36, 36, 36, 36, 
	36,	36, 36, 36, 43, 43, 43, 43, 
	43,	43, 43, 49, 49, 49, 49, 49,
	49, 54, 54, 54, 54, 54, 58, 58, 
	58, 58,	61, 61, 61, 63, 63, 64
};

#define AHI(p) ((ahi*)((char*)(p)-(int)&(((ahi*)0)->IdctVMT)))

// input: Block, Length
#define AHIDCBLOCK(Offset)					\
{											\
	if (Length < 1)							\
	{										\
		Block[0] = (idct_block_t)(Offset);	\
		Length = 1;							\
	}										\
	else									\
		Block[0] = (idct_block_t)(Block[0] + (Offset));\
}

// input: Count, Length
#define AHIDCATI(Offset)					\
{											\
	if (Count < 1)							\
	{										\
		*(int32_t*)Data = 0x00FFFFFF;		\
		Data->data0 = (short)(Offset);		\
		Count = 1;							\
	}										\
	else									\
	if (Data->skip0)						\
	{ /* we have to shift everything */		\
		idct_ati *i = Data + Count;			\
		assert(Data->skip0 != -1);			\
		if (i[-1].skip3 == -1)				\
			goto noskip3;					\
		else								\
		{									\
			++Count;						\
			*(int32_t*)i = -1;				\
		}									\
		do									\
		{									\
			i->skip0 = i[-1].skip3;			\
			i->data0 = i[-1].data3;			\
noskip3:									\
			--i;							\
			*(uint32_t*)i >>= 8;			\
			i->data2 = i->data3;			\
			i->data1 = i->data2;			\
			i->data0 = i->data1;			\
		}									\
		while (i != Data);					\
		--Data->skip0;						\
		Data->data0 = (short)(Offset);		\
	}										\
	else									\
		Data->data0 = (short)(Data->data0 + (Offset));\
}

// ADJUST: don't care about MVFwd (only in b-frames which are not accumulated)
#define AHIMVADJBLOCK						\
	if ((MVBack & 0x00010001)==0x00010001)	\
		AHIDCBLOCK(2);						

#define AHIMVADJATI							\
	if ((MVBack & 0x00010001)==0x00010001)	\
		AHIDCATI(2);					

#define AHIMVNORMAL					\
	Back = Fwd = NULL;				\
	if (MVBack != NOMV)				\
	{								\
		MVBack <<= 1;				\
		MVBack &= ~0x00010001;		\
		AHI(p)->MVBack = MVBack;	\
		Back = &AHI(p)->MVBack;		\
	}								\
	if (MVFwd != NOMV)		 		\
	{								\
		MVFwd <<= 1;				\
		MVFwd &= ~0x00010001;		\
		AHI(p)->MVFwd = MVFwd;		\
		Fwd = &AHI(p)->MVFwd;		\
	}

#define AHIMVSWAP					\
	Back = Fwd = NULL;				\
	if (MVBack != NOMV)				\
	{								\
		MVBack = (MVBack << 17) |	\
		 ((uint32_t)MVBack >> 15);	\
		MVBack &= ~0x00010001;		\
		AHI(p)->MVBack = MVBack;	\
		Back = &AHI(p)->MVBack;		\
	}								\
	if (MVFwd != NOMV)				\
	{								\
		MVFwd = (MVFwd << 17) |		\
		 ((uint32_t)MVFwd >> 15);	\
		MVFwd &= ~0x00010001;		\
		AHI(p)->MVFwd = MVFwd;		\
		Fwd = &AHI(p)->MVFwd;		\
	}

// Data, Count -> Block, Length
#define AHIUNPACK											\
{															\
	const unsigned char *Scan = ScanTable[IDCTSCAN_ZIGZAG];	\
	for (;Count>0;--Count)									\
	{														\
		assert(Data->skip0 != -1);							\
		Scan += Data->skip0;								\
		Block[*(Scan++)] = Data->data0;						\
		if (Data->skip1 == -1) break;						\
		Scan += Data->skip1;								\
		Block[*(Scan++)] = Data->data1;						\
		if (Data->skip2 == -1) break;						\
		Scan += Data->skip2;								\
		Block[*(Scan++)] = Data->data2;						\
		if (Data->skip3 == -1) break;						\
		Scan += Data->skip3;								\
		Block[*(Scan++)] = Data->data3;						\
	}														\
	Length = Scan - ScanTable[IDCTSCAN_ZIGZAG];				\
}

// Block, Length -> Dst, Count
#define AHIPACK(Dst)										\
	ScanEnd = Scan + Length;								\
	for (Data=Dst;;++Data)									\
	{														\
		*(int32_t*)Data = -1;								\
															\
		Last = Scan;										\
		while (Scan != ScanEnd && !Block[*Scan]) ++Scan;	\
		if (Scan == ScanEnd) break;							\
		Data->skip0 = (char)(Scan - Last);					\
		Data->data0 = Block[*(Scan++)];						\
															\
		Count++;											\
		Last = Scan;										\
		while (Scan != ScanEnd && !Block[*Scan]) ++Scan;	\
		if (Scan == ScanEnd) break;							\
		Data->skip1 = (char)(Scan - Last);					\
		Data->data1 = Block[*(Scan++)];						\
															\
		Last = Scan;										\
		while (Scan != ScanEnd && !Block[*Scan]) ++Scan;	\
		if (Scan == ScanEnd) break;							\
		Data->skip2 = (char)(Scan - Last);					\
		Data->data2 = Block[*(Scan++)];						\
															\
		Last = Scan;										\
		while (Scan != ScanEnd && !Block[*Scan]) ++Scan;	\
		if (Scan == ScanEnd) break;							\
		Data->skip3 = (char)(Scan - Last);					\
		Data->data3 = Block[*(Scan++)];						\
	}

#define AHIPACKNORMAL										\
{															\
	const unsigned char *Scan = ScanTable[ScanType];		\
	const unsigned char *ScanEnd;							\
	const unsigned char *Last;								\
	idct_ati *Data;											\
	AHIPACK(AHI(p)->Data)									\
}

#define AHIPACKSWAP(Dst)									\
{															\
	const unsigned char *Scan;								\
	const unsigned char *ScanEnd;							\
	const unsigned char *Last;								\
	idct_ati *Data;											\
	if (ScanType == IDCTSCAN_ZIGZAG)						\
	{														\
		Scan = ZigZagSwap;									\
		Length = ZigZagSwapLength[Length-1];				\
	}														\
	else													\
	{														\
		Scan = ScanTable[ScanType];							\
		ScanType = 3 - ScanType;							\
	}														\
	AHIPACK(Dst)											\
}

#define AHIBLOCKSWAP										\
{															\
	if (SubNo <= 4)											\
		AHIDCBLOCK(AHI(p)->DCAdd)							\
	Count = 0;												\
	if (SubNo == 2)											\
	{														\
		if (Length>0 && Length<=64)							\
			AHIPACKSWAP(AHI(p)->SaveData)					\
		AHI(p)->SaveCount = Count;							\
		AHI(p)->SaveScanType = ScanType;					\
	}														\
	else													\
	{														\
		if (Length>0 && Length<=64)							\
			AHIPACKSWAP(AHI(p)->Data)						\
		AHI(p)->AhiIDCT8x8(AHI(p)->IDCT,AHI(p)->Data,Count,AHI(p)->MacroNo,1 << ScanType); \
		if (SubNo == 3)										\
			AHI(p)->AhiIDCT8x8(AHI(p)->IDCT,AHI(p)->SaveData,AHI(p)->SaveCount,AHI(p)->MacroNo,1 << AHI(p)->SaveScanType); \
	}														\
}

#define AHIMCOMPBLOCKSWAP									\
{															\
	Count = 0;												\
	if (SubNo == 2)											\
	{														\
		if (Length>0 && Length<=64)							\
			AHIPACKSWAP(AHI(p)->SaveData)					\
		AHI(p)->SaveCount = Count;							\
		AHI(p)->SaveScanType = ScanType;					\
		AHI(p)->SaveMVBack = MVBack;						\
		AHI(p)->SaveMVFwd = MVFwd;							\
	}														\
	else													\
	{														\
		int* Back;											\
		int* Fwd;											\
		AHIMVSWAP;											\
		if (Length>0 && Length<=64)							\
			AHIPACKSWAP(AHI(p)->Data)						\
		AHI(p)->AhiIDCTMComp8x8(AHI(p)->IDCT,AHI(p)->Data,Count,Back,Fwd,AHI(p)->MacroNo,1 << ScanType);\
		if (SubNo == 3)										\
		{													\
			MVBack = AHI(p)->SaveMVBack;					\
			MVFwd = AHI(p)->SaveMVFwd;						\
			AHIMVSWAP;										\
			AHI(p)->AhiIDCTMComp8x8(AHI(p)->IDCT,AHI(p)->SaveData,AHI(p)->SaveCount,Back,Fwd,AHI(p)->MacroNo,1 << AHI(p)->SaveScanType);\
		}													\
	}														\
}

#define AHIATISWAP											\
	ScanType = 3 - ScanType;								\
	if (SubNo <= 4)											\
		AHIDCATI(AHI(p)->DCAdd);							\
	if (SubNo == 2)											\
	{														\
		memcpy(AHI(p)->SaveData,Data,sizeof(idct_ati)*Count);\
		AHI(p)->SaveCount = Count;							\
		AHI(p)->SaveScanType = ScanType;					\
	}														\
	else													\
	{														\
		AHI(p)->AhiIDCT8x8(AHI(p)->IDCT,Data,Count,AHI(p)->MacroNo,1 << ScanType);\
		if (SubNo == 3)										\
			AHI(p)->AhiIDCT8x8(AHI(p)->IDCT,AHI(p)->SaveData,AHI(p)->SaveCount,AHI(p)->MacroNo,1 << AHI(p)->SaveScanType);\
	}


#define AHIMCOMPATISWAP										\
	ScanType = 3 - ScanType;								\
	if (SubNo == 2)											\
	{														\
		memcpy(AHI(p)->SaveData,Data,sizeof(idct_ati)*Count);\
		AHI(p)->SaveCount = Count;							\
		AHI(p)->SaveScanType = ScanType;					\
		AHI(p)->SaveMVBack = MVBack;						\
		AHI(p)->SaveMVFwd = MVFwd;							\
	}														\
	else													\
	{														\
		int* Back;											\
		int* Fwd;											\
		AHIMVSWAP;											\
		AHI(p)->AhiIDCTMComp8x8(AHI(p)->IDCT,Data,Count,Back,Fwd,AHI(p)->MacroNo,1 << ScanType);\
		if (SubNo == 3)										\
		{													\
			MVBack = AHI(p)->SaveMVBack;					\
			MVFwd = AHI(p)->SaveMVFwd;						\
			AHIMVSWAP;										\
			AHI(p)->AhiIDCTMComp8x8(AHI(p)->IDCT,AHI(p)->SaveData,AHI(p)->SaveCount,Back,Fwd,AHI(p)->MacroNo,1 << AHI(p)->SaveScanType);\
		}													\
	}

static void IDCTProcess(void* p,int x,int y,const int *MVBackHint,const int *MVFwdHint )
{
	AHI(p)->MacroSubNo = 0;
	AHI(p)->MacroNo = x + y*AHI(p)->MacroWidth;
}

static void IDCTProcessSwap(void* p,int x,int y,const int *MVBackHint,const int *MVFwdHint )
{
	AHI(p)->MacroSubNo = 0;
	AHI(p)->MacroNo = y + x*AHI(p)->MacroWidth;
}

static void IDCTAti(void* p,idct_ati *Data,int Count,int ScanType)
{
	int SubNo = ++AHI(p)->MacroSubNo;
	if (SubNo <= 4)
		AHIDCATI(AHI(p)->DCAdd);
	AHI(p)->AhiIDCT8x8(AHI(p)->IDCT,Data,Count,AHI(p)->MacroNo,1 << ScanType);
}

static void IDCTMCompAti(void* p,int MVBack,int MVFwd,idct_ati *Data,int Count,int ScanType)
{
	int* Back;
	int* Fwd;
	AHIMVNORMAL;
	AHI(p)->AhiIDCTMComp8x8(AHI(p)->IDCT,Data,Count,Back,Fwd,AHI(p)->MacroNo,1 << ScanType);
}

static void IDCTMCompAtiAdj(void* p,int MVBack,int MVFwd,idct_ati *Data,int Count,int ScanType)
{
	int* Back;
	int* Fwd;
	AHIMVADJATI;
	AHIMVNORMAL;
	AHI(p)->AhiIDCTMComp8x8(AHI(p)->IDCT,Data,Count,Back,Fwd,AHI(p)->MacroNo,1 << ScanType);
}

static void IDCTAtiSwap(void* p,idct_ati *Data,int Count,int ScanType)
{
	int SubNo = ++AHI(p)->MacroSubNo;

	if (ScanType == IDCTSCAN_ZIGZAG)
	{
		int Length;
		idct_block_t* Block = AHI(p)->Block;

		AHIUNPACK
		AHIBLOCKSWAP
	}
	else
	{
		AHIATISWAP
	}
}

static void IDCTMCompAtiSwap(void* p,int MVBack,int MVFwd,idct_ati *Data,int Count,int ScanType)
{
	int SubNo = ++AHI(p)->MacroSubNo;

	if (ScanType == IDCTSCAN_ZIGZAG)
	{
		int Length;
		idct_block_t* Block = AHI(p)->Block;

		AHIUNPACK
		AHIMCOMPBLOCKSWAP
	}
	else
	{
		AHIMCOMPATISWAP;
	}
}

static void IDCTMCompAtiSwapAdj(void* p,int MVBack,int MVFwd,idct_ati *Data,int Count,int ScanType)
{
	int SubNo = ++AHI(p)->MacroSubNo;

	if (ScanType == IDCTSCAN_ZIGZAG)
	{
		int Length;
		idct_block_t* Block = AHI(p)->Block;

		AHIUNPACK
		AHIMVADJBLOCK
		AHIMCOMPBLOCKSWAP
	}
	else
	{
		AHIMVADJATI
		AHIMCOMPATISWAP;
	}
}

static void IDCTBlock(void* p,idct_block_t *Block,int Length,int ScanType)
{
	int Count = 0;
	int SubNo = ++AHI(p)->MacroSubNo;
	if (SubNo <= 4)
		AHIDCBLOCK(AHI(p)->DCAdd)
	
	if (Length>0 && Length<=64)
		AHIPACKNORMAL

	AHI(p)->AhiIDCT8x8(AHI(p)->IDCT,AHI(p)->Data,Count,AHI(p)->MacroNo,1 << ScanType);
}

static void IDCTMCompBlock(void* p,int MVBack,int MVFwd,idct_block_t *Block,int Length,int ScanType)
{
	int Count = 0;
	int* Back;
	int* Fwd;
	AHIMVNORMAL

	if (Length>0 && Length<=64)
		AHIPACKNORMAL

	AHI(p)->AhiIDCTMComp8x8(AHI(p)->IDCT,AHI(p)->Data,Count,Back,Fwd,AHI(p)->MacroNo,1 << ScanType);
}

static void IDCTMCompBlockAdj(void* p,int MVBack,int MVFwd,idct_block_t *Block,int Length,int ScanType)
{
	int Count = 0;
	int* Back;
	int* Fwd;
	AHIMVADJBLOCK
	AHIMVNORMAL

	if (Length>0 && Length<=64)
		AHIPACKNORMAL

	AHI(p)->AhiIDCTMComp8x8(AHI(p)->IDCT,AHI(p)->Data,Count,Back,Fwd,AHI(p)->MacroNo,1 << ScanType);
}

static void IDCTBlockSwap(void* p,idct_block_t *Block,int Length,int ScanType)
{
	int Count;
	int SubNo = ++AHI(p)->MacroSubNo;

	AHIBLOCKSWAP;
}

static void IDCTMCompBlockSwap(void* p,int MVBack,int MVFwd,idct_block_t *Block,int Length,int ScanType)
{
	int Count;
	int SubNo = ++AHI(p)->MacroSubNo;

	AHIMCOMPBLOCKSWAP;
}

static void IDCTMCompBlockSwapAdj(void* p,int MVBack,int MVFwd,idct_block_t *Block,int Length,int ScanType)
{
	int Count;
	int SubNo = ++AHI(p)->MacroSubNo;

	AHIMVADJBLOCK
	AHIMCOMPBLOCKSWAP
}

static void IDCTCopy8x8(void* p,int x,int y,int Forward)
{
	int MacroNo = x + y*AHI(p)->MacroWidth;
	int* Back = NULL;
	int* Fwd = NULL;

	if (Forward)
		Fwd = &AHI(p)->Zero;
	else
		Back = &AHI(p)->Zero;
	
	AHI(p)->AhiIDCTMComp8x8(AHI(p)->IDCT,NULL,0,Back,Fwd,MacroNo,0);
	AHI(p)->AhiIDCTMComp8x8(AHI(p)->IDCT,NULL,0,Back,Fwd,MacroNo,0);
	AHI(p)->AhiIDCTMComp8x8(AHI(p)->IDCT,NULL,0,Back,Fwd,MacroNo,0);
	AHI(p)->AhiIDCTMComp8x8(AHI(p)->IDCT,NULL,0,Back,Fwd,MacroNo,0);
	AHI(p)->AhiIDCTMComp8x8(AHI(p)->IDCT,NULL,0,Back,Fwd,MacroNo,0);
	AHI(p)->AhiIDCTMComp8x8(AHI(p)->IDCT,NULL,0,Back,Fwd,MacroNo,0);
}

static void IDCTCopy8x8Swap(void* p,int x,int y,int Forward)
{
	int MacroNo = y + x*AHI(p)->MacroWidth;
	int* Back = NULL;
	int* Fwd = NULL;

	if (Forward)
		Fwd = &AHI(p)->Zero;
	else
		Back = &AHI(p)->Zero;
	
	AHI(p)->AhiIDCTMComp8x8(AHI(p)->IDCT,NULL,0,Back,Fwd,MacroNo,0);
	AHI(p)->AhiIDCTMComp8x8(AHI(p)->IDCT,NULL,0,Back,Fwd,MacroNo,0);
	AHI(p)->AhiIDCTMComp8x8(AHI(p)->IDCT,NULL,0,Back,Fwd,MacroNo,0);
	AHI(p)->AhiIDCTMComp8x8(AHI(p)->IDCT,NULL,0,Back,Fwd,MacroNo,0);
	AHI(p)->AhiIDCTMComp8x8(AHI(p)->IDCT,NULL,0,Back,Fwd,MacroNo,0);
	AHI(p)->AhiIDCTMComp8x8(AHI(p)->IDCT,NULL,0,Back,Fwd,MacroNo,0);
}

static int IDCTSetFormat(ahi* p,int Format)
{
	if (Format == IDCTFORMAT_BLOCK)
	{
		p->IDCTFormat = Format;
		if (p->Overlay.Direction)
		{
			p->IdctVMT.Copy = IDCTCopy8x8Swap;
			p->IdctVMT.Process = IDCTProcessSwap;
			p->IdctVMT.Data = IDCTBlockSwap;
			p->IdctVMT.MCompData = p->IDCTBug ? IDCTMCompBlockSwapAdj : IDCTMCompBlockSwap; 
		}
		else 
		{
			p->IdctVMT.Copy = IDCTCopy8x8;
			p->IdctVMT.Process = IDCTProcess;
			p->IdctVMT.Data = IDCTBlock;
			p->IdctVMT.MCompData = p->IDCTBug ? IDCTMCompBlockAdj : IDCTMCompBlock;
		}
		return ERR_NONE;
	}
	else
	if (Format == IDCTFORMAT_ATI)
	{
		p->IDCTFormat = Format;
		if (p->Overlay.Direction)
		{
			p->IdctVMT.Copy = IDCTCopy8x8Swap;
			p->IdctVMT.Process = IDCTProcessSwap;
			p->IdctVMT.Data = IDCTAtiSwap;
			p->IdctVMT.MCompData = p->IDCTBug ? IDCTMCompAtiSwapAdj : IDCTMCompAtiSwap;
		}
		else 
		{
			p->IdctVMT.Copy = IDCTCopy8x8;
			p->IdctVMT.Process = IDCTProcess;
			p->IdctVMT.Data = IDCTAti;
			p->IdctVMT.MCompData = p->IDCTBug ? IDCTMCompAtiAdj : IDCTMCompAti;
		}
		return ERR_NONE;
	}
	return ERR_NOT_SUPPORTED;
}

static int IDCTOutputFormat(ahi* p, const void* Data, int Size )
{
	int Result;
	p->IDCTInit = 1;

	Result = p->p.VMT.Set((node*)p,VOUT_INPUT|PACKET_FORMAT,Data,Size);

	p->IDCTInit = 0;
	return Result;
}

static int IDCTSet(void* p, int No, const void* Data, int Size )
{
	int Result = ERR_INVALID_PARAM;
	switch (No)
	{
	case IDCT_FORMAT:
		if (Size == sizeof(int))
			Result = IDCTSetFormat(AHI(p),*(const int*)Data); 
		break;

	case IDCT_OUTPUT|PACKET_FORMAT:
		Result = IDCTOutputFormat(AHI(p),Data,Size);
		break;

	case IDCT_OUTPUT:
		if (Size == sizeof(pin))
			Result = NodeSetPin(&AHI(p)->IDCTOutput,(pin*)Data,&AHI(p)->p.Input,sizeof(video));
		break;

	case IDCT_ROUNDING:
		if (Size == sizeof(bool_t) && AHI(p)->IDCT)
		{
			AHI(p)->IDCTRounding = !*(const int*)Data;
			//AHI(p)->DCAdj = 3 - AHI(p)->IDCTRounding;
			AHI(p)->AhiIDCTPropSet(AHI(p)->IDCT,PROP_ROUNDING,&AHI(p)->IDCTRounding,sizeof(int));
			Result = ERR_NONE;
		}
		break;

	case IDCT_SHOW: SETVALUE(AHI(p)->ShowNext,int,ERR_NONE); break;

	case IDCT_BUFFERCOUNT:
		if (Size == sizeof(int))
		{
			Result = ERR_NONE;
			if (AHI(p)->BufCount != *(const int*)Data &&
				!UpdateOverlay(AHI(p),*(const int*)Data,0))
			{
				Result = ERR_OUT_OF_MEMORY;
			}
		}
		break;
	}
	return Result;
}

static int IDCTEnum( void* p, int No, datadef* Param )
{
	int Result = NodeEnumType(&No,Param,NodeParams,FlowParams,IDCTParams,NULL);
	if (Result == ERR_NONE && Param->No == IDCT_OUTPUT)
		Param->Flags |= DF_RDONLY;
	return Result;
}

static int IDCTGet(void* p, int No, void* Data, int Size )
{
	int Result = ERR_INVALID_PARAM;
	switch (No)
	{
	case NODE_ID: GETVALUE(AHIIDCT_ID,int); break;
	case IDCT_NATIVEFORMAT: GETVALUE(IDCTFORMAT_ATI,int); break;
	case IDCT_FORMAT: GETVALUE(AHI(p)->IDCTFormat,int); break;
	case IDCT_OUTPUT|PACKET_FORMAT: GETVALUE(AHI(p)->Overlay,video); break;
	case IDCT_OUTPUT: GETVALUE(AHI(p)->IDCTOutput,pin); break;
	case IDCT_ROUNDING: GETVALUE(!AHI(p)->IDCTRounding,bool_t); break;
	case IDCT_SHOW: GETVALUE(AHI(p)->ShowCurr,int); break;
	case IDCT_BUFFERCOUNT:
		if (Size == sizeof(int))
		{
			if (!AHI(p)->IDCT)
				*(int*)Data = 0;
			else
				*(int*)Data = AHI(p)->BufCount;
			Result = ERR_NONE;
		}
		break;
	}
	return Result;
}

int IDCTLock(void* p,int No,planes Planes,int* Brightness,int* Pitch)
{
	if (Brightness)
		*Brightness = AHI(p)->p.FX.Brightness;
	if (Pitch)
		*Pitch = AHI(p)->Overlay.Pitch;
	return Lock(AHI(p),No,Planes);
}

void IDCTUnlock(void* p,int No)
{
	Unlock(AHI(p),No);
}

int IDCTFrameStart(void* p,int FrameNo,int DstNo,int BackNo,int FwdNo,int ShowNo,bool_t Drop)
{
	int OldFrameNo;
	ahisurface* Back;
	ahisurface* Fwd;
	ahisurface* Dst;

	// we will try to avoid overwrite currently displayed buffers:
	//   ShowCurr currently displayed
	//   ShowLast may be still on screen (because flip occurs only during vblank)

	if (!AHI(p)->Buf[SCALE] && (AHI(p)->ShowLast == DstNo || AHI(p)->ShowCurr == DstNo))
	{
		// try to find a free buffer
		int SwapNo;
		for (SwapNo=0;SwapNo<AHI(p)->AllCount;++SwapNo)
			if (SwapNo != AHI(p)->ShowLast && SwapNo != AHI(p)->ShowCurr && 
				SwapNo != BackNo && SwapNo != FwdNo && SwapNo != ShowNo)
				break; 

		if (SwapNo < AHI(p)->AllCount || UpdateOverlay(AHI(p),AHI(p)->BufCount,AHI(p)->TempCount+1))
		{
			// use free buffer
			SwapPtr(AHI(p),SwapNo,DstNo,1);

			DEBUG_MSG2(T("ATI Swap %d,%d"),DstNo,SwapNo);
		}
		else
		if (AHI(p)->ShowLast >= 0 && AHI(p)->ShowCurr != AHI(p)->ShowLast &&
			AHI(p)->ShowLast != BackNo && AHI(p)->ShowLast != FwdNo) 
		{
			// no free buffer found and couldn't allocate new temp buffer
			// so we wait for vblank and use ShowLast
			if (!Drop)
				AHI(p)->AhiDispWaitVBlank(AHI(p)->Context,0);

			DEBUG_MSG2(T("ATI VBlack %d,%d"),DstNo,AHI(p)->ShowLast);

			if (DstNo != AHI(p)->ShowLast)
				SwapPtr(AHI(p),AHI(p)->ShowLast,DstNo,1);
		}
	}

	DEBUG_MSG3(T("ATI FrameStart %d,%d,%d"),DstNo,BackNo,FwdNo);

	AHI(p)->ShowNext = ShowNo;

	OldFrameNo = AHI(p)->BufFrameNo[DstNo];
	AHI(p)->BufFrameNo[DstNo] = FrameNo;

	Back = NULL;
	Fwd = NULL;
	Dst = AHI(p)->Buf[DstNo];
	if (BackNo >= 0)
		Back = AHI(p)->Buf[BackNo];
	if (FwdNo >= 0)
		Fwd = AHI(p)->Buf[FwdNo];

	AHI(p)->MacroWidth = AHI(p)->Overlay.Width >> 4;
	AHI(p)->AhiIDCTFrameStart(AHI(p)->IDCT,Dst,Back,Fwd,AHI(p)->FrameNo,0);

	return OldFrameNo;
}

void IDCTFrameEnd(void* p)
{
	AHI(p)->AhiIDCTFrameEnd(AHI(p)->IDCT,AHI(p)->FrameNo++);
}

void IDCTDrop(void* p)
{
	int No;
	for (No=0;No<AHI(p)->AllCount;++No)
		AHI(p)->BufFrameNo[No] = -1;
}

int IDCTSend(void* p,packet* Packet,int* Dropped)
{
	if (AHI(p)->ShowNext < 0)
		return ERR_NEED_MORE_DATA;

	if (Packet->CurrTime != -1 && 
		(Packet->RefTime > Packet->CurrTime + SHOWAHEAD || Packet->CurrTime == AHI(p)->LastTime))
		return ERR_BUFFER_FULL;

	AHI(p)->LastTime = (Packet->CurrTime>=0) ? Packet->CurrTime : Packet->RefTime;

	AHI(p)->ShowLast = AHI(p)->ShowCurr;
	AHI(p)->ShowCurr = AHI(p)->ShowNext;

	DEBUG_MSG2(T("ATI Send %d,%d"),AHI(p)->ShowCurr,AHI(p)->ShowLast);

	if (AHI(p)->Buf[SCALE])
	{
		if (!Packet->DropLevel)
			Stretch(AHI(p));
		else
		if (Dropped)
			++(*Dropped);
	}
	else
		UpdateOverlaySurf(AHI(p));

	return ERR_NONE;
}

static ahi AHI = 
{
	{
		OvlEnum,
		Get,
		Set,
		Init,
		Done,
		Blit,
		Update,
		UpdateShow,
		Orientation,
		Reset,
	},
	{
		IDCTEnum,
		IDCTGet,
		IDCTSet,

		IDCTSend,
		IDCTDrop,
		IDCTLock,
		IDCTUnlock,
		IDCTFrameStart,
		IDCTFrameEnd,
		NULL,
		NULL,
		NULL,
		NULL,
	}
};

void AHI_Init() 
{ 
	int No;
	ahi* p = &AHI;

	p->Context = NULL;
	p->p.Params = AHIParams;
	p->IDCTOutput.Node = (node*)p;
	p->IDCTOutput.No = VOUT_INPUT;
	p->IDCTFormat = IDCTFORMAT_ATI;
	p->IDCTBug = 0;
	p->IDCTInit = 0;

	p->p.DLL = LoadLibrary(T("ACE_DDI.DLL"));
	OvlGetProc(&p->p.DLL,&p->AhiInit,T("AhiInit"),0);
	OvlGetProc(&p->p.DLL,&p->AhiTerm,T("AhiTerm"),0);

	OvlGetProc(&p->p.DLL,&p->AhiDevEnum,T("AhiDevEnum"),0);
	OvlGetProc(&p->p.DLL,&p->AhiDevOpen,T("AhiDevOpen"),0);
	OvlGetProc(&p->p.DLL,&p->AhiDevClose,T("AhiDevClose"),0);
	OvlGetProc(&p->p.DLL,&p->AhiDevClientVersion,T("AhiDevClientVersion"),0);

	OvlGetProc(&p->p.DLL,&p->AhiDispModeGet,T("AhiDispModeGet"),0);
	OvlGetProc(&p->p.DLL,&p->AhiDispModeEnum,T("AhiDispModeEnum"),0);
	OvlGetProc(&p->p.DLL,&p->AhiDispSurfGet,T("AhiDispSurfGet"),0);

	OvlGetProc(&p->p.DLL,&p->AhiSurfGetLargestFreeBlockSize,T("AhiSurfGetLargestFreeBlockSize"),1);
	OvlGetProc(&p->p.DLL,&p->AhiSurfAlloc,T("AhiSurfAlloc"),0);
	OvlGetProc(&p->p.DLL,&p->AhiSurfReuse,T("AhiSurfReuse"),0);
	OvlGetProc(&p->p.DLL,&p->AhiSurfFree,T("AhiSurfFree"),0);
	OvlGetProc(&p->p.DLL,&p->AhiSurfLock,T("AhiSurfLock"),0);
	OvlGetProc(&p->p.DLL,&p->AhiSurfUnlock,T("AhiSurfUnlock"),0);
	OvlGetProc(&p->p.DLL,&p->AhiSurfInfo,T("AhiSurfInfo"),0);
	OvlGetProc(&p->p.DLL,&p->AhiSurfPlaneInfo,T("AhiSurfPlaneInfo"),0);

	OvlGetProc(&p->p.DLL,&p->AhiDrawRopSet,T("AhiDrawRopSet"),0);
	OvlGetProc(&p->p.DLL,&p->AhiDrawSurfSrcSet,T("AhiDrawSurfSrcSet"),0);
	OvlGetProc(&p->p.DLL,&p->AhiDrawSurfDstSet,T("AhiDrawSurfDstSet"),0);
	OvlGetProc(&p->p.DLL,&p->AhiDrawIdle,T("AhiDrawIdle"),0);
	OvlGetProc(&p->p.DLL,&p->AhiDispWaitVBlank,T("AhiDispWaitVBlank"),0);
	OvlGetProc(&p->p.DLL,&p->AhiDrawBitBlt,T("AhiDrawBitBlt"),0);
	OvlGetProc(&p->p.DLL,&p->AhiDrawStretchBlt,T("AhiDrawStretchBlt"),0);
	OvlGetProc(&p->p.DLL,&p->AhiDrawRotateBlt,T("AhiDrawRotateBlt"),0);

	OvlGetProc(&p->p.DLL,&p->AhiDispOverlayCaps,T("AhiDispOverlayCaps"),0);
	OvlGetProc(&p->p.DLL,&p->AhiDispOverlayState,T("AhiDispOverlayState"),0);
	OvlGetProc(&p->p.DLL,&p->AhiDispOverlayPos,T("AhiDispOverlayPos"),0);
	OvlGetProc(&p->p.DLL,&p->AhiDispOverlaySurf,T("AhiDispOverlaySurf"),0);

	OvlGetProc(&p->p.DLL,&p->AhiIDCTOpen,T("AhiIDCTOpen"),0);
	OvlGetProc(&p->p.DLL,&p->AhiIDCTClose,T("AhiIDCTClose"),0);
	OvlGetProc(&p->p.DLL,&p->AhiIDCTPropGet,T("AhiIDCTPropGet"),0);
	OvlGetProc(&p->p.DLL,&p->AhiIDCTPropSet,T("AhiIDCTPropSet"),0);
	OvlGetProc(&p->p.DLL,&p->AhiIDCTFrameStart,T("AhiIDCTFrameStart"),0);
	OvlGetProc(&p->p.DLL,&p->AhiIDCTFrameEnd,T("AhiIDCTFrameEnd"),0);
	OvlGetProc(&p->p.DLL,&p->AhiIDCTMComp8x8,T("AhiIDCTMComp8x8"),0);
	OvlGetProc(&p->p.DLL,&p->AhiIDCTMComp,T("AhiIDCTMComp"),0);
	OvlGetProc(&p->p.DLL,&p->AhiIDCT8x8,T("AhiIDCT8x8"),0);

	memset(&p->DevInfo,0,sizeof(p->DevInfo));

	if (p->p.DLL && p->AhiInit(0)==SUCCESS)
	{
		bool_t Found = 0;
		for (No=0;No<16;No++)
		{
			int Result = p->AhiDevEnum(&p->Device,&p->DevInfo,No);
			if (Result == SUCCESS)
			{
				Found = 1;
				if (p->DevInfo.ChipId == 0x57441002)
					p->IDCTBug = 1;
				break;
			}
			if (Result == ENDENUM)
				break;
		}
		if (!Found)
			p->p.DLL = NULL;

		p->AhiTerm();
	}

	if (p->p.DLL && p->DevInfo.ExternalMemSize)
	{
		p->p.AccelIDCT = &p->IdctVMT;
		NodeRegister((node*)&p->IdctVMT,PRI_DEFAULT+200);
	}

	OvlInit(&p->p,AHI_ID,1,1,PRI_DEFAULT+200);
}

void AHI_Done()
{
	NodeUnRegister((node*)&AHI.IdctVMT);
	OvlDone(&AHI.p);
}
#else
void AHI_Init() {}
void AHI_Done() {}
#endif
