/*****************************************************************************
 *
 * This program is free software ; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 *
 * $Id$
 *
 ****************************************************************************/

#include "stdafx.h"

void OvlInvalidate( overlay_base* p, const rect* Rect, bool_t Local )
{
	POINT o;
	RECT r;

	if (Rect->Height>0 && Rect->Width>0)
	{
		DEBUG_MSG5(T("Invalidate %d,%d,%d,%d (%d)"),Rect->x,Rect->y,Rect->Width,Rect->Height,Local);

		if (Local)
		{
			// only local invalidate
			o.x = 0;
			o.y = 0;
			ClientToScreen(p->Wnd,&o);

			r.left = Rect->x - o.x;
			r.top = Rect->y - o.y;
			r.right = r.left + Rect->Width;
			r.bottom = r.top + Rect->Height;
			InvalidateRect(p->Wnd,&r,0);
		}
		else
			GlobalInvalidate(Rect);
	}
}

void OvlValidate( overlay_base* p, const rect* Rect )
{
	POINT o;
	RECT r;

	if (Rect->Height>0 && Rect->Width>0)
	{
		DEBUG_MSG4(T("Validate %d,%d,%d,%d"),Rect->x,Rect->y,Rect->Width,Rect->Height);

		o.x = 0;
		o.y = 0;
		ClientToScreen(p->Wnd,&o);

		r.left = Rect->x - o.x;
		r.top = Rect->y - o.y;
		r.right = r.left+Rect->Width;
		r.bottom = r.top+Rect->Height;

		ValidateRect(p->Wnd,&r);
	}
} 

void OvlInit(overlay_base* p,int Id,int Primary,int UseDLL,int Priority)
{
	if (UseDLL && !p->DLL)
		return;

	p->Soft = NULL;
	p->Id = Id;
	p->Primary = Primary;

	memset(&p->OrigFX,0,sizeof(blitfx));
	p->OrigFX.ScaleX = SCALE_ONE;
	p->OrigFX.ScaleY = SCALE_ONE;
	p->AutoPrerotate = 1;
	p->ColorKey = -1;
	p->Aspect.Num = p->Aspect.Den = 1;

	NodeRegister(&p->VMT,Priority);
}

void OvlDone(overlay_base* p)
{
	BlitRelease(p->Soft);
	NodeUnRegister(&p->VMT);

	if (p->DLL)
		FreeLibrary(p->DLL);
	p->DLL = NULL;
}

int OvlEnum(overlay_base* p,int No,datadef* DataType)
{
	return NodeEnumType(&No,DataType,NodeParams,FlowParams,OutputParams,VOutputParams,p->Params,NULL);
}

bool_t OvlUpdateAlign( overlay_base* p )
{
	rect OldGUI = p->GUIAlignedRect;
	rect Old = p->DstAlignedRect;

	VirtToPhy(&p->Viewport,&p->DstAlignedRect,&p->Output);
	VirtToPhy(NULL,&p->SrcAlignedRect,&p->Input);

	BlitRelease(p->Soft);
	p->Soft = BlitCreate(&p->Output,&p->Input,&p->FX);
	BlitAlign(p->Soft,&p->DstAlignedRect, &p->SrcAlignedRect );

	PhyToVirt(&p->DstAlignedRect,&p->GUIAlignedRect,&p->Output);

	if (EqRect(&Old,&p->DstAlignedRect))
		return 0;

	if (p->Show && p->Primary)
	{
		OvlInvalidate(p,&OldGUI,0);
		OvlInvalidate(p,&p->Viewport,1);
		OvlValidate(p,&p->GUIAlignedRect);
	}
	return 1;
}

void OvlGetProc(void** DLL,void* Ptr,const tchar_t* ProcName,int Optional)
{
	FARPROC* Func = (FARPROC*)Ptr;
	if (*DLL)
	{
		*Func = GetProcAddress(*DLL,ProcName);
		if (!*Func && !Optional)
		{
			FreeLibrary(*DLL);
			*DLL = NULL;
		}
	}
	else
		*Func = NULL;
}

int OvlGet(overlay_base* p,int No,void* Data,int Size)
{
	int Result = ERR_INVALID_PARAM;
	switch (No)
	{
	case NODE_ID: GETVALUE(p->Id,int); break;
	case VOUT_PRIMARY: GETVALUE(p->Primary,bool_t); break;
	case VOUT_INPUT|PACKET_FORMAT: GETVALUE(p->Input,video); break;
	case VOUT_OUTPUT|PACKET_FORMAT: GETVALUE(p->Output,video); break;
	case VOUT_ACCEL_IDCT: GETVALUE(p->AccelIDCT,idct*); break;
	case VOUT_IDCT: GETVALUE(p->IDCT,idct*); break;
	case VOUT_WND: GETVALUE(p->Wnd,void*); break;
	case VOUT_VISIBLE: GETVALUE(p->Visible,bool_t); break;
	case VOUT_CLIPPING: GETVALUE(p->Clipping,bool_t); break;
	case VOUT_FX: GETVALUE(p->OrigFX,blitfx); break;
	case VOUT_VIEWPORT: GETVALUE(p->Viewport,rect); break;
	case VOUT_OUTPUTRECT: GETVALUE(p->GUIAlignedRect,rect); break;
	case VOUT_AUTOPREROTATE: GETVALUE(p->AutoPrerotate,bool_t); break;
	case VOUT_DIRTY: GETVALUE(p->Dirty,bool_t); break;
	case VOUT_COLORKEY: GETVALUECOND(p->ColorKey,int32_t,p->ColorKey!=-1); break;
	case VOUT_ASPECT: GETVALUE(p->Aspect,fraction); break;
	case VOUT_ORIENTATION:
		if (Size==sizeof(int) && p->QueryOrientation && !p->DisableOrientation)
		{
			*(int*)Data = p->QueryOrientation(p);
			Result = ERR_NONE;
		}
		break;
	}
	return Result;
}

static int UpdateFX(overlay_base* p, bool_t ForceUpdate)
{
	blitfx LastFX = p->FX;
	int Width,Height,ScaleX,ScaleY;

	if (!p->Inited) return ERR_INVALID_PARAM;

	p->FX = p->OrigFX;
	p->FX.Flags |= p->SetFX;
	p->FX.Flags &= ~p->ClearFX;
	p->FX.Direction = CombineDir(p->InputDirection,p->OrigFX.Direction,p->Output.Direction);

	if (p->OrigFX.ScaleX<=0)
	{
		int v = p->OrigFX.ScaleX;

		Width = p->Viewport.Width;
		Height = p->Viewport.Height;

		if ((p->InputDirection ^ p->OrigFX.Direction) & DIR_SWAPXY)
			SwapInt(&Width,&Height);

		if (p->PreRotate)
		{
			ScaleX = Scale(Width,SCALE_ONE,p->Input.Width);
			ScaleY = Scale(Height,SCALE_ONE,Scale(p->Input.Height,p->Aspect.Num,p->Aspect.Den));
		}
		else
		{
			ScaleX = Scale(Width,SCALE_ONE,Scale(p->Input.Width,p->Aspect.Num,p->Aspect.Den));
			ScaleY = Scale(Height,SCALE_ONE,p->Input.Height);
		}

		if (p->PreRotate && (v == -2*SCALE_ONE || v == -SCALE_ONE))
			v = -3*SCALE_ONE - v;

		if ((v == -2*SCALE_ONE) || (v != -SCALE_ONE && ScaleX>ScaleY))
			ScaleX = ScaleY;

		p->FX.ScaleY = ScaleX;
		p->FX.ScaleX = Scale(ScaleX,p->Aspect.Num,p->Aspect.Den);

		if (p->PreRotate)
			SwapInt(&p->FX.ScaleX,&p->FX.ScaleY);
	}
	else
	{
		// prefer source horizontal scaling (because of smooth scale option)
		if (p->PreRotate)
			p->FX.ScaleY = Scale(p->FX.ScaleY,p->Aspect.Num,p->Aspect.Den);
		else
			p->FX.ScaleX = Scale(p->FX.ScaleX,p->Aspect.Num,p->Aspect.Den);
	}

	if (p->Output.Pixel.Flags & PF_PIXELDOUBLE)
	{
		p->FX.ScaleX >>= 1;
		p->FX.ScaleY >>= 1;
	}

	if (ForceUpdate || !EqBlitFX(&p->FX,&LastFX))
	{
		p->Update(p);
		p->Dirty = 1;
		p->LastTime = -1;
	}

	return ERR_NONE;
}

static int UpdateInputDirection(overlay_base* p,bool_t Update)
{
	if (p->Inited)
	{
		p->InputDirection = p->Input.Direction;
		p->PreRotate = 0;

		if (p->AutoPrerotate)
		{
			rect r;
			PhyToVirt(NULL,&r,&p->Input);

			// portrait?
			if (r.Width < r.Height)
			{
				p->PreRotate = 1;
				p->InputDirection ^= DIR_SWAPXY | DIR_MIRRORUPDOWN;
			}
		}

		if (Update)
			UpdateFX(p,0);
	}
	return ERR_NONE;
}

static void Done(overlay_base* p)
{
	if (p->Inited)
	{
		if (p->Show)
		{
			p->Show = 0;
			if (p->UpdateShow)
				p->UpdateShow(p);
		}
		p->Done(p);
		BlitRelease(p->Soft);
		p->Soft = NULL;
		p->Inited = 0;

		if (p->Primary)
			OvlInvalidate(p,&p->Viewport,1);
	}
}

static int UpdateShow(overlay_base* p)
{
	bool_t Show = p->Visible && (p->ColorKey != -1 || !p->Clipping);

	if (p->Show != Show)
	{
		p->Show = Show;
		p->Dirty = 1;
		p->LastTime = -1;

		if (p->Inited)
		{
			if (p->UpdateShow)
				p->UpdateShow(p);
			else
			if (p->Primary)
			{
				if (p->Show)
				{
					OvlInvalidate(p,&p->Viewport,1); // redraw border (zoom may have changed)
					OvlValidate(p,&p->GUIAlignedRect);
				}
				else
				{
					OvlInvalidate(p,&p->Viewport,0); // redraw other windows
					OvlValidate(p,&p->Viewport); // own window is fine
				}
			}
		}
	}
	return ERR_NONE;
}

static int Init(overlay_base* p)
{
	memset(&p->FX,0,sizeof(blitfx));
	memset(&p->Output,0,sizeof(video));
	memset(&p->DstAlignedRect,0,sizeof(rect));
	memset(&p->SrcAlignedRect,0,sizeof(rect));
	memset(&p->GUIAlignedRect,0,sizeof(rect));

	p->Show = 0;
	p->Dirty = 1;
	p->ColorKey = -1;
	p->LastTime = -1;
	p->SetFX = 0;
	p->ClearFX = 0;

	if (p->Input.Pixel.Flags)
	{
		int Result;
		
		if (!p->Wnd)
			return ERR_INVALID_PARAM;

		if ((p->Input.Pixel.Flags & PF_FOURCC) && !AnyYUV(&p->Input.Pixel))
			return ERR_NOT_SUPPORTED;

		if (p->Input.Width<=0) 
			p->Input.Width = 2;

		if (p->Input.Height<=0) 
			p->Input.Height = 2;

		Result = p->Init(p);

		if (Result != ERR_NONE)
		{
			if (Result == ERR_DEVICE_ERROR)
				NodeError((node*)p,ERR_ID,ERR_DEVICE_ERROR);
			return Result;
		}

		p->Inited = 1;
			
		UpdateInputDirection(p,0);
		UpdateFX(p,1);
		UpdateShow(p);

		if (p->Primary)
			OvlInvalidate(p,&p->Viewport,1);
	}
	return ERR_NONE;
}

int OvlBlit(overlay_base* p,const planes Data,const planes LastData)
{
	int Result = ERR_NONE;

	if (p->Soft && p->Inited && p->Show)
	{
		if (p->Dirty)
			LastData = NULL;

		Result = p->Blit(p,Data,LastData);
		
		if (Result == ERR_NONE)
			p->Dirty = 0;
	}

	return Result;
}

void OvlAdjustOrientation(overlay_base* p)
{
	int Width = GetSystemMetrics(SM_CXSCREEN);
	int Height = GetSystemMetrics(SM_CYSCREEN);
	int Orientation;
	rect Virt;

	p->DisableOrientation = 1;

	if ((Width == p->Output.Width*2 && Height == p->Output.Height*2) || 
		(Width == p->Output.Height*2 && Height == p->Output.Width*2))
		p->Output.Pixel.Flags |= PF_PIXELDOUBLE;

	PhyToVirt(NULL,&Virt,&p->Output);

	// rotated?
	if (Virt.Width == Height && Virt.Height == Width)
	{
		Orientation = QueryOrientation();

		if (Orientation & DIR_MIRRORUPDOWN)
			p->Output.Direction ^= DIR_SWAPXY | DIR_MIRRORUPDOWN;
		else
			p->Output.Direction ^= DIR_SWAPXY | DIR_MIRRORLEFTRIGHT;
	}

	p->DisableOrientation = 0;
}

int OvlSet(overlay_base* p,int No,const void* Data,int Size)
{
	int Result = ERR_INVALID_PARAM;
	switch (No)
	{
	case VOUT_WND:
		if (Size == sizeof(HWND))
		{
			Result = ERR_NONE;
			if (p->Wnd != *(HWND*)Data)

			{
				Done(p);
				p->Wnd = *(HWND*)Data;
				Result = Init(p);
			}
		}
		break;

	case VOUT_INPUT:
		if (Size == sizeof(packet))
		{
			const packet* Packet = (const packet*)Data;

			//DEBUG_MSG3(T("VOUT currtime:%d reftime:%d lasttime:%d"),Packet->CurrTime,Packet->RefTime,p->LastTime);

			if (Packet->CurrTime != -1 && 
				(Packet->RefTime > Packet->CurrTime + SHOWAHEAD || Packet->CurrTime == p->LastTime))
				Result = ERR_BUFFER_FULL;
			else
			{
				p->LastTime = (Packet->CurrTime>=0) ? Packet->CurrTime : Packet->RefTime;
				Result = OvlBlit(p,Packet->Data,Packet->LastData);
			}
		}
		break;

	case VOUT_INPUT | PACKET_FORMAT:
		if (!Size || Size == sizeof(video))
		{
			Done(p);
			if (Size)
				p->Input = *(video*)Data;
			else
				memset(&p->Input,0,sizeof(video));
			Result = Init(p);
		}
		break;

	case VOUT_DIRTY: SETVALUE(p->Dirty,bool_t,ERR_NONE); break;
	case VOUT_CLIPPING: SETVALUE(p->Clipping,bool_t,UpdateShow(p)); break;
	case VOUT_VISIBLE: SETVALUE(p->Visible,bool_t,UpdateShow(p)); break;
	case VOUT_AUTOPREROTATE: SETVALUE(p->AutoPrerotate,bool_t,UpdateInputDirection(p,1)); break;
	case VOUT_ASPECT: SETVALUECMP(p->Aspect,fraction,UpdateFX(p,0),EqFrac); break;
	case VOUT_FX: SETVALUE(p->OrigFX,blitfx,UpdateFX(p,0)); break;

	case VOUT_VIEWPORT:
		if (Size == sizeof(rect))
		{
			Result = ERR_NONE;
			if (!EqRect(&p->Viewport,(const rect*)Data))
			{
				p->Viewport = *(const rect*)Data;
				if (p->Inited)
					UpdateFX(p,1);
			}
		}
		break;

	case VOUT_RESET:
		if (p->Inited && p->Reset)
		{
			p->Reset(p);
			p->Update(p);
			p->Dirty = 1;
			p->LastTime = -1;
		}
		Result = ERR_NONE;
		break;
	}
	return Result;
}