/*****************************************************************************
 *
 * This program is free software ; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 *
 * $Id$
 *
 ****************************************************************************/

#ifndef __OVERLAY_BASE_H
#define __OVERLAY_BASE_H

// default colorkey to use
#define COLORKEY	0x200020

typedef struct overlay_base
{
	node VMT;

	int (*Init)(void*);
	void (*Done)(void*);
	int (*Blit)(void*,const planes Data,const planes LastData);
	int (*Update)(void*);
	int (*UpdateShow)(void*);
	int (*QueryOrientation)(void*);
	int (*Reset)(void*);
	idct* AccelIDCT;
	const datadef* Params;

	// these parameters should be filled during Init()
	int SetFX;	
	int ClearFX;
	video Output;
	
	// these are handled by overlay_base
	int Id;
	bool_t Primary;
	idct* IDCT;
	void* Wnd;
	bool_t Inited;
	bool_t Dirty;
	bool_t Visible;
	bool_t Clipping;
	bool_t AutoPrerotate;
	bool_t Show;
	tick_t LastTime;
	int32_t ColorKey;
	video Input;
	void* Soft;
	rect Viewport;
	rect SrcAlignedRect;
	rect GUIAlignedRect;
	rect DstAlignedRect;
	blitfx FX;
	blitfx OrigFX;
	fraction Aspect;
	bool_t PreRotate;
	int InputDirection;
	bool_t DisableOrientation;
	void* DLL;

} overlay_base;

void OvlInit(overlay_base* p,int Id,int Primary,int UseDLL,int Priority);
void OvlDone(overlay_base* p);
int OvlEnum(overlay_base* p,int EnumNo,datadef* DataType);
int OvlGet(overlay_base* p,int No,void* Data,int Size);
int OvlSet(overlay_base* p,int No,const void* Data,int Size);
bool_t OvlUpdateAlign(overlay_base* p);
void OvlGetProc(void** DLL,void*,const tchar_t* ProcName,int Optional);
void OvlValidate( overlay_base* p, const rect* );
void OvlInvalidate( overlay_base* p, const rect*, bool_t Local );
void OvlAdjustOrientation( overlay_base* p );

#endif
