/*****************************************************************************
 *
 * This program is free software ; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 *
 * $Id$
 *
 ****************************************************************************/

#include "stdafx.h"

#if !defined(_WIN32_WCE)

#if _MSC_VER > 1000
#pragma warning( push, 3 )
#endif

#include <ddraw.h>

typedef struct ddraw
{
	overlay_base p;
	video Overlay;
	rect OverlayRect;
	blitfx SoftFX;
	blitfx OvlFX;

	int DstAlignSize;
	int DstAlignPos;
	int MinScale;
	int MaxScale;

	HRESULT (WINAPI* DirectDrawCreate)( GUID FAR*, LPDIRECTDRAW FAR*, IUnknown FAR* );

	LPDIRECTDRAW DD;
	LPDIRECTDRAWSURFACE DDPrimary;
	LPDIRECTDRAWSURFACE DDOverlay;
	LPDIRECTDRAWCOLORCONTROL DDColor;
	DDCAPS DDCaps;

} ddraw;

void Desc2Surface( const DDSURFACEDESC* Desc, video* p )
{
	memset(p,0,sizeof(video));
	p->Width = Desc->dwWidth;
	p->Height = Desc->dwHeight;
	p->Aspect = ASPECT_ONE;
	p->Direction = 0;

	if (Desc->ddpfPixelFormat.dwFlags & DDPF_FOURCC)
	{
		p->Pixel.Flags = PF_FOURCC;
		p->Pixel.FourCC = Desc->ddpfPixelFormat.dwFourCC;
	}
	else
	if (Desc->ddpfPixelFormat.dwFlags & DDPF_RGB)
	{
		p->Pixel.Flags = PF_RGB;
		p->Pixel.BitCount = Desc->ddpfPixelFormat.dwRGBBitCount;
		p->Pixel.RBitMask = Desc->ddpfPixelFormat.dwRBitMask;
		p->Pixel.GBitMask = Desc->ddpfPixelFormat.dwGBitMask;
		p->Pixel.BBitMask = Desc->ddpfPixelFormat.dwBBitMask;
	}
	else
	if (Desc->ddpfPixelFormat.dwFlags & DDPF_PALETTEINDEXED1)
	{
		p->Pixel.Flags = PF_PALETTE;
		p->Pixel.BitCount = 1;
	}
	else
	if (Desc->ddpfPixelFormat.dwFlags & DDPF_PALETTEINDEXED2)
	{
		p->Pixel.Flags = PF_PALETTE;
		p->Pixel.BitCount = 2;
	}
	else
	if (Desc->ddpfPixelFormat.dwFlags & DDPF_PALETTEINDEXED4)
	{
		p->Pixel.Flags = PF_PALETTE;
		p->Pixel.BitCount = 4;
	}
	else
	if (Desc->ddpfPixelFormat.dwFlags & DDPF_PALETTEINDEXED8)
	{
		p->Pixel.Flags = PF_PALETTE;
		p->Pixel.BitCount = 8;
	}
}

bool_t Surface2Desc( const video* p, DDSURFACEDESC* Desc, int PixelFormat )
{
	memset(Desc,0,sizeof(DDSURFACEDESC));
	Desc->dwSize = sizeof(DDSURFACEDESC);
	Desc->dwFlags = DDSD_WIDTH | DDSD_HEIGHT;

	Desc->dwWidth = p->Width;
	Desc->dwHeight = p->Height;

	if (PixelFormat)
	{
		Desc->dwFlags |= DDSD_PIXELFORMAT;
		Desc->ddpfPixelFormat.dwSize = sizeof(Desc->ddpfPixelFormat);

		if (p->Pixel.Flags & PF_FOURCC)
		{
			Desc->ddpfPixelFormat.dwFlags = DDPF_FOURCC;
			Desc->ddpfPixelFormat.dwFourCC = p->Pixel.FourCC;
		}
		else
		if (p->Pixel.Flags & PF_RGB)
		{
			Desc->ddpfPixelFormat.dwFlags = DDPF_RGB;
			Desc->ddpfPixelFormat.dwRGBBitCount = p->Pixel.BitCount;
			Desc->ddpfPixelFormat.dwRBitMask = p->Pixel.RBitMask;
			Desc->ddpfPixelFormat.dwGBitMask = p->Pixel.GBitMask;
			Desc->ddpfPixelFormat.dwBBitMask = p->Pixel.BBitMask;
		}
		else
		if (p->Pixel.Flags & PF_PALETTE)
		{
			switch (p->Pixel.BitCount)
			{
			case 1: Desc->ddpfPixelFormat.dwFlags = DDPF_PALETTEINDEXED1; break;
			case 2: Desc->ddpfPixelFormat.dwFlags = DDPF_PALETTEINDEXED2; break;
			case 4: Desc->ddpfPixelFormat.dwFlags = DDPF_PALETTEINDEXED4; break;
			case 8: Desc->ddpfPixelFormat.dwFlags = DDPF_PALETTEINDEXED8; break;
			}
		}
		else
			return 0;
	}

	return 1;
}

static bool_t CreateOverlay(ddraw* p,int PixelFormat)
{
	DDSURFACEDESC Desc;

	if (p->DDOverlay) 
	{ 
		IDirectDrawSurface_Release(p->DDOverlay); 
		p->DDOverlay = NULL;
	}

	if (!Surface2Desc(&p->Overlay,&Desc,PixelFormat))
		return 0;

	Desc.dwFlags |= DDSD_CAPS;
	Desc.ddsCaps.dwCaps = DDSCAPS_VIDEOMEMORY | DDSCAPS_OVERLAY;

	if (IDirectDraw_CreateSurface(p->DD,&Desc,&p->DDOverlay,NULL) != DD_OK)
		return 0;

	if (IDirectDrawSurface_GetSurfaceDesc(p->DDOverlay,&Desc) == DD_OK)
		Desc2Surface(&Desc,&p->Overlay);

	// fill with black
 
	if (IDirectDrawSurface_Lock(p->DDOverlay,NULL,&Desc,DDLOCK_WAIT,NULL) == DD_OK)
	{
		int v;

		FillInfo(&p->Overlay.Pixel);
		v = RGBToFormat(0,&p->Overlay.Pixel);

		if (PlanarYUV(&p->Overlay.Pixel) && p->Overlay.Pixel.BitCount==8)
		{
			uint8_t* i = (uint8_t*)Desc.lpSurface;
			FillColor(i,Desc.lPitch,p->Overlay.Width,p->Overlay.Height,8,v & 255);
			i += Desc.lPitch * p->Overlay.Height;
			FillColor(i,Desc.lPitch >> 1,p->Overlay.Width >> 1,p->Overlay.Height >> 1,8,(v>>8)&255);
			i += Desc.lPitch * p->Overlay.Height >> 2;
			FillColor(i,Desc.lPitch >> 1,p->Overlay.Width >> 1,p->Overlay.Height >> 1,8,(v>>16)&255);
		}
		else
		if (PackedYUV(&p->Overlay.Pixel) && p->Overlay.Pixel.BitCount==16)
			FillColor(Desc.lpSurface,Desc.lPitch,p->Overlay.Width >> 1,p->Overlay.Height,32,v);
		else
			FillColor(Desc.lpSurface,Desc.lPitch,p->Overlay.Width,p->Overlay.Height,
				p->Overlay.Pixel.BitCount,v);

		IDirectDrawSurface_Unlock(p->DDOverlay,NULL);
	}

	return 1;
}

static int Init(ddraw* p)
{
	DDSURFACEDESC Desc;

	if (p->DDCaps.dwCaps & DDCAPS_COLORKEY)
		p->p.ColorKey = COLORKEY;

	if (p->DDCaps.dwCaps & DDCAPS_OVERLAYSTRETCH)
	{
		p->MinScale = (p->DDCaps.dwMinOverlayStretch * SCALE_ONE) / 1000;
		p->MaxScale = (p->DDCaps.dwMaxOverlayStretch * SCALE_ONE) / 1000;
	}
	else
	{
		p->MinScale = SCALE_ONE;
		p->MaxScale = SCALE_ONE;
	}

	if (p->DDCaps.dwCaps & DDCAPS_ALIGNBOUNDARYDEST)
		p->DstAlignPos = p->DDCaps.dwAlignBoundaryDest;
	else
		p->DstAlignPos = 1;

	if (p->DDCaps.dwCaps & DDCAPS_ALIGNSIZEDEST)
		p->DstAlignSize = p->DDCaps.dwAlignSizeDest;
	else
		p->DstAlignSize = 1;

	// get mode
	memset(&Desc,0,sizeof(DDSURFACEDESC));
	Desc.dwSize = sizeof(DDSURFACEDESC);
	IDirectDraw_GetDisplayMode(p->DD,&Desc);
	Desc2Surface(&Desc,&p->p.Output);
	FillInfo(&p->p.Output.Pixel);

	if (IDirectDraw_SetCooperativeLevel(p->DD, p->p.Wnd, DDSCL_NORMAL ) != DD_OK)
		return ERR_DEVICE_ERROR;

	// get primary surface
	Desc.dwFlags = DDSD_CAPS;
	Desc.ddsCaps.dwCaps = DDSCAPS_PRIMARYSURFACE;
	if (IDirectDraw_CreateSurface(p->DD,&Desc,&p->DDPrimary,NULL) != DD_OK)
		return ERR_DEVICE_ERROR;

	p->Overlay = p->p.Input;

	// first try the optimal format
	if (!CreateOverlay(p,1))
	{
		if (PlanarYUV(&p->Overlay.Pixel))
		{
			// try other formats
			static const int FourCC[] = 
			{ 
				// prefer planar formats
				FOURCC_YV12, FOURCC_IYUV, 
				FOURCC_I420,
				// next packed formats
				FOURCC_YUY2, FOURCC_YUNV, 
				FOURCC_V422, FOURCC_YUYV, 
				FOURCC_YVYU, FOURCC_UYVY, 
				FOURCC_Y422, FOURCC_UYNV,
				0 
			};
			const int* i;

			for (i=FourCC;*i;++i)
			{
				p->Overlay.Pixel.Flags = PF_FOURCC;
				p->Overlay.Pixel.FourCC = *i;
				if (CreateOverlay(p,1))
					break;
			}
		}

		// last hope is the device's current RGB mode
		if (!p->DDOverlay && !CreateOverlay(p,0))
			return ERR_NOT_SUPPORTED;
	}

	return ERR_NONE;
}

static void Done(ddraw* p)
{
	if (p->DDOverlay) 
	{ 
		IDirectDrawSurface_Release(p->DDOverlay); 
		p->DDOverlay = NULL;
	}
	if (p->DDPrimary) 
	{ 
		IDirectDrawSurface_Release(p->DDPrimary); 
		p->DDPrimary = NULL;
	}
}

static int UpdateOverlay(ddraw* p)
{
	DDOVERLAYFX DDFX;
	RECT Src,Dst;
	int Flags;

	memset(&DDFX,0,sizeof(DDFX));
	DDFX.dwSize = sizeof(DDFX);
	Flags = 0;

	if (p->OvlFX.Flags & BLITFX_ARITHSTRETCH)
	{
		Flags |= DDOVER_DDFX;
		DDFX.dwDDFX |= DDOVERFX_ARITHSTRETCHY;
	}

	if (p->OvlFX.Flags & DIR_MIRRORLEFTRIGHT)
	{
		Flags |= DDOVER_DDFX;
		DDFX.dwDDFX ^= DDOVERFX_MIRRORLEFTRIGHT;
	}

	if (p->OvlFX.Direction & DIR_MIRRORUPDOWN)
	{
		Flags |= DDOVER_DDFX;
		DDFX.dwDDFX ^= DDOVERFX_MIRRORUPDOWN;
	}

	Src.left = p->OverlayRect.x;
	Src.top = p->OverlayRect.y;
	Src.right = p->OverlayRect.x + p->OverlayRect.Width;
	Src.bottom = p->OverlayRect.y + p->OverlayRect.Height;
 
	Dst.left = p->p.DstAlignedRect.x;
	Dst.top = p->p.DstAlignedRect.y;
	Dst.right = p->p.DstAlignedRect.x + p->p.DstAlignedRect.Width;
	Dst.bottom = p->p.DstAlignedRect.y + p->p.DstAlignedRect.Height;

	if (p->p.Show)
		Flags |= DDOVER_SHOW;
	else
		Flags |= DDOVER_HIDE;

	if (p->p.ColorKey != -1)
		Flags |= DDOVER_KEYDEST;

	IDirectDrawSurface_UpdateOverlay(p->DDOverlay,&Src,p->DDPrimary,&Dst,Flags,&DDFX);

	return ERR_NONE;
}

static int Update(ddraw* p)
{
	int OvlWidth = p->p.Input.Width;
	int OvlHeight = p->p.Input.Height;

	p->OvlFX = p->p.FX;
	p->SoftFX = p->p.FX;

	p->OvlFX.Brightness = 0;
	p->OvlFX.Contrast = 0;
	p->OvlFX.Saturation = 0;
	p->OvlFX.Direction &= ~DIR_SWAPXY; //swap handled by SoftFX

	p->SoftFX.ScaleX = SCALE_ONE;
	p->SoftFX.ScaleY = SCALE_ONE;

	if (!(p->DDCaps.dwFXCaps & DDFXCAPS_OVERLAYARITHSTRETCHY))
		p->OvlFX.Flags &= ~BLITFX_ARITHSTRETCH;

	//if (p->DDCaps.dwFXCaps & DDFXCAPS_OVERLAYMIRRORLEFTRIGHT)
	//	p->SoftFX.Direction &= ~DIR_MIRRORLEFTRIGHT;
	//else
		p->OvlFX.Direction &= ~DIR_MIRRORLEFTRIGHT;
		
	//if (p->DDCaps.dwFXCaps & DDFXCAPS_OVERLAYMIRRORUPDOWN)
	//	p->SoftFX.Direction &= ~DIR_MIRRORUPDOWN;
	//else
		p->OvlFX.Direction &= ~DIR_MIRRORUPDOWN;

	if (p->SoftFX.Direction & DIR_SWAPXY)
		SwapInt(&p->OvlFX.ScaleX,&p->OvlFX.ScaleY);

	if ((p->p.FX.Direction ^ p->p.Input.Direction) & DIR_SWAPXY)
		SwapInt(&OvlWidth,&OvlHeight);

	if (p->Overlay.Width != OvlWidth ||
		p->Overlay.Height != OvlHeight)
	{
		p->Overlay.Width = OvlWidth;
		p->Overlay.Height = OvlHeight;
		CreateOverlay(p,1);
	}

	VirtToPhy(&p->p.Viewport,&p->p.DstAlignedRect,&p->p.Output);
	VirtToPhy(NULL,&p->OverlayRect,&p->Overlay);

	AnyAlign(&p->p.DstAlignedRect, &p->OverlayRect, &p->OvlFX,
		p->DstAlignSize,p->DstAlignPos,p->MinScale,p->MaxScale);

	PhyToVirt(&p->p.DstAlignedRect,&p->p.GUIAlignedRect,&p->p.Output);

	VirtToPhy(NULL,&p->p.SrcAlignedRect,&p->p.Input);

	BlitRelease(p->p.Soft);
	p->p.Soft = BlitCreate(&p->Overlay,&p->p.Input,&p->SoftFX);

	BlitAlign(p->p.Soft,&p->OverlayRect,&p->p.SrcAlignedRect);

	if (p->p.ColorKey!=-1 && p->DDPrimary)
	{
		DDCOLORKEY Key;

		Key.dwColorSpaceLowValue = RGBToFormat(p->p.ColorKey,&p->p.Output.Pixel);
		Key.dwColorSpaceHighValue = 0;

		OvlInvalidate(&p->p,&p->p.Viewport,1);

		if (IDirectDrawSurface_SetColorKey(p->DDPrimary,DDCKEY_DESTOVERLAY,&Key)!=DD_OK)
			return ERR_NOT_SUPPORTED;
	}

	return UpdateOverlay(p);
}

static int Blit(ddraw* p, const planes Data, const planes DataLast )
{
	DDSURFACEDESC Desc;
	HRESULT hResult;
	planes Planes;

	if (!p->DDOverlay)
		return ERR_NOT_SUPPORTED;

	Desc.dwSize = sizeof(Desc);
	while ((hResult = IDirectDrawSurface_Lock(p->DDOverlay,NULL,&Desc,DDLOCK_WAIT,NULL)) != DD_OK)
		if (hResult != DDERR_SURFACELOST || IDirectDrawSurface_Restore(p->DDOverlay) != DD_OK)
			return ERR_NOT_SUPPORTED;

	Planes[0] = Desc.lpSurface;

	BlitImage(p->p.Soft,Planes,Data,DataLast,Desc.lPitch,-1);

	IDirectDrawSurface_Unlock(p->DDOverlay,NULL);

	return ERR_NONE;
}

static ddraw DDraw = 
{
	OvlEnum,
	OvlGet,
	OvlSet,
	Init,
	Done,
	Blit,
	Update,
	UpdateOverlay,
	NULL,
	NULL,
};

void DDraw_Init() 
{ 
	ddraw* p = &DDraw;

	p->DD = NULL;
	p->p.DLL = LoadLibrary(T("DDRAW.DLL"));
	OvlGetProc(&p->p.DLL,&p->DirectDrawCreate,T("DirectDrawCreate"),0);

	p->DDCaps.dwSize = sizeof(p->DDCaps);
	if (p->p.DLL && 
		p->DirectDrawCreate(NULL,&p->DD,NULL)==DD_OK &&			// create direct draw object
		IDirectDraw_GetCaps(p->DD,&p->DDCaps,NULL) == DD_OK &&	// get caps
		(p->DDCaps.dwCaps & DDCAPS_OVERLAY))					// overlay support?
	{
		OvlInit(&p->p,DDRAW_ID,1,1,PRI_DEFAULT+100);
		return;
	}

	if (p->DD) IDirectDraw_Release(p->DD);
	if (p->p.DLL) FreeLibrary(p->p.DLL);

	p->p.DLL = NULL;
	p->DD = NULL;
}

void DDraw_Done()
{
	ddraw* p = &DDraw;
	if (p->DD) 
		IDirectDraw_Release(p->DD);
	OvlDone(&p->p);
}

#else
void DDraw_Init() {}
void DDraw_Done() {}
#endif
