/*****************************************************************************
 *
 * This program is free software ; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 *
 * $Id$
 *
 ****************************************************************************/

#include "stdafx.h"

#define OUTPUT_VGA		17
#define Format320X240	2

typedef struct flytv
{
	overlay_base p;
	planes Planes;

	WCHAR* (*DriverVersion)();
	int (*Initial_FLY_TV)(int,int);
	void (*UnInitial_FLY_TV)();
	int (*Zoom)(int);
	int (*SendFrame)(const void*);
} flytv;

static int Init(flytv* p)
{
	p->p.Output.Width = 320;
	p->p.Output.Height = 240;
	p->p.Output.Direction = 0;
	p->p.Output.Aspect = ASPECT_ONE;
	p->p.Output.Pitch = p->p.Output.Width*2;
	DefaultRGB(&p->p.Output.Pixel,16,5,6,5,0,0,0);

	p->Planes[0] = Alloc(2 * p->p.Output.Height * p->p.Output.Width);
	if (!p->Planes[0])
		return ERR_OUT_OF_MEMORY;

	p->p.ClearFX = BLITFX_ONLYDIFF;
	return ERR_NONE;
}

static void Done(flytv* p)
{
	Free(p->Planes[0]); 
	p->Planes[0] = NULL;
}

static int Update(flytv* p)
{
	if (OvlUpdateAlign(&p->p))
		memset(p->Planes[0],0,p->p.Output.Width * p->p.Output.Height * 2);
	return ERR_NONE;
}

static int UpdateShow(flytv* p)
{
	if (p->p.Show)
	{
		if (!p->Initial_FLY_TV(OUTPUT_VGA,Format320X240))
		{
			NodeError((node*)p,ERR_ID,ERR_DEVICE_ERROR);
			return ERR_DEVICE_ERROR;
		}
		p->Zoom(2);
	}
	else
		p->UnInitial_FLY_TV();
	return ERR_NONE;
}

static int Blit(flytv* p, const planes Data, const planes DataLast )
{
	BlitImage(p->p.Soft,p->Planes,Data,DataLast,-1,-1);

	if (p->SendFrame(p->Planes[0]) != 0)
		return ERR_DEVICE_ERROR;

	return ERR_NONE;
}

static flytv FlyTV = 
{
	OvlEnum,
	OvlGet,
	OvlSet,
	Init,
	Done,
	Blit,
	Update,
	UpdateShow,
	NULL,
	NULL,
};

void FlyTV_Init() 
{ 
	flytv* p = &FlyTV;

	p->p.DLL = LoadLibrary(T("FlyPresenter_TV.DLL"));
	OvlGetProc(&p->p.DLL,&p->DriverVersion,T("DriverVersion"),1);
	OvlGetProc(&p->p.DLL,&p->Initial_FLY_TV,T("Initial_FLY_TV"),0);
	OvlGetProc(&p->p.DLL,&p->UnInitial_FLY_TV,T("UnInitial_FLY_TV"),0);
	OvlGetProc(&p->p.DLL,&p->Zoom,T("Zoom"),0);
	OvlGetProc(&p->p.DLL,&p->SendFrame,T("SendFrame"),0);

	OvlInit(&FlyTV.p,FLYTV_ID,0,1,PRI_DEFAULT-10);
}

void FlyTV_Done()
{
	OvlDone(&FlyTV.p);
}

