/*****************************************************************************
 *
 * This program is free software ; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 *
 * $Id$
 *
 ****************************************************************************/

#include "stdafx.h"

typedef struct GXDisplayProperties 
{
	int32_t cxWidth;
	int32_t cyHeight;
	int32_t cbxPitch;
	int32_t cbyPitch;
	int32_t cBPP;
	int32_t ffFormat;
} GXDisplayProperties;

#define SHFS_SHOWSTARTICON  0x0010
#define SHFS_SHOWTASKBAR    0x0001

#define GX_FULLSCREEN		0x01
#define kfLandscape			0x08
#define kfPalette			0x10
#define kfDirect			0x20
#define kfDirect555			0x40
#define kfDirect565			0x80
#define kfDirect888			0x100
#define kfDirect444			0x200
#define kfDirectInverted	0x400

typedef struct gapi
{
	overlay_base p;
	int AdjustPtr;
	int DRAM;
	bool_t FirstTime;
	bool_t Windows;
	bool_t Opened;
	GXDisplayProperties Info;
	HMODULE AygShell;

	int (*GXOpenDisplay)(HWND hWnd, DWORD dwFlags);
	int (*GXCloseDisplay)();
	void* (*GXBeginDraw)();
	int (*GXEndDraw)();
	GXDisplayProperties (*GXGetDisplayProperties)();
	int (*GXSuspend)();
	int (*GXResume)();
	int (*GXSetViewport)( DWORD dwTop, DWORD dwHeight, DWORD, DWORD );
	BOOL (*GXIsDisplayDRAMBuffer)();
	BOOL (WINAPI* SHFullScreen)( HWND, DWORD );

} gapi;

const datadef GAPIParams[] = 
{
	{ GAPI_ID,			TYPE_CLASS },
	{ GAPI_WIDTH,		TYPE_INT, DF_SETUP | DF_RDONLY },
	{ GAPI_HEIGHT,		TYPE_INT, DF_SETUP | DF_RDONLY },
	{ GAPI_PITCHX,		TYPE_INT, DF_SETUP | DF_RDONLY },
	{ GAPI_PITCHY,		TYPE_INT, DF_SETUP | DF_RDONLY },
	{ GAPI_BPP,			TYPE_INT, DF_SETUP | DF_RDONLY },
	{ GAPI_FORMAT,		TYPE_INT, DF_SETUP | DF_RDONLY|DF_HEX },

	DATADEF_END
};

static int Get(gapi* p,int No,void* Data,int Size)
{
	int Result = OvlGet(&p->p,No,Data,Size);
	switch (No)
	{
	case GAPI_WIDTH: GETVALUE(p->Info.cxWidth,int); break;
	case GAPI_HEIGHT: GETVALUE(p->Info.cyHeight,int); break;
	case GAPI_PITCHX: GETVALUE(p->Info.cbxPitch,int); break;
	case GAPI_PITCHY: GETVALUE(p->Info.cbyPitch,int); break;
	case GAPI_BPP: GETVALUE(p->Info.cBPP,int); break;
	case GAPI_FORMAT: GETVALUE(p->Info.ffFormat,int); break;
	}
	return Result;
}

static void LoadDLL(gapi* p, bool_t Windows )
{
	tchar_t Name[MAXPATH];
	HMODULE DLL;
	p->Windows = Windows;

	if (Windows)
	{
		tcscpy(Name,GetSystemPath());
		tcscat(Name,T("\\gx.dll"));
	}
	else
		tcscpy(Name,T("gx.dll"));

	DLL = LoadLibrary(Name);
	if (DLL)
	{
		if (p->p.DLL)
			FreeLibrary(p->p.DLL);

		p->p.DLL = DLL;
		OvlGetProc(&p->p.DLL,&p->GXOpenDisplay,T("?GXOpenDisplay@@YAHPAUHWND__@@K@Z"),0);
		OvlGetProc(&p->p.DLL,&p->GXCloseDisplay,T("?GXCloseDisplay@@YAHXZ"),0);
		OvlGetProc(&p->p.DLL,&p->GXBeginDraw,T("?GXBeginDraw@@YAPAXXZ"),0);
		OvlGetProc(&p->p.DLL,&p->GXEndDraw,T("?GXEndDraw@@YAHXZ"),0);
		OvlGetProc(&p->p.DLL,&p->GXGetDisplayProperties,T("?GXGetDisplayProperties@@YA?AUGXDisplayProperties@@XZ"),0);
		OvlGetProc(&p->p.DLL,&p->GXSuspend,T("?GXSuspend@@YAHXZ"),1);
		OvlGetProc(&p->p.DLL,&p->GXResume,T("?GXResume@@YAHXZ"),1);
		OvlGetProc(&p->p.DLL,&p->GXSetViewport,T("?GXSetViewport@@YAHKKKK@Z"),1);
		OvlGetProc(&p->p.DLL,&p->GXIsDisplayDRAMBuffer,T("?GXIsDisplayDRAMBuffer@@YAHXZ"),1);
	}
}

static int Init(gapi* p)
{
	int Caps = QueryPlatform(PLATFORM_CAPS);
	video GDI;
	GXDisplayProperties Info;

	if (p->FirstTime)
	{
		HWND WndTask;

		p->FirstTime = 0;

		// only open the device one time (WM2003SE?)

retry:	p->Opened = p->p.DLL && p->GXOpenDisplay(p->p.Wnd,GX_FULLSCREEN) != 0;

		if (!p->Opened && !p->Windows)
		{
			// try gx.dll in \windows directory (HPC with .NET)
			LoadDLL(p,1);
			goto retry;
		}

		if (!p->Opened)
			NodeError((node*)p,GAPI_ID,GAPI_OPEN_ERROR);
		else
		{
			// just in case GXOpenDisplay hides these things

			if (p->SHFullScreen)
			{
				p->SHFullScreen(p->p.Wnd,SHFS_SHOWSTARTICON);
				p->SHFullScreen(p->p.Wnd,SHFS_SHOWTASKBAR);
			}

			WndTask = FindWindow(T("HHTaskbar"),NULL);
			if (WndTask)
				ShowWindow(WndTask,SW_SHOWNA);
		}
	}

	if (!p->Opened)
		return ERR_DEVICE_ERROR;

	Info = p->Info = p->GXGetDisplayProperties();

	p->DRAM = 0;
	if (p->GXIsDisplayDRAMBuffer)
		p->DRAM = p->GXIsDisplayDRAMBuffer() != 0;	

	memset(&p->p.Output,0,sizeof(p->p.Output));
	p->p.Output.Width = Info.cxWidth;
	p->p.Output.Height = Info.cyHeight;
	p->p.Output.Direction = 0;
	p->p.Output.Aspect = ASPECT_ONE;

	if (Info.ffFormat & kfPalette)
	{
		QueryDesktop(&GDI);

		p->p.Output.Pixel.Flags = PF_PALETTE;
		p->p.Output.Pixel.BitCount = Info.cBPP;
		p->p.Output.Pixel.Palette = GDI.Pixel.Palette;
	}
	else
	if (Info.ffFormat & kfDirect444)
		DefaultRGB(&p->p.Output.Pixel,Info.cBPP,4,4,4,0,0,0);
	else
	if (Info.ffFormat & kfDirect565)
	{
		if (Caps & CAPS_ONLY12BITRGB)
			DefaultRGB(&p->p.Output.Pixel,Info.cBPP,4,4,4,1,2,1);
		else
			DefaultRGB(&p->p.Output.Pixel,Info.cBPP,5,6,5,0,0,0);
	}
	else
	if (Info.ffFormat & kfDirect555)
		DefaultRGB(&p->p.Output.Pixel,Info.cBPP,5,5,5,0,0,0);
	else
	if (Info.ffFormat & kfDirect888)
		DefaultRGB(&p->p.Output.Pixel,Info.cBPP,8,8,8,0,0,0);

	if (Info.ffFormat & kfDirectInverted)
		p->p.Output.Pixel.Flags |= PF_INVERTED;

	// get signed x/y pitches in bits
	Info.cbxPitch <<= 3;
	Info.cbyPitch <<= 3;

	// one pitch is probably zero when BPP<8
	if (!Info.cbxPitch) 
		if (Info.cbyPitch < 0)
			Info.cbxPitch = -Info.cBPP;
		else
			Info.cbxPitch = Info.cBPP;
	else
	if (!Info.cbyPitch)
		if (Info.cbxPitch < 0)
			Info.cbyPitch = Info.cBPP;
		else
			Info.cbyPitch = -Info.cBPP;

	// one of the pitches absolute value must be BPP
	if (abs(Info.cbyPitch) < abs(Info.cbxPitch))
	{
		Info.ffFormat |= kfLandscape;
		Info.cbyPitch = Info.cbyPitch<0 ? -Info.cBPP : Info.cBPP;
	}
	else
	{
		Info.ffFormat &= ~kfLandscape;
		Info.cbxPitch = Info.cbxPitch<0 ? -Info.cBPP : Info.cBPP;
	}

	// we need the physical start of the framebuffer
	p->AdjustPtr = 0;
	if (Info.cbxPitch<0) 
		p->AdjustPtr += (Info.cbxPitch * (Info.cxWidth-1)) >> 3;
	if (Info.cbyPitch<0) 
		p->AdjustPtr += (Info.cbyPitch * (Info.cyHeight-1)) >> 3;

	if (Info.ffFormat & kfLandscape)
	{
		p->p.Output.Direction |= DIR_SWAPXY;
		p->p.Output.Pitch = abs(Info.cbxPitch) >> 3;
		SwapInt(&p->p.Output.Width,&p->p.Output.Height);

		if (Info.cbxPitch<0) 
			p->p.Output.Direction |= DIR_MIRRORUPDOWN;
		if (Info.cbyPitch<0) 
			p->p.Output.Direction |= DIR_MIRRORLEFTRIGHT;
	}
	else
	{
		p->p.Output.Pitch = abs(Info.cbyPitch) >> 3;

		if (Info.cbxPitch<0) 
			p->p.Output.Direction |= DIR_MIRRORLEFTRIGHT;
		if (Info.cbyPitch<0) 
			p->p.Output.Direction |= DIR_MIRRORUPDOWN;
	}

	if (!p->DRAM)
	{
		p->p.SetFX = BLITFX_AVOIDTEARING;

		if ((Info.cxWidth > Info.cyHeight) && (Caps & CAPS_PORTRAIT))
			p->p.SetFX |= BLITFX_VMEMROTATED;

		if ((Info.cyHeight > Info.cxWidth) && (Caps & CAPS_LANDSCAPE))
			p->p.SetFX |= BLITFX_VMEMROTATED;
	}
	else
		p->p.ClearFX = BLITFX_ONLYDIFF;

	// wm2003se gapi emulation?
	if (QueryPlatform(PLATFORM_VER) >= 421 && Info.cxWidth == 240 && Info.cyHeight == 320)
		OvlAdjustOrientation(&p->p);

	return ERR_NONE;
}

static void Done(gapi* p)
{
}

static int Reset(gapi* p)
{
	Done(p);
	Init(p);
	return ERR_NONE;
}

static void Clear(gapi* p)
{
	uint8_t* Ptr;

	// clear screen

	if (p->DRAM && p->GXSetViewport)
		p->GXSetViewport(0,p->Info.cyHeight,0,0);

	Ptr = (uint8_t*)p->GXBeginDraw();
	if (Ptr)
	{
		FillColor(Ptr + p->AdjustPtr,p->p.Output.Pitch,p->p.Output.Width,
			p->p.Output.Height,p->p.Output.Pixel.BitCount,0);

		p->GXEndDraw();
	}
}

static int Update(gapi* p)
{
	rect FullScreen;
	PhyToVirt(NULL,&FullScreen,&p->p.Output);
	if (p->DRAM && EqRect(&FullScreen,&p->p.Viewport))
		Clear(p);

	OvlUpdateAlign(&p->p);
	return ERR_NONE;
}

static int Blit(gapi* p, const planes Data, const planes DataLast )
{
	char* Ptr;
	planes Planes;

	if (p->DRAM && p->GXSetViewport)
	{
		int y = p->p.GUIAlignedRect.y;
		int Height = p->p.GUIAlignedRect.Height;

		if (p->p.Output.Pixel.Flags & PF_PIXELDOUBLE)
		{
			y >>= 1;
			Height >>= 1;
		}

		p->GXSetViewport(y,Height,0,0); // coords in windows space
	}

	Ptr = (char*)p->GXBeginDraw();
	if (!Ptr)
		return ERR_NOT_SUPPORTED;

	Planes[0] = Ptr + p->AdjustPtr;

	BlitImage(p->p.Soft,Planes,Data,DataLast,-1,-1);

	p->GXEndDraw();

	return ERR_NONE;
}

static gapi GAPI = 
{
	OvlEnum,
	Get,
	OvlSet,
	Init,
	Done,
	Blit,
	Update,
	NULL,
	NULL,
	Reset,
};

void GAPI_Init() 
{ 
	gapi* p = &GAPI;

	p->AygShell = LoadLibrary(T("aygshell.dll"));
	OvlGetProc(&p->AygShell,&p->SHFullScreen,T("SHFullScreen"),1);

	p->p.Params = GAPIParams;
	p->FirstTime = 1;

	LoadDLL(p,0);

	OvlInit(&GAPI.p,GAPI_ID,1,1,PRI_DEFAULT+50);
}

void GAPI_Done()
{
	gapi* p = &GAPI;

	if (p->Opened)
		p->GXCloseDisplay();

	OvlDone(&p->p);

	if (p->AygShell)
		FreeLibrary(p->AygShell);
}
