/*****************************************************************************
 *
 * This program is free software ; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 *
 * $Id$
 *
 ****************************************************************************/

#include "stdafx.h"

typedef struct gdi
{
	overlay_base p;
	int DIBSection;
	HBITMAP Bitmap;
	HGDIOBJ Bitmap0; //bitmap returned by selectobject
	HDC DC;
	HDC DC2;
	video Overlay;
	rect OverlayRect;
	planes Planes;
} gdi;

static int AllocBitmap(gdi* p)
{
	p->p.Dirty = 1;

	if (p->DIBSection)
	{
		struct 
		{
			BITMAPINFOHEADER Head;
			int RBitMask;
			int GBitMask;
			int BBitMask;
		} Info;

		memset(&Info,0,sizeof(Info));
		Info.Head.biSize = sizeof(Info.Head);
		Info.Head.biWidth = p->Overlay.Width;
		Info.Head.biHeight = -p->Overlay.Height;
		Info.Head.biPlanes = 1;
		Info.Head.biBitCount = (WORD)p->Overlay.Pixel.BitCount;
		Info.Head.biCompression = BI_BITFIELDS;
		Info.Head.biSizeImage = (p->Overlay.Width * p->Overlay.Height * p->Overlay.Pixel.BitCount) >> 3;
		Info.RBitMask = p->Overlay.Pixel.RBitMask;
		Info.GBitMask = p->Overlay.Pixel.GBitMask;
		Info.BBitMask = p->Overlay.Pixel.BBitMask;

		p->Bitmap = CreateDIBSection(p->DC,(BITMAPINFO*)&Info,DIB_RGB_COLORS,&p->Planes[0],NULL,0);

		if (p->Bitmap)
			p->Bitmap0 = SelectObject(p->DC2,p->Bitmap);
	}
	else
		p->Planes[0] = Alloc(p->Overlay.Pitch * p->Overlay.Height);

	return p->Planes[0] != NULL;
}

static void FreeBitmap(gdi* p)
{
	if (p->Bitmap) 
	{ 
		SelectObject(p->DC2,p->Bitmap0);
		DeleteObject(p->Bitmap); 
		p->Bitmap = 0;
	}

	if (p->Planes[0])
	{
		if (!p->DIBSection)
			Free(p->Planes[0]);

		p->Planes[0] = NULL;
	}
}

static void Done(gdi* p)
{
	FreeBitmap(p);

	if (p->DC2) 
	{
		DeleteDC(p->DC2); 
		p->DC2 = 0;
	}
	if (p->DC) 
	{
		ReleaseDC(NULL,p->DC);
		p->DC = 0;
	}
}

static int Init(gdi* p)
{
	QueryDesktop(&p->p.Output);

	p->Planes[0] = NULL;

	p->Overlay.Width = 0;
	p->Overlay.Height = 0;
	p->Overlay.Direction = 0;
	p->Overlay.Aspect = ASPECT_ONE;
	p->Overlay.Pixel = p->p.Output.Pixel;

	p->DIBSection = (p->Overlay.Pixel.Flags & PF_RGB) != 0;
	p->Bitmap = 0;
	p->Bitmap0 = 0;
	p->DC = GetDC(NULL);
	p->DC2 = CreateCompatibleDC(p->DC);

	p->p.ClearFX = BLITFX_ONLYDIFF;
	return ERR_NONE;
}

static int Reset(gdi* p)
{
	Done(p);
	Init(p);
	return ERR_NONE;
}

static int Update(gdi* p)
{
	rect OldGUI = p->p.GUIAlignedRect;
	rect Old = p->p.DstAlignedRect;

	int OldWidth = p->Overlay.Width;
	int OldHeight = p->Overlay.Height;

	VirtToPhy(&p->p.Viewport,&p->p.DstAlignedRect,&p->p.Output);
	VirtToPhy(NULL,&p->p.SrcAlignedRect,&p->p.Input);

	AnyAlign(&p->p.DstAlignedRect, &p->p.SrcAlignedRect, &p->p.FX, 2, 2, 1, SCALE_ONE*1024 ); 

	PhyToVirt(&p->p.DstAlignedRect,&p->p.GUIAlignedRect,&p->p.Output);

	p->Overlay.Width = p->p.GUIAlignedRect.Width;
	p->Overlay.Height = p->p.GUIAlignedRect.Height;
	p->Overlay.Pitch = (p->Overlay.Width * p->Overlay.Pixel.BitCount) >> 3;
	p->Overlay.Pitch = (p->Overlay.Pitch+1) & ~1; //word aligned

	p->OverlayRect.x = p->OverlayRect.y = 0;
	p->OverlayRect.Width = p->Overlay.Width;
	p->OverlayRect.Height = p->Overlay.Height;

	BlitRelease(p->p.Soft);
	p->p.Soft = BlitCreate(&p->Overlay,&p->p.Input,&p->p.FX);
		
	BlitAlign(p->p.Soft, &p->OverlayRect, &p->p.SrcAlignedRect);

	p->p.GUIAlignedRect.Width = p->OverlayRect.Width;
	p->p.GUIAlignedRect.Height = p->OverlayRect.Height;

	VirtToPhy(&p->p.GUIAlignedRect,&p->p.DstAlignedRect,&p->p.Output);

	if (OldWidth != p->Overlay.Width || OldHeight != p->Overlay.Height)
		FreeBitmap(p);

	if (p->p.Show && !EqRect(&Old,&p->p.DstAlignedRect))
	{
		OvlInvalidate(&p->p,&OldGUI,0);
		OvlInvalidate(&p->p,&p->p.Viewport,1);
		OvlValidate(&p->p,&p->p.GUIAlignedRect);
	}
	return ERR_NONE;
}

static int Blit(gdi* p, const planes Data, const planes DataLast )
{
	if (!p->Planes[0] && !AllocBitmap(p))
		return ERR_OUT_OF_MEMORY;

	BlitImage(p->p.Soft,p->Planes,Data,DataLast,-1,-1);

	if (!p->DIBSection)
	{
		if (p->Bitmap)
		{
			SelectObject(p->DC2,p->Bitmap0);
			DeleteObject(p->Bitmap);
		}

		p->Bitmap = CreateBitmap( p->Overlay.Width, p->Overlay.Height, 1, 
			p->Overlay.Pixel.BitCount, (char*)p->Planes[0]);

		if (!p->Bitmap)
			return ERR_OUT_OF_MEMORY;

		p->Bitmap0 = SelectObject(p->DC2,p->Bitmap);
	}

	BitBlt(p->DC,p->p.GUIAlignedRect.x,p->p.GUIAlignedRect.y,
		   p->OverlayRect.Width,p->OverlayRect.Height,p->DC2,p->OverlayRect.x,p->OverlayRect.y,SRCCOPY);

	return ERR_NONE;
}

static gdi GDI = 
{
	OvlEnum,
	OvlGet,
	OvlSet,
	Init,
	Done,
	Blit,
	Update,
	NULL,
	NULL,
	Reset,
};

void GDI_Init() 
{ 
	OvlInit(&GDI.p,GDI_ID,1,0,PRI_DEFAULT);
}

void GDI_Done() 
{ 
	OvlDone(&GDI.p);
}
