/*****************************************************************************
 *
 * This program is free software ; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 *
 * $Id$
 *
 ****************************************************************************/

#include "stdafx.h"

#define GETRAWFRAMEBUFFER   0x00020001

#define FORMAT_565 1
#define FORMAT_555 2
#define FORMAT_OTHER 3

typedef struct RawFrameBufferInfo
{
	WORD wFormat;
	WORD wBPP;
	char *pFramePointer;
	int  cxStride;
	int  cyStride;
	int  cxPixels;
	int  cyPixels;

} RawFrameBufferInfo;

typedef struct raw
{
	overlay_base p;
	int AdjustPtr;
	RawFrameBufferInfo Info;
	HDC DC;

} raw;

const datadef RAWParams[] = 
{
	{ RAW_ID,			TYPE_CLASS },
	{ RAW_WIDTH,		TYPE_INT, DF_SETUP | DF_RDONLY },
	{ RAW_HEIGHT,		TYPE_INT, DF_SETUP | DF_RDONLY },
	{ RAW_PITCHX,		TYPE_INT, DF_SETUP | DF_RDONLY },
	{ RAW_PITCHY,		TYPE_INT, DF_SETUP | DF_RDONLY },
	{ RAW_BPP,			TYPE_INT, DF_SETUP | DF_RDONLY },
	{ RAW_FORMAT,		TYPE_INT, DF_SETUP | DF_RDONLY },
	{ RAW_POINTER,		TYPE_INT, DF_SETUP | DF_RDONLY | DF_HEX },

	DATADEF_END
};

static raw* Query(raw* p)
{
	HDC DC = p->DC;
	bool_t OwnDC = !DC;
	if (OwnDC)
		DC = GetDC(NULL);

	if (ExtEscape(DC, GETRAWFRAMEBUFFER, 0, NULL, sizeof(RawFrameBufferInfo), (char*)&p->Info) < 0)
		p->Info.pFramePointer = NULL;

	if (OwnDC)
		ReleaseDC(NULL,DC);
	return p;
}

static int Get(raw* p,int No,void* Data,int Size)
{
	int Result = OvlGet(&p->p,No,Data,Size);
	switch (No)
	{
	case RAW_WIDTH: GETVALUE(Query(p)->Info.cxPixels,int); break;
	case RAW_HEIGHT: GETVALUE(Query(p)->Info.cyPixels,int); break;
	case RAW_PITCHX: GETVALUE(Query(p)->Info.cxStride,int); break;
	case RAW_PITCHY: GETVALUE(Query(p)->Info.cyStride,int); break;
	case RAW_BPP: GETVALUE(Query(p)->Info.wBPP,int); break;
	case RAW_FORMAT: GETVALUE(Query(p)->Info.wFormat,int); break;
	case RAW_POINTER: GETVALUE((int)Query(p)->Info.pFramePointer,int); break;
	}
	return Result;
}

static int Init(raw* p)
{
	p->DC = GetDC(NULL);
	Query(p);

	if (!p->Info.pFramePointer)
		return ERR_DEVICE_ERROR;

	memset(&p->p.Output,0,sizeof(p->p.Output));
	p->p.Output.Width = p->Info.cxPixels;
	p->p.Output.Height = p->Info.cyPixels;
	p->p.Output.Direction = 0;
	p->p.Output.Aspect = ASPECT_ONE;

	switch (p->Info.wFormat)
	{
	case FORMAT_555:
		DefaultRGB(&p->p.Output.Pixel,p->Info.wBPP,5,5,5,0,0,0);
		break;
	case FORMAT_565:
		DefaultRGB(&p->p.Output.Pixel,p->Info.wBPP,5,6,5,0,0,0);
		break;
	default:
		return ERR_DEVICE_ERROR;
	}

	// we need the physical start of the framebuffer
	p->AdjustPtr = 0;
	if (p->Info.cxStride<0) 
		p->AdjustPtr += (p->Info.cxStride * (p->Info.cxPixels-1));
	if (p->Info.cyStride<0) 
		p->AdjustPtr += (p->Info.cyStride * (p->Info.cyPixels-1));

	if (abs(p->Info.cyStride) < abs(p->Info.cxStride))
	{
		p->p.Output.Direction |= DIR_SWAPXY;
		p->p.Output.Pitch = abs(p->Info.cxStride);
		SwapInt(&p->p.Output.Width,&p->p.Output.Height);

		if (p->Info.cxStride<0) 
			p->p.Output.Direction |= DIR_MIRRORUPDOWN;
		if (p->Info.cyStride<0) 
			p->p.Output.Direction |= DIR_MIRRORLEFTRIGHT;
	}
	else
	{
		p->p.Output.Pitch = abs(p->Info.cyStride);

		if (p->Info.cxStride<0) 
			p->p.Output.Direction |= DIR_MIRRORLEFTRIGHT;
		if (p->Info.cyStride<0) 
			p->p.Output.Direction |= DIR_MIRRORUPDOWN;
	}

	p->p.SetFX = BLITFX_AVOIDTEARING;

	OvlAdjustOrientation(&p->p);

	return ERR_NONE;
}

static void Done(raw* p)
{
	if (p->DC)
	{
		ReleaseDC(NULL, p->DC);
		p->DC = 0;
	}
}

static int Reset(raw* p)
{
	Done(p);
	Init(p);
	return ERR_NONE;
}

static int Update(raw* p)
{
	OvlUpdateAlign(&p->p);
	return ERR_NONE;
}

static int Blit(raw* p, const planes Data, const planes DataLast )
{
	planes Planes;

/*	is this neccessary???

	RawFrameBufferInfo Info;
	if (ExtEscape(p->DC, GETRAWFRAMEBUFFER, 0, NULL, sizeof(RawFrameBufferInfo), (char*)&Info) < 0)
		return ERR_DEVICE_ERROR;

	p->Info.pFramePointer = Info.pFramePointer;
	if (memcmp(&Info,&p->Info,sizeof(Info))!=0)
		return ERR_DEVICE_ERROR;
*/

	Planes[0] = p->Info.pFramePointer + p->AdjustPtr;

	BlitImage(p->p.Soft,Planes,Data,DataLast,-1,-1);

	return ERR_NONE;
}

static raw RAW = 
{
	OvlEnum,
	Get,
	OvlSet,
	Init,
	Done,
	Blit,
	Update,
	NULL,
	NULL,
	Reset,
};

void RAW_Init() 
{ 
	raw* p = &RAW;
	Query(p);
	if (p->Info.pFramePointer)
	{
		p->p.Params = RAWParams;
		OvlInit(&p->p,RAW_ID,1,0,PRI_DEFAULT+95);
	}
}

void RAW_Done()
{
	raw* p = &RAW;
	OvlDone(&p->p);
}
