/*****************************************************************************
 *
 * This program is free software ; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 *
 * $Id$
 *
 ****************************************************************************/

#include "stdafx.h"

#define LOGSIZE	8192

typedef struct benchresult
{
	win Win;
	tchar_t* Log;

} benchresult;

static int AddItem( benchresult* p, int y, int Id, const tchar_t* Buffer )
{
	const tchar_t* Label = String(BENCHRESULT_ID,Id);
	if (y>=0)
		y = WinPropLabel(&p->Win,y,Label,Buffer,NULL);
	stprintf(p->Log+tcslen(p->Log),T("%-30s%s\n"),Label,Buffer);
	return y;
}

static bool_t Proc(benchresult* p,int Msg, WPARAM wParam, LPARAM lParam, int* Result)
{
	node* Player;
	node* Format;
	node* Input;
	int y;
	int Frames;
	int Samples;
	int Bytes;
	tick_t Tick;
	tick_t OrigTick;
	video Video;
	audio Audio;
	tchar_t Buffer[256];
	fraction f;

	switch (Msg)
	{
	case WM_COMMAND:
		if (LOWORD(wParam) == BENCHRESULT_SAVE)
		{
			char* s;
			SYSTEMTIME t;
			FILE* f;

			GetSystemTime(&t);
			stprintf(Buffer,T("%s\\bench_%d%02d%02d_%02d%02d%02d.txt"),
				GetDocumentPath(),t.wYear,t.wMonth,t.wDay,t.wHour,t.wMinute,t.wSecond);

			f = tfopen(Buffer,T("w+"));
			if (f)
			{
				s = (char*) Alloc(LOGSIZE);
				if (s)
				{
					tcstostr(s,p->Log,LOGSIZE);
					fputs(s,f);
					ShowMessage(String(BENCHRESULT_ID,0),String(BENCHRESULT_ID,BENCHRESULT_SAVED),Buffer);
					Free(s);
				}
				fclose(f);
			}
		}
		if (LOWORD(wParam) == PLATFORM_OK)
			PostMessage(p->Win.Wnd,WM_CLOSE,0,0);
		break;

	case WM_DESTROY:
		Free(p->Log);
		p->Log = NULL;
		break;

	case WM_CREATE:
		WinCreateDialog(&p->Win);

		Player = NodeFind(PLAYER_ID);
		Player->Get(Player,PLAYER_FORMAT,&Format,sizeof(Format));
		Player->Get(Player,PLAYER_INPUT,&Input,sizeof(Input));

		OrigTick = 0;
		Frames = 0;
		Samples = 0;
		Bytes = 0;
		memset(&Video,0,sizeof(Video));
		memset(&Audio,0,sizeof(Audio));

		p->Log = (tchar_t*)Alloc(LOGSIZE);

		stprintf(p->Log,String(BENCHRESULT_ID,BENCHRESULT_LOG),NodeProgramId(),NodeVersion());
		stprintf(p->Log+tcslen(p->Log),T("\n\n"));

		if (Input && Input->Get(Input,STREAM_URL,Buffer,sizeof(Buffer))==ERR_NONE)
			AddItem(p,-1,BENCHRESULT_URL,Buffer);

		if (Format)
		{
			int No;
			pin Pin;
			datadef DataDef;

			Format->Get(Format,FORMAT_FILEPOS,&Bytes,sizeof(Bytes));

			for (No=0;Format->Get(Format,FORMAT_STREAM+No,&Pin,sizeof(Pin))==ERR_NONE;++No)
				if (Pin.Node && NodeDataDef(Format,FORMAT_STREAM+No,&DataDef))
				{
					if (DataDef.Format1 == PACKET_VIDEO && !Frames)
					{
						Format->Get(Format,(FORMAT_STREAM+No)|PACKET_FORMAT,&Video,sizeof(Video));
						Pin.Node->Get(Pin.Node,FLOW_TOTAL,&Frames,sizeof(Frames));
					}
					if (DataDef.Format1 == PACKET_AUDIO && !Samples)
					{
						Format->Get(Format,(FORMAT_STREAM+No)|PACKET_FORMAT,&Audio,sizeof(Audio));
						Pin.Node->Get(Pin.Node,FLOW_TOTAL,&Samples,sizeof(Samples));
					}
				}
		}

		y = 4;

		Player->Get(Player,PLAYER_BENCHMARK,&Tick,sizeof(Tick));

		if (Frames && Video.FrameRate.Num)
		{
			f.Num = Video.FrameRate.Den;
			f.Den = Video.FrameRate.Num;
			Simplify(&f,MAX_INT/TICKSPERSEC,MAX_INT);
			OrigTick = Scale(Frames,f.Num*TICKSPERSEC,f.Den);
		}
		else
		if (Samples && Audio.SampleRate)
			OrigTick = Scale(Samples,TICKSPERSEC,Audio.SampleRate);

		if (Tick && OrigTick)
		{
			f.Num = OrigTick;
			f.Den = Tick;

			FractionToString(Buffer,&f,sizeof(Buffer),1,2);
			y = AddItem(p,y,BENCHRESULT_SPEED,Buffer);
		}

		if (Frames)
		{
			IntToString(Buffer,Frames,sizeof(Buffer),0);
			y = AddItem(p,y,BENCHRESULT_FRAMES,Buffer);
		}

		if (Samples)
		{
			IntToString(Buffer,Samples,sizeof(Buffer),0);
			y = AddItem(p,y,BENCHRESULT_SAMPLES,Buffer);
		}

		if (Bytes)
		{
			IntToString(Buffer,Bytes/1024,sizeof(Buffer),0);
			tcscat(Buffer,T(" KBytes"));
			y = AddItem(p,y,BENCHRESULT_BYTES,Buffer);
		}

		y += 6;
		stprintf(p->Log+tcslen(p->Log),T("\n"));

		TickToString(Buffer,Tick,sizeof(Buffer),0,1);
		y = AddItem(p,y,BENCHRESULT_TIME,Buffer);

		if (Frames && Tick)
		{
			f.Num = Frames;
			f.Den = Tick;
			Simplify(&f,MAX_INT/TICKSPERSEC,MAX_INT);
			f.Num *= TICKSPERSEC;
			FractionToString(Buffer,&f,sizeof(Buffer),0,2);
			y = AddItem(p,y,BENCHRESULT_FPS,Buffer);
		}

		if (Samples && Tick)
		{
			f.Num = Samples;
			f.Den = Tick;
			Simplify(&f,MAX_INT/TICKSPERSEC,MAX_INT);
			f.Num *= TICKSPERSEC;
			FractionToString(Buffer,&f,sizeof(Buffer),0,0);
			y = AddItem(p,y,BENCHRESULT_SRATE,Buffer);
		}

		if (Bytes && Tick)
		{
			f.Num = Scale(Bytes,8,1000);
			f.Den = Tick;
			Simplify(&f,MAX_INT/TICKSPERSEC,MAX_INT);
			f.Num *= TICKSPERSEC;
			if (f.Den && (f.Num/f.Den) > 1000)
			{
				f.Den *= 1000;
				FractionToString(Buffer,&f,sizeof(Buffer),0,1);
				tcscat(Buffer,T(" Mbit/s"));
			}
			else
			{
				FractionToString(Buffer,&f,sizeof(Buffer),0,0);
				tcscat(Buffer,T(" kbit/s"));
			}
			y = AddItem(p,y,BENCHRESULT_BANDWIDTH,Buffer);
		}

		y += 6;
		stprintf(p->Log+tcslen(p->Log),T("\n"));

		if (OrigTick)
		{
			TickToString(Buffer,OrigTick,sizeof(Buffer),0,1);
			y = AddItem(p,y,BENCHRESULT_ORIG_TIME,Buffer);
		}

		if (Video.FrameRate.Num)
		{
			FractionToString(Buffer,&Video.FrameRate,sizeof(Buffer),0,2);
			y = AddItem(p,y,BENCHRESULT_ORIG_FPS,Buffer);
		}

		if (Audio.SampleRate)
		{
			IntToString(Buffer,Audio.SampleRate,sizeof(Buffer),0);
			y = AddItem(p,y,BENCHRESULT_ORIG_SRATE,Buffer);
		}

		if (Bytes && OrigTick)
		{
			f.Num = Scale(Bytes,8,1000);
			f.Den = OrigTick;
			Simplify(&f,MAX_INT/TICKSPERSEC,MAX_INT);
			f.Num *= TICKSPERSEC;
			if (f.Den && (f.Num/f.Den) > 1000)
			{
				f.Den *= 1000;
				FractionToString(Buffer,&f,sizeof(Buffer),0,1);
				tcscat(Buffer,T(" Mbit/s"));
			}
			else
			{
				FractionToString(Buffer,&f,sizeof(Buffer),0,0);
				tcscat(Buffer,T(" kbit/s"));
			}
			y = AddItem(p,y,BENCHRESULT_ORIG_BANDWIDTH,Buffer);
		}

		if (Frames && Samples)
			y = WinLabel(&p->Win,y+9,2,p->Win.DlgWidth-2,String(BENCHRESULT_ID,BENCHRESULT_MSG),11,0,NULL);

		break;
	}
	return 0;
}

static menudef MenuDef[] =
{
	{ 0, PLATFORM_ID, PLATFORM_OK, },
	{ 0, BENCHRESULT_ID, BENCHRESULT_SAVE, },

	MENUDEF_END
};

static benchresult BenchResult;

void BenchResult_Init()
{
	memset(&BenchResult,0,sizeof(BenchResult));
	BenchResult.Win.Id = BENCHRESULT_ID;
	BenchResult.Win.Width = 240;
	BenchResult.Win.Height = 320;
	BenchResult.Win.Dialog = 1;
	BenchResult.Win.Title = String(BENCHRESULT_ID,0);
	BenchResult.Win.MenuDef = MenuDef;
	BenchResult.Win.Proc = Proc;
	BenchResult.Win.LabelWidth = 88;

	if (QueryPlatform(PLATFORM_TYPENO)==TYPE_SMARTPHONE)
		BenchResult.Win.LabelWidth -= 5;

	WinRegister(&BenchResult.Win,PRI_DEFAULT);
}

void BenchResult_Done()
{
	WinUnRegister(&BenchResult.Win);
}