/*****************************************************************************
 *
 * This program is free software ; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 *
 * $Id$
 *
 ****************************************************************************/

#include "stdafx.h"

#define TIMER_REFRESH	500
#define TIMER_SEEK		501
#define TIMER_SEEKFORCE 503

#define SEEK_CYCLE		200
#define SEEKFORCE_CYCLE	1000
#define REFRESH_CYCLE	200

#define TRACKMAX		30000
#define TRACKHEIGHT		16
#define TRACKTHUMB		12

#define VOLTHUMB		12

#define IF_AUDIO		50000
#define IF_VIDEO		51000
#define IF_VIDEOACCEL	52000
#define IF_CHAPTER		53000
#define IF_STREAM_AUDIO	54000
#define IF_STREAM_VIDEO	55000

static int HotKey[] = 
{
	IF_FILE_OPENFILE,
	IF_FILE_PLAYLIST,
	IF_PLAY,
	IF_PLAY_FULLSCREEN,
	IF_STOP,
	IF_MOVE_FFWD,
	IF_MOVE_BACK,
	IF_NEXT,			
	IF_PREV,
	IF_FASTFORWARD,
	IF_OPTIONS_VOLUME_UP,
	IF_OPTIONS_VOLUME_DOWN,
	IF_OPTIONS_MUTE,
	IF_OPTIONS_FULLSCREEN,
	IF_OPTIONS_ROTATE,
	IF_OPTIONS_BRIGHTNESS_UP,
	IF_OPTIONS_BRIGHTNESS_DOWN,
	IF_OPTIONS_CONTRAST_UP,
	IF_OPTIONS_CONTRAST_DOWN,
	IF_OPTIONS_SATURATION_UP,
	IF_OPTIONS_SATURATION_DOWN,
	IF_OPTIONS_SCREEN,
	IF_FILE_EXIT,
};

#define HOTKEYCOUNT (sizeof(HotKey)/sizeof(int))


typedef struct intface
{
	win Win;
	DWORD ThreadId;
	player* Player;
	node* Color;
	WNDPROC DefTrackProc;
	WNDPROC DefVolProc;
	HWND WndTrack;
	HWND WndVol;
	HWND WndVolBack;
	RECT ClientRect;
	rect ClientViewport;
	int InSeek; // 0:no seek 1:moving 2:ending
	bool_t InSeekForce;
	bool_t InVol;
	int LastPos; // last trackbar pos (do not seek two times at the same place)
	tick_t ForcePos; // next seek will use this instead of trackbar pos
	bool_t Buttons;
	bool_t PlayAfterSeek;
	bool_t Focus;
	bool_t Clipping;
	bool_t Overlap;
	bool_t VolResizeNeeded;
	bool_t Exited;
	int ClipCountdown;
	POINT Offset;
	int IdleReset;
	int Initing;
	bool_t Bench;
	bool_t Wait;
	int HotKey[HOTKEYCOUNT];

	// refresh states
	int TrackPos;
	bool_t Play;
	bool_t FFwd;
	bool_t FullScreen;
	bool_t Mute;

	int MenuStreams;
	int MenuAStream;
	int MenuVStream;

} intface;

static void ResizeVolume(intface* p)
{
	if (p->WndVolBack)
	{
		RECT rv;
		RECT r;
		POINT o;
		int Width;
		int i;

		GetClientRect(p->Win.WndTB,&r);
		SendMessage(p->Win.WndTB,TB_GETRECT,IF_OPTIONS_MUTE,(LPARAM)&rv);

		Width = r.right-rv.right;
		if (Width < 20)
			Width = 20;

		i = Scale(50,WinPixelSY(),72);
		if (Width > i)
			Width = i;

		o.x = o.y = 0;
		if (!p->Win.ToolBarHeight)
			ClientToScreen(p->Win.WndTB,&o);

		MoveWindow(p->WndVolBack,rv.right+o.x,o.y+1,Width,r.bottom-1,TRUE);

		GetClientRect(p->WndVolBack,&r);
		i = Scale(2,WinPixelSX(),72);
		MoveWindow(p->WndVol,-i,0,r.right+2*i,r.bottom,TRUE);

		SendMessage(p->WndVol, TBM_SETTHUMBLENGTH,Scale(VOLTHUMB,WinPixelSY(),72),0);

		p->VolResizeNeeded = 0;
	}
}

static bool_t Toggle(intface* p,int Param,int Force);
static void UpdateVolume(intface* p);

static void Resize(intface* p)
{
	rect Viewport;
	RECT r;

	if (IsOrientationChanged())
	{
		p->Player->Set(p->Player,PLAYER_RESETVIDEO,NULL,0);
		p->VolResizeNeeded = 1;

		if (p->Win.FullScreen)
		{
			Toggle(p,PLAYER_FULLSCREEN,0);
			return;
		}
	}

	p->Offset.x = 0;
	p->Offset.y = 0;
	ClientToScreen(p->Win.Wnd,&p->Offset);

	if (!p->Win.FullScreen)
	{
#if !defined(_WIN32_WCE)
		WINDOWPLACEMENT Place;
		Place.length = sizeof(Place);
		GetWindowPlacement(p->Win.Wnd,&Place);
		if (Place.showCmd != SW_MAXIMIZE)
#endif
		{
			GetClientRect(p->Win.Wnd,&r);

			if (r.right != p->ClientRect.right || r.bottom != p->ClientRect.bottom)
			{
				int TrackHeight = Scale(TRACKHEIGHT,WinPixelSY(),72);

				p->ClientRect = r;

				MoveWindow(p->WndTrack,r.left,r.bottom-TrackHeight,r.right,TrackHeight,TRUE);
				SendMessage(p->WndTrack, TBM_SETTHUMBLENGTH,Scale(TRACKTHUMB,WinPixelSY(),72),0);

				p->ClientViewport.x = 0;
				p->ClientViewport.y = p->Win.ToolBarHeight;
				p->ClientViewport.Width = r.right - r.left;
				p->ClientViewport.Height = (r.bottom-TrackHeight) - p->Win.ToolBarHeight;

				if (p->Win.ToolBarHeight)
					ResizeVolume(p);
			}

			Viewport = p->ClientViewport;
			Viewport.x += p->Offset.x;
			Viewport.y += p->Offset.y;

			p->Player->Set(p->Player,PLAYER_SKIN_VIEWPORT,&Viewport,sizeof(Viewport));

			if (p->VolResizeNeeded)
				ResizeVolume(p);
		}
	}
}

static void StartSeek(intface* p)
{
	SetTimer(p->Win.Wnd,TIMER_SEEK,SEEK_CYCLE,NULL);
	if (!p->InSeekForce)
	{
		p->InSeekForce = 1;
		SetTimer(p->Win.Wnd,TIMER_SEEKFORCE,SEEKFORCE_CYCLE,NULL);
	}
}

static void Seek(intface* p)
{
	int Pos = SendMessage(p->WndTrack,TBM_GETPOS,0,0);
	fraction Percent;

	//DEBUG_MSG1("SEEK %d",Pos);

	if (p->ForcePos != -1)
	{
		p->Player->Set(p->Player,PLAYER_POSITION,&p->ForcePos,sizeof(tick_t));
		p->LastPos = -1;
	}
	else
	if (Pos != p->LastPos)
	{
		p->LastPos = Pos;
		Percent.Num = Pos;
		Percent.Den = TRACKMAX;
		p->Player->Set(p->Player,PLAYER_PERCENT,&Percent,sizeof(Percent));
	}

	if (p->InSeek == 2)
	{
		p->Player->Set(p->Player,PLAYER_PLAY,&p->PlayAfterSeek,sizeof(bool_t));
		p->InSeek = 0;
		p->ForcePos = -1;
	}
}

static BOOL CALLBACK EnumCheck(HWND Wnd, LPARAM lParam )
{
	intface* p = (intface*) lParam;
	RECT r;

	if (Wnd == p->Win.Wnd)
		return 0;

	if (IsWindowVisible(Wnd))
	{
		GetWindowRect(Wnd,&r);

		if (p->ClientViewport.x + p->Offset.x < r.right &&
			p->ClientViewport.x + p->Offset.x + p->ClientViewport.Width > r.left &&
			p->ClientViewport.y + p->Offset.y < r.bottom &&
			p->ClientViewport.y + p->Offset.y + p->ClientViewport.Height > r.top)
		{
			p->Overlap = 1;
			return 0;
		}
	}
	return 1;
}

static void UpdateHotKey(intface* p,bool_t State)
{
	int i;
	for (i=0;i<HOTKEYCOUNT;++i)
		if (p->HotKey[i])
		{
			if (State)
				WinRegisterHotKey(&p->Win,HotKey[i],p->HotKey[i]);
			else
				WinUnRegisterHotKey(&p->Win,HotKey[i]);
		}
}

static void UpdateClipping(intface* p,bool_t Suggest,bool_t HotKey)
{
	if (p->Clipping != Suggest)
	{
		if (!Suggest)
		{
			if (!p->Win.FullScreen || p->Wait)
			{
				// check for overlapping top-level windows
				p->Overlap = 0;
				EnumWindows(EnumCheck,(LPARAM)p);
				if (p->Overlap)
					return; // overlapping window -> clipping still needed
			}
		}
		else
			p->ClipCountdown = 3; // refresh cycles
		p->Clipping = Suggest;
		p->Player->Set(p->Player,PLAYER_CLIPPING,&p->Clipping,sizeof(bool_t));

		if (HotKey)
			UpdateHotKey(p,!p->Clipping);
	}
}

static void RefreshButton(intface* p,int No,bool_t* State,int Id,int Off,int On)
{
	bool_t b;

	if (p->Player->Get(p->Player,No,&b,sizeof(b)) != ERR_NONE)
		b = -1;

	if (b != *State)
	{
		if (b == -1 || *State == -1)
			SendMessage(p->Win.WndTB,TB_ENABLEBUTTON,Id,MAKELPARAM(b!=-1,0));

		*State = b;

		if (b != -1)
		{
			SendMessage(p->Win.WndTB,TB_CHECKBUTTON,Id,MAKELPARAM(b,0));

			if (On>=0)
				SendMessage(p->Win.WndTB,TB_CHANGEBITMAP,Id,MAKELPARAM(p->Win.BitmapNo+(b?On:Off),0));
		}
	}
}

static void SetTrackThumb(HWND Wnd, bool_t State)
{
	int Style = GetWindowLong(Wnd,GWL_STYLE);

	if ((Style & TBS_NOTHUMB) != (State?0:TBS_NOTHUMB))
	{
		Style ^= TBS_NOTHUMB;
		SetWindowLong(Wnd,GWL_STYLE,Style);
	}
}

static void Refresh(intface* p)
{
	fraction f;
	tick_t t;

	if (p->Bench && p->Player->Get(p->Player,PLAYER_BENCHMARK,&t,sizeof(tick_t))==ERR_NONE)
	{
		p->Bench = 0;
		Toggle(p,PLAYER_FULLSCREEN,0);

		WaitEnd();
		SetTrackThumb(p->WndTrack,1);

		WinOpen((win*)NodeFind(BENCHRESULT_ID),&p->Win,1);
	}

	if (p->Initing>0)
	{
		if (--p->Initing==0)
			NodeSaveValue(IF_REG_INITING,NULL,0,TYPE_INT);
	}

	if (p->Focus && p->Clipping)
		if (p->ClipCountdown)
			--p->ClipCountdown;
		else
			UpdateClipping(p,0,1);

	if (!p->Win.FullScreen && !p->Bench)
	{
		if (!p->InSeek)
		{
			int TrackPos;
			if (p->Player->Get(p->Player,PLAYER_PERCENT,&f,sizeof(fraction)) == ERR_NONE)
				TrackPos = Scale(TRACKMAX,f.Num,f.Den);
			else
				TrackPos = -1;

			//DEBUG_MSG1("TRACKPOS %d",TrackPos);

			if (TrackPos != p->TrackPos)
			{
				if (TrackPos == -1)
					EnableWindow(p->WndTrack,0);
				if (p->TrackPos == -1)
					EnableWindow(p->WndTrack,1);
				p->TrackPos = TrackPos;
				SendMessage(p->WndTrack,TBM_SETPOS,1,TrackPos);
			}
		}

		RefreshButton(p,PLAYER_FULLSCREEN,&p->FullScreen,IF_OPTIONS_FULLSCREEN,-1,-1);
	}

	RefreshButton(p,PLAYER_PLAY,&p->Play,IF_PLAY,0,1);
	RefreshButton(p,PLAYER_FFWD,&p->FFwd,IF_FASTFORWARD,4,1);

#if defined(_WIN32_WCE)
	if (++p->IdleReset == (8000/REFRESH_CYCLE)) // 8sec (SHIdleTimerReset min. 10sec)
	{
		p->IdleReset = 0;
		if (p->Play>0 || p->FFwd>0)
			IdleTimerReset();
	}
#endif
}

static void EnableMenu(intface* p,int Id,bool_t State)
{
	EnableMenuItem(p->Win.Menu,Id,(State ? MF_ENABLED:MF_GRAYED)|MF_BYCOMMAND);
}

static void UpdateMenuBool(intface* p,int MenuId,int Param)
{
	bool_t b;
	p->Player->Get(p->Player,Param,&b,sizeof(b));
	CheckMenuItem(p->Win.Menu,MenuId,b ? MF_CHECKED:MF_UNCHECKED);
}

static void UpdateMenuInt(intface* p,int MenuId,int Param,int Value)
{
	int v;
	p->Player->Get(p->Player,Param,&v,sizeof(v));
	CheckMenuItem(p->Win.Menu,MenuId,v==Value ? MF_CHECKED:MF_UNCHECKED);
}

static void UpdateMenuFrac(intface* p,int MenuId,int Param,int Num,int Den)
{
	fraction f;
	p->Player->Get(p->Player,Param,&f,sizeof(f));
	CheckMenuItem(p->Win.Menu,MenuId,f.Num*Den==f.Den*Num ? MF_CHECKED:MF_UNCHECKED);
}

static void GetStreamName(intface* p,int No,tchar_t* Name,int Num,int DefName)
{
	tchar_t s[256];

	if (p->Player->CommentByName(p->Player,No,PlayerComment(COMMENT_LANGUAGE),s,sizeof(s)) ||
	    p->Player->CommentByName(p->Player,No,PlayerComment(COMMENT_TITLE),s,sizeof(s)))
	{
		tcsncpy(Name,s,20);
		Name[20]=0;
	}
	else
		stprintf(Name,T("%s %d"),String(INTERFACE_ID,DefName),Num);
}

static tick_t GetChapter(intface* p,int No)
{
	tchar_t Chapter[64];
	tchar_t s[64];
	int Hour,Min,Sec;

	stprintf(Chapter,T("CHAPTER%02d"),No);
	if (p->Player->CommentByName(p->Player,-1,Chapter,s,sizeof(s)) && 
		stscanf(s,T("%d:%d:%d"),&Hour,&Min,&Sec)==3)
		return (Hour*60+Min*60+Sec)*TICKSPERSEC;

	return -1;
}

static void UpdateMenu(intface* p)
{
	audio Audio;
	bool_t Stereo;
	node* Node;
	node* i;
	int No;
	bool_t Accel;
	tchar_t Chapter[64];
	tchar_t Name[256];
	node* Format;
	int ACurrent = -1;
	int VCurrent = -1;

	// remove old chapters

	if (EnableMenuItem(p->Win.Menu,IF_FILE_CHAPTERS_NONE,MF_BYCOMMAND|MF_GRAYED)==-1)
	{
		InsertMenu(p->Win.Menu,IF_CHAPTER+1,MF_BYCOMMAND|MF_STRING,
			IF_FILE_CHAPTERS_NONE, String(INTERFACE_ID,IF_FILE_CHAPTERS_NONE));

		for (No=1;No<100;++No)
			if (!DeleteMenu(p->Win.Menu,IF_CHAPTER+No,MF_BYCOMMAND))
				break;
	}

	// add new chapters

	for (No=1;No<100;++No)
	{
		if (GetChapter(p,No)<0)
			break;

		stprintf(Chapter,T("CHAPTER%02dNAME"),No);
		if (!p->Player->CommentByName(p->Player,-1,Chapter,Name,sizeof(Name)))
			stprintf(Name,T("%d"),No);

		InsertMenu(p->Win.Menu,IF_FILE_CHAPTERS_NONE,MF_BYCOMMAND|MF_STRING,
			IF_CHAPTER+No,Name);
	}

	if (No>1)
		DeleteMenu(p->Win.Menu,IF_FILE_CHAPTERS_NONE,MF_BYCOMMAND);
	else
		EnableMenuItem(p->Win.Menu,IF_FILE_CHAPTERS_NONE,MF_BYCOMMAND|MF_GRAYED);

	// remove old streams

	if (p->MenuVStream >= 0)
	{
		InsertMenu(p->Win.Menu,p->MenuVStream,MF_BYCOMMAND|MF_STRING,
			IF_OPTIONS_VIDEO_STREAM_NONE, String(INTERFACE_ID,IF_OPTIONS_VIDEO_STREAM_NONE));

		for (No=0;No<p->MenuStreams;++No)
			DeleteMenu(p->Win.Menu,IF_STREAM_VIDEO+No,MF_BYCOMMAND);
	}
	if (p->MenuAStream >= 0)
	{
		InsertMenu(p->Win.Menu,p->MenuAStream,MF_BYCOMMAND|MF_STRING,
			IF_OPTIONS_AUDIO_STREAM_NONE, String(INTERFACE_ID,IF_OPTIONS_AUDIO_STREAM_NONE));

		for (No=0;No<p->MenuStreams;++No)
			DeleteMenu(p->Win.Menu,IF_STREAM_AUDIO+No,MF_BYCOMMAND);
	}

	// add new streams

	p->MenuStreams = 0;
	p->MenuAStream = -1;
	p->MenuVStream = -1;

	if (p->Player->Get(p->Player,PLAYER_FORMAT,&Format,sizeof(Format)) == ERR_NONE && Format)
	{
		int VNo = 0;
		int ANo	= 0;
		pin Pin;
		datadef DataDef;

		for (No=0;Format->Get(Format,FORMAT_STREAM+No,&Pin,sizeof(Pin))==ERR_NONE;++No)
			if (NodeDataDef(Format,FORMAT_STREAM+No,&DataDef))
			{
				if (DataDef.Format1 == PACKET_VIDEO)
				{
					p->MenuVStream = IF_STREAM_VIDEO+No;
					GetStreamName(p,No,Name,++VNo,IF_OPTIONS_VIDEO);
					InsertMenu(p->Win.Menu,IF_OPTIONS_VIDEO_STREAM_NONE,MF_BYCOMMAND|MF_STRING,
						IF_STREAM_VIDEO+No,Name);
				}
				else
				if (DataDef.Format1 == PACKET_AUDIO)
				{
					p->MenuAStream = IF_STREAM_AUDIO+No;
					GetStreamName(p,No,Name,++ANo,IF_OPTIONS_AUDIO);
					InsertMenu(p->Win.Menu,IF_OPTIONS_AUDIO_STREAM_NONE,MF_BYCOMMAND|MF_STRING,
						IF_STREAM_AUDIO+No,Name);
				}
			}

		p->MenuStreams = No;
	}

	p->Player->Get(p->Player,PLAYER_VSTREAM,&VCurrent,sizeof(int));
	p->Player->Get(p->Player,PLAYER_ASTREAM,&ACurrent,sizeof(int));

	if (p->MenuVStream<0)
		EnableMenuItem(p->Win.Menu,IF_OPTIONS_VIDEO_STREAM_NONE,MF_BYCOMMAND|MF_GRAYED);
	else
	{
		CheckMenuItem(p->Win.Menu,IF_STREAM_VIDEO+VCurrent,MF_CHECKED);
		DeleteMenu(p->Win.Menu,IF_OPTIONS_VIDEO_STREAM_NONE,MF_BYCOMMAND);
	}

	if (p->MenuAStream<0)
		EnableMenuItem(p->Win.Menu,IF_OPTIONS_AUDIO_STREAM_NONE,MF_BYCOMMAND|MF_GRAYED);
	else
	{
		CheckMenuItem(p->Win.Menu,IF_STREAM_AUDIO+ACurrent,MF_CHECKED);
		DeleteMenu(p->Win.Menu,IF_OPTIONS_AUDIO_STREAM_NONE,MF_BYCOMMAND);
	}

	UpdateMenuFrac(p,IF_OPTIONS_ASPECT_SQUARE,PLAYER_ASPECT,1,1);
	UpdateMenuFrac(p,IF_OPTIONS_ASPECT_4_3_NTSC,PLAYER_ASPECT,10,11);
	UpdateMenuFrac(p,IF_OPTIONS_ASPECT_4_3_PAL,PLAYER_ASPECT,12,11);
	UpdateMenuFrac(p,IF_OPTIONS_ASPECT_16_9_NTSC,PLAYER_ASPECT,40,33);
	UpdateMenuFrac(p,IF_OPTIONS_ASPECT_16_9_PAL,PLAYER_ASPECT,16,11);

	UpdateMenuFrac(p,IF_OPTIONS_ZOOM_FIT_SCREEN,PLAYER_ZOOM,0,1);
	UpdateMenuFrac(p,IF_OPTIONS_ZOOM_FIT_WIDTH,PLAYER_ZOOM,-1,1);
	UpdateMenuFrac(p,IF_OPTIONS_ZOOM_FIT_HEIGHT,PLAYER_ZOOM,-2,1);
	UpdateMenuFrac(p,IF_OPTIONS_ZOOM_50,PLAYER_ZOOM,1,2);
	UpdateMenuFrac(p,IF_OPTIONS_ZOOM_100,PLAYER_ZOOM,1,1);
	UpdateMenuFrac(p,IF_OPTIONS_ZOOM_150,PLAYER_ZOOM,3,2);
	UpdateMenuFrac(p,IF_OPTIONS_ZOOM_200,PLAYER_ZOOM,2,1);
	UpdateMenuBool(p,IF_OPTIONS_VIDEO_ZOOM_SMOOTH,PLAYER_SMOOTH);

	UpdateMenuInt(p,IF_OPTIONS_AUDIO_QUALITY_LOW,PLAYER_AUDIO_QUALITY,0);
	UpdateMenuInt(p,IF_OPTIONS_AUDIO_QUALITY_MEDIUM,PLAYER_AUDIO_QUALITY,1);
	UpdateMenuInt(p,IF_OPTIONS_AUDIO_QUALITY_HIGH,PLAYER_AUDIO_QUALITY,2);

	UpdateMenuInt(p,IF_OPTIONS_ROTATE_0,PLAYER_FULL_DIR,0);
	UpdateMenuInt(p,IF_OPTIONS_ROTATE_90,PLAYER_FULL_DIR,DIR_SWAPXY | DIR_MIRRORLEFTRIGHT);
	UpdateMenuInt(p,IF_OPTIONS_ROTATE_270,PLAYER_FULL_DIR,DIR_SWAPXY | DIR_MIRRORUPDOWN);
	UpdateMenuInt(p,IF_OPTIONS_ROTATE_180,PLAYER_FULL_DIR,DIR_MIRRORUPDOWN | DIR_MIRRORLEFTRIGHT);

	UpdateMenuBool(p,IF_OPTIONS_VIDEO_UPDATE,PLAYER_VIDEO_UPDATE);

	UpdateMenuFrac(p,IF_OPTIONS_SPEED_10,PLAYER_PLAY_SPEED,1,10);
	UpdateMenuFrac(p,IF_OPTIONS_SPEED_25,PLAYER_PLAY_SPEED,1,4);
	UpdateMenuFrac(p,IF_OPTIONS_SPEED_50,PLAYER_PLAY_SPEED,1,2);
	UpdateMenuFrac(p,IF_OPTIONS_SPEED_100,PLAYER_PLAY_SPEED,1,1);
	UpdateMenuFrac(p,IF_OPTIONS_SPEED_150,PLAYER_PLAY_SPEED,3,2);
	UpdateMenuFrac(p,IF_OPTIONS_SPEED_200,PLAYER_PLAY_SPEED,2,1);
	UpdateMenuFrac(p,IF_OPTIONS_SPEED_300,PLAYER_PLAY_SPEED,3,1);
	UpdateMenuFrac(p,IF_OPTIONS_SPEED_BENCHMARK,PLAYER_PLAY_SPEED,0,1);

	UpdateMenuBool(p,IF_PLAY,PLAYER_PLAY);

	UpdateMenuBool(p,IF_OPTIONS_VIDEO_DITHER,PLAYER_DITHER);
	UpdateMenuBool(p,IF_OPTIONS_VIDEO_QUALITYCOLOR,PLAYER_QUALITYCOLOR);

	UpdateMenuInt(p,IF_OPTIONS_AUDIO_STEREO,PLAYER_STEREO,0);
	UpdateMenuInt(p,IF_OPTIONS_AUDIO_MONO_LEFT,PLAYER_STEREO,1);
	UpdateMenuInt(p,IF_OPTIONS_AUDIO_MONO_RIGHT,PLAYER_STEREO,2);
	UpdateMenuBool(p,IF_OPTIONS_AUDIO_SWAPSTEREO,PLAYER_AUDIO_SWAP);

	UpdateMenuBool(p,IF_OPTIONS_FULLSCREEN,PLAYER_FULLSCREEN);
	UpdateMenuBool(p,IF_OPTIONS_REPEAT,PLAYER_REPEAT);
	UpdateMenuBool(p,IF_OPTIONS_SHUFFLE,PLAYER_SHUFFLE);
	UpdateMenuBool(p,IF_OPTIONS_EQUALIZER,PLAYER_EQENABLE);

	p->Player->Get(p->Player,PLAYER_VIDEO_ACCEL,&Accel,sizeof(Accel));
	p->Player->Get(p->Player,PLAYER_VOUTPUT,&Node,sizeof(node*));

	CheckMenuItem(p->Win.Menu,IF_OPTIONS_VIDEO_TURNOFF,!Node ? MF_CHECKED:MF_UNCHECKED);

	for (No=0;(i = NodeEnum(VOUT_CLASS,No))!=NULL;++No)
	{
		CheckMenuItem(p->Win.Menu,IF_VIDEO+No,i==Node && !Accel ? MF_CHECKED:MF_UNCHECKED);
		CheckMenuItem(p->Win.Menu,IF_VIDEOACCEL+No,i==Node && Accel ? MF_CHECKED:MF_UNCHECKED);
	}

	p->Player->Get(p->Player,PLAYER_AOUTPUT,&Node,sizeof(node*));
	CheckMenuItem(p->Win.Menu,IF_OPTIONS_AUDIO_TURNOFF,!Node ? MF_CHECKED:MF_UNCHECKED);

	for (No=0;(i = NodeEnum(AOUT_CLASS,No))!=NULL;++No)
		CheckMenuItem(p->Win.Menu,IF_AUDIO+No,i==Node ? MF_CHECKED:MF_UNCHECKED);

	Stereo = !Node || Node->Get(Node,AOUT_INPUT|PACKET_FORMAT,&Audio,sizeof(Audio))!=ERR_NONE || Audio.Channels!=1; //channels can be zero when audio not used

	EnableMenu(p,IF_OPTIONS_AUDIO_STEREO,Stereo);
	EnableMenu(p,IF_OPTIONS_AUDIO_MONO_LEFT,Stereo);
	EnableMenu(p,IF_OPTIONS_AUDIO_MONO_RIGHT,Stereo);
	EnableMenu(p,IF_OPTIONS_AUDIO_SWAPSTEREO,Stereo);
}

static int Delta(intface* p,node* Node,int Param,int d,int Min,int Max)
{
	int State;
	if (Node && Node->Get(Node,Param,&State,sizeof(State))==ERR_NONE)
	{
		State += d;
		if (State < Min)
			State = Min;
		if (State > Max)
			State = Max;

		Node->Set(Node,Param,&State,sizeof(State));

		if (Param == PLAYER_VOLUME)
		{
			if (State > Min) Toggle(p,PLAYER_MUTE,0);
			UpdateVolume(p);
		}
	}
	return State;
}

static void ShowVol(intface* p,bool_t Show)
{
	if (p->WndVolBack && !p->Win.ToolBarHeight)
	{
		if (Show)
			SetWindowPos(p->WndVolBack,HWND_TOPMOST,0,0,0,0,SWP_NOSIZE|SWP_NOMOVE|SWP_NOACTIVATE|SWP_SHOWWINDOW);
		else
			ShowWindow(p->WndVolBack,SW_HIDE);
	}
}

static bool_t Toggle(intface* p,int Param,int Force)
{
	bool_t State = 0;
	if (p->Player->Get(p->Player,Param,&State,sizeof(State))==ERR_NONE)
	{
		if (State==Force)
			return State;
		State = !State;
		if (Param == PLAYER_FULLSCREEN)
		{
			node* VOutput;
			bool_t Primary;
			video Format;

			if (p->Player->Get(p->Player,PLAYER_VOUTPUT,&VOutput,sizeof(VOutput))!=ERR_NONE ||
				!VOutput ||
				VOutput->Get(VOutput,VOUT_PRIMARY,&Primary,sizeof(Primary))!=ERR_NONE ||
				!Primary ||
				VOutput->Get(VOutput,VOUT_INPUT|PACKET_FORMAT,&Format,sizeof(Format))!=ERR_NONE ||
				!Format.Width)
				State = 0;

			ShowWindow(p->WndTrack,State ? SW_HIDE:SW_SHOWNA);
			WinSetFullScreen(&p->Win,State);
			ShowVol(p,!State); // called after WinSetFullScreen so TOPMOST is over taskbar
			Refresh(p);
		}
		p->Player->Set(p->Player,Param,&State,sizeof(State));

		Refresh(p);
	}
	return State;
}

void OpenSingle( intface* p, const tchar_t* URL)
{
	int n = 1;
	p->Player->Set(p->Player,PLAYER_LIST_COUNT,&n,sizeof(n));

	PlayerAdd(p->Player,0,URL);

	n = 0;
	p->Player->Set(p->Player,PLAYER_LIST_CURRENT,&n,sizeof(n));

	if (!IsWindowEnabled(p->Win.Wnd))
		Toggle(p,PLAYER_PLAY,0);
}

void SetTrackPos(intface* p,int x)
{
	RECT r,rt;
	int Adj;
	SendMessage(p->WndTrack,TBM_GETCHANNELRECT,0,(LONG)&r);
	SendMessage(p->WndTrack,TBM_GETTHUMBRECT,0,(LONG)&rt);
	Adj = (rt.right-rt.left) >> 1;
	r.left += Adj;
	r.right -= (rt.right-rt.left)-Adj;

	if (x > 32768) x -= 65536;
	if (r.right <= r.left) r.right = r.left+1;
	if (x < r.left) x = r.left;
	if (x > r.right) x = r.right;
	x = Scale(TRACKMAX,x-r.left,r.right-r.left);
	SendMessage(p->WndTrack,TBM_SETPOS,1,x);

	p->ForcePos = -1;
	StartSeek(p);
}

LRESULT CALLBACK TrackProc(HWND WndTrack, UINT Msg, WPARAM wParam, LPARAM lParam)
{
	intface* p = (intface*) WinGetObject(WndTrack);
	if (Msg == WM_LBUTTONDOWN)
	{
		if (!p->InSeek)
		{
			bool_t b = 0;
			p->Player->Get(p->Player,PLAYER_PLAY,&p->PlayAfterSeek,sizeof(bool_t));
			p->Player->Set(p->Player,PLAYER_PLAY,&b,sizeof(bool_t));
			p->LastPos = -1;
		}
		p->InSeek = 1;
		p->TrackPos = -2;

		SetTrackPos(p,LOWORD(lParam));
		SetCapture(WndTrack);
		return 0;
	}
	if (Msg == WM_MOUSEMOVE)
	{
		if (p->InSeek && (wParam & MK_LBUTTON))
			SetTrackPos(p,LOWORD(lParam));
		return 0;
	}
	if (Msg == WM_LBUTTONUP)
	{
		if (p->InSeek)
		{
			p->InSeek = 2;
			SetTrackPos(p,LOWORD(lParam));
			ReleaseCapture();
		}
		return 0;
	}
	return CallWindowProc(p->DefTrackProc,WndTrack,Msg,wParam,lParam);
}

void SetVolPos(intface* p,int x)
{
	RECT r;
	int Adj;
	SendMessage(p->WndVol,TBM_GETCHANNELRECT,0,(LONG)&r);
	Adj = Scale(VOLTHUMB,WinPixelSX(),72*4);
	r.left += Adj;
	r.right -= Adj;

	if (x > 32768) x -= 65536;
	if (r.right <= r.left) r.right = r.left+1;
	if (x < r.left) x = r.left;
	if (x > r.right) x = r.right;
	x = (100*(x-r.left))/(r.right-r.left);
	SendMessage(p->WndVol,TBM_SETPOS,1,x);
	p->Player->Set(p->Player,PLAYER_VOLUME,&x,sizeof(x));
}

LRESULT CALLBACK VolBackProc(HWND Wnd, UINT Msg, WPARAM wParam, LPARAM lParam)
{
	//DEBUG_MSG1(T("MSG %04x"),Msg);

	switch (Msg)
	{
#if defined(_WIN32_WCE)
	case WM_CTLCOLORSTATIC:
		return (LRESULT) GetSysColorBrush(COLOR_BTNFACE);
#endif
	}
	return DefWindowProc(Wnd,Msg,wParam,lParam);
}

LRESULT CALLBACK VolProc(HWND WndVol, UINT Msg, WPARAM wParam, LPARAM lParam)
{
	intface* p = (intface*) WinGetObject(WndVol);

	switch (Msg)
	{
	case WM_MOUSEMOVE:
		if ((wParam & MK_LBUTTON)==0)
			break;

	case WM_LBUTTONDOWN:
	case WM_LBUTTONDBLCLK:
		if (!p->InVol)
		{
			p->InVol = 1;
			SetForegroundWindow(p->Win.Wnd);
			SetCapture(WndVol);
		}

		SetVolPos(p,LOWORD(lParam));
		return 0;

	case WM_LBUTTONUP:
		if (!p->InVol)
			SetForegroundWindow(p->Win.Wnd);
		else
		{
			p->InVol = 0;
			ReleaseCapture();
		}
		SetVolPos(p,LOWORD(lParam));
		return 0;
	
		break;
	}

	return CallWindowProc(p->DefVolProc,WndVol,Msg,wParam,lParam);
}

void ClearStats(intface* p)
{
	int Zero=0;
	int No;
	pin Pin;
	node* Format;

	if (p->Player->Get(p->Player,PLAYER_FORMAT,&Format,sizeof(Format))==ERR_NONE && Format)
		for (No=0;Format->Get(Format,FORMAT_STREAM+No,&Pin,sizeof(Pin))==ERR_NONE;++No)
			if (Pin.Node && NodeIsClass(Pin.Node,FLOW_CLASS))
			{
				Pin.Node->Set(Pin.Node,FLOW_TOTAL,&Zero,sizeof(Zero));
				Pin.Node->Set(Pin.Node,FLOW_DROPPED,&Zero,sizeof(Zero));
			}
}

void CreateDeviceMenu(intface* p)
{
	int No;
	idct* IDCT;
	node* Node;
	HMENU Menu = p->Win.Menu;
	int Id;

	for (No=0;(Node = NodeEnum(VOUT_CLASS,No))!=NULL;++No)
		if (Node->Get(Node,NODE_ID,&Id,sizeof(Id))==ERR_NONE)
		{
			if (Node->Get(Node,VOUT_ACCEL_IDCT,&IDCT,sizeof(IDCT)) == ERR_NONE && IDCT)
			{
				tchar_t s[256];
				stprintf(s,String(INTERFACE_ID,IF_OPTIONS_VIDEO_ACCEL),String(Id,0));
				InsertMenu(Menu,IF_OPTIONS_VIDEO_TURNOFF,MF_BYCOMMAND|MF_STRING,IF_VIDEOACCEL+No,s);
			}
			InsertMenu(Menu,IF_OPTIONS_VIDEO_TURNOFF,MF_BYCOMMAND|MF_STRING,IF_VIDEO+No,String(Id,0));
		}

	for (No=0;(Node = NodeEnum(AOUT_CLASS,No))!=NULL;++No)
		if (Node->Get(Node,NODE_ID,&Id,sizeof(Id))==ERR_NONE)
			InsertMenu(Menu,IF_OPTIONS_AUDIO_TURNOFF,MF_BYCOMMAND|MF_STRING,IF_AUDIO+No,String(Id,0));
}

static menudef MenuDef[] =
{
	{ 0,INTERFACE_ID, IF_FILE },
	{ 1,INTERFACE_ID, IF_FILE_OPENFILE },
	{ 1,INTERFACE_ID, IF_FILE_PLAYLIST },
	{ 1,-1,-1 },
	{ 1,INTERFACE_ID, IF_PLAY },
	{ 1,INTERFACE_ID, IF_PREV },
	{ 1,INTERFACE_ID, IF_NEXT },
	{ 1,-1,-1 },
	{ 1,INTERFACE_ID, IF_FILE_CHAPTERS },
	{ 2,INTERFACE_ID, IF_FILE_CHAPTERS_NONE },
	{ 1,INTERFACE_ID, IF_FILE_MEDIAINFO },
	{ 1,INTERFACE_ID, IF_FILE_BENCHMARK },
	{ 1,INTERFACE_ID, IF_FILE_ABOUT },
	{ 1,-1,-1 },
	{ 1,INTERFACE_ID, IF_FILE_EXIT },
	{ 0,INTERFACE_ID, IF_OPTIONS },
	{ 1,INTERFACE_ID, IF_OPTIONS_SPEED },
	{ 2,INTERFACE_ID, IF_OPTIONS_SPEED_10 },
	{ 2,INTERFACE_ID, IF_OPTIONS_SPEED_25 },
	{ 2,INTERFACE_ID, IF_OPTIONS_SPEED_50 }, 
	{ 2,INTERFACE_ID, IF_OPTIONS_SPEED_100 },
	{ 2,INTERFACE_ID, IF_OPTIONS_SPEED_150 },
	{ 2,INTERFACE_ID, IF_OPTIONS_SPEED_200 },
	{ 2,INTERFACE_ID, IF_OPTIONS_SPEED_300 },
	{ 1,INTERFACE_ID, IF_OPTIONS_VIEW, MENU_SMALL },	
	{ 2,INTERFACE_ID, IF_OPTIONS_ZOOM },
	{ 3,INTERFACE_ID, IF_OPTIONS_ZOOM_FIT_SCREEN },
	{ 3,INTERFACE_ID, IF_OPTIONS_ZOOM_FIT_WIDTH },
	{ 3,INTERFACE_ID, IF_OPTIONS_ZOOM_FIT_HEIGHT },
	{ 3,INTERFACE_ID, IF_OPTIONS_ZOOM_50 },  
	{ 3,INTERFACE_ID, IF_OPTIONS_ZOOM_100 },
	{ 3,INTERFACE_ID, IF_OPTIONS_ZOOM_150 },
	{ 3,INTERFACE_ID, IF_OPTIONS_ZOOM_200 },
	{ 2,INTERFACE_ID, IF_OPTIONS_ASPECT },
	{ 3,INTERFACE_ID, IF_OPTIONS_ASPECT_SQUARE },
	{ 3,INTERFACE_ID, IF_OPTIONS_ASPECT_4_3_NTSC },
	{ 3,INTERFACE_ID, IF_OPTIONS_ASPECT_4_3_PAL },
	{ 3,INTERFACE_ID, IF_OPTIONS_ASPECT_16_9_NTSC },
	{ 3,INTERFACE_ID, IF_OPTIONS_ASPECT_16_9_PAL },
	{ 2,INTERFACE_ID, IF_OPTIONS_ORIENTATION },
	{ 3,INTERFACE_ID, IF_OPTIONS_ORIENTATION_FULLSCREEN },
	{ 3,INTERFACE_ID, IF_OPTIONS_ROTATE_0 },
	{ 3,INTERFACE_ID, IF_OPTIONS_ROTATE_90 },
	{ 3,INTERFACE_ID, IF_OPTIONS_ROTATE_270 },
	{ 3,INTERFACE_ID, IF_OPTIONS_ROTATE_180 },
	{ 1,INTERFACE_ID, IF_OPTIONS_VIDEO },
#if !defined(SH3) && !defined(MIPS)
	{ 2,INTERFACE_ID, IF_OPTIONS_VIDEO_ZOOM_SMOOTH },
#else
	{ 2,INTERFACE_ID, IF_OPTIONS_VIDEO_QUALITYCOLOR },
#endif
	{ 2,INTERFACE_ID, IF_OPTIONS_VIDEO_DITHER },
	{ 2,INTERFACE_ID, IF_OPTIONS_VIDEO_UPDATE },
	{ 2,-1,-1 },
	{ 2,INTERFACE_ID, IF_OPTIONS_VIDEO_DEVICE, MENU_SMALL|MENU_GRAYED },
	{ 3,INTERFACE_ID, IF_OPTIONS_VIDEO_TURNOFF },
	{ 2,-1,-1 },
	{ 2,INTERFACE_ID, IF_OPTIONS_VIDEO_STREAM, MENU_SMALL|MENU_GRAYED },
	{ 3,INTERFACE_ID, IF_OPTIONS_VIDEO_STREAM_NONE },

	{ 1,INTERFACE_ID, IF_OPTIONS_AUDIO },
	{ 2,INTERFACE_ID, IF_OPTIONS_AUDIO_SWAPSTEREO },
	{ 2,-1,-1 },
	{ 2,INTERFACE_ID, IF_OPTIONS_AUDIO_CHANNELS, MENU_SMALL },
	{ 3,INTERFACE_ID, IF_OPTIONS_AUDIO_STEREO },
	{ 3,INTERFACE_ID, IF_OPTIONS_AUDIO_MONO_LEFT },
	{ 3,INTERFACE_ID, IF_OPTIONS_AUDIO_MONO_RIGHT },
	{ 2,-1,-1 },
	{ 2,INTERFACE_ID, IF_OPTIONS_AUDIO_QUALITY, MENU_SMALL|MENU_GRAYED },
	{ 3,INTERFACE_ID, IF_OPTIONS_AUDIO_QUALITY_LOW },
	{ 3,INTERFACE_ID, IF_OPTIONS_AUDIO_QUALITY_MEDIUM },
	{ 3,INTERFACE_ID, IF_OPTIONS_AUDIO_QUALITY_HIGH }, 
	{ 2,-1,-1 },
	{ 2,INTERFACE_ID, IF_OPTIONS_AUDIO_DEVICE, MENU_SMALL|MENU_GRAYED },
	{ 3,INTERFACE_ID, IF_OPTIONS_AUDIO_TURNOFF },
	{ 2,-1,-1 },
	{ 2,INTERFACE_ID, IF_OPTIONS_AUDIO_STREAM, MENU_SMALL|MENU_GRAYED },
	{ 3,INTERFACE_ID, IF_OPTIONS_AUDIO_STREAM_NONE },

	{ 1,INTERFACE_ID, IF_OPTIONS_REPEAT },
	{ 1,INTERFACE_ID, IF_OPTIONS_SHUFFLE },
	{ 1,INTERFACE_ID, IF_OPTIONS_EQUALIZER },
	{ 1,INTERFACE_ID, IF_OPTIONS_FULLSCREEN },
	{ 1,INTERFACE_ID, IF_OPTIONS_SETTINGS },

	MENUDEF_END
};

void CreateButtons(intface* p)
{
	if (QueryPlatform(PLATFORM_TYPENO) != TYPE_SMARTPHONE)
	{
		// remove from menu
		HMENU Menu = p->Win.Menu;
		DeleteMenu(Menu,IF_PLAY,MF_BYCOMMAND);
		DeleteMenu(Menu,IF_OPTIONS_FULLSCREEN,MF_BYCOMMAND);
		DeleteMenu(Menu,IF_OPTIONS_MUTE,MF_BYCOMMAND);

		WinBitmap(&p->Win,IDB_TOOLBAR16,IDB_TOOLBAR32,10);
		WinAddButton(&p->Win,-1,-1,NULL,0);
		WinAddButton(&p->Win,IF_PLAY,0,NULL,1);
		WinAddButton(&p->Win,IF_FASTFORWARD,4,NULL,1);
		WinAddButton(&p->Win,IF_STOP,5,NULL,0);
		WinAddButton(&p->Win,-1,-1,NULL,0);
		WinAddButton(&p->Win,IF_OPTIONS_FULLSCREEN,6,NULL,1);
		WinAddButton(&p->Win,IF_OPTIONS_MUTE,7,NULL,1);
	}
}

static void UpdateVolume(intface* p)
{
	int x;
	if (p->WndVol)
	{
		bool_t Enable = p->Player->Get(p->Player,PLAYER_VOLUME,&x,sizeof(x))==ERR_NONE;
		EnableWindow(p->WndVol,Enable);
		SetTrackThumb(p->WndVol,Enable);
		if (Enable)
			SendMessage(p->WndVol,TBM_SETPOS,1,x);
	}
	RefreshButton(p,PLAYER_MUTE,&p->Mute,IF_OPTIONS_MUTE,7,8);
}

static void CreateVolumeTrack(intface* p)
{
	if (QueryPlatform(PLATFORM_TYPENO) != TYPE_SMARTPHONE)
	{
		WNDCLASS WinClass;

		memset(&WinClass,0,sizeof(WinClass));
		WinClass.style			= CS_HREDRAW | CS_VREDRAW;
		WinClass.lpfnWndProc	= (WNDPROC) VolBackProc;
		WinClass.cbClsExtra		= 0;
		WinClass.cbWndExtra		= 0;
		WinClass.hInstance		= p->Win.Module;
		WinClass.hCursor		= LoadCursor(NULL,IDC_ARROW);
		WinClass.hbrBackground	= GetSysColorBrush(COLOR_BTNFACE);
		WinClass.lpszMenuName	= 0;
		WinClass.lpszClassName	= T("VolumeBack");
		RegisterClass(&WinClass);

		p->WndVolBack = CreateWindowEx( 
			WS_EX_TOPMOST,                    
			WinClass.lpszClassName,       
			NodeProgramId(),   
			(p->Win.ToolBarHeight ? WS_CHILD:WS_POPUP) | WS_VISIBLE | 
			TBS_HORZ|TBS_BOTH|TBS_NOTICKS|TBS_FIXEDLENGTH, 
			200, 0,    
			40, 20,   
			(p->Win.ToolBarHeight ? p->Win.WndTB:p->Win.Wnd),	
			NULL,						
			p->Win.Module,                     
			NULL                      
			); 

		p->WndVol = CreateWindowEx( 
			0,                    
			TRACKBAR_CLASS,       
			NULL,   
			WS_CHILD | WS_VISIBLE | 
			TBS_HORZ|TBS_BOTH|TBS_NOTICKS|TBS_FIXEDLENGTH, 
			0, 0,    
			40, 20,   
			p->WndVolBack,
			NULL,						
			p->Win.Module,                     
			NULL                      
			); 

		p->DefVolProc = (WNDPROC)GetWindowLong(p->WndVol,GWL_WNDPROC);
		SetWindowLong(p->WndVol,GWL_WNDPROC,(LONG)VolProc);
		WinSetObject(p->WndVol,&p->Win);

		SendMessage(p->WndVol, TBM_SETRANGE, FALSE,MAKELONG(0,100));
		SendMessage(p->WndVol, TBM_SETPAGESIZE,0,10);
		SendMessage(p->WndVol, TBM_SETLINESIZE,0,10);

		ResizeVolume(p);
		UpdateVolume(p);
	}
}

static void CreateTrack(intface* p)
{
	p->WndTrack = CreateWindowEx( 
		0,                    
		TRACKBAR_CLASS,       
		T("Time"),   
		WS_CHILD | WS_VISIBLE | 
		TBS_HORZ|TBS_BOTH|TBS_NOTICKS|TBS_FIXEDLENGTH, 
		0, 0,    
		200, 20,   
		p->Win.Wnd,	
		NULL,						
		p->Win.Module,                     
		NULL                      
		); 

	p->DefTrackProc = (WNDPROC)GetWindowLong(p->WndTrack,GWL_WNDPROC);
	SetWindowLong(p->WndTrack,GWL_WNDPROC,(LONG)TrackProc);
	WinSetObject(p->WndTrack,&p->Win);

	SendMessage(p->WndTrack, TBM_SETRANGE, FALSE,MAKELONG(0,TRACKMAX));
	SendMessage(p->WndTrack, TBM_SETPAGESIZE,0,TRACKMAX/20);
	SendMessage(p->WndTrack, TBM_SETLINESIZE,0,TRACKMAX/100);
}

static void BeforeExit(intface* p)
{
	HWND Wnd = NULL;
	if (p->Player->Get(p->Player,PLAYER_WND,&Wnd,sizeof(Wnd))==ERR_NONE && Wnd)
	{
		bool_t b;
		if (p->Player->Get(p->Player,PLAYER_KEEPLIST,&b,sizeof(b))==ERR_NONE && !b)
		{
			int i = 0;
			p->Player->Set(p->Player,PLAYER_LIST_COUNT,&i,sizeof(i));
		}

		NodeSaveValue(IF_REG_INITING,NULL,0,TYPE_INT);
		NodeSaveSetup((node*)p->Player);
		UpdateHotKey(p,0);
		if (p->WndVolBack)
		{
			DestroyWindow(p->WndVolBack);
			p->WndVolBack = NULL;
		}

		Wnd = NULL;
		p->Player->Set(p->Player,PLAYER_WND,&Wnd,sizeof(Wnd));
	}
}

static void ShowError(intface* p, const tchar_t* s)
{
	win* Win;

	if (p->Player)
	{
		p->Player->Stop(p->Player); // full stop (no filling buffers afterwards)
		Toggle(p,PLAYER_FULLSCREEN,0);
	}
	UpdateClipping(p,1,1); // force clipping mode
	
	// find application foreground window
	Win = &p->Win;
	while (Win->Child) Win = Win->Child;

	MessageBox(Win->Wnd,s,String(PLATFORM_ID,PLATFORM_ERROR),MB_OK|MB_SETFOREGROUND|MB_TOPMOST);
}

static bool_t Proc(intface* p, int Msg, WPARAM wParam, LPARAM lParam, int* Result)
{
	bool_t b;
	int Pos;
	int i;
	tick_t t,t2;
	fraction f;
	PAINTSTRUCT Paint;
	tchar_t Path[MAXPATH];
	void* Null = NULL;
	NMHDR* Notify;
	node* Node;

	switch (Msg) 
	{
	case WM_CANCELMODE:
		if (p->InVol)
		{
			p->InVol = 0;
			ReleaseCapture();
		}

		if (p->InSeek)
		{
			p->InSeek = 2;
			StartSeek(p);
			ReleaseCapture();
		}
		// no break
#if !defined(_WIN32_WCE)
	case WM_INITMENU:
#endif
		UpdateClipping(p,1,1); // force clipping mode
		break;

	case WM_PAINT:
		BeginPaint(p->Win.Wnd,&Paint);
		p->Player->Paint(p->Player,Paint.hdc,p->Offset.x,p->Offset.y);
		EndPaint(p->Win.Wnd,&Paint);
		break;

#if defined(_WIN32_WCE)
	case WM_CTLCOLORSTATIC:
		*Result = (LRESULT) GetSysColorBrush(COLOR_BTNFACE);
		return 1;
#endif

	case WM_HSCROLL:
	case WM_VSCROLL:
		if (lParam == (LONG)p->WndTrack)
		{
			if (!p->InSeek)
			{
				b = 0;
				p->Player->Get(p->Player,PLAYER_PLAY,&p->PlayAfterSeek,sizeof(bool_t));
				p->Player->Set(p->Player,PLAYER_PLAY,&b,sizeof(bool_t));
				p->LastPos = -1;
			}

			if (LOWORD(wParam)!=TB_ENDTRACK && LOWORD(wParam)!=TB_THUMBPOSITION)
				p->InSeek = 1;
			else
				p->InSeek = 2;

			p->ForcePos = -1;
			StartSeek(p);
			return 0;
		}
		break;

	case WM_SYSCHAR:
		if (wParam == 13) //Alt+Enter
			PostMessage(p->Win.Wnd,WM_COMMAND,IF_OPTIONS_FULLSCREEN,0);
		break;

	case WM_TIMER:
		if (wParam == TIMER_REFRESH)
		{
			Refresh(p);
		}
		if (wParam == TIMER_SEEK || wParam == TIMER_SEEKFORCE)
		{
			KillTimer(p->Win.Wnd,TIMER_SEEK);
			KillTimer(p->Win.Wnd,TIMER_SEEKFORCE);
			p->InSeekForce = 0;

			//DEBUG_MSG1("SEEKTIMER %d",p->InSeek);
			Seek(p);
		}
		break;

	case WM_NOTIFY:
		Notify = (NMHDR*)lParam;
		break;

	case WM_INITMENUPOPUP:
		UpdateClipping(p,1,1); // force clipping mode
		UpdateMenu(p);
		break;

	case WM_HOTKEY:
		PostMessage(p->Win.Wnd,WM_COMMAND,wParam,0);
		break;

	case WM_LBUTTONDOWN:
		DEBUG_MSG(T("LBUTTON"));
		if ((wParam & MK_LBUTTON) && p->Focus)
		{
			SetDisplayPower(1);
			PostMessage(p->Win.Wnd,WM_COMMAND,IF_PLAY_FULLSCREEN,0);
		}
		SetForegroundWindow(p->Win.Wnd); //sometimes wince suck (triple click in opendialog)
		*Result = 0;
		return 1;

	case WM_MOVE:
		Resize(p);
		break;

	case WM_SIZE:
		if (wParam == SIZE_MAXIMIZED)
		{
			ShowWindow(p->Win.Wnd,SW_SHOWNORMAL);
			Toggle(p,PLAYER_FULLSCREEN,1);
		}
		else
			Resize(p);
		break;

	case WM_SETFOCUS:
		DEBUG_MSG(T("SETFOCUS"));
		p->Focus = 1;
		WinShowSip(0);
		if (p->Wait)
		{
			WaitBegin();
			UpdateHotKey(p,1);
			// clipping still needed
		}
		else
			UpdateClipping(p,0,1); // turn of clipping when no overlapping windows
		break;

	case WM_KILLFOCUS:
		DEBUG_MSG(T("KILLFOCUS"));
		p->Focus = 0;
		UpdateClipping(p,1,1); // force clipping mode
		if (p->Wait)
			WaitEnd();
		break;

	case WM_ACTIVATE:
		if ((HWND)lParam != p->WndVolBack)
		{
			if (LOWORD(wParam)==WA_INACTIVE) 
			{
//#ifdef NDEBUG
				Toggle(p,PLAYER_BACKGROUND,1);
				Toggle(p,PLAYER_FULLSCREEN,0);
//#endif
				ShowVol(p,0);

				p->InSeek = 0;
				p->InSeekForce = 0;
				KillTimer(p->Win.Wnd,TIMER_SEEK);
				KillTimer(p->Win.Wnd,TIMER_SEEKFORCE);

				SetDisplayPower(1);
			}
			else
			{
				SetTimer(p->Win.Wnd,TIMER_REFRESH,REFRESH_CYCLE,NULL);
				ShowVol(p,1);
				UpdateVolume(p);
				Toggle(p,PLAYER_BACKGROUND,0);
			}
		}
		break;

	case WM_DESTROY:
		BeforeExit(p);
		PostQuitMessage(0);

		NodeErrorPin.Node = NULL;
		break;

	case MSG_ERROR:
		ShowError(p,(const tchar_t*)lParam);
		Free((tchar_t*)lParam);
		return 1;

	case MSG_WAIT:
		if (p->Wait != (bool_t)wParam)
		{
			p->Wait = wParam;
			if (p->Wait)
			{
				WaitBegin();
				UpdateClipping(p,1,0);
			}
			else
				WaitEnd();
		}
		return 1;

	case MSG_OPEN:
		if (NodeLoadValue(IF_REG_OPEN,Path,sizeof(Path),NULL))
		{
			OpenSingle(p,Path);
			NodeSaveValue(IF_REG_OPEN,NULL,0,TYPE_STRING);
		}
		return 1;

	case MSG_INIT:
		if (NodeLoadValue(IF_REG_INITING,&i,sizeof(i),NULL))
		{
			i = 0; // last time crashed -> clear stored playlist
			p->Player->Set(p->Player,PLAYER_LIST_COUNT,&i,sizeof(i));
		}

		if (NodeCmdLine()[0])
		{
			i = 0; // clear stored playlist
			p->Player->Set(p->Player,PLAYER_LIST_COUNT,&i,sizeof(i));
		}

#ifdef NDEBUG
		p->Initing = 10;
		i = 1;
		NodeSaveValue(IF_REG_INITING,&i,sizeof(int),TYPE_INT);
#endif

		p->Player->Set(p->Player,PLAYER_WND,&p->Win.Wnd,sizeof(HWND));

		if (NodeCmdLine()[0])
			OpenSingle(p,NodeCmdLine());

		if (String(LANG_ID,LANG_DEFAULT)[0]==0 &&
			String(LANG_ID,StringLang)[0]==0)
			ShowMessage(NULL,T("Language files (*.txt) are missing!"));

		return 1;

	case WM_CREATE:

		p->ThreadId = GetCurrentThreadId();
		p->Player = (player*)NodeFind(PLAYER_ID);
		p->Color = NodeFind(COLOR_ID);
		p->Focus = 0;
		p->Clipping = 0;
		p->LastPos = -1;
		p->ForcePos = -1;

		NodeErrorPin.No = INTERFACE_ERROR;
		NodeErrorPin.Node = (node*)p;

		EnableMenu(p,IF_OPTIONS_VIDEO_STREAM,0);
		EnableMenu(p,IF_OPTIONS_VIDEO_STREAM,0);
		EnableMenu(p,IF_OPTIONS_AUDIO_STREAM,0);
		EnableMenu(p,IF_OPTIONS_VIDEO_DEVICE,0);
		EnableMenu(p,IF_OPTIONS_AUDIO_DEVICE,0);
		EnableMenu(p,IF_OPTIONS_ORIENTATION_FULLSCREEN,0);
		EnableMenu(p,IF_OPTIONS_AUDIO_QUALITY,0);
		
		CreateButtons(p);
		CreateTrack(p);
		CreateVolumeTrack(p);
		CreateDeviceMenu(p);

		SetTimer(p->Win.Wnd,TIMER_REFRESH,REFRESH_CYCLE,NULL);
		Resize(p);

		// first finish window creation and postpone loading playlist and etc... 
		PostMessage(p->Win.Wnd,MSG_INIT,0,0);

		UpdateHotKey(p,1);
		break;

	case WM_COMMAND:

		switch (LOWORD(wParam))
		{
		case IF_OPTIONS_VOLUME_UP:
			Delta(p,(node*)p->Player,PLAYER_VOLUME,10,0,100);
			break;
		case IF_OPTIONS_VOLUME_DOWN:
			Delta(p,(node*)p->Player,PLAYER_VOLUME,-10,0,100);
			break;
		case IF_OPTIONS_BRIGHTNESS_UP:
			Delta(p,p->Color,COLOR_BRIGHTNESS,6,-128,1278);
			break;
		case IF_OPTIONS_BRIGHTNESS_DOWN:
			Delta(p,p->Color,COLOR_BRIGHTNESS,-6,-128,127);
			break;
		case IF_OPTIONS_CONTRAST_UP:
			Delta(p,p->Color,COLOR_CONTRAST,6,-128,127);
			break;
		case IF_OPTIONS_CONTRAST_DOWN:
			Delta(p,p->Color,COLOR_CONTRAST,-6,-128,127);
			break;
		case IF_OPTIONS_SATURATION_UP:
			Delta(p,p->Color,COLOR_SATURATION,6,-128,127);
			break;
		case IF_OPTIONS_SATURATION_DOWN:
			Delta(p,p->Color,COLOR_SATURATION,-6,-128,127);
			break;

		case IF_OPTIONS_SCREEN:
			SetDisplayPower(!GetDisplayPower());
			break;

		case IF_PLAY_FULLSCREEN:
			i = -1;
			if (p->Player->Get(p->Player,PLAYER_FULLSCREEN,&b,sizeof(b))==ERR_NONE && b)
				i = 0;
			b = Toggle(p,PLAYER_PLAY,i);
			Toggle(p,PLAYER_FULLSCREEN,b);
			break;

		case IF_MOVE_FFWD:
		case IF_MOVE_BACK:

			if (!p->InSeek)
			{
				bool_t b = 0;
				p->Player->Get(p->Player,PLAYER_PLAY,&p->PlayAfterSeek,sizeof(bool_t));
				p->Player->Set(p->Player,PLAYER_PLAY,&b,sizeof(bool_t));
				p->LastPos = -1;
				p->InSeek = 2;
			}

			t = p->ForcePos;
			if ((t!=-1 || p->Player->Get(p->Player,PLAYER_POSITION,&t,sizeof(tick_t))==ERR_NONE) &&
				p->Player->Get(p->Player,PLAYER_MOVESTEP,&t2,sizeof(tick_t))==ERR_NONE)
			{
				if (LOWORD(wParam) == IF_MOVE_BACK)
					t -= t2;
				else
					t += t2;
				p->ForcePos = t;

				if (p->Player->Get(p->Player,PLAYER_DURATION,&t2,sizeof(tick_t))==ERR_NONE)
				{
					Pos = Scale(TRACKMAX,t,t2);
					SendMessage(p->WndTrack,TBM_SETPOS,1,Pos);
				}
				StartSeek(p);
			}
			else
			{
				if (p->Player->Get(p->Player,PLAYER_PERCENT,&f,sizeof(fraction)) == ERR_NONE)
				{
					Pos = Scale(TRACKMAX,f.Num,f.Den);
					if (LOWORD(wParam) == IF_MOVE_BACK)
						Pos -= TRACKMAX/256;
					else
						Pos += TRACKMAX/256;
					SendMessage(p->WndTrack,TBM_SETPOS,1,Pos);
					p->ForcePos = -1;
					StartSeek(p);
				}
			}
			break;

		case IF_NEXT:
			p->Player->ListGo(p->Player,1);
			break;

		case IF_PREV:
			p->Player->ListGo(p->Player,-1);
			break;

		case IF_PLAY:
			p->Play = -1; // force button refresh
			Toggle(p,PLAYER_PLAY,-1);
			break;

		case IF_FASTFORWARD:
			p->FFwd = -1; // force button refresh
			Toggle(p,PLAYER_FFWD,-1);
			break;

		case IF_FILE_BENCHMARK:
			if (!Toggle(p,PLAYER_FULLSCREEN,1))
			{
				SetTrackThumb(p->WndTrack,0);
				WaitBegin();
			}
			t = 0;
			p->Player->Set(p->Player,PLAYER_BENCHMARK,&t,sizeof(t));
			p->Bench = 1;
			break;

		case IF_STOP:
			p->Player->Stop(p->Player);
			ClearStats(p); // clear total/dropped stats
			p->Player->ListGo(p->Player,0); // reload if current playlist item changed 
			f.Num = 0;
			f.Den = 1;
			p->Player->Set(p->Player,PLAYER_PERCENT,&f,sizeof(f));
			Refresh(p);
			break;

		case IF_FILE_OPENFILE:
			Node = NodeFind(OPENFILE_ID);
			b = 0;
			Node->Set(Node,OPENFILE_ADD,&b,sizeof(b));
			i = FORMAT_CLASS;
			Node->Set(Node,OPENFILE_CLASS,&i,sizeof(i));
			b = 0;
			Node->Set(Node,OPENFILE_SINGLE,&b,sizeof(i));
			WinOpen((win*)Node,&p->Win,1);
			break;

		case IF_FILE_ABOUT:
			WinOpen((win*)NodeFind(ABOUT_ID),&p->Win,1);
			break;

		case IF_OPTIONS_SETTINGS:
			WinOpen((win*)NodeFind(SETTINGS_ID),&p->Win,1);
			break;

		case IF_FILE_PLAYLIST:
			WinOpen((win*)NodeFind(PLAYLIST_ID),&p->Win,1);
			break;

		case IF_FILE_MEDIAINFO:
			WinOpen((win*)NodeFind(MEDIAINFO_ID),&p->Win,1);
			break;

		case IF_FILE_EXIT:
			// just to be sure to stop overlay update (WM_DESTROY comes after window is hidden)
			BeforeExit(p);
			DestroyWindow(p->Win.Wnd);
			break;

		case IF_OPTIONS_REPEAT:
			Toggle(p,PLAYER_REPEAT,-1);
			break;

		case IF_OPTIONS_SHUFFLE:
			Toggle(p,PLAYER_SHUFFLE,-1);
			break;

		case IF_OPTIONS_EQUALIZER:
			Toggle(p,PLAYER_EQENABLE,-1);
			p->Player->ReSync(p->Player);
			break;

		case IF_OPTIONS_FULLSCREEN:
			Toggle(p,PLAYER_FULLSCREEN,-1);
			break;

		case IF_OPTIONS_MUTE:
			if (p->Player->Get(p->Player,PLAYER_VOLUME,&i,sizeof(i))==ERR_NONE && i==0)
			{
				i = 1;
				p->Player->Set(p->Player,PLAYER_VOLUME,&i,sizeof(i));
			}
			Toggle(p,PLAYER_MUTE,-1);
			UpdateVolume(p);
			break;

		case IF_OPTIONS_VIDEO_DITHER:
			Toggle(p,PLAYER_DITHER,-1);
			break;

		case IF_OPTIONS_VIDEO_QUALITYCOLOR:
			Toggle(p,PLAYER_QUALITYCOLOR,-1);
			break;

		case IF_OPTIONS_ROTATE:
			p->Player->Get(p->Player,PLAYER_FULL_DIR,&i,sizeof(i));
			switch (i)
			{
			case 0:	i = DIR_SWAPXY | DIR_MIRRORLEFTRIGHT; break;
			case DIR_SWAPXY | DIR_MIRRORLEFTRIGHT: i = DIR_MIRRORUPDOWN | DIR_MIRRORLEFTRIGHT; break;
			case DIR_MIRRORUPDOWN | DIR_MIRRORLEFTRIGHT: i = DIR_SWAPXY | DIR_MIRRORUPDOWN; break;
			case DIR_SWAPXY | DIR_MIRRORUPDOWN:	i = 0; break;
			default: i = 0; break;
			}
			p->Player->Set(p->Player,PLAYER_FULL_DIR,&i,sizeof(i));
			break;

		case IF_OPTIONS_ROTATE_0:
			i = 0;
			p->Player->Set(p->Player,PLAYER_FULL_DIR,&i,sizeof(i));
			break;

		case IF_OPTIONS_ROTATE_270:
			i = DIR_SWAPXY | DIR_MIRRORUPDOWN;
			p->Player->Set(p->Player,PLAYER_FULL_DIR,&i,sizeof(i));
			break;

		case IF_OPTIONS_ROTATE_180:
			i = DIR_MIRRORUPDOWN | DIR_MIRRORLEFTRIGHT;
			p->Player->Set(p->Player,PLAYER_FULL_DIR,&i,sizeof(i));
			break;

		case IF_OPTIONS_ROTATE_90:
			i = DIR_SWAPXY | DIR_MIRRORLEFTRIGHT;
			p->Player->Set(p->Player,PLAYER_FULL_DIR,&i,sizeof(i));
			break;

		case IF_OPTIONS_ASPECT_SQUARE:
			f.Num = 1;
			f.Den = 1;
			p->Player->Set(p->Player,PLAYER_ASPECT,&f,sizeof(f));
			break;

		case IF_OPTIONS_ASPECT_4_3_NTSC:
			f.Num = 10;
			f.Den = 11;
			p->Player->Set(p->Player,PLAYER_ASPECT,&f,sizeof(f));
			break;

		case IF_OPTIONS_ASPECT_4_3_PAL:
			f.Num = 12;
			f.Den = 11;
			p->Player->Set(p->Player,PLAYER_ASPECT,&f,sizeof(f));
			break;

		case IF_OPTIONS_ASPECT_16_9_NTSC:
			f.Num = 40;
			f.Den = 33;
			p->Player->Set(p->Player,PLAYER_ASPECT,&f,sizeof(f));
			break;

		case IF_OPTIONS_ASPECT_16_9_PAL:
			f.Num = 16;
			f.Den = 11;
			p->Player->Set(p->Player,PLAYER_ASPECT,&f,sizeof(f));
			break;

		case IF_OPTIONS_ZOOM_FIT_SCREEN:
			f.Num = 0;
			f.Den = 1;
			p->Player->Set(p->Player,PLAYER_ZOOM,&f,sizeof(f));
			break;

		case IF_OPTIONS_ZOOM_FIT_WIDTH:
			f.Num = -1;
			f.Den = 1;
			p->Player->Set(p->Player,PLAYER_ZOOM,&f,sizeof(f));
			break;

		case IF_OPTIONS_ZOOM_FIT_HEIGHT:
			f.Num = -2;
			f.Den = 1;
			p->Player->Set(p->Player,PLAYER_ZOOM,&f,sizeof(f));
			break;

		case IF_OPTIONS_ZOOM_50:
			f.Num = 1;
			f.Den = 2;
			p->Player->Set(p->Player,PLAYER_ZOOM,&f,sizeof(f));
			break;

		case IF_OPTIONS_ZOOM_100:
			f.Num = 1;
			f.Den = 1;
			p->Player->Set(p->Player,PLAYER_ZOOM,&f,sizeof(f));
			break;

		case IF_OPTIONS_ZOOM_150:
			f.Num = 3;
			f.Den = 2;
			p->Player->Set(p->Player,PLAYER_ZOOM,&f,sizeof(f));
			break;

		case IF_OPTIONS_ZOOM_200:
			f.Num = 2;
			f.Den = 1;
			p->Player->Set(p->Player,PLAYER_ZOOM,&f,sizeof(f));
			break;

		case IF_OPTIONS_VIDEO_ZOOM_SMOOTH:
			Toggle(p,PLAYER_SMOOTH,-1);
			break;

		case IF_OPTIONS_VIDEO_UPDATE:
			Toggle(p,PLAYER_VIDEO_UPDATE,-1);
			break;

		case IF_OPTIONS_AUDIO_SWAPSTEREO:
			Toggle(p,PLAYER_AUDIO_SWAP,-1);
			break;

		case IF_OPTIONS_AUDIO_STEREO:
			i = 0;
			p->Player->Set(p->Player,PLAYER_STEREO,&i,sizeof(i));
			break;

		case IF_OPTIONS_AUDIO_MONO_LEFT:
			i = 1;
			p->Player->Set(p->Player,PLAYER_STEREO,&i,sizeof(i));
			break;

		case IF_OPTIONS_AUDIO_MONO_RIGHT:
			i = 2;
			p->Player->Set(p->Player,PLAYER_STEREO,&i,sizeof(i));
			break;

		case IF_OPTIONS_AUDIO_QUALITY_LOW:
			i = 0;
			p->Player->Set(p->Player,PLAYER_AUDIO_QUALITY,&i,sizeof(i));
			break;

		case IF_OPTIONS_AUDIO_QUALITY_MEDIUM:
			i = 1;
			p->Player->Set(p->Player,PLAYER_AUDIO_QUALITY,&i,sizeof(i));
			break;

		case IF_OPTIONS_AUDIO_QUALITY_HIGH:
			i = 2;
			p->Player->Set(p->Player,PLAYER_AUDIO_QUALITY,&i,sizeof(i));
			break;

		case IF_OPTIONS_SPEED_10:
			f.Num = 1;
			f.Den = 10;
			p->Player->Set(p->Player,PLAYER_PLAY_SPEED,&f,sizeof(f));
			break;

		case IF_OPTIONS_SPEED_25:
			f.Num = 1;
			f.Den = 4;
			p->Player->Set(p->Player,PLAYER_PLAY_SPEED,&f,sizeof(f));
			break;

		case IF_OPTIONS_SPEED_50:
			f.Num = 1;
			f.Den = 2;
			p->Player->Set(p->Player,PLAYER_PLAY_SPEED,&f,sizeof(f));
			break;

		case IF_OPTIONS_SPEED_100:
			f.Num = 1;
			f.Den = 1;
			p->Player->Set(p->Player,PLAYER_PLAY_SPEED,&f,sizeof(f));
			break;

		case IF_OPTIONS_SPEED_150:
			f.Num = 3;
			f.Den = 2;
			p->Player->Set(p->Player,PLAYER_PLAY_SPEED,&f,sizeof(f));
			break;

		case IF_OPTIONS_SPEED_200:
			f.Num = 2;
			f.Den = 1;
			p->Player->Set(p->Player,PLAYER_PLAY_SPEED,&f,sizeof(f));
			break;

		case IF_OPTIONS_SPEED_300:
			f.Num = 3;
			f.Den = 1;
			p->Player->Set(p->Player,PLAYER_PLAY_SPEED,&f,sizeof(f));
			break;

		case IF_OPTIONS_SPEED_BENCHMARK:
			f.Num = 0;
			f.Den = 1;
			p->Player->Set(p->Player,PLAYER_PLAY_SPEED,&f,sizeof(f));
			break;

		case IF_OPTIONS_VIDEO_TURNOFF:
			p->Player->Set(p->Player,PLAYER_VOUTPUT,&Null,sizeof(Null));
			break;

		case IF_OPTIONS_AUDIO_TURNOFF:
			p->Player->Set(p->Player,PLAYER_AOUTPUT,&Null,sizeof(Null));
			UpdateVolume(p);
			break;
		}

		if (LOWORD(wParam) >= IF_CHAPTER && LOWORD(wParam) < IF_CHAPTER+100)
		{
			tick_t Time = GetChapter(p,LOWORD(wParam)-IF_CHAPTER);
			if (Time>=0)
				p->Player->Set(p->Player,PLAYER_POSITION,&Time,sizeof(tick_t));
		}

		if (LOWORD(wParam) >= IF_STREAM_AUDIO && LOWORD(wParam) < IF_STREAM_AUDIO+MAXSTREAM)
		{
			i = LOWORD(wParam)-IF_STREAM_AUDIO;
			p->Player->Set(p->Player,PLAYER_ASTREAM,&i,sizeof(i));
			p->Player->ReSync(p->Player);
		}

		if (LOWORD(wParam) >= IF_STREAM_VIDEO && LOWORD(wParam) < IF_STREAM_VIDEO+MAXSTREAM)
		{
			i = LOWORD(wParam)-IF_STREAM_VIDEO;
			p->Player->Set(p->Player,PLAYER_VSTREAM,&i,sizeof(i));
			p->Player->ReSync(p->Player);
		}

		if (LOWORD(wParam) >= IF_VIDEO && LOWORD(wParam) < IF_VIDEO+MAXNODE)
		{
			b = 0;
			Node = NodeEnum(VOUT_CLASS,LOWORD(wParam)-IF_VIDEO);
			p->Player->Set(p->Player,PLAYER_VIDEO_ACCEL,&b,sizeof(b));
			p->Player->Set(p->Player,PLAYER_VOUTPUT,&Node,sizeof(node*));
		}

		if (LOWORD(wParam) >= IF_VIDEOACCEL && LOWORD(wParam) < IF_VIDEOACCEL+MAXNODE)
		{
			b = 1;
			Node = NodeEnum(VOUT_CLASS,LOWORD(wParam)-IF_VIDEOACCEL);
			p->Player->Set(p->Player,PLAYER_VIDEO_ACCEL,&b,sizeof(b));
			p->Player->Set(p->Player,PLAYER_VOUTPUT,&Node,sizeof(node*));
		}

		if (LOWORD(wParam) >= IF_AUDIO && LOWORD(wParam) < IF_AUDIO+MAXNODE)
		{
			Node = NodeEnum(AOUT_CLASS,LOWORD(wParam)-IF_AUDIO);
			p->Player->Set(p->Player,PLAYER_AOUTPUT,&Node,sizeof(node*));
			UpdateVolume(p);
		}

		if (p->Focus && p->Clipping)
			UpdateClipping(p,0,1); // maybe overlapping window is gone
		break;
	}

	return 0;
}

static const datadef Params[] =
{
	{ INTERFACE_ID,			TYPE_CLASS },
	{ INTERFACE_ERROR,		TYPE_STRING, DF_HIDDEN },

	DATADEF_END
};

static int Enum( intface* p, int No, datadef* Param )
{
	if (NodeEnumType(&No,Param,Params,NULL)==ERR_NONE)
		return ERR_NONE;

	if (No>=0 && No<HOTKEYCOUNT)
	{
		memset(Param,0,sizeof(datadef));
		Param->No = HotKey[No];
		Param->Type = TYPE_HOTKEY;
		Param->Size = sizeof(int);
		Param->Flags = DF_SETUP;
		Param->Class = INTERFACE_ID;
		Param->Name = String(INTERFACE_ID,Param->No);
		return ERR_NONE;
	}
	return ERR_INVALID_PARAM;
}

static int Get(intface* p,int No,void* Data,int Size)
{
	int Result = ERR_INVALID_PARAM;
	int i;

	for (i=0;i<HOTKEYCOUNT;++i)
		if (HotKey[i]==No)
		{
			GETVALUE(p->HotKey[i],int); 
			break;
		}

	return Result;
}

static void SetHotKey(intface* p,int i,int Key)
{
	if (p->Focus && p->HotKey[i])
		WinUnRegisterHotKey(&p->Win,HotKey[i]);

	p->HotKey[i] = Key;

	if (p->Focus && p->HotKey[i])
		WinRegisterHotKey(&p->Win,HotKey[i],p->HotKey[i]);
}

static int Set(intface* p,int No,const void* Data,int Size)
{
	int Result = ERR_INVALID_PARAM;
	int i,j;
	const tchar_t* Error;

	for (i=0;i<HOTKEYCOUNT;++i)
		if (HotKey[i]==No)
		{
			SetHotKey(p,i,0);
			if (*(int*)Data)
			{
				for (j=0;j<HOTKEYCOUNT;++j)
					if (p->HotKey[j] == *(int*)Data)
						SetHotKey(p,j,0);
				SetHotKey(p,i,*(int*)Data);
			}
			Result = ERR_NONE;
			break;
		}

	switch (No)
	{
	case INTERFACE_ERROR:

		Error = (const tchar_t*)Data;
		if (GetCurrentThreadId() != p->ThreadId)
		{
			tchar_t* s = (tchar_t*) Alloc((tcslen(Error)+1)*sizeof(tchar_t));
			tcscpy(s,Error);
			PostThreadMessage(p->ThreadId,MSG_ERROR,0,(LPARAM)s);
		}
		else
			ShowError(p,(const tchar_t*)Data);

		Result = ERR_NONE;
		break;
	}
	return Result;
}

static intface Interface;

void Interface_Init()
{
	int Key;
	intface* p = &Interface;
	p->Win.Id = INTERFACE_ID;
	p->Win.Width = 480;
	p->Win.Height = 320;
	p->Win.Proc = Proc;
	p->Win.MenuDef = MenuDef;
	p->Win.Title = NodeProgramId();
	p->Win.Enum = Enum;
	p->Win.Get = Get;
	p->Win.Set = Set;

	p->MenuStreams = 0;
	p->MenuAStream = -1;
	p->MenuVStream = -1;

	Key = VK_RIGHT;
	Set(p,IF_MOVE_FFWD,&Key,sizeof(Key));
	Key = VK_LEFT;
	Set(p,IF_MOVE_BACK,&Key,sizeof(Key));
	Key = VK_UP;
	Set(p,IF_OPTIONS_VOLUME_UP,&Key,sizeof(Key));
	Key = VK_DOWN;
	Set(p,IF_OPTIONS_VOLUME_DOWN,&Key,sizeof(Key));
	Key = VK_RETURN;
	Set(p,QueryPlatform(PLATFORM_TYPENO) == TYPE_SMARTPHONE?IF_PLAY_FULLSCREEN:IF_PLAY,&Key,sizeof(Key));

	WinRegister(&p->Win,PRI_MAXIMUM+570);
}

void Interface_Done()
{
	WinUnRegister(&Interface.Win);
}