/*****************************************************************************
 *
 * This program is free software ; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 *
 * $Id$
 *
 ****************************************************************************/

#include "stdafx.h"

#define OPENFILE_TYPES		50000

#define IMG_FILE		0
#define IMG_DIR			1
#define IMG_PLAYLIST	2
#define IMG_VIDEO		3
#define IMG_AUDIO		4
#define IMG_DIRUP		5

#define MAXHISTORY		20

typedef struct openitem
{
	int Image;
	tchar_t FileName[MAXPATH];
	tchar_t Name[MAXPATH];
	tchar_t Ext[MAXPATH];
	int Size;
	
} openitem;

typedef struct openfile
{
	win Win;
	HWND WndList;
	HWND WndURL;
	HWND WndGo;
	int FileType;
	int SortCol;
	int Class;
	bool_t SortDir;
	bool_t Multiple;
	bool_t Single;
	bool_t Add;
	bool_t OwnSelect;
	bool_t NextFocus;
	int LastClick;
	int DlgWidth[3];
	tchar_t Path[MAXPATH];
	tchar_t Exts[1024];
	tchar_t Last[MAXPATH];

	int HistoryCount;
	tchar_t* History[MAXHISTORY];

} openfile;

static const datadef Params[] =
{
	{ OPENFILE_ID,			TYPE_CLASS },
	{ OPENFILE_ADD,			TYPE_BOOL, DF_HIDDEN },
	{ OPENFILE_SINGLE,		TYPE_BOOL, DF_HIDDEN },
	{ OPENFILE_CLASS,		TYPE_INT, DF_HIDDEN },
	{ OPENFILE_MULTIPLE,	TYPE_BOOL, DF_SETUP|DF_HIDDEN },
	{ OPENFILE_FILETYPE,	TYPE_INT, DF_SETUP|DF_HIDDEN },
	{ OPENFILE_PATH,		TYPE_STRING, DF_SETUP|DF_HIDDEN },
	{ OPENFILE_SORTCOL,		TYPE_INT, DF_SETUP|DF_HIDDEN },
	{ OPENFILE_SORTDIR,		TYPE_BOOL, DF_SETUP|DF_HIDDEN },
	{ OPENFILE_WIDTH_NAME,	TYPE_INT, DF_SETUP|DF_HIDDEN },
	{ OPENFILE_WIDTH_TYPE,	TYPE_INT, DF_SETUP|DF_HIDDEN },
	{ OPENFILE_WIDTH_SIZE,	TYPE_INT, DF_SETUP|DF_HIDDEN },
	{ OPENFILE_HISTORYCOUNT,TYPE_INT, DF_SETUP|DF_HIDDEN },
	DATADEF_END
};

static int Enum(openfile* p, int No, datadef* Param )
{
	if (NodeEnumType(&No,Param,Params,NULL)==ERR_NONE)
		return ERR_NONE;

	if (No >= 0 && No < p->HistoryCount)
	{
		memset(Param,0,sizeof(datadef));
		Param->No = OPENFILE_HISTORY + No;
		Param->Type = TYPE_STRING;
		Param->Size = MAXDATA;
		Param->Flags = DF_HIDDEN | DF_SETUP;
		Param->Class = OPENFILE_ID;
		return ERR_NONE;
	}

	return ERR_INVALID_PARAM;
}

static int Get(openfile* p,int No,void* Data,int Size)
{
	int Result = ERR_INVALID_PARAM;
	switch (No)
	{
	case OPENFILE_MULTIPLE: GETVALUE(p->Multiple,bool_t); break;
	case OPENFILE_SINGLE: GETVALUE(p->Single,bool_t); break;
	case OPENFILE_CLASS: GETVALUE(p->Class,int); break;
	case OPENFILE_ADD: GETVALUE(p->Add,bool_t); break;
	case OPENFILE_FILETYPE: GETVALUE(p->FileType,int); break;
	case OPENFILE_PATH: GETSTRING(p->Path); break;
	case OPENFILE_SORTCOL: GETVALUE(p->SortCol,int); break;
	case OPENFILE_SORTDIR: GETVALUE(p->SortDir,int); break;
	case OPENFILE_WIDTH_NAME: GETVALUE(p->DlgWidth[0],int); break;
	case OPENFILE_WIDTH_TYPE: GETVALUE(p->DlgWidth[1],int); break;
	case OPENFILE_WIDTH_SIZE: GETVALUE(p->DlgWidth[2],int); break;
	case OPENFILE_HISTORYCOUNT: GETVALUE(p->HistoryCount,int); break;
	}
	
	if (No >= OPENFILE_HISTORY && No < OPENFILE_HISTORY+p->HistoryCount && p->History[No-OPENFILE_HISTORY])
		GETSTRING(p->History[No-OPENFILE_HISTORY]);

	return Result;
}

static int Set(openfile* p,int No,const void* Data,int Size)
{
	int Result = ERR_INVALID_PARAM;
	switch (No)
	{
	case OPENFILE_MULTIPLE: SETVALUE(p->Multiple,bool_t,ERR_NONE); break;
	case OPENFILE_SINGLE: SETVALUE(p->Single,bool_t,ERR_NONE); break;
	case OPENFILE_CLASS: SETVALUE(p->Class,int,ERR_NONE); break;
	case OPENFILE_ADD: SETVALUE(p->Add,bool_t,ERR_NONE); break;
	case OPENFILE_FILETYPE: SETVALUE(p->FileType,int,ERR_NONE); break;
	case OPENFILE_PATH: SETSTRING(p->Path); break;
	case OPENFILE_SORTCOL: SETVALUE(p->SortCol,int,ERR_NONE); break;
	case OPENFILE_SORTDIR: SETVALUE(p->SortDir,int,ERR_NONE); break;
	case OPENFILE_WIDTH_NAME: SETVALUE(p->DlgWidth[0],int,ERR_NONE); break;
	case OPENFILE_WIDTH_TYPE: SETVALUE(p->DlgWidth[1],int,ERR_NONE); break;
	case OPENFILE_WIDTH_SIZE: SETVALUE(p->DlgWidth[2],int,ERR_NONE); break;
	case OPENFILE_HISTORYCOUNT: SETVALUE(p->HistoryCount,int,ERR_NONE); break;
	}

	if (No >= OPENFILE_HISTORY && No < OPENFILE_HISTORY+p->HistoryCount)
	{
		tchar_t* s = (tchar_t*) Alloc(sizeof(tchar_t)*(tcslen((tchar_t*)Data)+1));
		tcscpy(s,(tchar_t*)Data);
		Free(p->History[No-OPENFILE_HISTORY]);
		p->History[No-OPENFILE_HISTORY] = s;
		Result = ERR_NONE;
	}

	return Result;
}

static int GetColWidth(openfile* p, int Index)
{
	int cx = ListView_GetColumnWidth(p->WndList,Index);
	return Scale(cx,72,WinPixelSX());
}

static void AddCol(openfile* p, int Index, const tchar_t* Name, int Width, int Format )
{
	LVCOLUMN Col;
	Col.mask=LVCF_TEXT | LVCF_WIDTH | LVCF_FMT;
	Col.cx=Scale(Width,WinPixelSX(),72);
	Col.pszText=(tchar_t*)Name;
	Col.cchTextMax=tcslen(Name)+1;
	Col.fmt=Format;

	ListView_InsertColumn(p->WndList,Index,&Col);
}

static int CALLBACK CompareProc(LPARAM lParam1, LPARAM lParam2, LPARAM lParamSort)
{
	openfile* p = (openfile*)lParamSort;
	openitem* a = (openitem*)lParam1;
	openitem* b = (openitem*)lParam2;
	int Result = 0;

	if (!a) return -1;
	if (!b) return 1;

	if (a->Image==IMG_DIR)
		if (b->Image==IMG_DIR)
			Result = tcscmp(a->Name,b->Name);
		else
			Result = -1;
	else
	if (b->Image==IMG_DIR)
		Result = 1;
	else
	switch(p->SortCol)
	{
	case 0: 
		Result = tcsicmp(a->Name,b->Name); 
		break;
	case 1: 
		Result = tcsicmp(a->Ext,b->Ext); 
		if (!Result) 
			Result = tcsicmp(a->Name,b->Name);
		break;
	case 2: 
		if (a->Size == b->Size) 
			Result = tcsicmp(a->Name,b->Name);
		else
			Result = a->Size > b->Size ? 1:-1;
		break;
	}

	if (p->SortDir)
		Result = -Result;
	return Result;
}

static void SavePos(openfile* p)
{
	LVITEM Item;
	int n;

	p->Last[0] = 0;

	for (n=0;n<ListView_GetItemCount(p->WndList);++n)
		if (ListView_GetItemState(p->WndList,n,LVIS_FOCUSED)==LVIS_FOCUSED)
		{
			Item.iItem=n;
			Item.iSubItem = 0;
			Item.mask=LVIF_PARAM;
			ListView_GetItem(p->WndList,&Item);

			if (Item.lParam)
				tcscpy(p->Last,((openitem*)Item.lParam)->Name);

			break;
		}
}

static void ClearList(openfile* p)
{
	LVITEM Item;
	int n;

	for (n=0;n<ListView_GetItemCount(p->WndList);++n)
	{
		Item.iItem=n;
		Item.iSubItem=0;
		Item.mask=LVIF_PARAM;
		ListView_GetItem(p->WndList,&Item);
		if (Item.lParam)
			Free((void*)Item.lParam);
	}	

	ListView_DeleteAllItems(p->WndList);
}

static void SelectAll(openfile* p)
{
	LVITEM Item;
	int n;
	int State;

	p->OwnSelect = 1;

	for (n=0;n<ListView_GetItemCount(p->WndList);++n)
	{
		Item.iItem=n;
		Item.iSubItem=0;
		Item.mask=LVIF_PARAM;
		ListView_GetItem(p->WndList,&Item);

		State = 0;
		if (Item.lParam && ((openitem*)Item.lParam)->Image != IMG_DIR)
			State = LVIS_SELECTED;

		ListView_SetItemState(p->WndList,n,State,LVIS_SELECTED);
	}	

	p->OwnSelect = 0;
}

static void UpdateMultiple(openfile* p)
{
	int Count,i;

	if (p->Single) 
		return;

	if (QueryPlatform(PLATFORM_TYPENO)==TYPE_SMARTPHONE)
	{
		CheckMenuItem(p->Win.Menu,OPENFILE_MORE,p->Multiple ? MF_CHECKED:MF_UNCHECKED);
	}
	else
	{
		TBBUTTONINFO Button;
		Button.cbSize = sizeof(Button);
		Button.dwMask = TBIF_STATE;
		Button.fsState = TBSTATE_ENABLED;
		if (p->Multiple)
			Button.fsState |= TBSTATE_CHECKED;

		SendMessage(p->Win.WndTB,TB_SETBUTTONINFO,OPENFILE_MORE,(LPARAM)&Button);
	}

	p->OwnSelect = 1;

	Count = ListView_GetItemCount(p->WndList);
	for (i=0;i<Count;++i)
	{
		int State = 0;
		if (!p->Multiple && ListView_GetItemState(p->WndList,i,LVIS_FOCUSED)==LVIS_FOCUSED)
			State = LVIS_SELECTED;
		ListView_SetItemState(p->WndList,i,State,LVIS_SELECTED);
	}

	p->OwnSelect = 0;
}

static void AddHistory(openfile* p,const tchar_t* s)
{
	int i;
	for (i=0;i<p->HistoryCount;++i)
		if (p->History[i] && tcsicmp(p->History[i],s)==0)
		{
			for (;i>0;--i)
				SwapPChar(&p->History[i],&p->History[i-1]);
			break;
		}

	if (i==p->HistoryCount)
	{
		if (i<MAXHISTORY)
			p->HistoryCount++;
		else
			Free(p->History[--i]);

		for (;i>0;--i)
			p->History[i] = p->History[i-1];

		p->History[0] = (tchar_t*) Alloc(sizeof(tchar_t)*(tcslen(s)+1));
		tcscpy(p->History[0],s);
	}

	if (QueryPlatform(PLATFORM_TYPENO)!=TYPE_SMARTPHONE)
	{
		int i;
		SendMessage(p->WndURL,CB_RESETCONTENT,0,0);
		for (i=0;i<p->HistoryCount;++i)
			SendMessage(p->WndURL,CB_ADDSTRING,0,(LPARAM)p->History[i]);
	}

	SetWindowText(p->WndURL,s);
}

static void UpdateList(openfile* p, bool_t Silent)
{
	streamdir DirItem;
	stream* Stream;
	int Result;
	openitem New;
	LVITEM Item;
	int No=0;
	int Pos=0;
	int State = LVIS_FOCUSED;
	int Len;
	const tchar_t* s;

	if (!p->Multiple || p->Single)
		State |= LVIS_SELECTED; 

	Len = tcslen(p->Path);
	if (Len && p->Path[Len-1] == '\\')
		p->Path[--Len] = 0;

	s = p->Path[0] ? p->Path : T("\\");
	AddHistory(p,s);

	WaitBegin();
	ShowWindow(p->WndList,SW_HIDE);
	ClearList(p);

	Item.mask=LVIF_TEXT|LVIF_IMAGE|LVIF_PARAM;
	Item.pszText=T("");
	Item.cchTextMax=1;
	Item.iSubItem=0;

	Stream = GetStream(p->Path,Silent);

	if (Stream)
	{
		s = GetMime(p->Path,NULL);
		if (*s)
		{
			Item.iImage=IMG_DIRUP;
			Item.iItem = No;
			Item.lParam = 0;

			ListView_InsertItem(p->WndList,&Item);
			ListView_SetItemText(p->WndList,No,0,T(".."));
			ListView_SetItemText(p->WndList,No,1,(tchar_t*)String(OPENFILE_ID,OPENFILE_UP));
			++No;
		}

		Result = Stream->EnumDir(Stream,p->Path,p->Exts,p->FileType!=-1,&DirItem);

		if (Result == ERR_NOT_DIRECTORY && !Silent)
		{
			New.FileName[0] = 0;
			New.Ext[0] = 0;
			New.Image = IMG_FILE;
			New.Size = 0;

			Item.lParam = (DWORD) Alloc(sizeof(openitem));
			if (Item.lParam)
			{
				p->Last[0] = 0;
				Pos = No;
				State = LVIS_SELECTED;
				PostMessage(p->Win.Wnd,WM_COMMAND,PLATFORM_OK,0);

				*(openitem*)Item.lParam = New;
				Item.iItem = No;
				Item.iImage = New.Image;
				ListView_InsertItem(p->WndList,&Item);
				++No;
			}
		}
	
		while (Result == ERR_NONE)
		{
			tchar_t Size[32];

			tcscpy(New.FileName,DirItem.FileName);

			if (DirItem.Size < 0)
			{
				New.Image = IMG_DIR;
				New.Size = 0;
				tcscpy(New.Name,New.FileName);
				tcscpy(New.Ext,String(OPENFILE_ID,OPENFILE_DIR));
				Size[0] = 0;
			}
			else
			{
				switch (DirItem.Type)
				{
				case FTYPE_AUDIO: New.Image = IMG_AUDIO; break;
				case FTYPE_PLAYLIST: New.Image = IMG_PLAYLIST; break;
				case FTYPE_VIDEO: New.Image = IMG_VIDEO; break;
				default: New.Image = IMG_FILE; break;
				}
				New.Size = DirItem.Size;
				SplitURL(New.FileName,NULL,NULL,New.Name,New.Ext);
				tcsupr(New.Ext);
				if (New.Size < 10000*1024)
					stprintf(Size,T("%d KB"),New.Size/1024);
				else
					stprintf(Size,T("%d MB"),New.Size/(1024*1024));
			}
	
			Item.lParam = (DWORD) Alloc(sizeof(openitem));
			if (Item.lParam)
			{
				*(openitem*)Item.lParam = New;
				Item.iItem = No;
				Item.iImage = New.Image;
				ListView_InsertItem(p->WndList,&Item);
				ListView_SetItemText(p->WndList,No,0,New.Name);
				ListView_SetItemText(p->WndList,No,1,New.Ext);
				ListView_SetItemText(p->WndList,No,2,Size);
				++No;
			}

			Result = Stream->EnumDir(Stream,NULL,NULL,0,&DirItem);
		}

		Stream->Delete(Stream);
	}

	ListView_SortItems(p->WndList,CompareProc,p);
	ShowWindow(p->WndList,SW_SHOWNA);

	if (p->Last[0])
	{
		int i;
		for (i=0;i<No;++i)
		{
			LVITEM Item;
			Item.iItem=i;
			Item.iSubItem = 0;
			Item.mask=LVIF_PARAM;
			ListView_GetItem(p->WndList,&Item);

			if (Item.lParam && tcsicmp(((openitem*)Item.lParam)->Name,p->Last)==0)
			{
				Pos = i;
				break;
			}
		}
	}

	SetFocus(p->WndList);

	p->OwnSelect = 1;
	ListView_SetItemState(p->WndList,Pos,State,State);
	p->OwnSelect = 0;

	ListView_EnsureVisible(p->WndList,Pos,TRUE);

	WaitEnd();
}

static void UpdateMenu(openfile* p)
{
	int No;
	int Id;
	node* Node;

	InsertMenu(p->Win.Menu,OPENFILE_ALL_FILES,MF_BYCOMMAND|MF_STRING,
		OPENFILE_TYPES-1,String(p->Class,0));

	for (No=0;(Node = NodeEnum(p->Class,No))!=NULL;++No)
		if (Node->Get(Node,NODE_ID,&Id,sizeof(Id))==ERR_NONE)
		{
			const tchar_t* Name = String(Id,0);
			if (Name[0])
				InsertMenu(p->Win.Menu,OPENFILE_ALL_FILES,MF_BYCOMMAND|MF_STRING,
					OPENFILE_TYPES+No,Name);
		}
}

static void UpdateSort(openfile* p)
{
	CheckMenuItem(p->Win.Menu,OPENFILE_LABEL_NAME,p->SortCol==0 ? MF_CHECKED:MF_UNCHECKED);
	CheckMenuItem(p->Win.Menu,OPENFILE_LABEL_TYPE,p->SortCol==1 ? MF_CHECKED:MF_UNCHECKED);
	CheckMenuItem(p->Win.Menu,OPENFILE_LABEL_SIZE,p->SortCol==2 ? MF_CHECKED:MF_UNCHECKED);
	ListView_SortItems( p->WndList, CompareProc, (LPARAM)p);
}

static void UpdateFileType(openfile* p)
{
	int Id;
	int No;
	node* Node;

	if (p->FileType && p->FileType!=-1 && !NodeFind(p->FileType))
		p->FileType = 0;

	CheckMenuItem(p->Win.Menu,OPENFILE_TYPES-1,p->FileType==0 ? MF_CHECKED:MF_UNCHECKED);
	CheckMenuItem(p->Win.Menu,OPENFILE_ALL_FILES,p->FileType==-1 ? MF_CHECKED:MF_UNCHECKED);

	p->Exts[0]=0;

	for (No=0;(Node = NodeEnum(p->Class,No))!=NULL;++No)
	{
		Node->Get(Node,NODE_ID,&Id,sizeof(Id));
		CheckMenuItem(p->Win.Menu,OPENFILE_TYPES+No,p->FileType==Id ? MF_CHECKED:MF_UNCHECKED);

		if (p->FileType==Id || p->FileType==0 || p->FileType==-1)
		{
			tchar_t s[MAXPATH];
			if (Node->Get(Node,FORMAT_EXTS,s,sizeof(s))==ERR_NONE)
			{
				if (p->Exts[0]) tcscat(p->Exts,T(";"));
				tcscat(p->Exts,s);
			}
		}
	}

	p->Last[0] = 0;

	UpdateList(p,1);
}

static void Add(openfile* p,int PosIfNone)
{
	player* Player = (player*) NodeFind(PLAYER_ID);
	bool_t Add = p->Add;
	bool_t b;
	int Count,n,i;

	if (Player)
	{
		Count = ListView_GetItemCount(p->WndList);

		for (i=0;i<Count;++i)
			if (ListView_GetItemState(p->WndList,i,LVIS_SELECTED)==LVIS_SELECTED)
			{
				LVITEM Item;
				Item.iItem=i;
				Item.iSubItem=0;
				Item.mask=LVIF_PARAM;
				if (ListView_GetItem(p->WndList,&Item) && Item.lParam && ((openitem*)Item.lParam)->Image != IMG_DIR)
					break;
			}

		if (i==Count && PosIfNone>=0)
		{
			p->OwnSelect = 1;
			ListView_SetItemState(p->WndList,PosIfNone,LVIS_SELECTED,LVIS_SELECTED);
			p->OwnSelect = 0;

			i = 0;
		}

		if (i<Count)
		{
			if (Add)
			{
				Player->Get(Player,PLAYER_LIST_COUNT,&n,sizeof(n));
				Add = n>0;
			}
			else
			{
				b = 0;
				Player->Set(Player,PLAYER_PLAY,&b,sizeof(b));
				n = 0;
				Player->Set(Player,PLAYER_LIST_COUNT,&n,sizeof(n));
			}

			for (i=0;i<Count;++i)
				if (ListView_GetItemState(p->WndList,i,LVIS_SELECTED)==LVIS_SELECTED)
				{
					LVITEM Item;
					Item.iItem=i;
					Item.iSubItem=0;
					Item.mask=LVIF_PARAM;
					if (ListView_GetItem(p->WndList,&Item) && Item.lParam && ((openitem*)Item.lParam)->Image != IMG_DIR)
					{
						tchar_t URL[MAXPATH];
						tcscpy(URL,p->Path);
						if (((openitem*)Item.lParam)->FileName[0])
							stprintf(URL+tcslen(URL),T("\\%s"),((openitem*)Item.lParam)->FileName);
						n = PlayerAdd(Player,n,URL);
					}
				}

			if (!Add)
			{
				n=0;
				Player->Set(Player,PLAYER_LIST_CURRENT,&n,sizeof(n));
			}
		}
	}

	PostMessage(p->Win.Wnd,WM_CLOSE,0,0);
}

static void Resize(openfile* p)
{
	RECT r;
	int BorderX = WinPixelSX()/72;
	int BorderY = WinPixelSY()/72;
	int URLHeight;

	GetClientRect(p->Win.Wnd,&r);

	r.left += BorderX;
	r.right -= BorderX;

	if (QueryPlatform(PLATFORM_TYPENO)==TYPE_SMARTPHONE)
	{
		URLHeight = Scale(14,WinPixelSY(),72);
		MoveWindow(p->WndURL,r.left,BorderY,r.right-r.left,URLHeight-2*BorderY,0);
	}
	else
	{
		RECT Combo;
		int ButtonWidth = Scale(30,WinPixelSX(),72);

		SendMessage(p->WndURL,CB_SHOWDROPDOWN,0,0);
		r.right -= ButtonWidth;
		MoveWindow(p->WndURL,r.left,BorderY,r.right-r.left-BorderX,Scale(200,WinPixelSY(),72),0);
		GetWindowRect(p->WndURL,&Combo);
		URLHeight = Combo.bottom - Combo.top + BorderY*2;
		MoveWindow(p->WndGo,r.right,BorderY,ButtonWidth,URLHeight-BorderY*2,0);
	}

	GetClientRect(p->Win.Wnd,&r);

	r.top += p->Win.ToolBarHeight;
	MoveWindow(p->Win.WndDialog,r.left,r.top,r.right-r.left,URLHeight,0);

	r.top += URLHeight;
	MoveWindow(p->WndList,r.left,r.top,r.right-r.left,r.bottom-r.top,0);
}

static void InitMenu(openfile* p)
{
	int i;

	if (EnableMenuItem(p->Win.Menu,OPENFILE_EMPTY,MF_BYCOMMAND|MF_GRAYED)==-1)
	{
		InsertMenu(p->Win.Menu,OPENFILE_HISTORY+1,MF_BYCOMMAND|MF_STRING,
			OPENFILE_EMPTY,String(OPENFILE_ID,OPENFILE_EMPTY));

		for (i=1;i<MAXHISTORY;++i)
			DeleteMenu(p->Win.Menu,OPENFILE_HISTORY+i,MF_BYCOMMAND);
	}

	if (p->HistoryCount>1)
	{
		for (i=1;i<MAXHISTORY;++i)
			InsertMenu(p->Win.Menu,OPENFILE_EMPTY,MF_BYCOMMAND|MF_STRING,
				OPENFILE_HISTORY+i,p->History[i]);

		DeleteMenu(p->Win.Menu,OPENFILE_EMPTY,MF_BYCOMMAND);
	}
	else
		EnableMenuItem(p->Win.Menu,OPENFILE_EMPTY,MF_BYCOMMAND|MF_GRAYED);
}

static bool_t DialogProc(openfile* p,int Msg, WPARAM wParam, LPARAM lParam, int* Result)
{
	int i;

	switch (Msg)
	{
	case WM_COMMAND:

		if (p->WndGo == (HWND)lParam && HIWORD(wParam)==BN_CLICKED)
			SendMessage(p->Win.Wnd,WM_COMMAND,OPENFILE_GO,0);

		if (p->WndURL == (HWND)lParam)
		{
			switch (HIWORD(wParam))
			{
			case CBN_DROPDOWN:
				//SendMessage(p->WndURL,CB_SETCURSEL,(WPARAM)-1,0);
				break;

			case CBN_CLOSEUP:
				i = SendMessage(p->WndURL,CB_GETCURSEL,0,0);
				if (i!=CB_ERR && i>0 && i < p->HistoryCount)
				{
					SetWindowText(p->WndURL,p->History[i]);					
					PostMessage(p->Win.Wnd,WM_COMMAND,OPENFILE_GO,0);
				}
				break;

			case CBN_SETFOCUS:
			case EN_SETFOCUS:
				//WinShowSip(1); no sip because of combobox dropdown
				p->Win.CaptureKeys = 1;
				break;

			case CBN_KILLFOCUS:
			case EN_KILLFOCUS:
				p->Win.CaptureKeys = 0;
				//WinShowSip(0);
				break;
			}
		}
	}
	return 0;
}

static menudef MenuDef[] =
{
	{ 0,PLATFORM_ID, PLATFORM_OK },
	{ 0,OPENFILE_ID, OPENFILE_FILTER },
	{ 1,OPENFILE_ID, OPENFILE_ALL_FILES },
	{ 0,PLATFORM_ID, PLATFORM_CANCEL },
	{ 0,-1,-1 },
	{ 0,OPENFILE_ID, OPENFILE_MORE },
	{ 0,-1,-1 },
	{ 0,OPENFILE_ID, OPENFILE_SELECTALL },

	MENUDEF_END
};

static menudef MenuDefSingle[] =
{
	{ 0,PLATFORM_ID, PLATFORM_OK },
	{ 0,OPENFILE_ID, OPENFILE_FILTER },
	{ 1,OPENFILE_ID, OPENFILE_ALL_FILES },
	{ 0,PLATFORM_ID, PLATFORM_CANCEL },

	MENUDEF_END
};

static menudef MenuDefSmart[] =
{
	{ 0,PLATFORM_ID, PLATFORM_OK },
	{ 0,OPENFILE_ID, OPENFILE_OPTIONS },
	{ 1,OPENFILE_ID, OPENFILE_ENTERURL },
	{ 1,OPENFILE_ID, OPENFILE_HISTORYMENU },
	{ 2,OPENFILE_ID, OPENFILE_EMPTY },
	{ 1,OPENFILE_ID, OPENFILE_MORE },
	{ 1,OPENFILE_ID, OPENFILE_SELECTALL },
	{ 1,OPENFILE_ID, OPENFILE_SORT },
	{ 2,OPENFILE_ID, OPENFILE_LABEL_NAME },
	{ 2,OPENFILE_ID, OPENFILE_LABEL_TYPE },
	{ 2,OPENFILE_ID, OPENFILE_LABEL_SIZE },
	{ 1,OPENFILE_ID, OPENFILE_FILTER },
	{ 2,OPENFILE_ID, OPENFILE_ALL_FILES },

	MENUDEF_END
};

static menudef MenuDefSmartSingle[] =
{
	{ 0,PLATFORM_ID, PLATFORM_OK },
	{ 0,OPENFILE_ID, OPENFILE_OPTIONS },
	{ 1,OPENFILE_ID, OPENFILE_ENTERURL },
	{ 1,OPENFILE_ID, OPENFILE_HISTORYMENU },
	{ 2,OPENFILE_ID, OPENFILE_EMPTY },
	{ 1,OPENFILE_ID, OPENFILE_SORT },
	{ 2,OPENFILE_ID, OPENFILE_LABEL_NAME },
	{ 2,OPENFILE_ID, OPENFILE_LABEL_TYPE },
	{ 2,OPENFILE_ID, OPENFILE_LABEL_SIZE },
	{ 1,OPENFILE_ID, OPENFILE_FILTER },
	{ 2,OPENFILE_ID, OPENFILE_ALL_FILES },

	MENUDEF_END
};

static bool_t Proc(openfile* p,int Msg, WPARAM wParam, LPARAM lParam, int* Result)
{
	LPNMLISTVIEW Notify;
	int i;
	int Style;
	int FontSize;

	switch (Msg)
	{
	case WM_KEYDOWN:
		if (wParam == VK_RETURN)
		{
			SendMessage(p->Win.Wnd,WM_COMMAND,OPENFILE_GO,0);
			return 1;
		}

		if (wParam == VK_ESCAPE)
			PostMessage(p->Win.Wnd,WM_CLOSE,0,0);
		break;

	case WM_COMMAND:

		i = LOWORD(wParam);

		if (i>=OPENFILE_HISTORY && i<OPENFILE_HISTORY+p->HistoryCount)
		{
			SetWindowText(p->WndURL,p->History[i-OPENFILE_HISTORY]);					
			PostMessage(p->Win.Wnd,WM_COMMAND,OPENFILE_GO,0);
		}
		else
		switch (i)
		{
		case PLATFORM_OK:
			Add(p,-1); 
			break;
		case PLATFORM_CANCEL:
			PostMessage(p->Win.Wnd,WM_CLOSE,0,0);
			break;
		case OPENFILE_TYPES-1:
			p->FileType = 0;
			UpdateFileType(p);
			break;
		case OPENFILE_ALL_FILES:
			p->FileType = -1;
			UpdateFileType(p);
			break;

		case OPENFILE_LABEL_NAME:
		case OPENFILE_LABEL_TYPE:
		case OPENFILE_LABEL_SIZE:
			if (p->SortCol != i-OPENFILE_LABEL_NAME)
				p->SortDir = 0;
			else
				p->SortDir = !p->SortDir;
			p->SortCol = i-OPENFILE_LABEL_NAME;
			UpdateSort(p);
			break;

		case OPENFILE_GO:
			tcscpy(p->Last,p->Path);
			GetWindowText(p->WndURL,p->Path,sizeof(p->Path)/sizeof(tchar_t));
			if (tcsicmp(p->Last,p->Path)==0)
				SavePos(p);
			else
				p->Last[0] = 0;

			UpdateList(p,0);
			break;

		case OPENFILE_ENTERURL:
			SetFocus(p->WndURL);
			break;

		case OPENFILE_SELECTALL:
			p->Multiple = 1;
			UpdateMultiple(p);
			SelectAll(p);
			break;

		case OPENFILE_MORE:
			p->Multiple = !p->Multiple;
			UpdateMultiple(p);
			break;
		}
		if (i >= OPENFILE_TYPES && i < OPENFILE_TYPES+MAXNODE)
		{
			node* Node = NodeEnum(p->Class,i-OPENFILE_TYPES);
			if (Node && Node->Get(Node,NODE_ID,&p->FileType,sizeof(int))==ERR_NONE)
				UpdateFileType(p);
		}
		break;

	case WM_INITMENUPOPUP:
		InitMenu(p);
		break;

	case MSG_PREPARE:
		if (QueryPlatform(PLATFORM_TYPENO)==TYPE_SMARTPHONE)
			p->Win.MenuDef = p->Single ? MenuDefSmartSingle : MenuDefSmart;
		else
			p->Win.MenuDef = p->Single ? MenuDefSingle : MenuDef;
		break;

	case WM_CREATE:

		p->Last[0] = 0;

		WinCreateDialog(&p->Win);

		if (QueryPlatform(PLATFORM_TYPENO)==TYPE_SMARTPHONE)
		{
			p->WndGo = NULL;
			p->WndURL = CreateWindow(T("EDIT"), NULL, WS_TABSTOP|WS_VISIBLE|WS_CHILD|WS_BORDER, 
				0,0,10,10,p->Win.WndDialog, NULL, p->Win.Module, NULL);
		}
		else
		{
			p->WndURL = CreateWindow(T("COMBOBOX"), NULL, WS_TABSTOP|WS_VISIBLE|WS_CHILD|CBS_DROPDOWN, 
				0,0,10,10,p->Win.WndDialog, NULL, p->Win.Module, 0L);

			p->WndGo = CreateWindow(T("BUTTON"),NULL,WS_TABSTOP|WS_VISIBLE|WS_CHILD|BS_PUSHBUTTON|BS_NOTIFY,
				0,0,10,10,p->Win.WndDialog, NULL, p->Win.Module, NULL );

			SetWindowText(p->WndGo,String(OPENFILE_ID,OPENFILE_GO));

			FontSize = 11;
			SendMessage(p->WndGo,WM_SETFONT,(WPARAM)WinFont(&FontSize,0),0);
		}

		FontSize = 12;
		SendMessage(p->WndURL,WM_SETFONT,(WPARAM)WinFont(&FontSize,0),0);

		Style = WS_TABSTOP|WS_VISIBLE|WS_CHILD|WS_VSCROLL|WS_HSCROLL|LVS_REPORT|LVS_SHOWSELALWAYS;
		if (QueryPlatform(PLATFORM_TYPENO)==TYPE_SMARTPHONE)
			Style |= LVS_NOCOLUMNHEADER;

		p->WndList = CreateWindow(T("SysListView32"), NULL, Style,
			0,0,20,20, p->Win.Wnd, NULL, p->Win.Module, NULL);

		ListView_SetExtendedListViewStyle(p->WndList,LVS_EX_FULLROWSELECT);

		if (WinPixelSX() >= 96*2)
			ListView_SetImageList(p->WndList,
				ImageList_LoadBitmap(p->Win.Module,MAKEINTRESOURCE(IDB_FICON32),32,0,0xFF00FF), LVSIL_SMALL);
		else
			ListView_SetImageList(p->WndList,
				ImageList_LoadBitmap(p->Win.Module,MAKEINTRESOURCE(IDB_FICON16),16,0,0xFF00FF), LVSIL_SMALL);

		AddCol(p,0,String(OPENFILE_ID,OPENFILE_LABEL_NAME),max(15,p->DlgWidth[0]),LVCFMT_LEFT);
		AddCol(p,1,String(OPENFILE_ID,OPENFILE_LABEL_TYPE),max(15,p->DlgWidth[1]),LVCFMT_LEFT);
		AddCol(p,2,String(OPENFILE_ID,OPENFILE_LABEL_SIZE),max(15,p->DlgWidth[2]),LVCFMT_RIGHT);

		p->LastClick = -1;

		Resize(p);

		UpdateMenu(p);
		UpdateFileType(p);
		UpdateSort(p);
		UpdateMultiple(p);
		break;

	case WM_SETFOCUS:
		SetFocus(p->WndList);
		break;

	case WM_DESTROY:
		p->DlgWidth[0] = GetColWidth(p,0);
		p->DlgWidth[1] = GetColWidth(p,1);
		p->DlgWidth[2] = GetColWidth(p,2);
		ClearList(p);
		break;

	case WM_SIZE:
		Resize(p);
		break;

	case WM_NOTIFY:
		Notify = (LPNMLISTVIEW) lParam;

		// mouse double click in multiple mode
		if (p->Multiple && !p->Single && 
			Notify->hdr.code==LVN_ITEMACTIVATE && Notify->iItem==p->LastClick &&
			Notify->ptAction.x >= 0 && Notify->ptAction.y >= 0)
		{
			LVITEM Item;
			Item.iItem=Notify->iItem;
			Item.iSubItem=0;
			Item.mask=LVIF_PARAM;

			if (ListView_GetItem(p->WndList,&Item) && Item.lParam && ((openitem*)Item.lParam)->Image != IMG_DIR)
			{
				Add(p,Notify->iItem);
				return 0;
			}
		}
	
		if (Notify->hdr.code==LVN_ITEMACTIVATE || Notify->hdr.code==NM_CLICK)
		{
			LVITEM Item;
			Item.iItem=Notify->iItem;
			Item.iSubItem=0;
			Item.mask=LVIF_PARAM;

			p->LastClick = Notify->iItem;

			if (ListView_GetItem(p->WndList,&Item))
			{
				if (!Item.lParam) // ".."
				{
					tchar_t* i = tcsrchr(p->Path,'\\');
					if (i) 
					{
						tcscpy(p->Last,i+1);
						*i = 0;
						if (i>p->Path && i[-1]=='\\')
							i[-1]=0;

						p->LastClick = -1;
						UpdateList(p,1);
					}
				}
				else
				if (((openitem*)Item.lParam)->Image==IMG_DIR) // directory
				{
					tcscat(p->Path,T("\\"));
					tcscat(p->Path,((openitem*)Item.lParam)->FileName);
					p->Last[0] = 0;
					p->LastClick = -1;
					UpdateList(p,1);
				}
				else
				{
					if (p->Multiple && !p->Single)
					{
						int State = ListView_GetItemState(p->WndList,Notify->iItem,LVIS_SELECTED) & LVIS_SELECTED;

						p->OwnSelect = 1;
						ListView_SetItemState(p->WndList,Notify->iItem,(State ^ LVIS_SELECTED)|LVIS_FOCUSED,LVIS_SELECTED|LVIS_FOCUSED);
						p->OwnSelect = 0;
					}
					else
						Add(p,-1);
				}
			}

		}

		if (Notify->hdr.code==LVN_KEYDOWN && ((LPNMLVKEYDOWN)lParam)->wVKey == VK_ESCAPE)
			PostMessage(p->Win.Wnd,WM_CLOSE,0,0);

		if (p->Multiple && !p->Single)
		{
			// manual focus and selection

			if (Notify->hdr.code==LVN_KEYDOWN)
			{
				// pass one focus change after keyboard
				
				LPNMLVKEYDOWN Key = (LPNMLVKEYDOWN) lParam;

				if (Key->wVKey == VK_UP ||
					Key->wVKey == VK_PRIOR  ||
					Key->wVKey == VK_HOME)
				{
					if (ListView_GetItemState(p->WndList,0,LVIS_FOCUSED)!=LVIS_FOCUSED)
						p->NextFocus = 1;
				}

				if (Key->wVKey == VK_DOWN ||
					Key->wVKey == VK_NEXT ||
					Key->wVKey == VK_END)
				{
					if (ListView_GetItemState(p->WndList,ListView_GetItemCount(p->WndList)-1,LVIS_FOCUSED)!=LVIS_FOCUSED)
						p->NextFocus = 1;
				}
			}

			if (p->NextFocus && Notify->hdr.code==LVN_ITEMCHANGED && (Notify->uChanged & LVIF_STATE) && 
				((Notify->uOldState ^ Notify->uNewState) & LVIS_FOCUSED))
			{
				// clear NextFocus if the new focus has been set
				if (Notify->uNewState & LVIS_FOCUSED)
					p->NextFocus = 0;

				if ((Notify->uOldState ^ Notify->uNewState) & LVIS_SELECTED)
				{
					// restore old selection
					p->OwnSelect = 1;
					ListView_SetItemState(p->WndList,Notify->iItem,Notify->uOldState & LVIS_SELECTED,LVIS_SELECTED);
					p->OwnSelect = 0;
				}
			}

			if (!p->OwnSelect && Notify->hdr.code==LVN_ITEMCHANGING && (Notify->uChanged & LVIF_STATE))
			{
				// let focus change pass (if NextFocus is set)
				if (!p->NextFocus || !((Notify->uOldState ^ Notify->uNewState) & LVIS_FOCUSED))
				{
					*Result = 1;
					return 1;
				}
			}
		}

		if (Notify->hdr.code==LVN_COLUMNCLICK)
		{
			if (p->SortCol != Notify->iSubItem)
				p->SortDir = 0;
			else
				p->SortDir = !p->SortDir;
			p->SortCol = Notify->iSubItem;

			UpdateSort(p);
		}
		break;
	}
	return 0;
}

static openfile OFile;

void OpenFile_Init()
{
	openfile* p = &OFile;
	memset(p,0,sizeof(openfile));

	p->DlgWidth[0] = 90;
	p->DlgWidth[1] = 40;
	p->DlgWidth[2] = 46;
	p->Class = FORMAT_CLASS;

	p->Win.Id = OPENFILE_ID;
	p->Win.Width = 240;
	p->Win.Height = 320;
	p->Win.Dialog = 1;
	p->Win.OwnDialogSize = 1;
	p->Win.Proc = Proc;
	p->Win.DialogProc = DialogProc;
	p->Win.Enum = Enum;
	p->Win.Get = Get;
	p->Win.Set = Set;

	WinRegister(&p->Win,PRI_DEFAULT);
}

void OpenFile_Done()
{
	int i;

	NodeSaveSetup(&OFile.Win.VMT);
	WinUnRegister(&OFile.Win);

	for (i=0;i<MAXHISTORY;++i)
		Free(OFile.History[i]);
}

