/*****************************************************************************
 *
 * This program is free software ; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 *
 * $Id$
 *
 ****************************************************************************/

#ifndef __PLAYER_H
#define __PLAYER_H

#define PLAYER_ID			FOURCC('P','L','A','Y')

// buffer size in bytes (int)
#define PLAYER_BUFFER_SIZE	0x20
// microdrive mode (bool_t)
#define PLAYER_MICRODRIVE	0x21
// repeat (bool_t)
#define PLAYER_REPEAT		0x22
// shuffle (bool_t)
#define PLAYER_SHUFFLE		0x23
// start play after open (bool_t)
#define PLAYER_PLAYATOPEN	0x24
// keep playlist after WND set null (bool_t)
#define PLAYER_KEEPLIST		0x48
// drop tolerance (tick_t) 
#define PLAYER_DROPTOL		0x26
// skip tolerance (tick_t) 
#define PLAYER_SKIPTOL		0x27
// play or pause (bool_t) 
#define PLAYER_PLAY			0x32
// play speed (fraction)
#define PLAYER_PLAY_SPEED	0x3F
// fastforward (bool_t) 
#define PLAYER_FFWD			0x49
// fastforward speed (fraction) 
#define PLAYER_FFWD_SPEED	0x4A
// position in fraction (fraction)
#define PLAYER_PERCENT		0x25
// position in time (tick_t) 
#define PLAYER_POSITION		0x28
// duration (tick_t) 
#define PLAYER_DURATION		0x46
// benchmark duration (tick_t*) 
#define PLAYER_BENCHMARK	0x55
// display name (tchar_t[])
#define PLAYER_TITLE		0x29
// current format (format*)
#define PLAYER_FORMAT		0x2B
// current stream (stream*)
#define PLAYER_INPUT		0x2A
// current audio stream (int)
#define PLAYER_VSTREAM		0x76
// current video stream (int)
#define PLAYER_ASTREAM		0x77
// current audio output (node*)
#define PLAYER_AOUTPUT		0x2C
// current video output (node*)
#define PLAYER_VOUTPUT		0x2D
// number of files in playlist (int)
#define PLAYER_LIST_COUNT	0x2E
// current file in playlist (int) 
#define PLAYER_LIST_CURRENT	0x2F
// equalizer
#define PLAYER_EQUALIZER	0x33
// equalizer enabled (from equalizer struct)
#define PLAYER_EQENABLE		0x69
// dithering enabled (bool_t)
#define PLAYER_DITHER		0x34
// quality color enabled (bool_t)
#define PLAYER_QUALITYCOLOR	0x61
// zoom factor (fraction_t)
#define PLAYER_ZOOM			0x35
// bilinear zoom (bool_t)
#define PLAYER_SMOOTH		0x47
// optimal pixel update (bool_t)
#define PLAYER_VIDEO_UPDATE 0x36
// fullscreen direction flags (int)
#define PLAYER_FULL_DIR		0x39
// non fullscreen direction flags (int)
#define PLAYER_SKIN_DIR		0x45
// window handle (int)
#define PLAYER_WND			0x3A
// overlay is on top or clipping needed (bool_t)
#define PLAYER_CLIPPING		0x3B
// skin viewport rectangle (rect)
#define PLAYER_SKIN_VIEWPORT 0x3C
// prerotate portrait movies
#define PLAYER_AUTOPREROTATE 0x3D
// fullscreen mode (bool_t)
#define PLAYER_FULLSCREEN	0x3E
// volume volume (int 0..100)
#define PLAYER_VOLUME		0x40
// volume mute (bool_t)
#define PLAYER_MUTE			0x41
// audio quality (int 0..2)
#define PLAYER_AUDIO_QUALITY 0x42
// swap stereo channels
#define PLAYER_AUDIO_SWAP	0x43
// video idct acceleration (bool_t)
#define PLAYER_VIDEO_ACCEL	0x44
// reset and update video settings
#define PLAYER_RESETVIDEO	0x60
// keep audio playing in background
#define PLAYER_KEEPPLAY_AUDIO 0x63
// keep video playing in background
#define PLAYER_KEEPPLAY_VIDEO 0x72
// player in background
#define PLAYER_BACKGROUND	0x64
// manual A/V offset
#define PLAYER_AVOFFSET		0x65
// microdrive start at (%)
#define PLAYER_BURSTSTART	0x66
// how much to load (%)
#define PLAYER_UNDERRUN		0x67
// move step time (tick_t)
#define PLAYER_MOVESTEP		0x68
// stereo enum (int)
#define PLAYER_STEREO		0x74
// aspect ratio (frac)		
#define PLAYER_ASPECT		0x75
// on list changed event
#define PLAYER_LIST_CHANGED 0x78

// array type params (just a hint)
#define PLAYER_ARRAY		0x1000
// url in playlist (0x1000,0x1001,0x1002...)
#define PLAYER_LIST_URL		0x1000
// title in playlist (0x2000,0x2001,0x2002...)
#define PLAYER_LIST_TITLE	0x2000
// title or filename in playlist
#define PLAYER_LIST_AUTOTITLE 0x5000
// length in playlist (0x4000,0x4001,0x4002...)
#define PLAYER_LIST_LENGTH	0x4000
// input comments 
#define PLAYER_COMMENT		0x3000

typedef void (*playerpaint)(void* This,void* DC,int x0,int y0);
typedef void (*playercmd)(void* This);
typedef bool_t (*playercomment)(void* This,int Stream,const tchar_t* Name,tchar_t* Value,int Size);
typedef void (*playergo)(void* This,int Delta);
typedef void (*playerswap)(void* This,int a,int b);

typedef struct player_t
{
	nodeenum		Enum;
	nodeget			Get;
	nodeset			Set;
	playerpaint		Paint;
	playercmd		Stop;
	playercmd		Unload;
	playercmd		ReSync;
	playercomment	CommentByName;
	playergo		ListGo;
	playerswap		ListSwap;

} player;

extern void Player_Init();
extern void Player_Done();

extern int PlayerAdd(player* Player,int Index, const tchar_t* Path);

// strings
#define PLAYER_LOW				0x100
#define PLAYER_MEDIUM			0x101
#define PLAYER_HIGH				0x102
#define PLAYER_AUDIO_NOT_FOUND	0x103
#define PLAYER_VIDEO_NOT_FOUND	0x104
#define PLAYER_MIME_NOT_FOUND	0x105
#define PLAYER_FORMAT_NOT_FOUND	0x106
#define PLAYER_DEADLOCK			0x107

// translated comments
#define COMMENT_TITLE			0x110
#define COMMENT_ARTIST			0x111
#define COMMENT_ALBUM			0x112
#define COMMENT_LANGUAGE		0x113

extern const tchar_t* PlayerComment(int Code); // standard english comment tags

#endif